# readme catalog
>```
>I, PG4ML Machine Learning Framework Installation Guidelines
>    1. Installation conditions and pre work
>    2. Installation steps
>II, Processing of MNIST handwritten digit dataset, construction and training of LENET-5 model
>    1. MNIST Processing of Handwritten Digit Datasets
>    2. LENET-5 Model construction
>    3. LENET-5 Model training
>    4. Several explanations
>III, Basic Principles of Machine Learning Framework Design Based on Database
>    1. Specification for schemas, broadcast operations, grouping of slices/blocks, multicast operations, polymorphic pseudotypes, etc
>    2. Suggestions for improving operational efficiency
>    3. Library Function Planning Classification
>      1). Matrix Basic Operations 
>      2). Classification, regression, neural networks
>      3). Cluster
>      4). Scientific computing
>      5). Other functions
>    4. Method of adding operator types
>    5. Method of increasing the type of loss function
>IV, Appendix
>  (I), Project Directory Structure Description
>  (II), Common abbreviations or explanations for naming database objects
>  (III), Schedule
>    1. Function List
>    2. Operator List
>    3. Introduction to Stored Procedures
>    4. Neural network nodes support operation types
>```

# I, PG4ML Machine Learning Framework Installation Guidelines 
PG4ML Machine Learning Framework is an extension of PostgreSQL, so it's very simple to use it.

>##  1. Installation conditions and pre work
>>>1). Linux (CentOS/RHEL/Ubuntu);

>>>2). PostgreSQL has been installed (version >= 14 with python supported);

>##  2. Installation steps
>>>1). Unzip the project:
>>>>Assuming to directory /var/lib/pgsql/pg4ml.
>>>>```
>>>>├─/var/lib/pgsql/pg4ml/
>>>>│  ├─sql
>>>>│  ├─README.md
>>>>│  ├─README_en.md
>>>>│  ├─pg4ml.control
>>>>│  ├─gen_sql_files.sh
>>>>│  ├─Makefile
>>>>│  ├─...
>>>>```

>>>2). Prepare before install: 

>>>>a. Find postgresql install directory, Assuming install at /usr/pgsql-14/bin/postgres

>>>>> find / -name postgres

>>>>b. Confirm that the system variable PATH contains the postgreSQL installation directory and includes pg_ Config directory:

>>>>> which postgres
>>>>> which pg_config
>>>>> export PATH=$PATH:/usr/pgsql-14/bin    #If not include, add it to PATH

>>>>c. If installed through a package, it is also necessary to install the development toolkit;

>>>>> yum install postgresql14-devel

>>>>> yum install postgresql14-contrib

>>>3). Compile & Install:

>>>> export USE_PGXS=1

>>>> make && make install

>>>4). Create the pg4ml extension
>>>> CREATE EXTENSION pg4ml CASCADE;

# II, Processing of MNIST handwritten digit dataset, construction and training of LENET-5 model
>> Model examples of pg4ml are stored in catelog ./02_nn_model/.  
>>> As a model example ./02_nn_model/000000002_mnist_lenet5_batchnorm, please refer ./02_nn_model/000000002_mnist_lenet5_batchnorm/readme.txt

# III, Basic Principles of Machine Learning Framework Design Based on Database 
>##  1. Specification for schemas, broadcast operations, grouping of slices/blocks, multicast operations, polymorphic pseudotypes, etc

>A schema named sm_sc has been defined to load framework objects and avoid conflicts between postgresql native object names or other extended object names;

><br>

>Protocol: The order of matrix dimensions increases from 1 and decreases from -1, and the 0th dimension is not supported. The positive dimension is different from the matrix dimension in Python and numpy, which starts counting from 0.  

><br>

>The design of array operation functions and the planning of operation libraries take into account the granularity of operations. The operation object may be one or more of the following:
>>1). The internal structure of each element in the matrix. For example, taking the real parts of each element in a complex matrix;

>>2). The one-to-one corresponding element of two or more input matrices. For example, four operations of element (broadcast);

>>3). Slice of the input parameter matrix. For example, slice aggregation;

>>4). Enter the parameter matrix. For example: transpose;

>>5). Aggregation of multiple input parameter matrices. For example, averaging multiple matrices at each corresponding position;

>>6). After slicing the two-dimensional matrix of a three-dimensional or four-dimensional array, it is stacked and aggregated into a three-dimensional or four-dimensional array. For example, multi-channel operation of two-dimensional images;

>>7). The operation of a three-dimensional or four-dimensional array. For example, 3D or 4D element addition, 3D convolution;

><br>

>There are two expressions for several matrix operators in elementary operations of element granularity: one expression does not have reverse quotes as suffixes and is applicable to any element type; Another type is to use reverse quotes as suffixes, usually only for float element types, and adopt optimization strategies such as Python numpy SIMD. Including: addition, subtraction, multiplication, division, exponentiation, logarithm, comparison size, etc.

>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Example: The following are two operator expressions for element addition operation:

>```
>  select 
>    array[[1.2, 3.4]] + array[[5.6,7.8]]
>  , array[[1.2, 3.4]] +` array[[5.6,7.8]]
>```

><br>

>Most slicing and block granularity aggregation operations support grouping;

>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Example 1: Grouping and averaging the vertical slices of a matrix, with one group for every two elements:

>```
>  select 
>    sm_sc.fv_aggr_slice_avg
>    (
>      array[[1.5, 11.5]
>           ,[2.1, 12.1]
>           ,[3.3, 13.3]
>           ,[4.3, 14.3]
>           ,[5.5, 15.5]
>           ,[6.1, 16.1]]
>    , array[2, 1]
>    )
>  return:
>    array[[1.8, 11.8]
>         ,[3.8, 13.8]
>         ,[5.8, 15.8]]
>  Among them, the element 1.8 of the returned result matrix comes from (1.5 + 2.1) / 2 of the input parameter matrix.
>```

>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Example 2: Total grouping of horizontal slices of a matrix, with each group consisting of three elements:

>```
>  select 
>    sm_sc.fv_aggr_slice_sum
>    (
>      array[[1.5, 11.5, 4.3, 14.3, 24.3, 34.3]
>           ,[2.1, 12.1, 5.5, 15.5, 25.5, 35.5]
>           ,[3.3, 13.3, 6.1, 16.1, 26.1, 36.1]]
>    , array[1, 3]
>    )
>  return:
>    array[[17.3, 72.9]
>         ,[19.7, 76.5]
>         ,[22.7, 78.3]]
>  Among them, element 17.3 of the returned result matrix comes from 1.5 + 11.5 + 4.3 of the input parameter matrix.
>```

>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Example 3: Average the grouping of matrix blocks, with each matrix block with a height and width specification of 2\*3  being a group;

>```
>  select 
>    sm_sc.fv_aggr_slice_max
>    (
>      array[[2.3, 5.1, 8.2, 2.56, 3.33, -1.9]
>           ,[3.25, 6.4, 6.6, 6.9, -2.65, -4.6]
>           ,[-2.3, 5.1, -8.2, 2.56, -3.33, -1.9]
>           ,[3.25, -6.4, -6.6, 6.9, -2.65, -4.6]
>           ]
>      , array[2, 3]
>    )
>  return:
>    array[[8.2, 6.9]
>         ,[5.1, 6.9]]
>  Among them, element 8.2 of the returned result matrix comes from greatest(2.3, 5.1, 8.2, 3.25, 6.4, 6.6) of the input parameter matrix.
>```

><br>

>The input parameter matrix multicast of matrix binocular operation is suitable for addition, subtraction, multiplication, division, power operation, logarithmic operation, and matrix multiplication, which refers to the operation of the sub matrix of the background matrix within the sliding window and the window matrix. The sliding window principle is similar to that of convolutional sliding windows.

>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Example 1: Roll down sm_sc.fv_conv_sub

>```
>  select 
>    sm_sc.fv_conv_sub
>    (
>      array[[1.0,2.0,3.0,4.0,5.0,6.0,7.0]
>           ,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]
>           ,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]
>           ,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]
>           ,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]
>           ]
>     , array[[1.0, 2.0, 3.0]
>            ,[-1.0, -2.0, -3.0]
>            ,[3.0, -2.0, 1.0]]
>     , array[2, 2]
>    )
>  return:
>    array[[0.0, 0.0, 0.0, 2.0, 2.0, 2.0, 4.0, 4.0, 4.0]
>         ,[11.0, 22.0, 33.0, 31.0, 42.0, 53.0, 51.0, 62.0, 73.0]
>         ,[97.0, 202.0, 299.0, 297.0, 402.0, 499.0, 497.0, 602.0, 699.0]
>         ,[99.0, 198.0, 297.0, 299.0, 398.0, 497.0, 499.0, 598.0, 697.0]
>         ,[0.0, 0.0, 0.0, -2.0, -2.0, -2.0, -4.0, -4.0, -4.0]
>         ,[-13.0, -18.0, -31.0, -33.0, -38.0, -51.0, -53.0, -58.0, -71.0]]
>  Among them, input parameter one is the background matrix, input parameter two is the window matrix, and input parameter three is the vertical and horizontal sliding step size.
>```

>For a certain type of binary operation that does not follow the commutative law, such as matrix multicast subtraction, there are two types of multicast operations: convolutional subtraction and deconvolution subtraction. The first input parameter of the convolution subtraction operation is the background matrix, and the second input parameter is the window matrix; The first input parameter of the deconvolution operation is the window matrix, and the second input parameter is the background matrix;

>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Example 2: Deconvolving sm_sc.fv_conv_de_sub

>```
>  select 
>    sm_sc.fv_conv_de_sub
>    (
>      array[[1.0, 2.0, 3.0]
>           ,[-1.0, -2.0, -3.0]
>           ,[3.0, -2.0, 1.0]]
>      , array[[1.0,2.0,3.0,4.0,5.0,6.0,7.0]
>             ,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]
>             ,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]
>             ,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]
>             ,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]
>             ]
>     , array[2, 2]
>    )
>  return:
>    array[[0.0, 0.0, 0.0, -2.0, -2.0, -2.0, -4.0, -4.0, -4.0]
>         ,[-11.0, -22.0, -33.0, -31.0, -42.0, -53.0, -51.0, -62.0, -73.0]
>         ,[-97.0, -202.0, -299.0, -297.0, -402.0, -499.0, -497.0, -602.0, -699.0]
>         ,[-99.0, -198.0, -297.0, -299.0, -398.0, -497.0, -499.0, -598.0, -697.0]
>         ,[0.0, 0.0, 0.0, 2.0, 2.0, 2.0, 4.0, 4.0, 4.0]
>         ,[13.0, 18.0, 31.0, 33.0, 38.0, 51.0, 53.0, 58.0, 71.0]]
>  Among them, input parameter one is the window matrix, input parameter two is the background matrix, and input parameter three is the vertical and horizontal sliding step size.
>```

><br>

>This set of matrix basic operation function library, with functions explained by "element (broadcast)", usually supports broadcast function. Including the following scenarios:
>>1). Single element values and one-dimensional arrays (regardless of the order of input parameters);

>>2). Single element values and 2D arrays (regardless of the order of input parameters);

>>3). One dimensional array and 2D array (regardless of the order of input parameters);

>>4). A single row of 2D arrays and a two-dimensional array (regardless of the order of input parameters);

>>5). 2D array single column and 2D array (input parameters in no particular order);

>>6). Single element values and 3D arrays (regardless of the order of input parameters);

>>7). Single element values and 4D arrays (regardless of the order of input parameters);

>>8). 3D slicing and 3D array (regardless of the order of input parameters);

>>9). 4D slicing and 4D array (regardless of the order of input parameters);

>Among them, the results of 3) and 4) are equivalent.

>For operators constructing neural networks, 2_Opr input parameters are usually also supported, with a broadcast function where one input parameter has several dimensions with a length of 1, but does not support broadcast functions with different dimensions. Refer to the specifications mentioned in sections 4), 5), 8), and 9) above.

><br>

>Many functions in the PG4ML matrix operation function library utilize the pseudo type feature of postgresql, using anyarray as input and output parameter types, similar to Java generics or C++ class templates, to support the polymorphism of these functions on matrices of any element type.



><br>

>##  2. Suggestions for improving operational efficiency

>>1). Turn off the debugging review switch:

>>>Many functions in the PG4ML matrix operation function library review the dimensions and lengths of array type parameters.

>>>This review can be enabled by setting the session variable pg4ml._v_is_debug_check. This review is deemed invalid by default.

>>>     set session pg4ml._v_is_debug_check = '1';   -- open
>>>     set session pg4ml._v_is_debug_check = '0';   -- close
>>>     select current_setting('pg4ml._v_is_debug_check', true)   -- Viewing Variable Values

>>2). Matrix operation library optimization:

>>>In the current version, matrix multiplication is implemented through pg's extension pgplpython3 calling numpy.

>>>It is recommended to also use other programming language matrix operation libraries with higher running efficiency, such as Openblas;

>>3). Multi threaded training process:

>>>a. Adjust the system kernel parameter /etc/sysctl.conf, /etc/security/limits.conf to allocate more system resources for the pg database;

>>>b. Set pg parallel parameters;

>>>     min_parallel_table_scan_size
>>>     min_parallel_index_scan_size
>>>     force_parallel_mode
>>>     max_parallel_workers_per_gather
>>>     parallel_setup_cost
>>>     parallel_tuple_cost

>>4). Distributed computing transformation:

>>>It is recommended to modify the lambda call of neural network nodes to RDD or MPP;

>>5). Using GPU:

>>>It is recommended to create an pg_strom extension and introduce GPU computing;

>>6). Modification of setting/getting middle tensor during model training and inference:

>>>In current version, middle tensor during model training and inference is got/set in table sm_sc.__vt_array_kv through below 4 function interface: 
>>>  sm_sc.__fv_set_kv(), sm_sc.__fv_set_kvs(), sm_sc.__fv_get_kv(), sm_sc.__fv_get_kvs()
>>>For this table and four interfaces, modifications can be made through methods such as PG extension or kernel storage engine.  

<br>

>##  3. Library Function Planning Classification

>###  1). Matrix Basic Operations 
>>Specific library function classification planning:
>>>    fn_a. Arithmetic, comparison, carry, concatenation, Boolean, bitwise, complex, and broadcast operations of matrix elements:

>>>    fn_b. Matrix multiplication, transpose, flip, turn, concatenation:

>>>    fn_c. The construction, sampling, replacement, filling, character matching, random generation, and shelling of matrices (as the PG13 version does not currently support array high-dimensional slicing):

>>>    fn_d. Trigonometric function broadcast operation of matrix:

>>>    fn_f. Slice granularity aggregation of matrices, matrix granularity aggregation, sorting by a certain slice value, slice granularity positioning extreme value position:

><br>

>###  2). Classification, regression, neural networks
>>1). Table - sm_sc.tb_nn_node / sm_sc.__vt_nn_node (the former is used for training, while the latter is used for testing and validation) and Table - sm_sc.tb_nn_path are used to carry the node and path structures of deep neural networks;

>>2).  Table - sm_sc.tb_nn_train_input_buff, used for public reception training set;

>>3). Table - sm_sc.tb_classify_task, used for deploying and managing training tasks;

>>4). Defined relevant functions for machine learning and matrix operations, used for real-time execution of simple or atomic operations;

>>Specific operational function library classification planning:
>>>    fn_g. The redistribution function of machine learning, including standardization, normalization, normalization, centralization, and so on;

>>>    fn_h. Activation function of machine learning;

>>>    fn_j. Convolutional, pooling, matrix multicast for machine learning;

>>>    fn_k. Lambda for machine learning;

>>>    fn_l. Neural network specific for machine learning;

>>>    fn_m. Encoding and decoding of machine learning;

>>>    fn_n. The loss function of machine learning;

>>>    fn_q. Derivative formula derivation;

>>>    fn_r. Derivation of complex functions;

>>   5). Defined procedure; 

>>>    st_a. procedure for ML train;

>>>    st_b. procedure for DL train;

>>>    st_c. procedure for manage DL session;

>>>    st_d. procedure for manage DL model;

>>   6). Function - sm_sc.ft_nn_in_out, used to input a single piece of data from a test or validation set, and output test or validation results;

><br>

>###  3). Cluster
>>   1). Hard clustering - kmeans++
>>>    a. table - sm_sc.tb_cluster_task, used for deploying and managing training tasks;
>>>    b. procedure - sm_sc.prc_kmeans_pp, used to execute kmeans++tasks;

>>   2). Soft clustering - dbscan
>>>    a. table - sm_sc.tb_cluster_task, used for deploying and managing training tasks;
>>>    b. procedure - sm_sc.prc_dbscan_pp, used to execute dbscan tasks;
><br>

>###  4). Scientific Computing
>>   1). Defined the relevant functions of scientific computing for real-time execution of simple or atomic operations;

>>Specific operational function library classification planning:
>>>    fn_e. Waveform processing:

>>>    fn_i. JSON serialization and deserialization of computational graphs:

>>>    fn_o. Complex operations:

>>>    fn_s. Linear Algebra:

<!--
>>   2). Defined the relevant processes of scientific computing for asynchronous execution of complex or time-consuming operations;
>>>    a. procedure - sm_sc.prc_mx_qr, used for qr decomposition;
>>>    b. procedure - sm_sc.prc_mx_evd_value, used to obtain matrix eigenvalues;
>>>    c. procedure - sm_sc.prc_mx_evd, used for EVD decomposition;
>>>    d. procedure - sm_sc.prc_mx_max_abs_evd_value, used to obtain the main eigenvalue of a matrix, which is the modulus (absolute value) of the maximum eigenvalue;
>>>    e. procedure - sm_sc.prc_mx_min_abs_evd_value, used to calculate the modulus (absolute value) of the minimum eigenvalue;
>>>    f. procedure - sm_sc.prc_mx_svd, used for singular value decomposition;
-->
><br>

>###  5). Other functions
>>    The udf public function supplemented by fn_p. for postgresql:
>>    fn_q. (secondary development) user defined operator functions:
>>>     This type of function is defined by the user, extended for differentiation, and registered as an operator, and is called by a unified lambda extension interface.
>>    fn_r. (secondary development) user defined composite operator functions:
>>>     This type of function is defined by the user, only in the form of JSON configuration, to assist in the design and construction of the model, and does not participate in model training and deployment.

<br>

>##  4. Method of adding operator types
>>The following are guidelines for defining operators within the framework:

>>>1). Write operator functions and their derivatives, and plan to store them in the directory where the source code belongs (if there are derivatives);
>>>>  Taking convolution as an example,
>>>>  Function source code storage path:

>>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/06_conv_pool/sm_sc.fv_conv_2d_im2col_py.sql

>>>>  Function derivative source code storage path:
>>>>
>>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/06_conv_pool/01_delta/sm_sc.fv_d_conv_2d_dloss_dindepdt_1_ex.sql
>>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/06_conv_pool/01_delta/sm_sc.fv_d_conv_2d_dloss_dindepdt_2_py.sql
>>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/06_conv_pool/01_delta/sm_sc.fv_d_conv_2d_dloss_dindepdt_3.sql

>>>2). Define the lambda name of the newly added operator;
>>>>  Taking convolution as an example, the lambda name of the convolution is '05_conv_prod_mx'.

>>>3). In the dictionary table, set the configuration structure and meaning of each element in the hyperparameter array field of the operator;
>>>>  Dictionary table location:
>>>>
>>>>     plpgsql_pg4ml/sql/06_data_initial/table_data_initial.sql
>>>>       Record of enum_name = 'node_fn_asso_value' in Table sm_sc.tb_dic_enum

>>>>  Taking convolution as an example, the configuration structure and meaning of function hyperparameters:
>>>>
>>>>     select trim('window_len_heigh__1d         '), trim('Sliding window height specifications                   '), trim('05_conv_prod_mx        '),  1, numrange(2.0, 2.0, '[]') union all
>>>>     select trim('stride__1d                   '), trim('Vertical and horizontal step size                     '), trim('05_conv_prod_mx        '),  2, numrange(4.0, 5.0, '[]') union all
>>>>     select trim('padding__1d                  '), trim('Complete the number of rows/columns from top to bottom, left to right              '), trim('05_conv_prod_mx        '),  3, numrange(6.0, 9.0, '[]') union all
>>>>     select trim('padding_value__1d            '), trim('Completing Fill Element Values                     '), trim('05_conv_prod_mx        '),  4, numrange(10.0, 10.0, '[]') union all


>>>4). In the dictionary table, set other feature configurations for the operator;
>>>>  Taking convolution as an example,
>>>>
>>>>     plpgsql_pg4ml/sql/06_data_initial/table_data_initial.sql
>>>>       Record of enum_name = 'node_fn_type' in Table sm_sc.tb_dic_enum
>>>>       Record of enum_name = 'node_fn_type_delta' in Table sm_sc.tb_dic_enum
>>>>       Record of enum_name = 'node_fn_type_delta_method' in Table sm_sc.tb_dic_enum

>>>5). Register the operator and its derivative anonymous calls in the lambda function related to neural networks;
>>>>  Path to the lamdda function:
>>>>
>>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/07_lambda/sm_sc.fv_lambda_arr_len.sql
>>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/07_lambda/sm_sc.fv_lambda_arr.sql
>>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/07_lambda/sm_sc.fv_lambda_arr_ddepdt_dindepdt.sql
>>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/07_lambda/sm_sc.fv_lambda_arr_dloss_dindepdt.sql
>>>>  Among them,
>>>>    a. Derive lambda, select one of fv_lambda_arr_ddepdt_dindepdt and fv_lambda_arr_dloss_dindepdt based on the derivative method of the newly added operator;
>>>>    b. In fv_lambda_arr_len, it may not be necessary to configure new operator types. There is a default audit strategy for monocular operations, where the specification of the dependent variable array is equal to the specification of the unique independent variable array;

>>>6). If an operator is specified for this operator function, create it in the following source code file;
>>>>  Array operator source code path:
>>>>
>>>>     plpgsql_pg4ml/sql/09_operator/operator_mx.sql

>>>7). Add relevant explanations for the newly added operator in the readme.md document and multi-language version;
>>>>    a. List of operator functions:
>>>>>    IV, appendix -> (III), Schedule -> 1. Function list;
>>>>    b. List of derivative functions for operators:
>>>>>    IV, appendix -> (III), Schedule -> 1. Function List -> r. Derivation of some matrix operations;
>>>>    c. If Operator list is established:
>>>>>    IV, appendix -> (III), Schedule -> 2. Operator list;
>>>>    d. List of neural network node types:
>>>>>    IV, appendix -> (III), Schedule -> 4. Neural network nodes support operation types;

>>>8). Edit testcase
>>>>  This includes use cases such as operator functions, derivative functions, lambda anonymous calls, etc. Use case script path:
>>>>     plpgsql_pg4ml/sql/unit_test.sql

>>>9). Incrementally deploy the above code changes and configuration data initialization section to the database where the PG4ML framework is located;

>>The following are guidelines for custom operators:

>>>1). In the PG4ML deep learning framework, it is recommended to store the custom operator named with "81_" as the prefix in the following path:

>>>     plpgsql_pg4ml/sql/03_function/81_ufv/81_xxxxx/

>>>2). In the dictionary table, set other feature configurations for the operator;
>>>>  Taking convolution as an example,
>>>>
>>>>     plpgsql_pg4ml/sql/06_data_initial/01_ufv_data_initial_81_xxxxx.sql
>>>>       Table sm_sc.tb_dic_enum: record with enum_name = 'node_fn_asso_value'
>>>>       Table sm_sc.tb_dic_enum: record with enum_name = 'node_fn_type'
>>>>       Table sm_sc.tb_dic_enum: record with enum_name = 'node_fn_type_delta'
>>>>       Table sm_sc.tb_dic_enum: record with enum_name = 'node_fn_type_delta_method'

>>>3). The operator with this named prefix can be recognized by the lambda function in the following path and called anonymously;

>>>>     plpgsql_pg4ml/sql/03_function/81_ufv/sm_sc.ufv_lambda_...

>>>4). Similar to the operators within the framework, custom operators also require corresponding derivative functions, configurations, and documentation. It is recommended to store them separately in the same path.

>>>>     plpgsql_pg4ml/sql/03_function/81_ufv/readme_udf_lambda.md

>##  5. Method of increasing the type of loss function
>>1). Write loss functions and their derivatives, and plan to store them in the directory where the source code of the loss function belongs;
>>>  Taking the least squares loss function as an example,
>>>  Function source code storage path:

>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/10_loss_fn/sm_sc.fv_loss_least_square.sql

>>>  Function derivative source code storage path:
>>>
>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/10_loss_fn/01_delta/sm_sc.fv_dloss_dz_least_square.sql

>>2). Define the lambda name for the newly added loss function type;
>>>  Taking the least squares loss function as an example, its lambda name code is' 101 '.

>>3). In the dictionary table, set the explanation for the loss function of the operator;
>>>  Dictionary table location:
>>>
>>>     plpgsql_pg4ml/sql/06_data_initial/table_data_initial.sql
>>>       Record of enum_name = 'loss_fn_type' in Table sm_sc.tb_dic_enum

>>>  Taking the least squares loss function as an example, the configuration structure and meaning of the hyperparameters of the function are as follows:
>>>
>>>     select 
>>>       'loss_fn_type'      as enum_name, 
>>>       '101'               as enum_key, 
>>>       'least square method'        as enum_value, 
>>>       row_number() over() as enum_order

>>4). Register the operator and its derivative anonymous calls in the lambda function related to neural networks;
>>>  Path to the lamdda function:
>>>
>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/07_lambda/sm_sc.fv_lambda_loss.sql
>>>     plpgsql_pg4ml/sql/03_function/04_machine_learn/07_lambda/sm_sc.fv_lambda_dloss_dz.sql

>>5). Incrementally deploy the above code changes and configuration data initialization section to the database where the PG4ML framework is located;


# IV, appendix
## (I), Project Directory Structure Description
>```
>├─plpgsql_pg4ml
>│  ├─sql
>│  │  ├─01_type_schema_extension                       : Create a schema and extension
>│  │  │  └─04_create_type                              : Type Created
>│  │  │     ├─01                                       : The serial number here is only used to control the deployment execution order
>│  │  │     │  ├─01_typ                                : Create a regular type
>│  │  │     │  └─02___type                             : Create a global internal usage type
>│  │  │     └─02                                       : The serial number here is only used to control the deployment execution order
>│  │  ├─02_table_view                                  : Create tables and views
>│  │  ├─03_function                                    : Create function
>│  │  │  ├─03_matrix                                   : Matrix correlation function
>│  │  │  │  ├─01_calcu_bit_concat_compare              : Arithmetic operations, bit operations, concatenation, comparison
>│  │  │  │  │  └─01_delta                              : Corresponding derivative function
>│  │  │  │  ├─03_prod_trans_turn_mirror                : Matrix multiplication, transform, transform, flip
>│  │  │  │  │  └─01_delta                              : Corresponding derivative function
>│  │  │  │  ├─04_construct_sample_match_replace        : Build, sample, match, replace
>│  │  │  │  │  └─01_delta                              : Corresponding derivative function
>│  │  │  │  ├─05_trigono                               : Trigonometric operation
>│  │  │  │  │  └─01_delta                              : Corresponding derivative function
>│  │  │  │  ├─06_wave                                  : Waveform transformation
>│  │  │  │  └─09_slice_aggr_ord                        : Slice, aggregate, sort
>│  │  │  │     │─01_fv                                 : Functions returning type
>│  │  │  │     │  └─01_delta                           : Corresponding derivative function
>│  │  │  │     └─02_fa                                 : Aggregate function
>│  │  │  │        └─01_delta                           : Corresponding derivative function
>│  │  │  ├─04_machine_learn                            : The Correlation Function of Machine Learning
>│  │  │  │  ├─01_udf_calcu                             : Customized simple type functions
>│  │  │  │  │  └─01_delta                              : Corresponding derivative function
>│  │  │  │  ├─02_redistribute                          : Standardized functions
>│  │  │  │  │  └─01_delta                              : Corresponding derivative function
>│  │  │  │  ├─03_grad_reduction                        : Derivative derivation
>│  │  │  │  ├─04_activate                              : Activation function
>│  │  │  │  │  └─01_delta                              : Corresponding derivative function
>│  │  │  │  ├─05_computational_graph_json              : Calculation diagram of JSON type
>│  │  │  │  ├─06_conv_pool                             : Convolutional, pooling, matrix multicast
>│  │  │  │  │  └─01_delta                              : Corresponding derivative function
>│  │  │  │  ├─07_lambda                                : Lambda function
>│  │  │  │  ├─08_nn                                    : Neural network
>│  │  │  │  ├─09_encode_decode                         : Encoding and decoding
>│  │  │  │  └─10_loss_fn                               : Loss function
>│  │  │  ├─05_complex                                  : Complex
>│  │  │  │  ├─01_calcu                                 : Complex elementary operation function
>│  │  │  │  ├─02_alian                                 : Alias for complex elementary arithmetic functions
>│  │  │  │  └─03_aggr                                  : Complex aggregation operation
>│  │  │  ├─06_common                                   : Universal function
>│  │  │  ├─07_math_array                               : Linear algebra
>│  │  │  ├─81_ufv                                      : Custom operator
>│  │  │  └─83_ufv_conbine                              : Custom composite operator
>│  │  ├─04_procedure                                   : Create Stored Procedure
>│  │  │  └─01_machine_learn                            : Machine learning
>│  │  │     ├─01_classify_regress                      : Classification, regression
>│  │  │     ├─02_neural_network                        : Neural network
>│  │  │     └─03_cluster                               : Cluster
>│  │  ├─06_data_initial                                : Initialized Data
>│  │  ├─09_operator                                    : Create Operators
>│  │  └─10_install                                     : Installation and deployment
>│  └─02_nn_model                                       : Model
>│     ├─000000002_mnist_lenet5_batchnorm               : MNIST dataset, LENET-5 neural network, BN standardization
>│     │  ├─00_for_design_and_pre_train                 : Model design and pre-train
>│     │  │ ├─01_mnist_dataset                          : MNIST dataset
>│     │  │ ├─02_import_to_buffer_table                 : Import buff table
>│     │  │ ├─03_create_train_nn_task                   : Create neural network structure and training tasks
>│     │  │ ├─04_train                                  : Execute training
>│     │  │ ├─05_validation                             : Checking
>│     │  │ └─06_model_backup                           : Export training model backup
>│     │  ├─01_import_pre_train_model                   : Export pre-training model
>│     │  ├─02_import_infer_model                       : Export infer model
>│     │  ├─03_export_model                             : Import model
>│     │  └─04_delete_model                             : Delete model
>│     ├─00xxxxxxx_xxxx                                 : Other models
>│     │  ├─01_xxx_dataset                              : Xxx dataset
>│     │  ├─...                                         : Other directories for the model
>│     ├─...                                            : Other models
>```

## (Ⅱ), Common abbreviations or explanations for naming database objects
>* \_            : The common abbreviation or explanation for naming database objects as a naming prefix means temporary objects used internally within the execution body (including code blocks, functions, stored procedures)
>* \_\_          : As a naming prefix, it means that objects are used globally internally
>* typ\_         : type
>* sm\_          : Schema
>* seq\_         : sequence
>* tb\_          : table
>* idx\_         : index
>* vw\_          : view
>* xv\_          : x\_view
>* fv\_          : function\_value
>* ft\_          : function\_table
>* fa\_          : function\_aggregate
>* vf\_          : function\_void
>* prc\_         : procedure
>* m\_           : member
>* p\_           : point
>* is\_          : bool 
>* i\_           : input
>* o\_           : output
>* v\_           : val
>* cte\_         : common\_table\_expression
>* enum\_        : enum
>* sess\_        : session
>* \_no          : number
>* \_id          : id
>* \_a\_         : alias
>* \_cur\_       : cursor
>* \_co\_        : cooperative
>* \_asso\_      : associate
>* \_ex\_        : extra
>* \_2\_         : to
>* \_4\_         : for
>* \_l\_         : long
>* \_s\_         : short
>* \_tmp\_       : temp
>* \_ord\_       : order
>* \_aggr\_      : aggregate
>* \_idx\_       : position
>* \_dic\_       : dictionary
>* \_opr\_       : operator
>* \_sum\_       : summary
>* \_lsum\_      : low\_summary
>* \_hsum\_      : high\_summary
>* \_cfg\_       : configure
>* \_fn\_        : function
>* \_time\_      : time
>* \_lang\_      : language
>* \_gis\_       : geographic\_information\_system
>* \_sys\_       : system
>* \_map\_       : map
>* \_snap\_      : snapshot
>* \_trs\_       : transaction
>* \_ods\_       : operational\_data\_store
>* \_dw\_        : data\_warehouse
>* \_dm\_        : data\_market
>* \_sour\_      : source
>* \_tar\_       : target
>* \_his\_       : history
>* \_slice\_     : slice
>* \_chunk\_     : chunk
>* \_mx\_        : matrix
>* \_arr\_       : array
>* \_y\_         : Y-axis, the first dimension of an array. If it is a two-dimensional array, that is, in the vertical direction
>* \_x\_         : X-axis, the second dimension of an array. If it is a two-dimensional array, that is, in the horizontal direction
>* \_x3\_        : The third dimension of an array.
>* \_x4\_        : The fourth dimension of an array.
>* \_heith\_     : The penultimate dimension of an array, usually the height dimension of a visual operation plane slice.
>* \_width\_     : The reciprocal first dimension of an array, usually the width dimension of a visual operation plane slice.
>* \_mx\_        : Matrix
>* \_1d\_        : 1st\_dimension, One-dimensional (array)
>* \_2d\_        : 2nd\_dimension, 2D (array)
>* \_3d\_        : 3rd\_dimension, 3D (array)
>* \_4d\_        : 4th\_dimension, 4D (array)
>* \_ele\_       : (Array) Element
>* \_area\_      : Designated Area
>* \_pos\_       : Specified position
>* \_nn\_        : neural\_network
>* \_fore\_      : Forward neural network
>* \_back\_      : (Neural Network) Reverse
>* \_loss\_      : loss,loss function related 
>* \_d\_         : delta
>* \_de\_        : de-, no, opposite
>* \_ne\_        : negative
>* \_nega\_      : negative
>* \_posi\_      : positive
>* \_bi\_        : bis-, pairwise
>* \_tri\_       : triangular, three
>* \_depdt\_     : Dependent variable
>* \_indepdt\_   : Independent variable
>* \_redistr\_   : Redistribution, including standardization, normalization, normalization, and centralization related
>* \_45\_        : degree of angle 45
>* \_90\_        : degree of angle 90
>* \_180\_       : degree of angle 180
>* \_dddi\_      : delta dependent / delta independent, the derivative of the dependent variable over the independent variable
>* \_dldi\_      : delta loss / delta independent, derivative of loss function with respect to independent variable

## (Ⅲ), Schedule
>###  1. Function List
>>####   a. Arithmetic, comparison, carry, concatenation, Boolean, bitwise, complex, and broadcast operations of matrix elements:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_opr_add<br>sm_sc.fv_opr_add_py               |Element (broadcast) add.<br>Input param:<br>&nbsp;&nbsp;1. Addition 1;<br>&nbsp;&nbsp;2. Addition 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_add            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-1.3, 2.5],[-2.4, 2.1]]
>>> sm_sc.fv_opr_sub<br>sm_sc.fv_opr_sub_py               |Element (broadcast) sub. <br>Input param:<br>&nbsp;&nbsp;1. Subtracted number;<br>&nbsp;&nbsp;2. Subtraction; |select<br>&nbsp;&nbsp;sm_sc.fv_opr_sub            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[3.9, -20.1],[6.8, -8.1]]
>>> sm_sc.fv_opr_sub<br>sm_sc.fv_opr_sub_py               |Element (broadcast) opposite number.<br>Input param:<br>&nbsp;&nbsp;1. Original array;|select<br>&nbsp;&nbsp;sm_sc.fv_opr_sub            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-1.3,8.8],[-2.2,3.0]]
>>> sm_sc.fv_opr_mul<br>sm_sc.fv_opr_mul_py               |Element (broadcast) mul.<br>Input param:<br>&nbsp;&nbsp;1. Multiplier 1;<br>&nbsp;&nbsp;2. Multiplier 2; |select<br>&nbsp;&nbsp;sm_sc.fv_opr_mul            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-3.38, -99.44],[-10.12, -15.30]]
>>> sm_sc.fv_opr_div<br>sm_sc.fv_opr_div_py               |Element (broadcast) div.<br>Input param:<br>&nbsp;&nbsp;1. Dividend;<br>&nbsp;&nbsp;2. Divisor;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_div            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-0.5, -0.77876],[-0.47826, -0.58824]]
>>> sm_sc.fv_opr_div<br>sm_sc.fv_opr_div_py               |Element (broadcast) reciprocal.<br>Input param:<br>&nbsp;&nbsp;1. Original array; |select<br>&nbsp;&nbsp;sm_sc.fv_opr_div            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.76923, -0.11364],[0.45455, -0.33333]]
>>> sm_sc.fv_opr_mod<br>sm_sc.fv_opr_mod_posi_py               |Element (broadcast) positive remainder.<br>Input param:<br>&nbsp;&nbsp;1. Dividend;<br>&nbsp;&nbsp;2. Divisor; |select<br>&nbsp;&nbsp;sm_sc.fv_opr_mod_posi_py            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[10, -29],[12, -33]]<br>&nbsp;&nbsp;, array[[-3, 11],[-5, 5]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1, 4],[2, 2]]
>>> sm_sc.fv_opr_mod<br>sm_sc.fv_opr_mod_nega_py               |Element (broadcast) negative remainder.<br>Input param:<br>&nbsp;&nbsp;1. Dividend;<br>&nbsp;&nbsp;2. Divisor; |select<br>&nbsp;&nbsp;sm_sc.fv_opr_mod_nega_py            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[10, -29],[12, -33]]<br>&nbsp;&nbsp;, array[[-3, 11],[-5, 5]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-2, -7],[-3, -3]]
>>> sm_sc.fv_opr_pow<br>sm_sc.fv_opr_pow_py               |Element (broadcast) pow.<br>Input param:<br>&nbsp;&nbsp;1. Base; <br>&nbsp;&nbsp;2. Exponent;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_pow            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 8.8],[2.2, 3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 1.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.50553, 16.89770],[0.02660, 271.21793]]
>>> sm_sc.fv_opr_exp<br>sm_sc.fv_opr_exp_py               |Element (broadcast) natural constant power.<br>Input param:<br>&nbsp;&nbsp;1. Exponent;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_exp            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[3.66930, 0.00015],[9.02501, 0.04979]]
>>> sm_sc.fv_opr_log<br>sm_sc.fv_opr_log_py               |Element (broadcast) log.<br>Input param:<br>&nbsp;&nbsp;1. Base; <br>&nbsp;&nbsp;2. Antilogarithm;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_log            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 8.8],[2.2, 3.0]]<br>&nbsp;&nbsp;, array[[2.6, 11.3],[4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[3.64193, 1.11498],[1.93550, 1.48300]]
>>> sm_sc.fv_opr_ln<br>sm_sc.fv_opr_ln_py                |Element (broadcast) natural log.<br>Input param:<br>&nbsp;&nbsp;1. Antilogarithm;   |select<br>&nbsp;&nbsp;sm_sc.fv_opr_ln             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 8.8],[2.2, 3.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.26236, 2.17475],[0.78846, 1.09861]]
>>> sm_sc.fv_opr_or                |Element (broadcast) bit OR, logical OR operation.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_or             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;, array[[false, true],[false, true]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[true, true],[true, true]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_or             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;, array[[B'011', B'101'],[B'011', B'101']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[B'011', B'111'],[B'111', B'111']]
>>> sm_sc.fv_opr_and               |Element (broadcast) bit AND, logical AND operation.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_and            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;, array[[false, true],[false, true]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[false, false],[false, false]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_and        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;, array[[B'011', B'101'],[B'011', B'101']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[B'010', B'001'],[B'001', B'001']]
>>> sm_sc.fv_opr_not               |Element (broadcast) bit Not, logical No operation.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_not            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[false, true],[false, true]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_not        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[B'101', B'100'],[B'010', B'100']]
>>> sm_sc.fv_opr_is_regexp_match   |Element (broadcast) does the text regex match.<br>Input param:<br>&nbsp;&nbsp;1. Original text matrix; <br>&nbsp;&nbsp;2. Regular expression matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_is_regexp_match<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[['abbbbbc122223', 'abc123'],['abc123', 'ac13']]<br>&nbsp;&nbsp;, array[['a.c', '1.\*?3'],['1.3', 'a.\*?c']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[false, true],[true, true]]
>>> sm_sc.fv_opr_conjugate         |Element (broadcast) complex element conjugation.<br>Input param:<br>&nbsp;&nbsp;1. Original array;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_conjugate      <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[(12.3, 25.1), (-2.56, -3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, -3.25)]] :: sm_sc.typ_l_complex[]
>>> sm_sc.fv_opr_conjugate_i       |Complex matrix conjugation.<br>Input param:<br>&nbsp;&nbsp;1. Original array;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_conjugate_i    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[(12.3, 25.1), (12.3, 0.0)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(-2.56, -3.25), (0.0, -3.25)]] :: sm_sc.typ_l_complex[]
>>> sm_sc.fv_opr_real              |Element (broadcast) Complex element real part.<br>Input param:<br>&nbsp;&nbsp;1. Original array;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_real           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[12.3, -2.56],[12.3, 0.0]]
>>> sm_sc.fv_opr_imaginary         |Element (broadcast) complex element imaginary part.<br>Input param:<br>&nbsp;&nbsp;1. Original array;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_imaginary      <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-25.1, 3.25],[0.0, 3.25]]
>>> sm_sc.fv_opr_conjugate_45      |Element (broadcast) exchange of real and imaginary parts of complex elements.<br>Input param:<br>&nbsp;&nbsp;1. Original array;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_conjugate_45   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[(-25.1, 12.3), (3.25, -2.56)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(0.0, 12.3), (3.25, 0.0)]] :: sm_sc.typ_l_complex[]
>>> sm_sc.fv_opr_xor               |Element (broadcast) bit XOR, logical XOR operation.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_xor            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[true, true],[true, true]]<br>&nbsp;&nbsp;, array[[false, true],[false, true]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[true, true],[true, true]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_xor             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;, array[[B'011', B'101'],[B'011', B'101']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[B'001', B'110'],[B'110', B'110']]
>>> sm_sc.fv_opr_xnor              |Element (broadcast) bit XNOR, logical XNOR operation.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_xnor           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;, array[[false, true],[false, true]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[false, false],[false, false]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_xnor             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;, array[[B'011', B'101'],[B'011', B'101']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[B'110', B'001'],[B'001', B'001']]
>>> sm_sc.fv_opr_concat            |Element (broadcast) concat.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;   |select<br>&nbsp;&nbsp;sm_sc.fv_opr_concat         <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[['abc', 'efgh'],['hijk', 'lm']]<br>&nbsp;&nbsp;, array[['no', 'pqr'],['stu', 'vwxyz']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[['abcno', 'efghpqr'],['hijkstu', 'lmvwxyz']]
>>> sm_sc.fv_opr_shift_left        |Element (broadcast) bit shift left.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_shift_left     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;, array[[1, 2],[-1, -2]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[B'100', B'100'],[B'010', B'000']]
>>> sm_sc.fv_opr_shift_right       |Element (broadcast) bit shift right.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_shift_right    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;, array[[1, 2],[-1, -2]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[B'001', B'000'],[B'010', B'100']]
>>> sm_sc.fv_opr_is_equal<br>sm_sc.fv_opr_is_equal_py          |Element (broadcast) equality judgment.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_is_equal       <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, -8.8],[2.2, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[false, true],[true, false]]
>>> sm_sc.fv_opr_is_less<br>sm_sc.fv_opr_is_less_py           |Element (broadcast) less than judgment.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_is_less        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[false, true],[false, true]]
>>> sm_sc.fv_opr_is_greater<br>sm_sc.fv_opr_is_greater_py        |Element (broadcast) greater than judgment.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_is_greater     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 3.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[true, false],[true, false]]
>>> sm_sc.fv_opr_is_less_ex<br>sm_sc.fv_opr_is_less_ex_py        |Element (broadcast) less than or equal to judgment.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_is_less_ex     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[1.3, -8.8],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[true, true],[false, true]]
>>> sm_sc.fv_opr_is_greater_ex<br>sm_sc.fv_opr_is_greater_ex_py     |Element (broadcast) greater than or equal to judgment.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_is_greater_ex  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 11.3],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[2.2, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[true, true],[true, false]]
>>> sm_sc.fv_opr_compare           |Element (broadcast) compare.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_compare        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1, -1],[1, -1]]
>>> sm_sc.fv_opr_sign<br>sm_sc.fv_opr_sign_py              |Element (broadcast) positive and negative judgment.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_sign           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0],[-2.6, 11.3]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1, -1],[1, -1],[-1, 1]]
>>> sm_sc.fv_opr_least             |Element (broadcast) take the minimum value.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_least          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-2.6, -8.8],[-4.6, -3.0]]
>>> sm_sc.fv_opr_greatest          |Element (broadcast) Take the maximum value.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_greatest       <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.3, 11.3],[2.2, 5.1]]
>>> sm_sc.fv_opr_abs               |Element (broadcast) take absolute value.<br>Input param:<br>&nbsp;&nbsp;1. Original array;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_abs            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0],[-2.6, 11.3]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.3, 8.8],[2.2, 3.0],[2.6, 11.3]]
>>> sm_sc.fv_opr_round             |Element (broadcast) rounding.<br>Input param:<br>&nbsp;&nbsp;1. Original array; <br>&nbsp;&nbsp;2. Retain the number of decimal places;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_round          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1244.3753, -8.8443],[225.2153, -30.03821]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1240, -10],[230, -30]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_round          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1244.3753, -8.8443],[225.2153, -30.03821]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1244, -9],[225, -30]]
>>> sm_sc.fv_opr_floor             |Element (broadcast) digital floor value.<br>Input param:<br>&nbsp;&nbsp;1. Original array; <br>&nbsp;&nbsp;2. Retain the number of decimal places;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_floor          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1244.3753, -8.8443],[225.2153, -30.03821]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1244.37, -8.85],[225.21, -30.04]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_floor          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1244.3753, -8.8443],[225.2153, -30.03821]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1244, -9],[225, -31]]
>>> sm_sc.fv_opr_ceil              |Element (broadcast) digital ceiling value.<br>Input param:<br>&nbsp;&nbsp;1. Original array; <br>&nbsp;&nbsp;2. Retain the number of decimal places;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_ceil           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1244.3753, -8.8443],[225.2153, -30.03821]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1244.38, -8.84],[225.22, -30.03]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_ceil           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1244.3753, -8.8443],[225.2153, -30.03821]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1245, -8],[226, -30]]
>>> sm_sc.fv_opr_trunc             |Element (broadcast) digital truncation value.<br>Input param:<br>&nbsp;&nbsp;1. Original array; <br>&nbsp;&nbsp;2. Retain the number of decimal places;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_trunc          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1244.3753, -8.8443],[225.2153, -30.03821]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1244.37, -8.84],[225.21, -30.03]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_opr_trunc          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1244.3753, -8.8443],[225.2153, -30.03821]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1244, -8],[225, -30]]
>>
>>####   b. Matrix multiplication, transpose, flip, turn, concat:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_concat_py         |array concat.  <br>Input: <br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2; <br>&nbsp;&nbsp;3. The dimension axes where concat is located;  |select<br>&nbsp;&nbsp;sm_sc.fv_concat_py      <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1],[2.6, 5.8]]<br>&nbsp;&nbsp;, 1<br>&nbsp;&nbsp;);<br><br>return: <br>&nbsp;&nbsp;array[[1.3, -8.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2.2, -3.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-2.6, 11.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-4.6, 5.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2.6, 5.8]]
>>> sm_sc.fv_concat_y         |2D array above and below concat, the first dimension of a multidimensional array concat.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_concat_y      <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1],[2.6, 5.8]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.3, -8.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2.2, -3.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-2.6, 11.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-4.6, 5.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2.6, 5.8]]
>>> sm_sc.fv_concat_x         |2D array left and right concat, the second dimension of a multidimensional array concat.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_concat_x      <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3, 1.1],[-4.6, 5.1, 0.4]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.3, -8.8, -2.6, 11.3, 1.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2.2, -3.0, -4.6, 5.1, 0.4]]
>>> sm_sc.fv_concat_x3            |The third dimension of a multidimensional array concat.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_concat_x3<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,2,3,4],[11,12,13,14],[111,112,113,114]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[5,6,7,8],[15,16,17,18],[115,116,117,118]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,2,3],[11,12,13],[111,112,113]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[5,6,7],[15,16,17],[115,116,117]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,3,4,1,2,3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[11,12,13,14,11,12,13]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[111,112,113,114,111,112,113]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[5,6,7,8,5,6,7]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[15,16,17,18,15,16,17]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[115,116,117,118,115,116,117]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_concat_x4            |The fourth dimension of a multidimensional array concat.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_concat_x4<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,2,3],[11,12,13],[111,112,113]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[5,6,7],[15,16,17],[115,116,117]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[21,22,23],[31,32,33],[131,132,133]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[25,26,27],[35,36,37],[135,136,137]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[-1,-2,-3,-4],[-11,-12,-13,-14],[-111,-112,-113,-114]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[-5,-6,-7,-8],[-15,-16,-17,-18],[-115,-116,-117,-118]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[-21,-22,-23,-24],[-31,-32,-33,-34],[-121,-122,-123,-124]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[-25,-26,-27,-28],[-35,-36,-37,-38],[-125,-126,-127,-128]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,3,-1,-2,-3,-4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[11,12,13,-11,-12,-13,-14]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[111,112,113,-111,-112,-113,-114]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[5,6,7,-5,-6,-7,-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[15,16,17,-15,-16,-17,-18]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[115,116,117,-115,-116,-117,-118]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,22,23,-21,-22,-23,-24]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[31,32,33,-31,-32,-33,-34]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[131,132,133,-121,-122,-123,-124]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[25,26,27,-25,-26,-27,-28]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[35,36,37,-35,-36,-37,-38]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[135,136,137,-125,-126,-127,-128]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_opr_concat_heigh     |2D background plane above and below (height direction) concat, the reciprocal second dimension of a multidimensional array concat.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_concat_heigh  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1],[2.6, 5.8]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.3, -8.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2.2, -3.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-2.6, 11.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-4.6, 5.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2.6, 5.8]]
>>> sm_sc.fv_opr_concat_width     |2D background plane left and right (width direction) concat,the reciprocal first dimension of a multidimensional array concat.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_concat_width  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3, 1.1],[-4.6, 5.1, 0.4]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.3, -8.8, -2.6, 11.3, 1.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2.2, -3.0, -4.6, 5.1, 0.4]]
>>> sm_sc.fv_opr_mirror_heigh         |Flip up and down a 2D array, the reciprocal second dimension of a multi-dimensional array flip up and down.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_mirror_heigh      <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[2.2, -3.0, -0.95],[1.3, -8.8, 2.25]]
>>> sm_sc.fv_opr_mirror_width         |Flip left and right a 2D array, the reciprocal first dimension of a multi-dimensional array flip up and down.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_mirror_width      <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[2.25, -8.8, 1.3],[-0.95, -3.0, 2.2]]
>>> sm_sc.fv_opr_mirror<br>sm_sc.fv_opr_mirror_py           |Multidimensional array specifies the reverse order of dimensions.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Specify the dimension. The highest 4D;  |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_opr_mirror<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,2,3,4],[11,12,13,14]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[5,6,7,8],[15,16,17,18]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[21,22,23,24],[31,32,33,34]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[25,26,27,28],[35,36,37,38]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;3<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[11,12,13,14],[1,2,3,4]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[15,16,17,18],[5,6,7,8]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[31,32,33,34],[21,22,23,24]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[35,36,37,38],[25,26,27,28]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_opr_transpose<br>sm_sc.fv_opr_transpose_py        |2D array transpose.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_transpose     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[1.3, 2.2],[-8.8,-3.0],[2.25,-0.95]]
>>> sm_sc.fv_opr_transpose_nd<br>sm_sc.fv_opr_transpose_nd_py     |3D matrix transpose.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Two dimensions of transposition;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_transpose_nd     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]],[[-1.3, 8.8, -2.25],[-2.2, 3.0, 0.95]]]<br>&nbsp;&nbsp;, array[2, 3]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[[1.3,2.2],[-8.8,-3.0],[2.25,-0.95]],[[-1.3,-2.2],[8.8,3.0],[-2.25,0.95]]]
>>> sm_sc.fv_opr_transpose_i<br>sm_sc.fv_opr_transpose_i_py      |Matrix inversion transpose, equivalent to flipping left and right before transposing.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_transpose_i   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[-0.95,2.25],[-3.0,-8.8],[2.2,1.3]]
>>> sm_sc.fv_chunk_transpose      |Matrix transpose chunk by chunk。<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Chunk height and width specifications; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_chunk_transpose<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,&nbsp;&nbsp;6,&nbsp;&nbsp;15])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:2][1:10][1:9]
>>> sm_sc.fv_opr_turn_180<br>fv_opr_turn_180_py     |Turn the two dimensions specified in the array by 180 degrees. <br>Input: <br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. two turn dimensions； |select&nbsp;&nbsp;sm_sc.fv_opr_turn_180<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[[1,2,3,4],[11,12,13,14]],[[5,6,7,8],[15,16,17,18]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[[14,13,12,11],[4,3,2,1]],[[18,17,16,15],[8,7,6,5]]]
>>> sm_sc.fv_opr_turn_90<br>fv_opr_turn_90_py     |The array is turn 90 degrees from the positive half axis direction of a specified dimension to the positive half axis direction of another specified dimension. <br>Input: <br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. The starting dimension of the turning half axis; |select&nbsp;&nbsp;sm_sc.fv_opr_turn_90<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[[1,2,3,4],[11,12,13,14]],[[5,6,7,8],[15,16,17,18]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[[11,1],[12,2],[13,3],[14,4]],[[15,5],[16,6],[17,7],[18,8]]]
>>> sm_sc.fv_opr_turn_heigh_width_180     |Rotate the heigh and width dimensions of the array by 180 °, rotate the matrix by 180 °, or equivalent to flipping left and right and then up and down.<br>Equivalent to transposing and then inverting,<br>or equivalent to flipping the heigh and width dimensions in sequence.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_turn_heigh_width_180     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[2.25, -8.8, 1.3],[-0.95, -3.0, 2.2]]
>>> sm_sc.fv_opr_turn_width_heigh_90      |Rotate the width dimension of the array 90 ° in the positive direction to the heigh dimension in the positive direction.<br>Rotate the matrix counterclockwise by 90 °.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_turn_width_heigh_90     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[2.2, 1.3],[-3.0, -8.8],[-0.95, 2.25]]
>>> sm_sc.fv_opr_turn_heigh_width_90      |Rotate the matrix clockwise by 90 °.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_turn_heigh_width_90    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[2.25, -0.95],[-8.8, -3.0],[1.3, 2.2]]
>>> sm_sc.fv_opr_prod_inner       |Matrix inner product; Total product of corresponding elements.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_prod_inner    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3, 3.3],[-4.6, 5.1, -6.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;-115.020
>>> sm_sc.fv_opr_prod_inner_pow   |Matrix power product; Corresponding element power total.<br>Input param:<br>&nbsp;&nbsp;1. Base of the power product; <br>&nbsp;&nbsp;2. Exponent of power product;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_prod_inner_pow<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;, 3.0<br>&nbsp;&nbsp;);                             <br><br>return:<br>&nbsp;&nbsp;-685.09375
>>> sm_sc.fv_opr_prod_kronecker   |Kronecker product and outer product.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_prod_kronecker<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-3.38, 14.69, 22.88, -99.44, -5.850, 25.425]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-5.98, 6.63, 40.48, -44.88, -10.350, 11.475]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-5.72, 24.86, 7.80, -33.90, 2.470, -10.735]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-10.12, 11.22, 13.80, -15.30, 4.370, -4.845]]
>>> sm_sc.fv_opr_prod_mx_py       |Matrix multiplication implemented in Python: multiplication, dot product and quantity product.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_prod_mx_py    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]]<br>&nbsp;&nbsp;, array[[-2.6, 11.3],[-4.6, 5.1],[4.1, -3.1]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[46.3250, -37.1650],[4.1850, 12.5050]]
>>> sm_sc.fv_chunk_prod_mx        |Multiplication of block grouping matrices.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; <br>&nbsp;&nbsp;2. Operand 2; <br>&nbsp;&nbsp;3. The height and width specifications of the two input parameters of the matrix multiplication block; Respectively:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a. i_chunk_len[1]: The height of parameter one in the multiplication of partitioned matrices;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;b. i_chunk_len[2]: The width of block matrix multiplication into parameter one, that is, the height of block matrix multiplication into parameter two;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;c. i_chunk_len[3]: The width of block matrix multiplication into parameter two;|select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_chunk_prod_mx<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array_fill(random(),&nbsp;&nbsp;array[2&nbsp;&nbsp;*&nbsp;&nbsp;7,&nbsp;&nbsp;3&nbsp;&nbsp;*&nbsp;&nbsp;11])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array_fill(random(),&nbsp;&nbsp;array[3&nbsp;&nbsp;*&nbsp;&nbsp;7,&nbsp;&nbsp;5&nbsp;&nbsp;*&nbsp;&nbsp;11])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)&nbsp;&nbsp;&nbsp;&nbsp;--&nbsp;&nbsp;Expected specifications:array[2&nbsp;&nbsp;*&nbsp;&nbsp;7,&nbsp;&nbsp;5&nbsp;&nbsp;*&nbsp;&nbsp;11]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:14][1:55]
>>> sm_sc.fv_opr_prod_mx_pow      |Square matrix multiplication power.<br>Input param:<br>&nbsp;&nbsp;1. Base of matrix multiplication; <br>&nbsp;&nbsp;2.Exponent of matrix multiplication; |select<br>&nbsp;&nbsp;sm_sc.fv_opr_prod_mx_pow   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);                               <br><br>return:<br>&nbsp;&nbsp;array[[1.3000, 2.25],[2.2000, -3.0]]
>>> sm_sc.fv_opr_prod_mx_left     |Square matrix multiplied by its transpose.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_prod_mx_left  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[-8.8, 2.25],[2.2, -0.95]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[82.5025, -21.4975],[-21.4975, 5.7425]]
>>> sm_sc.fv_opr_prod_mx_right    |Multiply the transpose of a square matrix by that matrix.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1;  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_prod_mx_right <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[6.5300, -3.6750],[-3.6750, 14.0625]]
>>
>>####   c. The construction, sampling, replacement, filling, character matching, random generation, and shelling of matrices (as the PG13 version does not currently support array high-dimensional slicing):
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_mx_ele_2d_2_1d<br>sm_sc.fv_mx_ele_2d_2_1d_py       |Flatten a 2D array into a 1D array.  Original dimension disappear.<br>Input param:<br>&nbsp;&nbsp;1. Operand 1; |select<br>&nbsp;&nbsp;sm_sc.fv_mx_ele_2d_2_1d<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]<br>&nbsp;&nbsp;);                                             <br><br>return:<br>&nbsp;&nbsp;array[1.3, -8.8, 2.2, -3.0]
>>> sm_sc.fv_mx_ele_1d_2_2d<br>sm_sc.fv_mx_ele_1d_2_2d_py       |A 1D array is stacked into a 2D array according to the specified width.<br>Input param:<br>&nbsp;&nbsp;1. Original 1D array; <br>&nbsp;&nbsp;2. Specified width; |select<br>&nbsp;&nbsp;sm_sc.fv_mx_ele_1d_2_2d<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -8.8, 2.2, -3.0, 6.1, 0.23]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;); <br><br>return:<br>&nbsp;&nbsp;array[[1.3, -8.8, 2.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-3.0, 6.1, 0.23]]
>>> sm_sc.fv_mx_ele_3d_2_2d       |A 3D array flattens two specified dimensions into a 2D array.  Original dimension disappear.<br>Input param:<br>&nbsp;&nbsp;1. Original 4D array; <br>&nbsp;&nbsp;2. Two dimensions to be merged. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The merged new dimension is in the order and position of to; <br>&nbsp;&nbsp;3. Old dimensions with fixed element order. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The order of elements under this old dimension will be retained until the new dimension. |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_mx_ele_3d_2_2d<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[5,&nbsp;6,&nbsp;7,&nbsp;8],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[9,&nbsp;10,&nbsp;11,&nbsp;12]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[25,&nbsp;26,&nbsp;27,&nbsp;28],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[29,&nbsp;30,&nbsp;31,&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;2],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4]<br>&nbsp;&nbsp;,&nbsp;[5,&nbsp;6,&nbsp;7,&nbsp;8]<br>&nbsp;&nbsp;,&nbsp;[9,&nbsp;10,&nbsp;11,&nbsp;12]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24]<br>&nbsp;&nbsp;,&nbsp;[25,&nbsp;26,&nbsp;27,&nbsp;28]<br>&nbsp;&nbsp;,&nbsp;[29,&nbsp;30,&nbsp;31,&nbsp;32]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_mx_ele_2d_2_3d<br>sm_sc.fv_mx_ele_2d_2_3d_py       |A 2D array is stacked into a 3D array based on the specified dimension and width.<br>Input param:<br>&nbsp;&nbsp;1. Original 3D array; <br>&nbsp;&nbsp;2. The number of elements in each segmentation group. <br>&nbsp;&nbsp;3. Split dimension; <br>&nbsp;&nbsp;4. New dimension; <br>&nbsp;&nbsp;3. Whether to retain element order in the split dimension. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Otherwise, preserve the order of elements in the new dimension. |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_mx_ele_2d_2_3d<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9,&nbsp;10,&nbsp;11,&nbsp;12]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29,&nbsp;30,&nbsp;31,&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;true<br>&nbsp;&nbsp;)<br><br>return: <br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3],&nbsp;[4,&nbsp;5,&nbsp;6],&nbsp;[7,&nbsp;8,&nbsp;9],&nbsp;[10,&nbsp;11,&nbsp;12]]<br>&nbsp;&nbsp;,&nbsp;[[21,&nbsp;22,&nbsp;23],&nbsp;[24,&nbsp;25,&nbsp;26],&nbsp;[27,&nbsp;28,&nbsp;29],&nbsp;[30,&nbsp;31,&nbsp;32]]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_mx_ele_4d_2_3d<br>sm_sc.fv_mx_ele_4d_2_3d_py       |A 4D array flattens two specified dimensions into a 3D array.  Original dimension disappear.<br>Input param:<br>&nbsp;&nbsp;1. Original 4D array; <br>&nbsp;&nbsp;2. Two dimensions to be merged. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The merged new dimension is in the order and position of to; <br>&nbsp;&nbsp;3. Old dimensions with fixed element order. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The order of elements under this old dimension will be retained until the new dimension. |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_ele_4d_2_3d<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4],&nbsp;&nbsp;[5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8],&nbsp;&nbsp;[9,&nbsp;&nbsp;10,&nbsp;&nbsp;11,&nbsp;&nbsp;12]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24],&nbsp;&nbsp;[25,&nbsp;&nbsp;26,&nbsp;&nbsp;27,&nbsp;&nbsp;28],&nbsp;&nbsp;[29,&nbsp;&nbsp;30,&nbsp;&nbsp;31,&nbsp;&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4],&nbsp;&nbsp;[-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7,&nbsp;&nbsp;-8],&nbsp;&nbsp;[-9,&nbsp;&nbsp;-10,&nbsp;&nbsp;-11,&nbsp;&nbsp;-12]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-21,&nbsp;&nbsp;-22,&nbsp;&nbsp;-23,&nbsp;&nbsp;-24],&nbsp;&nbsp;[-25,&nbsp;&nbsp;-26,&nbsp;&nbsp;-27,&nbsp;&nbsp;-28],&nbsp;&nbsp;[-29,&nbsp;&nbsp;-30,&nbsp;&nbsp;-31,&nbsp;&nbsp;-32]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[[[1,2,3,4],[21,22,23,24]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[[5,6,7,8],[25,26,27,28]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[[9,10,11,12],[29,30,31,32]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[[-1,-2,-3,-4],[-21,-22,-23,-24]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[[-5,-6,-7,-8],[-25,-26,-27,-28]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[[-9,-10,-11,-12],[-29,-30,-31,-32]]]
>>> sm_sc.fv_mx_ele_3d_2_4d<br>fv_mx_ele_3d_2_4d_py       |A 3D array is stacked into a 4D array based on the specified dimension and width.<br>Input param:<br>&nbsp;&nbsp;1. Original 3D array; <br>&nbsp;&nbsp;2. The number of elements in each segmentation group. <br>&nbsp;&nbsp;3. Split dimension; <br>&nbsp;&nbsp;4. New dimension; <br>&nbsp;&nbsp;3. Whether to retain element order in the split dimension. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Otherwise, preserve the order of elements in the new dimension. |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_ele_3d_2_4d<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[6,8,10])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;4<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;true<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;"[1:6][1:2][1:10][1:4]"
>>> sm_sc.fv_mx_ele_flatten_2dims<br>fv_mx_ele_flatten_2dims_py |A 2D, 3D or 4D array flattens two specified dimensions into a array with same ndims.  Original dimension remain length = 1.<br>Input param:<br>&nbsp;&nbsp;1. Original N_D array; <br>&nbsp;&nbsp;2. Two dimensions to be merged. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The merged new dimension is in the order and position of to; <br>&nbsp;&nbsp;3. Old dimensions with fixed element order. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The order of elements under this old dimension will be retained until the new dimension. |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_ele_flatten_2dims<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,3,5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>返回:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:2][1:1][1:15]
>>> sm_sc.fv_mx_slice_4d_2_3d<br>fv_mx_slice_4d_2_3d_py     |4D slicing to remove shells into 3D.<br>Input param:<br>&nbsp;&nbsp;1. Original 4D array; <br>&nbsp;&nbsp;2. Sliced dimension; <br>&nbsp;&nbsp;3. The serial number of the sliced position. The default is 1; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_slice_4d_2_3d<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array_fill(random(),&nbsp;&nbsp;array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;4,&nbsp;&nbsp;5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)&nbsp;&nbsp;&nbsp;&nbsp;--&nbsp;&nbsp;Expected specifications: array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:3][1:2][1:5]
>>> sm_sc.fv_mx_slice_3d_2_2d<br>fv_mx_slice_3d_2_2d_py     |3D slicing to remove shells into 2D.<br>Input param:<br>&nbsp;&nbsp;1. Original 3D array; <br>&nbsp;&nbsp;2. Sliced dimension; <br>&nbsp;&nbsp;3. The serial number of the sliced position. The default is 1; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_slice_3d_2_2d<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array_fill(random(),&nbsp;&nbsp;array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)&nbsp;&nbsp;&nbsp;&nbsp;--&nbsp;&nbsp;Expected specifications: array[3,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:3][1:2]
>>> sm_sc.fv_mx_slice_4d_2_2d<br>fv_mx_slice_4d_2_2d_py     |4D slicing to remove shells into 2D.<br>Input param:<br>&nbsp;&nbsp;1. Original 4D array; <br>&nbsp;&nbsp;2. Two dimensions sliced; <br>&nbsp;&nbsp;3. The serial number of the sliced position. The default is array[1, 1]; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_slice_4d_2_2d<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array_fill(random(),&nbsp;&nbsp;array[4,&nbsp;&nbsp;3,&nbsp;&nbsp;2,&nbsp;&nbsp;5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)&nbsp;&nbsp;&nbsp;&nbsp;--&nbsp;&nbsp;Expected specifications: array[4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:4][1:5]
>>> sm_sc.fv_mx_ascend_dim        |Add shell to increase dimension.<br>Input param:<br>&nbsp;&nbsp;1. Original array; <br>&nbsp;&nbsp;2. Number of shells added; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_ascend_dim<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array_fill(random(),&nbsp;&nbsp;array[2,&nbsp;&nbsp;5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:1][1:1][1:2][1:5]
>>> sm_sc.fv_mx_descend_dim<br>sm_sc.fv_mx_descend_dim_py       |remove shell to decrease dimension.<br>Input param:<br>&nbsp;&nbsp;1. Original array; <br>&nbsp;&nbsp;2. Number of shells removed; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_descend_dim<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array_fill(random(),&nbsp;&nbsp;array[1,&nbsp;&nbsp;1,&nbsp;&nbsp;2,&nbsp;&nbsp;5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:2][1:5]
>>> sm_sc.fv_new                  |Construct an array with given element values.<br>Input param:<br>&nbsp;&nbsp;1. Element value;<br>&nbsp;&nbsp;2. Specification of array dimensions and length; <br>---------------------<br>Construct a new array with given array units.<br>Input param:<br>&nbsp;&nbsp;1. Array unit; <br>&nbsp;&nbsp;2. Number of repetitions of array units;|select<br>&nbsp;&nbsp;sm_sc.fv_new<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;3.6<br>&nbsp;&nbsp;, array[3, 4]<br>&nbsp;&nbsp;);                                                             <br><br>return:<br>&nbsp;&nbsp;array[[3.6, 3.6, 3.6, 3.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[3.6, 3.6, 3.6, 3.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[3.6, 3.6, 3.6, 3.6]]<br>---------------------<br>select<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[2,&nbsp;&nbsp;4],&nbsp;&nbsp;[1,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2,&nbsp;&nbsp;4,&nbsp;&nbsp;2,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,&nbsp;&nbsp;3,&nbsp;&nbsp;1,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2,&nbsp;&nbsp;4,&nbsp;&nbsp;2,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,&nbsp;&nbsp;3,&nbsp;&nbsp;1,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2,&nbsp;&nbsp;4,&nbsp;&nbsp;2,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,&nbsp;&nbsp;3,&nbsp;&nbsp;1,&nbsp;&nbsp;3]]
>>> sm_sc.fv_repeat_axis_py       |Given an array, construct a new array where each element is duplicated along a single axis.  <br>Input：<br>&nbsp;&nbsp;1. Array unit; <br>&nbsp;&nbsp;2. Extension axis for repeat element; <br>&nbsp;&nbsp;3. Repetitions; <br>---------------------<br>Given an array, construct a new array where each element is duplicated along multiple of axes.  <br>Input：<br>&nbsp;&nbsp;1. Array unit; <br>&nbsp;&nbsp;2. multiple of axes for repeat elements; <br>&nbsp;&nbsp;3. Element repetitions on each axis; |select<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_repeat_axis_py<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[2,&nbsp;&nbsp;&nbsp;&nbsp;4],&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>return: <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2,&nbsp;&nbsp;2,&nbsp;&nbsp;4,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,&nbsp;&nbsp;1,&nbsp;&nbsp;3,&nbsp;&nbsp;3]]<br>---------------------<br>select<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_repeat_axis_py<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[2,&nbsp;&nbsp;&nbsp;&nbsp;4],&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;&nbsp;&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;&nbsp;&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>return: <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[2,&nbsp;&nbsp;&nbsp;&nbsp;4,&nbsp;&nbsp;&nbsp;&nbsp;2,&nbsp;&nbsp;&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2,&nbsp;&nbsp;&nbsp;&nbsp;4,&nbsp;&nbsp;&nbsp;&nbsp;2,&nbsp;&nbsp;&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[2,&nbsp;&nbsp;&nbsp;&nbsp;4,&nbsp;&nbsp;&nbsp;&nbsp;2,&nbsp;&nbsp;&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,&nbsp;&nbsp;&nbsp;&nbsp;1,&nbsp;&nbsp;&nbsp;&nbsp;3,&nbsp;&nbsp;&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,&nbsp;&nbsp;&nbsp;&nbsp;1,&nbsp;&nbsp;&nbsp;&nbsp;3,&nbsp;&nbsp;&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,&nbsp;&nbsp;&nbsp;&nbsp;1,&nbsp;&nbsp;&nbsp;&nbsp;3,&nbsp;&nbsp;&nbsp;&nbsp;3]]
>>> sm_sc.fv_opr_reshape_py   |Given that the storage order of array elements remains unchanged, rearrange the dimension axis and dimension length. <br>Input: <br>&nbsp;&nbsp;1. Array unit; <br>&nbsp;&nbsp;2. Dimension and length specifications; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_reshape_py<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,&nbsp;&nbsp;2.4,&nbsp;&nbsp;3.5],&nbsp;&nbsp;[1.6,&nbsp;&nbsp;2.5,&nbsp;&nbsp;3.9]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,2.4],[3.5,1.6],[2.5,3.9]]
>>> sm_sc.fv_strided_float_py   |Given an array, slide to take the cut pieces and combine them in order to form a new array.  <br>Input: <br>&nbsp;&nbsp;1. Array unit; <br>&nbsp;&nbsp;2. The dimension axis and dimension length specifications of the new array; <br>&nbsp;&nbsp;3. The distance between a certain element in the new matrix and the number of adjacent elements across different dimensions that are accessed in the original matrix; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_strided_float_py<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1,2,3,4],[5,6,7,8],[9,10,11,12],[13,14,15,16]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[4,&nbsp;&nbsp;1,&nbsp;&nbsp;4,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[[[1,2,3],[5,6,7],[9,10,11]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[[2,3,4],[6,7,8],[10,11,12]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[[[5,6,7],[9,10,11],[13,14,15]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[[6,7,8],[10,11,12],[14,15,16]]]]
>>> sm_sc.fv_new_rand             |Construct a random array within the range of 0.0 to the given value.<br>Input param:<br>&nbsp;&nbsp;1. Upper limit of random values; <br>&nbsp;&nbsp;2. Array specifications; <br>---------------------<br>Construct a random array within the range of 0.0 to 1.0.<br>Input param:<br>&nbsp;&nbsp;1. Array specifications; |select<br>&nbsp;&nbsp;sm_sc.fv_new_rand<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;2.0<br>&nbsp;&nbsp;, array[3, 2]<br>&nbsp;&nbsp;);                                        <br><br>return:<br>&nbsp;&nbsp;array[[0.070740982239343, 0.0899943754998205]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.629551500745066, 0.369479334510409]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.159237183526308, 1.29839220669395]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_new_rand<br>&nbsp;&nbsp;( <br>&nbsp;&nbsp;&nbsp;&nbsp;array[3, 2]<br>&nbsp;&nbsp;);                                        <br><br>return:<br>&nbsp;&nbsp;array[[0.070740982239343, 0.0899943754998205]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.629551500745066, 0.369479334510409]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.159237183526308, 0.29839220669395]]
>>> sm_sc.fv_new_rand_enum        |Constructing enumerated random matrix.<br>Input param:<br>&nbsp;&nbsp;1. 1D array of original enumeration values; <br>&nbsp;&nbsp;2. New array specification;|select<br>&nbsp;&nbsp;sm_sc.fv_new_rand_enum<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']<br>&nbsp;&nbsp;, array[2, 3]<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[['Sun', 'Sat', 'Sun ],['Sun', 'Tue', 'Wed']]
>>> sm_sc.fv_new_randn            |Constructing a normal distribution random matrix.<br>Input param:<br>&nbsp;&nbsp;1. Normal distribution mean (expected value); <br>&nbsp;&nbsp;2. Normal distribution standard deviation; <br>&nbsp;&nbsp;3. New array specification; |select<br>&nbsp;&nbsp;sm_sc.fv_new_randn<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;3.6<br>&nbsp;&nbsp;&nbsp;&nbsp;, 0.2<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 2]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[3.66170684, 3.40963440]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[3.60093244, 3.51437251]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[3.30444215, 3.35682136]]
>>> sm_sc.fv_eye                  |Construct diagonal matrices for given sequences and background elements.<br>Input param:<br>&nbsp;&nbsp;1. Background element value; <br>&nbsp;&nbsp;2. array dims; <br>&nbsp;&nbsp;3...n.Diagonal element values arranged in order can be variant arrays; |select<br>&nbsp;&nbsp;sm_sc.fv_eye<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;1.0<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;, variadic array[1.5, 2.3, 5.6]<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[[1.5, 1.0, 1.0],[1.0, 2.3, 1.0],[1.0, 1.0, 5.6]]
>>> sm_sc.fv_eye_arr_dense              |Construction dense matrix for given diagonal blocks.<br>Input param:<br>&nbsp;&nbsp;1...n. Diagonal block matrix queue; <br>--------------------------<br>Construction dense matrix for given diagonal blocks and background element values.<br>Input param:<br>&nbsp;&nbsp;1. Background element value (default to 0.0); <br>&nbsp;&nbsp;2...n. Diagonal block matrix queue; |select<br>&nbsp;&nbsp;sm_sc.fv_eye_arr_dense<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, 2.4], [11.2, 12.4]]<br>&nbsp;&nbsp;, array[[1.3, 2.6, 1.4, 2.8], [11.3, 12.6, 11.4, 12.8]]<br>&nbsp;&nbsp;, array[[1.1, 2.2], [11.1, 12.2], [21.1, 22.2]]<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[[1.2, 2.4, 0.0, 0.0, 0.0, 0.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[11.2, 12.4, 0.0, 0.0, 0.0, 0.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0, 0.0, 1.3, 2.6, 1.4, 2.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0, 0.0, 11.3, 12.6, 11.4, 12.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0, 0.0, 0.0, 0.0, 1.1, 2.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0, 0.0, 0.0, 0.0, 11.1, 12.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0, 0.0, 0.0, 0.0, 21.1, 22.2]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_eye_arr_dense<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;1.0, <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, 2.4], [11.2, 12.4]]<br>&nbsp;&nbsp;, array[[1.3, 2.6, 1.4, 2.8], [11.3, 12.6, 11.4, 12.8]]<br>&nbsp;&nbsp;, array[[1.1, 2.2], [11.1, 12.2], [21.1, 22.2]]<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[[1.2, 2.4, 1.0, 1.0, 1.0, 1.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[11.2, 12.4, 1.0, 1.0, 1.0, 1.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1.0, 1.0, 1.3, 2.6, 1.4, 2.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1.0, 1.0, 11.3, 12.6, 11.4, 12.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1.0, 1.0, 1.0, 1.0, 1.1, 2.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1.0, 1.0, 1.0, 1.0, 11.1, 12.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1.0, 1.0, 1.0, 1.0, 21.1, 22.2]]
>>> sm_sc.fv_eye_arr_sparse       |Construction sparse matrix for given diagonal blocks.<br>Input param:<br>&nbsp;&nbsp;1...n. Diagonal block matrix queue; <br>--------------------------<br>Construction sparse matrix for given diagonal blocks and background element values.<br>Input param:<br>&nbsp;&nbsp;1. Background element value (default to 0.0); <br>&nbsp;&nbsp;2...n. Diagonal block matrix queue; |select<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_eye_arr_sparse<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,&nbsp;&nbsp;2.4],&nbsp;&nbsp;[11.2,&nbsp;&nbsp;12.4]]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.3,&nbsp;&nbsp;2.6,&nbsp;&nbsp;1.4,&nbsp;&nbsp;2.8],&nbsp;&nbsp;[11.3,&nbsp;&nbsp;12.6,&nbsp;&nbsp;11.4,&nbsp;&nbsp;12.8]]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.1,&nbsp;&nbsp;2.2],&nbsp;&nbsp;[11.1,&nbsp;&nbsp;12.2],&nbsp;&nbsp;[21.1,&nbsp;&nbsp;22.2]]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;);&nbsp;&nbsp;<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,2.4,0,0,0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[11.2,12.4,0,0,0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0,0,1.3,2.6,1.4,2.8,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0,0,11.3,12.6,11.4,12.8,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0,0,0,0,0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0,0,0,0,0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0,0,0,0,0,0,1.1,2.2,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0,0,0,0,0,0,11.1,12.2,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0,0,0,0,0,0,21.1,22.2,0]]<br>--------------------------<br>select<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_eye_arr_sparse<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.0,&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,&nbsp;&nbsp;2.4],&nbsp;&nbsp;[11.2,&nbsp;&nbsp;12.4]]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.3,&nbsp;&nbsp;2.6,&nbsp;&nbsp;1.4,&nbsp;&nbsp;2.8],&nbsp;&nbsp;[11.3,&nbsp;&nbsp;12.6,&nbsp;&nbsp;11.4,&nbsp;&nbsp;12.8]]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.1,&nbsp;&nbsp;2.2],&nbsp;&nbsp;[11.1,&nbsp;&nbsp;12.2],&nbsp;&nbsp;[21.1,&nbsp;&nbsp;22.2]]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;);&nbsp;&nbsp;<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,2.4,1,1,1,1,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[11.2,12.4,1,1,1,1,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,1,1.3,2.6,1.4,2.8,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,1,11.3,12.6,11.4,12.8,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,1,1,1,1,1,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,1,1,1,1,1,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,1,1,1,1,1,1.1,2.2,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,1,1,1,1,1,11.1,12.2,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1,1,1,1,1,1,21.1,22.2,1]]
>>> sm_sc.fv_eye_unit             |Construct a diagonal matrix with a given value.<br>Input param:<br>&nbsp;&nbsp;1. Diagonal matrix height and width specifications; <br>&nbsp;&nbsp;2. Diagonal element value; <br>&nbsp;&nbsp;3. Specify the length of the first or first and second dimensions when stacking in three or four dimensions; <br>--------------------------<br>Construct a diagonal matrix with a value of 1.0 for the unit (left diagonal element).<br>Input param:<br>&nbsp;&nbsp;1. Diagonal matrix height and width specifications; |select<br>&nbsp;&nbsp;sm_sc.fv_eye_unit<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;3<br>&nbsp;&nbsp;, 3.6<br>&nbsp;&nbsp;);                                                  <br><br>return:<br>&nbsp;&nbsp;array[[3.6, 0, 0],[0, 3.6, 0],[0, 0, 3.6]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_eye_unit<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[[1.0, 0, 0],[0, 1.0, 0],[0, 0, 1.0]]
>>> sm_sc.fv_augmented            |Widening or chunking of high-width array filling values.<br>Input param:<br>&nbsp;&nbsp;1. Original height width array; <br>&nbsp;&nbsp;2. The coordinates of the upper left corner of the augmented or cut block in the original matrix; <br>&nbsp;&nbsp;3. Enlarged or cut height and width specifications; <br>&nbsp;&nbsp;4. Increase the value of background elements in the area; |select<br>&nbsp;&nbsp;sm_sc.fv_augmented<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4, 5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[6, 7, 8, 9, 10]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[11, 12, 13, 14, 15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[16, 17, 18, 19, 20]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[21, 22, 23, 24, 25]]<br>&nbsp;&nbsp;, array[2, -1]<br>&nbsp;&nbsp;, array[8, 4]<br>&nbsp;&nbsp;, 0<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[[0, 0, 6, 7, 8, 9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 11, 12, 13, 14]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 16, 17, 18, 19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 21, 22, 23, 24]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0, 0, 0]]
>>> sm_sc.fv_array_1d_set_ele     |Assigning values to one-dimensional array positioning,<br>supports specifying negative or zero position numbers.<br>Input param:<br>&nbsp;&nbsp;1. Original one-dimensional array; <br>&nbsp;&nbsp;2. A position number or range of position numbers; <br>&nbsp;&nbsp;3. Assign or assign to a given array;  |select<br>&nbsp;&nbsp;sm_sc.fv_array_1d_set_ele<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3, 4]<br>&nbsp;&nbsp;, -2<br>&nbsp;&nbsp;, 5<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[1, 2, 5, 4]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_array_1d_set_ele<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3, 4]<br>&nbsp;&nbsp;, 4<br>&nbsp;&nbsp;, array[5, 6, 7]<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[1, 2, 3, 5, 6, 7]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_array_1d_set_ele<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3, 4]<br>&nbsp;&nbsp;, int4range(1, 2, '[]')<br>&nbsp;&nbsp;, array[5, 6, 7]<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[5, 6, 7, 3, 4]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_array_1d_set_ele<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3, 4]<br>&nbsp;&nbsp;, -2<br>&nbsp;&nbsp;, 5<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[1, 2, 5, 4]
>>> sm_sc.fv_arr_remove_1d          |1d array remove elements with some values。<br>Input param:: <br>&nbsp;&nbsp;1. Original array; <br>&nbsp;&nbsp;2. elements values to be removed; |select<br>&nbsp;&nbsp;sm_sc.fv_arr_remove_1d<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;[1, 2, 3, 4, 5, 3, 4, 5, 6, 7]<br>&nbsp;&nbsp;, array[3, 5]<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[1, 2, 4, 4, 6, 7]
>>> sm_sc.fv_rand_1d_ele          |Random combination, with repeatable serial numbers.<br>Input param:<br>&nbsp;&nbsp;1. Overall quantity; <br>&nbsp;&nbsp;2. Sample quantity; |select<br>&nbsp;&nbsp;sm_sc.fv_rand_1d_ele<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;8<br>&nbsp;&nbsp;, 5<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[1, 1, 7, 3, 7]
>>> sm_sc.fv_rand_1d_ele_pick     |Random combination with no duplicate sequence numbers.<br>Input param:<br>&nbsp;&nbsp;1. Overall quantity; <br>&nbsp;&nbsp;2. Sample quantity; |select<br>&nbsp;&nbsp;sm_sc.fv_rand_1d_ele_pick<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;8<br>&nbsp;&nbsp;, 5<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[3, 6, 1, 7, 5]
>>> sm_sc.ft_rand_slice_y         |Randomly obtain slice rows in the y-direction of the array,<br>element number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>&nbsp;&nbsp;o_ord_nos, o_slices <br>from<br>&nbsp;&nbsp;sm_sc.ft_rand_slice_y<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 2.0],[3.0, 4.0],[5.0, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.0, 8.0],[9.0, 10.0],[11.0, 12.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[6, 2, 6]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[11.0, 12.0],[3.0, 4.0],[11.0, 12.0]]
>>> sm_sc.ft_rand_slice_y_pick    |Randomly pick slice rows in the y-direction of the array,<br>element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>&nbsp;&nbsp;o_ord_nos, o_slices <br>from<br>&nbsp;&nbsp;sm_sc.ft_rand_slice_y_pick<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 2.0],[3.0, 4.0],[5.0, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.0, 8.0],[9.0, 10.0],[11.0, 12.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[6, 2, 6]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[5.0, 6.0],[3.0, 4.0],[11.0, 12.0]]
>>> sm_sc.ft_rand_slice_x         |Randomly obtain slice columns in the x-direction of the array,<br>element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>&nbsp;&nbsp;o_ord_nos, o_slices <br>from<br>&nbsp;&nbsp;sm_sc.ft_rand_slice_x<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.0, 8.0, 9.0, 10.0, 11.0, 12.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 4, 1]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 4.0, 1.0],[7.0, 10.0, 7.0]]
>>> sm_sc.ft_rand_slice_x_pick    |Randomly pick slice columns in the x-direction of the array,<br>element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>&nbsp;&nbsp;o_ord_nos, o_slices <br>from<br>&nbsp;&nbsp;sm_sc.ft_rand_slice_x_pick<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.0, 8.0, 9.0, 10.0, 11.0, 12.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[4, 5, 3]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[4.0, 5.0, 3.0],[10.0, 11.0, 9.0]]
>>> sm_sc.ft_rand_slice_x3        |Randomly obtain slice rows in the x3 direction of the array,<br>element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select&nbsp;<br>&nbsp;&nbsp;o_ord_nos,<br>&nbsp;&nbsp;o_slices<br>from&nbsp;sm_sc.ft_rand_slice_x3<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[11,&nbsp;21,&nbsp;31]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[12,&nbsp;22,&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[13,&nbsp;23,&nbsp;33]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[14,&nbsp;24,&nbsp;34]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[15,&nbsp;25,&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[16,&nbsp;26,&nbsp;36]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[17,&nbsp;27,&nbsp;37]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[18,&nbsp;28,&nbsp;38]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[19,&nbsp;29,&nbsp;39]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[61,&nbsp;62,&nbsp;63]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[81,&nbsp;82,&nbsp;83]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;2<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;1]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[2,&nbsp;1],[21,&nbsp;11],[22,&nbsp;12],[23,&nbsp;13],[24,&nbsp;14],[25,&nbsp;15]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[26,&nbsp;16],[27,&nbsp;17],[28,&nbsp;18],[29,&nbsp;19],[62,&nbsp;61],[82,&nbsp;81]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.ft_rand_slice_x3_pick   |Randomly pick slice rows in the x3 direction of the array,<br>element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select&nbsp;<br>&nbsp;&nbsp;o_ord_nos,<br>&nbsp;&nbsp;o_slices<br>from&nbsp;sm_sc.ft_rand_slice_x3_pick<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],&nbsp;[2,&nbsp;-2],&nbsp;[3,&nbsp;-3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[11,&nbsp;-11],&nbsp;[21,&nbsp;-21],&nbsp;[31,&nbsp;-31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[12,&nbsp;-12],&nbsp;[22,&nbsp;-22],&nbsp;[32,&nbsp;-32]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[13,&nbsp;-13],&nbsp;[23,&nbsp;-23],&nbsp;[33,&nbsp;-33]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[14,&nbsp;-14],&nbsp;[24,&nbsp;-24],&nbsp;[34,&nbsp;-34]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[15,&nbsp;-15],&nbsp;[25,&nbsp;-25],&nbsp;[35,&nbsp;-35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[16,&nbsp;-16],&nbsp;[26,&nbsp;-26],&nbsp;[36,&nbsp;-36]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[27,&nbsp;-27],&nbsp;[27,&nbsp;-27],&nbsp;[37,&nbsp;-37]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[18,&nbsp;-18],&nbsp;[28,&nbsp;-28],&nbsp;[38,&nbsp;-38]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[19,&nbsp;-19],&nbsp;[29,&nbsp;-29],&nbsp;[39,&nbsp;-39]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[61,&nbsp;-61],&nbsp;[62,&nbsp;-62],&nbsp;[63,&nbsp;-63]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[81,&nbsp;-81],&nbsp;[82,&nbsp;-82],&nbsp;[83,&nbsp;-83]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;2<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;3]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],&nbsp;[3,&nbsp;-3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[11,&nbsp;-11],&nbsp;[31,&nbsp;-31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[12,&nbsp;-12],&nbsp;[32,&nbsp;-32]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[13,&nbsp;-13],&nbsp;[33,&nbsp;-33]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[14,&nbsp;-14],&nbsp;[34,&nbsp;-34]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[15,&nbsp;-15],&nbsp;[35,&nbsp;-35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[16,&nbsp;-16],&nbsp;[36,&nbsp;-36]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[27,&nbsp;-27],&nbsp;[37,&nbsp;-37]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[18,&nbsp;-18],&nbsp;[38,&nbsp;-38]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[19,&nbsp;-19],&nbsp;[39,&nbsp;-39]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[61,&nbsp;-61],&nbsp;[63,&nbsp;-63]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[81,&nbsp;-81],&nbsp;[83,&nbsp;-83]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.ft_rand_slice_x4        |Randomly obtain slice columns in the x4 direction of the array,<br>element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;o_ord_nos,<br>&nbsp;&nbsp;&nbsp;&nbsp;o_slices<br>from&nbsp;&nbsp;sm_sc.ft_rand_slice_x4<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;-1],&nbsp;&nbsp;[2,&nbsp;&nbsp;-2],&nbsp;&nbsp;[3,&nbsp;&nbsp;-3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[11,&nbsp;&nbsp;-11],&nbsp;&nbsp;[21,&nbsp;&nbsp;-21],&nbsp;&nbsp;[31,&nbsp;&nbsp;-31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[12,&nbsp;&nbsp;-12],&nbsp;&nbsp;[22,&nbsp;&nbsp;-22],&nbsp;&nbsp;[32,&nbsp;&nbsp;-32]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[13,&nbsp;&nbsp;-13],&nbsp;&nbsp;[23,&nbsp;&nbsp;-23],&nbsp;&nbsp;[33,&nbsp;&nbsp;-33]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[14,&nbsp;&nbsp;-14],&nbsp;&nbsp;[24,&nbsp;&nbsp;-24],&nbsp;&nbsp;[34,&nbsp;&nbsp;-34]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[15,&nbsp;&nbsp;-15],&nbsp;&nbsp;[25,&nbsp;&nbsp;-25],&nbsp;&nbsp;[35,&nbsp;&nbsp;-35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[16,&nbsp;&nbsp;-16],&nbsp;&nbsp;[26,&nbsp;&nbsp;-26],&nbsp;&nbsp;[36,&nbsp;&nbsp;-36]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[17,&nbsp;&nbsp;-17],&nbsp;&nbsp;[27,&nbsp;&nbsp;-27],&nbsp;&nbsp;[37,&nbsp;&nbsp;-37]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[18,&nbsp;&nbsp;-18],&nbsp;&nbsp;[28,&nbsp;&nbsp;-28],&nbsp;&nbsp;[38,&nbsp;&nbsp;-38]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[19,&nbsp;&nbsp;-19],&nbsp;&nbsp;[29,&nbsp;&nbsp;-29],&nbsp;&nbsp;[39,&nbsp;&nbsp;-39]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[61,&nbsp;&nbsp;-61],&nbsp;&nbsp;[62,&nbsp;&nbsp;-62],&nbsp;&nbsp;[63,&nbsp;&nbsp;-63]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[81,&nbsp;&nbsp;-81],&nbsp;&nbsp;[82,&nbsp;&nbsp;-82],&nbsp;&nbsp;[83,&nbsp;&nbsp;-83]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[-1,&nbsp;&nbsp;-1],&nbsp;&nbsp;[-2,&nbsp;&nbsp;-2],&nbsp;&nbsp;[-3,&nbsp;&nbsp;-3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-11,&nbsp;&nbsp;-11],&nbsp;&nbsp;[-21,&nbsp;&nbsp;-21],&nbsp;&nbsp;[-31,&nbsp;&nbsp;-31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-12,&nbsp;&nbsp;-12],&nbsp;&nbsp;[-22,&nbsp;&nbsp;-22],&nbsp;&nbsp;[-32,&nbsp;&nbsp;-32]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-13,&nbsp;&nbsp;-13],&nbsp;&nbsp;[-23,&nbsp;&nbsp;-23],&nbsp;&nbsp;[-33,&nbsp;&nbsp;-33]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-14,&nbsp;&nbsp;-14],&nbsp;&nbsp;[-24,&nbsp;&nbsp;-24],&nbsp;&nbsp;[-34,&nbsp;&nbsp;-34]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-15,&nbsp;&nbsp;-15],&nbsp;&nbsp;[-25,&nbsp;&nbsp;-25],&nbsp;&nbsp;[-35,&nbsp;&nbsp;-35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[-16,&nbsp;&nbsp;-16],&nbsp;&nbsp;[-26,&nbsp;&nbsp;-26],&nbsp;&nbsp;[-36,&nbsp;&nbsp;-36]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-17,&nbsp;&nbsp;-17],&nbsp;&nbsp;[-27,&nbsp;&nbsp;-27],&nbsp;&nbsp;[-37,&nbsp;&nbsp;-37]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-18,&nbsp;&nbsp;-18],&nbsp;&nbsp;[-28,&nbsp;&nbsp;-28],&nbsp;&nbsp;[-38,&nbsp;&nbsp;-38]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-19,&nbsp;&nbsp;-19],&nbsp;&nbsp;[-29,&nbsp;&nbsp;-29],&nbsp;&nbsp;[-39,&nbsp;&nbsp;-39]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-61,&nbsp;&nbsp;-61],&nbsp;&nbsp;[-62,&nbsp;&nbsp;-62],&nbsp;&nbsp;[-63,&nbsp;&nbsp;-63]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-81,&nbsp;&nbsp;-81],&nbsp;&nbsp;[-82,&nbsp;&nbsp;-82],&nbsp;&nbsp;[-83,&nbsp;&nbsp;-83]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.ft_rand_slice_x4_pick   |Randomly pick slice columns in the x4 direction of the array,<br>element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;o_ord_nos,<br>&nbsp;&nbsp;&nbsp;&nbsp;o_slices<br>from&nbsp;&nbsp;sm_sc.ft_rand_slice_x4_pick<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;-1],&nbsp;&nbsp;[2,&nbsp;&nbsp;-2],&nbsp;&nbsp;[3,&nbsp;&nbsp;-3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[11,&nbsp;&nbsp;-11],&nbsp;&nbsp;[21,&nbsp;&nbsp;-21],&nbsp;&nbsp;[31,&nbsp;&nbsp;-31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[12,&nbsp;&nbsp;-12],&nbsp;&nbsp;[22,&nbsp;&nbsp;-22],&nbsp;&nbsp;[32,&nbsp;&nbsp;-32]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[13,&nbsp;&nbsp;-13],&nbsp;&nbsp;[23,&nbsp;&nbsp;-23],&nbsp;&nbsp;[33,&nbsp;&nbsp;-33]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[14,&nbsp;&nbsp;-14],&nbsp;&nbsp;[24,&nbsp;&nbsp;-24],&nbsp;&nbsp;[34,&nbsp;&nbsp;-34]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[15,&nbsp;&nbsp;-15],&nbsp;&nbsp;[25,&nbsp;&nbsp;-25],&nbsp;&nbsp;[35,&nbsp;&nbsp;-35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[16,&nbsp;&nbsp;-16],&nbsp;&nbsp;[26,&nbsp;&nbsp;-26],&nbsp;&nbsp;[36,&nbsp;&nbsp;-36]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[27,&nbsp;&nbsp;-27],&nbsp;&nbsp;[27,&nbsp;&nbsp;-27],&nbsp;&nbsp;[37,&nbsp;&nbsp;-37]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[18,&nbsp;&nbsp;-18],&nbsp;&nbsp;[28,&nbsp;&nbsp;-28],&nbsp;&nbsp;[38,&nbsp;&nbsp;-38]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[19,&nbsp;&nbsp;-19],&nbsp;&nbsp;[29,&nbsp;&nbsp;-29],&nbsp;&nbsp;[39,&nbsp;&nbsp;-39]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[61,&nbsp;&nbsp;-61],&nbsp;&nbsp;[62,&nbsp;&nbsp;-62],&nbsp;&nbsp;[63,&nbsp;&nbsp;-63]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[81,&nbsp;&nbsp;-81],&nbsp;&nbsp;[82,&nbsp;&nbsp;-82],&nbsp;&nbsp;[83,&nbsp;&nbsp;-83]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[-1&nbsp;&nbsp;,&nbsp;&nbsp;1&nbsp;&nbsp;],&nbsp;&nbsp;[-2&nbsp;&nbsp;,&nbsp;&nbsp;2&nbsp;&nbsp;],&nbsp;&nbsp;[-3&nbsp;&nbsp;,&nbsp;&nbsp;3&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-11,&nbsp;&nbsp;11],&nbsp;&nbsp;[-21,&nbsp;&nbsp;21],&nbsp;&nbsp;[-31,&nbsp;&nbsp;31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-12,&nbsp;&nbsp;12],&nbsp;&nbsp;[-22,&nbsp;&nbsp;22],&nbsp;&nbsp;[-32,&nbsp;&nbsp;32]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-13,&nbsp;&nbsp;13],&nbsp;&nbsp;[-23,&nbsp;&nbsp;23],&nbsp;&nbsp;[-33,&nbsp;&nbsp;33]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-14,&nbsp;&nbsp;14],&nbsp;&nbsp;[-24,&nbsp;&nbsp;24],&nbsp;&nbsp;[-34,&nbsp;&nbsp;34]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-15,&nbsp;&nbsp;15],&nbsp;&nbsp;[-25,&nbsp;&nbsp;25],&nbsp;&nbsp;[-35,&nbsp;&nbsp;35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[-16,&nbsp;&nbsp;16],&nbsp;&nbsp;[-26,&nbsp;&nbsp;26],&nbsp;&nbsp;[-36,&nbsp;&nbsp;36]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-27,&nbsp;&nbsp;27],&nbsp;&nbsp;[-27,&nbsp;&nbsp;27],&nbsp;&nbsp;[-37,&nbsp;&nbsp;37]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-18,&nbsp;&nbsp;18],&nbsp;&nbsp;[-28,&nbsp;&nbsp;28],&nbsp;&nbsp;[-38,&nbsp;&nbsp;38]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-19,&nbsp;&nbsp;19],&nbsp;&nbsp;[-29,&nbsp;&nbsp;29],&nbsp;&nbsp;[-39,&nbsp;&nbsp;39]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-61,&nbsp;&nbsp;61],&nbsp;&nbsp;[-62,&nbsp;&nbsp;62],&nbsp;&nbsp;[-63,&nbsp;&nbsp;63]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[-81,&nbsp;&nbsp;81],&nbsp;&nbsp;[-82,&nbsp;&nbsp;82],&nbsp;&nbsp;[-83,&nbsp;&nbsp;83]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_rand_slice_y         |Randomly obtain slice rows in the y-direction of the array,<br>element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity;   |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rand_slice_y<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[11,&nbsp;&nbsp;21,&nbsp;&nbsp;31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[16,&nbsp;&nbsp;26,&nbsp;&nbsp;36],&nbsp;&nbsp;[17,&nbsp;&nbsp;27,&nbsp;&nbsp;37]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;5<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[11,&nbsp;&nbsp;21,&nbsp;&nbsp;31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[11,&nbsp;&nbsp;21,&nbsp;&nbsp;31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[16,&nbsp;&nbsp;26,&nbsp;&nbsp;36],[17,&nbsp;&nbsp;27,&nbsp;&nbsp;37]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[11,&nbsp;&nbsp;21,&nbsp;&nbsp;31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[11,&nbsp;&nbsp;21,&nbsp;&nbsp;31]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_rand_slice_y_pick    |Randomly pick slice rows in the y-direction of the array,<br>element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity;   |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rand_slice_y_pick<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;21,&nbsp;&nbsp;31]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[16,&nbsp;&nbsp;26,&nbsp;&nbsp;36]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[17,&nbsp;&nbsp;27,&nbsp;&nbsp;37]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[16,&nbsp;&nbsp;26,&nbsp;&nbsp;36]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;21,&nbsp;&nbsp;31]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[17,&nbsp;&nbsp;27,&nbsp;&nbsp;37]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_rand_slice_x         |Randomly obtain slice columns in the x-direction of the array,<br>element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity;   |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rand_slice_x<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[7.0,&nbsp;&nbsp;8.0,&nbsp;&nbsp;9.0,&nbsp;&nbsp;10.0,&nbsp;&nbsp;11.0,&nbsp;&nbsp;12.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;4.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,&nbsp;&nbsp;11.0,&nbsp;&nbsp;10.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_rand_slice_x_pick    |Randomly pick slice columns in the x-direction of the array,<br>element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity;   |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rand_slice_x_pick<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[7.0,&nbsp;&nbsp;8.0,&nbsp;&nbsp;9.0,&nbsp;&nbsp;10.0,&nbsp;&nbsp;11.0,&nbsp;&nbsp;12.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;4<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[5.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11.0,&nbsp;&nbsp;9.0,&nbsp;&nbsp;8.0,&nbsp;&nbsp;7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_rand_slice_x3        |Randomly obtain slice rows in the x3 direction of the array,<br>element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity;  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rand_slice_x3<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[7.0,&nbsp;&nbsp;8.0,&nbsp;&nbsp;9.0,&nbsp;&nbsp;10.0,&nbsp;&nbsp;11.0,&nbsp;&nbsp;12.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;4<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[6.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;1.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[12.0,&nbsp;&nbsp;11.0,&nbsp;&nbsp;7.0,&nbsp;&nbsp;12.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]]
>>> sm_sc.fv_rand_slice_x3_pick   |Randomly pick slice rows in the x3 direction of the array,<br>element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity;  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rand_slice_x3_pick<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[7.0,&nbsp;&nbsp;8.0,&nbsp;&nbsp;9.0,&nbsp;&nbsp;10.0,&nbsp;&nbsp;11.0,&nbsp;&nbsp;12.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;4<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[5.0,&nbsp;&nbsp;6.0,&nbsp;&nbsp;1.0,&nbsp;&nbsp;4.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11.0,&nbsp;&nbsp;12.0,&nbsp;&nbsp;7.0,&nbsp;&nbsp;10.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]]
>>> sm_sc.fv_rand_slice_x4        |Randomly obtain slice columns in the x4 direction of the array,<br>element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity;  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rand_slice_x4<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[7.0,&nbsp;&nbsp;8.0,&nbsp;&nbsp;9.0,&nbsp;&nbsp;10.0,&nbsp;&nbsp;11.0,&nbsp;&nbsp;12.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;4<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[6.0,&nbsp;&nbsp;6.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;1.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[12.0,&nbsp;&nbsp;12.0,&nbsp;&nbsp;9.0,&nbsp;&nbsp;7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]]]
>>> sm_sc.fv_rand_slice_x4_pick   |Randomly pick slice columns in the x4 direction of the array,<br>element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity;  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rand_slice_x4_pick<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[7.0,&nbsp;&nbsp;8.0,&nbsp;&nbsp;9.0,&nbsp;&nbsp;10.0,&nbsp;&nbsp;11.0,&nbsp;&nbsp;12.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;4<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3.0,&nbsp;&nbsp;1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;4.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[9.0,&nbsp;&nbsp;7.0,&nbsp;&nbsp;8.0,&nbsp;&nbsp;10.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]]]
>>> sm_sc.fv_chunk              |Multi-dimension slicing (chunking).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The upper and lower bounds of each dimension range for the slice position index;|select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_chunk<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1,2,3,4],[5,6,7,8]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[2,1],[3,2]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[5,6]]
>>> sm_sc.fv_slice_y              |The first dimension slicing (chunking).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Slice position subscript range set; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_slice_y<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1,&nbsp;2],[3,&nbsp;4],[5,&nbsp;6],[7,&nbsp;8]]<br>&nbsp;&nbsp;,&nbsp;array[int4range(1,&nbsp;3,&nbsp;'[]'),&nbsp;int4range(2,&nbsp;3,&nbsp;'[]')]<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1,&nbsp;2],[3,&nbsp;4],[5,&nbsp;6],[3,&nbsp;4],[5,&nbsp;6]]
>>> sm_sc.fv_slice_x              |The second dimension slicing (chunking).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Slice position subscript range set; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_slice_x<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1,&nbsp;2,&nbsp;3,&nbsp;4],[5,&nbsp;6,&nbsp;7,&nbsp;8]]<br>&nbsp;&nbsp;,&nbsp;array[int4range(1,&nbsp;3,&nbsp;'[]'),&nbsp;int4range(2,&nbsp;3,&nbsp;'[]')]<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;2,&nbsp;3],[5,&nbsp;6,&nbsp;7,&nbsp;6,&nbsp;7]]
>>> sm_sc.fv_slice_x3             |The third dimension slicing (chunking).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Slice position subscript range set; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_slice_x3<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[11,&nbsp;12,&nbsp;13,&nbsp;14],[21,&nbsp;22,&nbsp;23,&nbsp;24]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[31,&nbsp;32,&nbsp;33,&nbsp;34],[41,&nbsp;42,&nbsp;43,&nbsp;44]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;array[int4range(1,&nbsp;3,&nbsp;'[]'),&nbsp;int4range(2,&nbsp;3,&nbsp;'[]')]<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[[11,&nbsp;12,&nbsp;13,&nbsp;12,&nbsp;13],[21,&nbsp;22,&nbsp;23,&nbsp;22,&nbsp;23]]<br>&nbsp;&nbsp;,&nbsp;[[31,&nbsp;32,&nbsp;33,&nbsp;32,&nbsp;33],[41,&nbsp;42,&nbsp;43,&nbsp;42,&nbsp;43]]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_slice_x4             |The fourth dimension slicing (chunking).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Slice position subscript range set; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_slice_x4<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[11,&nbsp;12,&nbsp;13,&nbsp;14],[21,&nbsp;22,&nbsp;23,&nbsp;24]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[31,&nbsp;32,&nbsp;33,&nbsp;34],[41,&nbsp;42,&nbsp;43,&nbsp;44]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[51,&nbsp;52,&nbsp;53,&nbsp;54],[61,&nbsp;62,&nbsp;63,&nbsp;64]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[71,&nbsp;72,&nbsp;73,&nbsp;74],[81,&nbsp;82,&nbsp;83,&nbsp;84]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;array[int4range(1,&nbsp;3,&nbsp;'[]'),&nbsp;int4range(2,&nbsp;3,&nbsp;'[]')]<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[11,&nbsp;12,&nbsp;13,&nbsp;12,&nbsp;13],[21,&nbsp;22,&nbsp;23,&nbsp;22,&nbsp;23]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[31,&nbsp;32,&nbsp;33,&nbsp;32,&nbsp;33],[41,&nbsp;42,&nbsp;43,&nbsp;42,&nbsp;43]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[51,&nbsp;52,&nbsp;53,&nbsp;52,&nbsp;53],[61,&nbsp;62,&nbsp;63,&nbsp;62,&nbsp;63]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[71,&nbsp;72,&nbsp;73,&nbsp;72,&nbsp;73],[81,&nbsp;82,&nbsp;83,&nbsp;82,&nbsp;83]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_sample_y              |The first dimension interval sampling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sampling period; <br>&nbsp;&nbsp;3. Sampling window constraint.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Null value indicates no constraints; <br>&nbsp;&nbsp;4. Sample range.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Queue supports multiple ranges; <br>&nbsp;&nbsp;5. Fill in the value matrix with upsampling. Default to null; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Fill the sample range that exceeds the window constraint in the value matrix;|select<br>&nbsp;&nbsp;sm_sc.fv_sample_y<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2],[3, 4],[5, 6],[7, 8],[9, 10],[11, 12]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;, array[int4range(1,2, '[]'), int4range(2,4, '[]')]<br>&nbsp;&nbsp;, array[[13, 14], [15, 16]]<br>&nbsp;&nbsp;);                        <br><br>return:<br>&nbsp;&nbsp;array[[1, 2],[3, 4],[3, 4],[5, 6],[13, 14],[15, 16]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7, 8],[9, 10],[9, 10],[11, 12],[13, 14],[15, 16]]
>>> sm_sc.fv_sample_x              |The second dimension interval sampling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sampling period; <br>&nbsp;&nbsp;3. Sampling window constraint.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Null value indicates no constraints; <br>&nbsp;&nbsp;4. Sample range.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Queue supports multiple ranges; <br>&nbsp;&nbsp;5. Fill in the value matrix with upsampling. Default to null; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Fill the sample range that exceeds the window constraint in the value matrix; |select<br>&nbsp;&nbsp;sm_sc.fv_sample_x<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4, 5, 6],[7, 8, 9, 10, 11, 12]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;, array[int4range(1,2, '[]'), int4range(2,4, '[]')]<br>&nbsp;&nbsp;, array[[13, 14], [15, 16]]<br>&nbsp;&nbsp;);                        <br><br>return:<br>&nbsp;&nbsp;array[[1, 2, 2, 13, 14, 13, 14, 4, 5, 5, 13, 14, 13, 14]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7, 8, 8, 15, 16, 15, 16, 10, 11, 11, 15, 16, 15, 16]]
>>> sm_sc.fv_sample_x3            |The third dimension interval sampling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sampling period; <br>&nbsp;&nbsp;3. Sampling window constraint.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Null value indicates no constraints; <br>&nbsp;&nbsp;4. Sample range.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Queue supports multiple ranges; <br>&nbsp;&nbsp;5. Fill in the value matrix with upsampling. Default to null; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Fill the sample range that exceeds the window constraint in the value matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_sample_x3<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],[2,&nbsp;-2]],[[3,&nbsp;-3],[4,&nbsp;-4]],[[5,&nbsp;-5],[6,&nbsp;-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[7,&nbsp;-7],[8,&nbsp;-8]],[[9.7,&nbsp;-9],[10,&nbsp;-10]],[[11,&nbsp;-11],[12,&nbsp;-12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],[2,&nbsp;-2]],[[3,&nbsp;-3],[4,&nbsp;-4]],[[5,&nbsp;-5],[6,&nbsp;-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[7,&nbsp;-7],[8,&nbsp;-8]],[[9.7,&nbsp;-9],[10,&nbsp;-10]],[[11,&nbsp;-11],[12,&nbsp;-12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],[2,&nbsp;-2]],[[3,&nbsp;-3],[4,&nbsp;-4]],[[5,&nbsp;-5],[6,&nbsp;-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[7,&nbsp;-7],[8,&nbsp;-8]],[[9.7,&nbsp;-9],[10,&nbsp;-10]],[[11,&nbsp;-11],[12,&nbsp;-12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],[2,&nbsp;-2]],[[3,&nbsp;-3],[4,&nbsp;-4]],[[5,&nbsp;-5],[6,&nbsp;-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[7,&nbsp;-7],[8,&nbsp;-8]],[[9.7,&nbsp;-9],[10,&nbsp;-10]],[[11,&nbsp;-11],[12,&nbsp;-12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]::float[]<br>&nbsp;&nbsp;,&nbsp;3<br>&nbsp;&nbsp;,&nbsp;3<br>&nbsp;&nbsp;,&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(-1,&nbsp;2,&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;int4range(-2,&nbsp;0,&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;int4range(3,&nbsp;5,&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;int4range(4,&nbsp;5,&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[32,&nbsp;-42]],[[54,&nbsp;-64]],[[75,&nbsp;-85]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[97,&nbsp;-97]],[[19.7,&nbsp;-19]],[[110,&nbsp;-116]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[32,&nbsp;-42]],[[54,&nbsp;-64]],[[75,&nbsp;-85]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[97,&nbsp;-97]],[[19.7,&nbsp;-19]],[[110,&nbsp;-116]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[32,&nbsp;-42]],[[54,&nbsp;-64]],[[75,&nbsp;-85]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[97,&nbsp;-97]],[[19.7,&nbsp;-19]],[[110,&nbsp;-116]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[32,&nbsp;-42]],[[54,&nbsp;-64]],[[75,&nbsp;-85]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[97,&nbsp;-97]],[[19.7,&nbsp;-19]],[[110,&nbsp;-116]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]::float[]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[32,-42],[32,-42],[1,-1],[2,-2],[32,-42],[32,-42]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[32,-42],[32,-42],[32,-42],[32,-42],[32,-42]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[54,-64],[54,-64],[3,-3],[4,-4],[54,-64],[54,-64]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[54,-64],[54,-64],[54,-64],[54,-64],[54,-64]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[75,-85],[75,-85],[5,-5],[6,-6],[75,-85],[75,-85]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[75,-85],[75,-85],[75,-85],[75,-85],[75,-85]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[97,-97],[97,-97],[7,-7],[8,-8],[97,-97],[97,-97]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[97,-97],[97,-97],[97,-97],[97,-97],[97,-97]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[19.7,-19],[19.7,-19],[9.7,-9],[10,-10],[19.7,-19],[19.7,-19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[110,-116],[110,-116],[11,-11],[12,-12],[110,-116],[110,-116]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[110,-116],[110,-116],[110,-116],[110,-116],[110,-116]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[32,-42],[32,-42],[1,-1],[2,-2],[32,-42],[32,-42]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[32,-42],[32,-42],[32,-42],[32,-42],[32,-42]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[54,-64],[54,-64],[3,-3],[4,-4],[54,-64],[54,-64]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[54,-64],[54,-64],[54,-64],[54,-64],[54,-64]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[75,-85],[75,-85],[5,-5],[6,-6],[75,-85],[75,-85]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[75,-85],[75,-85],[75,-85],[75,-85],[75,-85]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[97,-97],[97,-97],[7,-7],[8,-8],[97,-97],[97,-97]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[97,-97],[97,-97],[97,-97],[97,-97],[97,-97]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[19.7,-19],[19.7,-19],[9.7,-9],[10,-10],[19.7,-19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[110,-116],[110,-116],[11,-11],[12,-12],[110,-116],[110,-116]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[110,-116],[110,-116],[110,-116],[110,-116],[110,-116]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[32,-42],[32,-42],[1,-1],[2,-2],[32,-42],[32,-42]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[32,-42],[32,-42],[32,-42],[32,-42],[32,-42]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[54,-64],[54,-64],[3,-3],[4,-4],[54,-64],[54,-64]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[54,-64],[54,-64],[54,-64],[54,-64],[54,-64]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[75,-85],[75,-85],[5,-5],[6,-6],[75,-85],[75,-85]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[75,-85],[75,-85],[75,-85],[75,-85],[75,-85]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[97,-97],[97,-97],[7,-7],[8,-8],[97,-97],[97,-97]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[97,-97],[97,-97],[97,-97],[97,-97],[97,-97]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[19.7,-19],[19.7,-19],[9.7,-9],[10,-10],[19.7,-19],[19.7,-19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[110,-116],[110,-116],[11,-11],[12,-12],[110,-116],[110,-116]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[110,-116],[110,-116],[110,-116],[110,-116],[110,-116]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[32,-42],[32,-42],[1,-1],[2,-2],[32,-42],[32,-42]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[32,-42],[32,-42],[32,-42],[32,-42],[32,-42]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[54,-64],[54,-64],[3,-3],[4,-4],[54,-64],[54,-64]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[54,-64],[54,-64],[54,-64],[54,-64],[54,-64]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[75,-85],[75,-85],[5,-5],[6,-6],[75,-85],[75,-85]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[75,-85],[75,-85],[75,-85],[75,-85],[75,-85]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[97,-97],[97,-97],[7,-7],[8,-8],[97,-97],[97,-97]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[97,-97],[97,-97],[97,-97],[97,-97],[97,-97]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[19.7,-19],[19.7,-19],[9.7,-9],[10,-10],[19.7,-19],[19.7,-19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19],[19.7,-19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[110,-116],[110,-116],[11,-11],[12,-12],[110,-116],[110,-116]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[110,-116],[110,-116],[110,-116],[110,-116],[110,-116]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_sample_x4            |The fourth dimension interval sampling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sampling period; <br>&nbsp;&nbsp;3. Sampling window constraint.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Null value indicates no constraints; <br>&nbsp;&nbsp;4. Sample range.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Queue supports multiple ranges; <br>&nbsp;&nbsp;5. Fill in the value matrix with upsampling. Default to null; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Fill the sample range that exceeds the window constraint in the value matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_sample_x4<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],[2,&nbsp;-2]],[[3,&nbsp;-3],[4,&nbsp;-4]],[[5,&nbsp;-5],[6,&nbsp;-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[7,&nbsp;-7],[8,&nbsp;-8]],[[9.7,&nbsp;-9],[10,&nbsp;-10]],[[11,&nbsp;-11],[12,&nbsp;-12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],[2,&nbsp;-2]],[[3,&nbsp;-3],[4,&nbsp;-4]],[[5,&nbsp;-5],[6,&nbsp;-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[7,&nbsp;-7],[8,&nbsp;-8]],[[9.7,&nbsp;-9],[10,&nbsp;-10]],[[11,&nbsp;-11],[12,&nbsp;-12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],[2,&nbsp;-2]],[[3,&nbsp;-3],[4,&nbsp;-4]],[[5,&nbsp;-5],[6,&nbsp;-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[7,&nbsp;-7],[8,&nbsp;-8]],[[9.7,&nbsp;-9],[10,&nbsp;-10]],[[11,&nbsp;-11],[12,&nbsp;-12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;-1],[2,&nbsp;-2]],[[3,&nbsp;-3],[4,&nbsp;-4]],[[5,&nbsp;-5],[6,&nbsp;-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[7,&nbsp;-7],[8,&nbsp;-8]],[[9.7,&nbsp;-9],[10,&nbsp;-10]],[[11,&nbsp;-11],[12,&nbsp;-12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]::float[]<br>&nbsp;&nbsp;,&nbsp;3<br>&nbsp;&nbsp;,&nbsp;3<br>&nbsp;&nbsp;,&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(-1,&nbsp;2,&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;int4range(-2,&nbsp;0,&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;int4range(3,&nbsp;5,&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;int4range(4,&nbsp;5,&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1],&nbsp;[2]],&nbsp;[[3],&nbsp;[-4]],&nbsp;[[-5],&nbsp;[-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[-7],&nbsp;[-8]],&nbsp;[[9.7],&nbsp;[10]],&nbsp;[[-11],&nbsp;[12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1],&nbsp;[2]],&nbsp;[[3],&nbsp;[-4]],&nbsp;[[-5],&nbsp;[-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[-7],&nbsp;[-8]],&nbsp;[[9.7],&nbsp;[10]],&nbsp;[[-11],&nbsp;[12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1],&nbsp;[2]],&nbsp;[[3],&nbsp;[-4]],&nbsp;[[-5],&nbsp;[-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[-7],&nbsp;[-8]],&nbsp;[[9.7],&nbsp;[10]],&nbsp;[[-11],&nbsp;[12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1],&nbsp;[2]],&nbsp;[[3],&nbsp;[-4]],&nbsp;[[-5],&nbsp;[-6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[[-7],&nbsp;[-8]],&nbsp;[[9.7],&nbsp;[10]],&nbsp;[[-11],&nbsp;[12]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]::float[]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,1,1,-1,1,1,1,1,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,2,2,-2,2,2,2,2,2,2,2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3,3,3,-3,3,3,3,3,3,3,3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-4,-4,4,-4,-4,-4,-4,-4,-4,-4,-4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-5,-5,5,-5,-5,-5,-5,-5,-5,-5,-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-6,-6,6,-6,-6,-6,-6,-6,-6,-6,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-7,-7,7,-7,-7,-7,-7,-7,-7,-7,-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-8,-8,8,-8,-8,-8,-8,-8,-8,-8,-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[9.7,9.7,9.7,-9,9.7,9.7,9.7,9.7,9.7,9.7,9.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[10,10,10,-10,10,10,10,10,10,10,10]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-11,-11,11,-11,-11,-11,-11,-11,-11,-11,-11]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[12,12,12,-12,12,12,12,12,12,12,12]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,1,1,-1,1,1,1,1,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,2,2,-2,2,2,2,2,2,2,2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3,3,3,-3,3,3,3,3,3,3,3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-4,-4,4,-4,-4,-4,-4,-4,-4,-4,-4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-5,-5,5,-5,-5,-5,-5,-5,-5,-5,-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-6,-6,6,-6,-6,-6,-6,-6,-6,-6,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-7,-7,7,-7,-7,-7,-7,-7,-7,-7,-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-8,-8,8,-8,-8,-8,-8,-8,-8,-8,-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[9.7,9.7,9.7,-9,9.7,9.7,9.7,9.7,9.7,9.7,9.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[10,10,10,-10,10,10,10,10,10,10,10]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-11,-11,11,-11,-11,-11,-11,-11,-11,-11,-11]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[12,12,12,-12,12,12,12,12,12,12,12]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,1,1,-1,1,1,1,1,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,2,2,-2,2,2,2,2,2,2,2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3,3,3,-3,3,3,3,3,3,3,3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-4,-4,4,-4,-4,-4,-4,-4,-4,-4,-4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-5,-5,5,-5,-5,-5,-5,-5,-5,-5,-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-6,-6,6,-6,-6,-6,-6,-6,-6,-6,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-7,-7,7,-7,-7,-7,-7,-7,-7,-7,-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-8,-8,8,-8,-8,-8,-8,-8,-8,-8,-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[9.7,9.7,9.7,-9,9.7,9.7,9.7,9.7,9.7,9.7,9.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[10,10,10,-10,10,10,10,10,10,10,10]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-11,-11,11,-11,-11,-11,-11,-11,-11,-11,-11]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[12,12,12,-12,12,12,12,12,12,12,12]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,1,1,-1,1,1,1,1,1,1,1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,2,2,-2,2,2,2,2,2,2,2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3,3,3,-3,3,3,3,3,3,3,3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-4,-4,4,-4,-4,-4,-4,-4,-4,-4,-4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-5,-5,5,-5,-5,-5,-5,-5,-5,-5,-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-6,-6,6,-6,-6,-6,-6,-6,-6,-6,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-7,-7,7,-7,-7,-7,-7,-7,-7,-7,-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[-8,-8,8,-8,-8,-8,-8,-8,-8,-8,-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[9.7,9.7,9.7,-9,9.7,9.7,9.7,9.7,9.7,9.7,9.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[10,10,10,-10,10,10,10,10,10,10,10]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-11,-11,11,-11,-11,-11,-11,-11,-11,-11,-11]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[12,12,12,-12,12,12,12,12,12,12,12]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_lower_tri_mx         |Lower triangular matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Upper triangle filling value. The default value is null; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_lower_tri_mx<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,-2,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,2,-3,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,-2,3,4,NULL,NULL]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,-2,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,2,-3,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,-2,3,4,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,-2,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,2,-3,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,-2,3,4,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_upper_tri_mx         |Upper triangular matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Lower triangle filling value. The default value is null; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_upper_tri_mx<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,-3,4,5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,NULL,4,-5,6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,-3,4,5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,NULL,4,-5,6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,-3,4,5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,NULL,4,-5,6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_lower_tri_mx_ex         |Degenerate lower triangular matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Fill the upper triangle with diagonal lines. The default value is null; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_lower_tri_mx_ex<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[NULL,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,2,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,-2,3,NULL,NULL,NULL]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[NULL,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,2,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,-2,3,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[NULL,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,NULL,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,2,NULL,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,-2,3,NULL,NULL,NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_upper_tri_mx_ex         |Degenerate upper triangular matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The filling value of the lower triangle containing the diagonal line. The default value is null; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_upper_tri_mx_ex<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;-5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[null,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,null,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,null,4,5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,NULL,null,-5,6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[null,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,null,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,null,4,5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,NULL,null,-5,6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[null,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,null,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,null,4,5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[NULL,NULL,NULL,null,-5,6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_lmask                |Mask the left part of the high width 2D face with a given length at each height position.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The length of each height position mask; <br>&nbsp;&nbsp;3. Padding value; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_lmask<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7,&nbsp;&nbsp;-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17,&nbsp;&nbsp;18]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,0,0,4,5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,5,6,7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,5,6,7,8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,-2,-3,-4,-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0,-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,-6,-7,-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,0,0,14,15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,15,16,17]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,15,16,17,18]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_rmask                |Mask the right part of the high wide 2D surface with a given length at each height position.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The length of each height position mask; <br>&nbsp;&nbsp;3. Padding value; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_rmask<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7,&nbsp;&nbsp;-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17,&nbsp;&nbsp;18]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,4,5,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,5,6,7,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,-2,-3,-4,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2,-3,-4,-5,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-4,-5,-6,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,12,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[13,14,15,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[14,15,16,17,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_amask                |Mask the upper part of the high wide two-dimensional surface with a given length at each width position.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The length of each width position mask; <br>&nbsp;&nbsp;3. Padding value; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_amask<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7,&nbsp;&nbsp;-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17,&nbsp;&nbsp;18]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3,&nbsp;&nbsp;0,&nbsp;&nbsp;2,&nbsp;&nbsp;1,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;1,&nbsp;&nbsp;4,&nbsp;&nbsp;2,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2,&nbsp;&nbsp;1,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,2,0,0,5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,3,0,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,4,5,6,7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,5,6,7,8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2,-3,0,0,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3,-4,0,-6,-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-4,-5,0,-7,-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,15,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,14,15,16,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[14,15,16,17,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_bmask                |Mask the lower part of the high width two-dimensional plane with a given length at each width position.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The length of each width position mask; <br>&nbsp;&nbsp;3. Padding value; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_bmask<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2,&nbsp;&nbsp;-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3,&nbsp;&nbsp;-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-4,&nbsp;&nbsp;-5,&nbsp;&nbsp;-6,&nbsp;&nbsp;-7,&nbsp;&nbsp;-8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17,&nbsp;&nbsp;18]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3,&nbsp;&nbsp;0,&nbsp;&nbsp;2,&nbsp;&nbsp;1,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;1,&nbsp;&nbsp;4,&nbsp;&nbsp;2,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2,&nbsp;&nbsp;1,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,3,4,5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,4,0,6,7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,5,0,0,8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,-2,0,-4,-5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2,-3,0,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3,-4,0,0,-7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,12,13,14,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,13,14,15,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,16,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_ele_replace          |Replace the specified element value with another specified element value.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. One or more element values replaced; <br>&nbsp;&nbsp;3. New element value; |select<br>&nbsp;&nbsp;sm_sc.fv_ele_replace<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.3, 25.1],[3.25, 6.1]]<br>&nbsp;&nbsp;, array[12.3, 3.25]<br>&nbsp;&nbsp;, 1.0<br>&nbsp;&nbsp;);    <br><br>return:<br>&nbsp;&nbsp;array[[1.0, 25.1],[1.0, 6.1]]
>>> sm_sc.fv_area_replace         |Replace slices or chunks in the specified location area.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The replaced chunk is at the beginning of the metamatrix; <br>&nbsp;&nbsp;3. Replace chunk matrix; |select<br>&nbsp;&nbsp;sm_sc.fv_area_replace<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[['a', 'b', 'c'],['e', 'f', 'g'],['h', 'i', 'j'],['x', 'y', 'z']]<br>&nbsp;&nbsp;, array[3, 2]<br>&nbsp;&nbsp;, array[['m', 'n'],['p', 'q']]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[['a', 'b', 'c'],['e', 'f', 'g'],['h', 'm', 'n'],['x', 'p', 'q']]
>>> sm_sc.fv_pos_replaces         |Replace several positions with the specified element values.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. List of replaced positions; <br>&nbsp;&nbsp;3. New element value; |select<br>&nbsp;&nbsp;sm_sc.fv_pos_replaces<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[['a', 'b', 'c'],['e', 'f', 'g']]<br>&nbsp;&nbsp;, array[[1, 3],[2, 2]]<br>&nbsp;&nbsp;, 'd'<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[['a', 'b', 'd'],['e', 'd', 'g']]
>>> sm_sc.fv_regexp_matches       |REGEX matching of text matrix element values.<br>Input param:<br>&nbsp;&nbsp;1. Original text matrix; <br>&nbsp;&nbsp;2. REGEX; <br>&nbsp;&nbsp;3. REGEX option marker. Default to 'g'; |select<br>&nbsp;&nbsp;sm_sc.fv_regexp_matches<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[['abbbbbc122223', 'abc123adc'],['abc123', 'ac13']]<br>&nbsp;&nbsp;, array['a.c']<br>&nbsp;&nbsp;, 'gi'<br>&nbsp;&nbsp;);         <br><br>return:<br>&nbsp;&nbsp;array[[null, '"{{abc},{adc}}"'],['"{{abc}}"', null]]
>>> sm_sc.fv_regexp_replace       |REGEX replacement of text matrix element values.<br>Input param:<br>&nbsp;&nbsp;1. Original text matrix; <br>&nbsp;&nbsp;2. REGEX; <br>&nbsp;&nbsp;3. Replace new text; <br>&nbsp;&nbsp;4. REGEX option marker. Default to 'g'; |select<br>&nbsp;&nbsp;sm_sc.fv_regexp_replace<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[['abbbbbc122223', 'abc123adc'],['abc123', 'ac13']]<br>&nbsp;&nbsp;, array[['a.c'], ['1.*?3']]<br>&nbsp;&nbsp;, 'fff'<br>&nbsp;&nbsp;, 'gi'<br>&nbsp;&nbsp;);         <br><br>return:<br>&nbsp;&nbsp;array[['abbbbbc122223', 'fff123fff'],['abcfff', 'acfff']]
>>> sm_sc.fv_nullif               |Element avoidance value.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Avoidance sensitivity Matrix; |select<br>&nbsp;&nbsp;sm_sc.fv_nullif<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.3, -12.3],[45.6, -45.6]]<br>&nbsp;&nbsp;, array[[12.3],[-45.6]]<br>&nbsp;&nbsp;);                         <br><br>return:<br>&nbsp;&nbsp;array[[null,-12.3],[45.6,null]]
>>> sm_sc.fv_coalesce<br>sm_sc.fv_coalesce_variadic|Element null replace.<br>&nbsp;&nbsp;sm_sc.fv_coalesce input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Priority first non null element acquisition; <br>&nbsp;&nbsp;&nbsp;&nbsp;2. Second priority first non null element acquisition; <br>--------------------------<br>&nbsp;&nbsp;sm_sc.fv_coalesce_variadic input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1...n. Obtain the first non null element in sequence; |select<br>&nbsp;&nbsp;sm_sc.fv_coalesce<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[null, 2.4, 5.6, 52.3]<br>&nbsp;&nbsp;, array[1.2, 2.3, null, 52.1]<br>&nbsp;&nbsp;);            <br><br>return:<br>&nbsp;&nbsp;array[[],[]]<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_coalesce_variadic<br>&nbsp;&nbsp;(variadic<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[null, 2.4, 5.6, 52.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[null, 2.3, null, 52.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1.2, 2.8, null, null]]<br>&nbsp;&nbsp;);            <br><br>return:<br>&nbsp;&nbsp;array[{1.2, 2.4, 5.6, 52.3}]
>>> sm_sc.fv_lpad                 |Align right and fill left.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Padding value matrix; <br>&nbsp;&nbsp;3. Count of repetitions; |select<br>&nbsp;&nbsp;sm_sc.fv_lpad<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4, 5, 6],[-1, -2, -3, -4, -5, -6]]<br>&nbsp;&nbsp;, array[[7],[8]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[7, 7, 1, 2, 3, 4, 5, 6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[8, 8, -1, -2, -3, -4, -5, -6]]
>>> sm_sc.fv_rpad                 |Left aligned and right filled.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Padding value matrix; <br>&nbsp;&nbsp;3. Count of repetitions; |select<br>&nbsp;&nbsp;sm_sc.fv_rpad<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4, 5, 6],[-1, -2, -3, -4, -5, -6]]<br>&nbsp;&nbsp;, array[[7],[8]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[1, 2, 3, 4, 5, 6, 7, 7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-1, -2, -3, -4, -5, -6, 8, 8]]
>>> sm_sc.fv_apad                 |Align Bottom and Fill Top.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Padding value matrix; <br>&nbsp;&nbsp;3. Count of repetitions; |select<br>&nbsp;&nbsp;sm_sc.fv_apad<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4, 5, 6],[-1, -2, -3, -4, -5, -6]]<br>&nbsp;&nbsp;, array[[7],[8]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[7, 7, 7, 7, 7, 7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[8, 8, 8, 8, 8, 8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7, 7, 7, 7, 7, 7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[8, 8, 8, 8, 8, 8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 2, 3, 4, 5, 6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-1, -2, -3, -4, -5, -6]]
>>> sm_sc.fv_bpad                 |Align Top and Fill Bottom.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Padding value matrix; <br>&nbsp;&nbsp;3. Count of repetitions; |select<br>&nbsp;&nbsp;sm_sc.fv_bpad<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4, 5, 6],[-1, -2, -3, -4, -5, -6]]<br>&nbsp;&nbsp;, array[[7],[8]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[1, 2, 3, 4, 5, 6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-1, -2, -3, -4, -5, -6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7, 7, 7, 7, 7, 7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[8, 8, 8, 8, 8, 8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7, 7, 7, 7, 7, 7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[8, 8, 8, 8, 8, 8]]
>>> sm_sc.fv_trim                 |Clear a certain element value up, down, left, right, and leave it blank.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The element value which is set null at tip.|select <br>&nbsp;&nbsp;sm_sc.fv_trim<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2, 0, 0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 2, 0, 4, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 1, 0, 2, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 1, 2, 2, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 0, 1, 0, 0, 3]]<br>&nbsp;&nbsp;, 0<br>&nbsp;&nbsp;);                 <br><br>return:<br>&nbsp;&nbsp;array[[2, null, null, null, null, null]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 2, null, 4, null, null]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[null, null, 1, 0, 2, null]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[null, 1, 2, 2, null, null]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, null, 1, null, null, 3]]
>>> sm_sc.fv_ltrim                |Left eliminate a certain element value and leave it blank.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The element value which is set null at tip.|select<br>&nbsp;&nbsp;sm_sc.fv_ltrim<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0, 1, 2, 3, 4, 5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 1, 2, 0, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 1, 2, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0, 1, 2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 1, 0, 0, 3]]<br>&nbsp;&nbsp;, 0<br>&nbsp;&nbsp;);                 <br><br>return:<br>&nbsp;&nbsp;array[[null, 1, 2, 3, 4, 5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[null, null, 1, 2, 0, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[null, null, null, 1, 2, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[null, null, null, null, 1, 2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[null, null, 1, 0, 0, 3]]
>>> sm_sc.fv_rtrim                |Right eliminate a certain element value and leave it blank.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The element value which is set null at tip.|select<br>&nbsp;&nbsp;sm_sc.fv_rtrim<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0, 1, 2, 3, 4, 5, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4, 0, 0, 1, 2, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 1, 2, 3, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 1, 2, 0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 1, 0, 0, 3, 0]]<br>&nbsp;&nbsp;, 0<br>&nbsp;&nbsp;);                 <br><br>return:<br>&nbsp;&nbsp;array[[0, 1, 2, 3, 4, 5, NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4, 0, 0, 1, 2, NULL, NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 1, 2, 3, NULL, NULL, NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 1, 2, NULL, NULL, NULL, NULL]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 1, 0, 0, 3, NULL]]
>>> sm_sc.fv_atrim                |Top eliminate a certain element value and leave it blank.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The element value which is set null at tip.|select<br>&nbsp;&nbsp;sm_sc.fv_atrim<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0, 0, 0, 0, 0, 2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 1, 2, 0, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 1, 0, 2, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0, 1, 2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 0, 1, 0, 0, 3]]<br>&nbsp;&nbsp;, 0<br>&nbsp;&nbsp;);                 <br><br>return:<br>&nbsp;&nbsp;array[[NULL, NULL, NULL, NULL, NULL, 2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[NULL, NULL, 1, 2, NULL, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[NULL, 1, 0, 2, NULL, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[NULL, 0, 0, 0, 1, 2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 0, 1, 0, 0, 3]]
>>> sm_sc.fv_btrim                |Bottom eliminate a certain element value and leave it blank.<br>For 3D and 4D arrays, the background is based on the deepest two dimensions as the orientation,<br>&nbsp;&nbsp;That is, the 3D x and x3; 4D x3 and x4.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. The element value which is set null at tip.|select<br>&nbsp;&nbsp;sm_sc.fv_btrim<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2, 0, 0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 2, 0, 4, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 1, 0, 2, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 1, 2, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 0, 1, 0, 0, 3]]<br>&nbsp;&nbsp;, 0<br>&nbsp;&nbsp;);                 <br><br>return:<br>&nbsp;&nbsp;array[[2, 0, 0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, 2, 0, 4, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 1, NULL, 2, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 1, 2, NULL, NULL, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1, NULL, 1, NULL, NULL, 3]]
<!--
>>> sm_sc.ft_rand_ele_y           |Obtain a specified number of random elements for each slice in the y-direction of a 2D array,<br>the number of rows remains unchanged, element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>&nbsp;&nbsp;o_ord_nos, o_slices <br>from<br>&nbsp;&nbsp;sm_sc.ft_rand_ele_y<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.0, 8.0, 9.0, 10.0, 11.0, 12.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[4, 5, 5],[4, 4, 3]]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[4.0, 5.0, 5.0],[10.0, 10.0, 9.0]]
>>> sm_sc.ft_rand_ele_y_pick      |Pick a specified number of random elements per slice in the y-direction of a 2D array,<br>the number of rows remains unchanged, element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>&nbsp;&nbsp;o_ord_nos, o_slices <br>from<br>&nbsp;&nbsp;sm_sc.ft_rand_ele_y_pick<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.0, 8.0, 9.0, 10.0, 11.0, 12.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[6, 2, 4],[3, 1, 5]]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[6.0, 2.0, 4.0],[9.0, 7.0, 11.0]]
>>> sm_sc.ft_rand_ele_x           |Obtain a specified number of random elements per slice in the x-direction of a 2D array,<br>the number of columns remains unchanged, element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>&nbsp;&nbsp;o_ord_nos, o_slices <br>from<br>&nbsp;&nbsp;sm_sc.ft_rand_ele_x<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 2.0],[3.0, 4.0],[5.0, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.0, 8.0],[9.0, 10.0],[11.0, 12.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[3, 2],[3, 1],[4, 3]]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[5.0, 4.0],[5.0, 2.0],[7.0, 6.0]]
>>> sm_sc.ft_rand_ele_x_pick      |Pick a specified number of random elements per slice in the x-direction of a 2D array,<br>the number of columns remains unchanged, element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>&nbsp;&nbsp;o_ord_nos, o_slices <br>from<br>&nbsp;&nbsp;sm_sc.ft_rand_ele_x_pick<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, 2.0],[3.0, 4.0],[5.0, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.0, 8.0],[9.0, 10.0],[11.0, 12.0]]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;o_ord_nos:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[3, 4],[5, 1],[4, 3]]<br>&nbsp;&nbsp;o_slices:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[5.0, 8.0],[9.0, 2.0],[7.0, 6.0]]
>>> sm_sc.ft_rand_ele_x3          |Obtain a specified number of random elements per slice in the x3 direction of a 2D array,<br>the number of rows remains unchanged, element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>
>>> sm_sc.ft_rand_ele_x3_pick     |Pick a specified number of random elements per slice in the x3 direction of a 2D array,<br>the number of rows remains unchanged, element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>
>>> sm_sc.ft_rand_ele_x4          |Obtain a specified number of random elements per slice in the x4 direction of a 2D array,<br>the number of columns remains unchanged, element serial number can be repeated.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>
>>> sm_sc.ft_rand_ele_x4_pick     |Pick a specified number of random elements per slice in the x4 direction of a 2D array,<br>the number of columns remains unchanged, element sequence number does not repeat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Sample quantity; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. New matrix composed of sampling slices; |select <br>
-->

>>
>>####   d. Trigonometric function broadcast operation of matrix:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_degrees   |Element (broadcast) radian to angle.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_degrees<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.64, -12.64],[3.16, -3.16],[0.1, -0.1]]<br>&nbsp;&nbsp;); <br><br>return:<br>&nbsp;&nbsp;array[[724.2187, -724.2187],[181.0547, -181.0547],[5.7296, -5.7296]]
>>> sm_sc.fv_radians   |Element (broadcast) angle to radian.<br>Input param:<br>&nbsp;&nbsp;1. Original angle matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_radians<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[89, -90],[181, -270],[0, -360.1]]<br>&nbsp;&nbsp;); <br><br>return:<br>&nbsp;&nbsp;array[[1.5533, -1.5708],[3.1590, -4.7124],[0, -6.2849]]
>>> sm_sc.fv_sin<br>sm_sc.fv_sin_py       |Element (broadcast) sin.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_sin    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[1.5533, -1.5708],[3.1590, -4.7124],[0, -6.2849]]
>>> sm_sc.fv_cos<br>sm_sc.fv_cos_py       |Element (broadcast) cos.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_cos    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[-1, -1],[0, 0.7071]]
>>> sm_sc.fv_tan<br>sm_sc.fv_tan_py       |Element (broadcast) tan.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_tan    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[0, 0],[1.6331e+16, 1]]
>>> sm_sc.fv_cot       |Element (broadcast) cot.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_cot    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[-8.1656e+15, 8.1656e+15],[0, 1]]
>>> sm_sc.fv_sec       |Element (broadcast) sec.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_sec    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[-1.0, -1.0],[1.6331e+16, 1.414]]
>>> sm_sc.fv_csc       |Element (broadcast) csc.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_csc    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[8.1656e+15, -8.1656e+15],[1.0, 1.414]]
>>> sm_sc.fv_sind      |Element (broadcast) sind.<br>Input param:<br>&nbsp;&nbsp;1. Original angle matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_sind   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[180, -180],[90, 45]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[0, -0],[1, 0.7071]]
>>> sm_sc.fv_cosd      |Element (broadcast) cosd.<br>Input param:<br>&nbsp;&nbsp;1. Original angle matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_cosd   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[180, -180],[90, 45]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[-1, -1],[1, 0.7071]]
>>> sm_sc.fv_tand      |Element (broadcast) tand.<br>Input param:<br>&nbsp;&nbsp;1. Original angle matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_tand   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[180, -180],[90, 45]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[0, 0],[Infinity, 1]]
>>> sm_sc.fv_cotd      |Element (broadcast) cotd.<br>Input param:<br>&nbsp;&nbsp;1. Original angle matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_cotd   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[180, -180],[90, 45]]::double precision[]<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[-Infinity, Infinity],[0, 1]]
>>> sm_sc.fv_asin      |Element (broadcast) asin.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_asin   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -1],[0.5, 0.25]]::double precision[]<br>&nbsp;&nbsp;);     <br><br>return:<br>&nbsp;&nbsp;array[[1.5708, -1.5708],[0.5236, 0.2527]]
>>> sm_sc.fv_acos      |Element (broadcast) acos.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_acos   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -1],[0.5, 0.25]]::double precision[]<br>&nbsp;&nbsp;);     <br><br>return:<br>&nbsp;&nbsp;array[[0, 3.1416],[1.0472, 1.3181]]
>>> sm_sc.fv_atan      |Element (broadcast) atan.<br>Input param:<br>&nbsp;&nbsp;1. Original radian matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_atan   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -1],[0.5, 0.25]]::double precision[]<br>&nbsp;&nbsp;);     <br><br>return:<br>&nbsp;&nbsp;array[[0.7854, -0.7854],[0.4636, 0.2450]]
>>> sm_sc.fv_asind     |Element (broadcast) asind.<br>Input param:<br>&nbsp;&nbsp;1. Original angle matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_asind  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -1], [1.0 :: float/ 2, 1.0 :: float/ 4]]::double precision[]<br>&nbsp;&nbsp;);     <br><br>return:<br>&nbsp;&nbsp;array[[90, -90],[30, 14.4775]]
>>> sm_sc.fv_acosd     |Element (broadcast) acosd.<br>Input param:<br>&nbsp;&nbsp;1. Original angle matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_acosd  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -1], [1.0 :: float/ 2, 1.0 :: float/ 4]]::double precision[]<br>&nbsp;&nbsp;);     <br><br>return:<br>&nbsp;&nbsp;array[[0, 180],[60, 75.5225]]
>>> sm_sc.fv_atand     |Element (broadcast) atand.<br>Input param:<br>&nbsp;&nbsp;1. Original angle matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_atand  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -1],[1.0 / 2, 1.0 / 4],['-Infinity', 'Infinity']]::double precision[]<br>&nbsp;&nbsp;);     <br><br>return:<br>&nbsp;&nbsp;array[[45, -45],[26.5651, 14.0362],[-90, 90]]
>>> sm_sc.fv_sinh<br>sm_sc.fv_sinh_py      |Element (broadcast) sinh.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_sinh   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);           <br><br>return:<br>&nbsp;&nbsp;array[[11.5487, -11.5487],[2.3013, 0.8687]]
>>> sm_sc.fv_cosh<br>sm_sc.fv_cosh_py      |Element (broadcast) cosh.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_cosh   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);           <br><br>return:<br>&nbsp;&nbsp;array[[11.5920, 11.5920],[2.5092, 1.3246]]
>>> sm_sc.fv_tanh<br>sm_sc.fv_tanh_py      |Element (broadcast) tanh.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_tanh   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[pi(), -pi()],[pi() / 2, pi() / 4]]::double precision[]<br>&nbsp;&nbsp;);           <br><br>return:<br>&nbsp;&nbsp;array[[0.9963, -0.9963],[0.9176, 0.6558]]
>>> sm_sc.fv_asinh     |Element (broadcast) asinh.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_asinh  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -1],[0.5, 0.25],['-Infinity', 'Infinity']]::double precision[]<br>&nbsp;&nbsp;);         <br><br>return:<br>&nbsp;&nbsp;array[[0.8814, -0.8814],[0.4812, 0.2475],[-Infinity,Infinity]]
>>> sm_sc.fv_acosh     |Element (broadcast) acosh.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_acosh  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, 1.3],[1.5, 1.25],[1.0, 'Infinity']]::double precision[]<br>&nbsp;&nbsp;);         <br><br>return:<br>&nbsp;&nbsp;array[[0.6224, 0.7564],[0.9624, 0.6931],[0,Infinity]]
>>> sm_sc.fv_atanh     |Element (broadcast) atanh.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  |select<br>&nbsp;&nbsp;sm_sc.fv_atanh  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -1],[0.5, 0.25]]::double precision[]<br>&nbsp;&nbsp;);         <br><br>return:<br>&nbsp;&nbsp;array[[Infinity, -Infinity],[0.5493, 0.2554]]
>>
>>####   e. Waveform processing:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_grad_heigh_1st   |First order difference in the y-direction of the matrix.<br>Output height is 1 less than input length.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;        |select<br>&nbsp;&nbsp;sm_sc.fv_grad_heigh_1st  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4],[-1, -2, -3, -4],[-3, -2, -4, -1]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[-2, -4, -6, -8],[-2, 0, -1, 3]]
>>> sm_sc.fv_grad_heigh_2nd   |Second order difference in the y-direction of the matrix.<br>Output height is 2 less than input length.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;        |select<br>&nbsp;&nbsp;sm_sc.fv_grad_heigh_2nd  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4],[-1, -2, -3, -4],[-3, -2, -4, -1]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[0, 4, 5, 11]]
>>> sm_sc.fv_grad_width_1st   |First order difference in the x-direction of the matrix.<br>Output length is 1 less than input length.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;        |select<br>&nbsp;&nbsp;sm_sc.fv_grad_width_1st  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4],[-1, -2, -3, -4],[-3, -2, -4, -1]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[1, 1, 1],[-1, -1, -1],[1, -2, 3]]
>>> sm_sc.fv_grad_width_2nd   |Second order difference in the x-direction of the matrix.<br>Output length is 2 less than input length.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;        |select<br>&nbsp;&nbsp;sm_sc.fv_grad_width_2nd  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4],[-1, -2, -3, -4],[-3, -2, -4, -1]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[0, 0],[-3, 5]]
>>> sm_sc.fv_sgn_fft      |1 Dimension FFT .<br>Output an integer power of length 2.<br>Input param:<br>&nbsp;&nbsp;1. Original time domain sequence;       |select<br>&nbsp;&nbsp;sm_sc.fv_sgn_fft     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[ <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  1.        ,  0.97952994,  0.91895781,  0.82076344,  0.68896692,  0.52896401,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  0.34730525,  0.15142778, -0.05064917, -0.25065253, -0.44039415, -0.61210598,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; -0.75875812, -0.87434662, -0.95413926, -0.99486932, -0.99486932, -0.95413926,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; -0.87434662, -0.75875812, -0.61210598, -0.44039415, -0.25065253, -0.05064917,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  0.15142778,  0.34730525,  0.52896401,  0.68896692,  0.82076344,  0.91895781,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  0.97952994,  1.        <br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;);    <br><br>return an array of complex types(sm_sc.typ_l_complex[]): <br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;1.00000000e+00+0.00000000e+00j,  1.61497357e+01+1.59061014e+00j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-3.53833224e-01-7.03818042e-02j, -1.26604779e-01-3.84051399e-02j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-6.41365736e-02-2.65662386e-02j, -3.75538211e-02-2.00729356e-02j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-2.37109631e-02-1.58431590e-02j, -1.55930224e-02-1.27968628e-02j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-1.04489194e-02-1.04489194e-02j, -7.01356383e-03-8.54605226e-03j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-4.63822579e-03-6.94159545e-03j, -2.96316279e-03-5.54368767e-03j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-1.77736596e-03-4.29094101e-03j, -9.52464878e-04-3.13985591e-03j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-4.09315616e-04-2.05776856e-03j, -1.00322243e-04-1.01858882e-03j,<br>&nbsp;&nbsp;&nbsp;&nbsp; 2.22044605e-16+0.00000000e+00j, -1.00322243e-04+1.01858882e-03j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-4.09315616e-04+2.05776856e-03j, -9.52464878e-04+3.13985591e-03j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-1.77736596e-03+4.29094101e-03j, -2.96316279e-03+5.54368767e-03j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-4.63822579e-03+6.94159545e-03j, -7.01356383e-03+8.54605226e-03j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-1.04489194e-02+1.04489194e-02j, -1.55930224e-02+1.27968628e-02j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-2.37109631e-02+1.58431590e-02j, -3.75538211e-02+2.00729356e-02j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-6.41365736e-02+2.65662386e-02j, -1.26604779e-01+3.84051399e-02j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-3.53833224e-01+7.03818042e-02j,  1.61497357e+01-1.59061014e+00j<br>&nbsp;&nbsp;]
>>> sm_sc.fv_sgn_ifft     |1 Dimension IFFT.<br>Output an integer power of length 2.<br>Input param:<br>&nbsp;&nbsp;1. Original frequency domain sequence;     |select<br>&nbsp;&nbsp;sm_sc.fv_sgn_ifft    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[ <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  1.        ,  0.97952994,  0.91895781,  0.82076344,  0.68896692,  0.52896401,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  0.34730525,  0.15142778, -0.05064917, -0.25065253, -0.44039415, -0.61210598,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; -0.75875812, -0.87434662, -0.95413926, -0.99486932, -0.99486932, -0.95413926,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; -0.87434662, -0.75875812, -0.61210598, -0.44039415, -0.25065253, -0.05064917,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  0.15142778,  0.34730525,  0.52896401,  0.68896692,  0.82076344,  0.91895781,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;  0.97952994,  1.        <br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;);    <br><br>return an array of complex types(sm_sc.typ_l_complex[]): <br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00014495-0.00021693j, -0.00009260-0.00017324j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00005554-0.00013409j, -0.00002977-0.00009812j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00001279-0.00006431j, -0.00000314-0.00003183j,<br>&nbsp;&nbsp;&nbsp;&nbsp;0.00000000+0.00000000j , -0.00000314+0.00003183j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00001279+0.00006431j, -0.00002977+0.00009812j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00005554+0.00013409j, -0.00009260+0.00017324j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00014495+0.00021693j, -0.00021917+0.00026707j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00032653+0.00032653j, -0.00048729+0.00039991j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00074097+0.00049510j, -0.00117356+0.00062728j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.00200427+0.00083020j, -0.00395641+0.00120017j,<br>&nbsp;&nbsp;&nbsp;&nbsp;-0.01105730+0.00219944j, 0.50467924-0.04970657j<br>&nbsp;&nbsp;]
>>> sm_sc.fv_sgn_fft2     |2D-FFT.<br>Output an integer power of 2 in height and width.<br>Input param:<br>&nbsp;&nbsp;1. Original time domain sequence;       |select<br>&nbsp;&nbsp;sm_sc.fv_sgn_fft2    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4],[4, 5, 6, 7],[8, 9, 10, 11]]<br>&nbsp;&nbsp;);    <br><br>return:<br>&nbsp;&nbsp;array[[70.0+0.0j, -6.0-6.0j, -6.0+0.0j, -6.0+6.0j]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-28.0+22.0j, 2.0-2.0j, 0.0-2.0j, -2.0-2.0j]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[26.0+0.0j, -2.0-2.0j, -2.0+0.0j, -2.0+2.0j]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-28.0-22.0j, -2.0+2.0j, 0.0+2.0j, 2.0+2.0j]]
>>> sm_sc.fv_sgn_ifft2    |2D-IFFT.<br>Output an integer power of 2 in height and width.<br>Input param:<br>&nbsp;&nbsp;1. Original frequency domain sequence;     |select<br>&nbsp;&nbsp;sm_sc.fv_sgn_ifft2   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4],[4, 5, 6, 7],[8, 9, 10, 11]]<br>&nbsp;&nbsp;);    <br><br>return:<br>&nbsp;&nbsp;array[[4.3750+0.0000j, -0.3750+0.3750j, -0.3750+0.0000j, -0.3750-0.3750j]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-1.7500-1.3750j, 0.1250+0.1250j, 0.0000+0.1250j, -0.1250+0.1250j]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1.6250+0.0000j, -0.1250+0.1250j, -0.1250+0.0000j, -0.1250-0.1250j]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-1.7500+1.3750j, -0.1250-0.1250j, 0.0000-0.1250j, 0.1250-0.1250j]]
>>> sm_sc.fv_sgn_kalman   |Kalman filtering transformation.<br>Output sequence is 1 longer than input sequence.<br>Input param:<br>&nbsp;&nbsp;1. Original time domain sequence;           |<br>-- create extension tablefunc;<br>with <br>-- construct precise sequences<br>cte_ods as<br>(<br>&nbsp;&nbsp;  select <br>&nbsp;&nbsp;&nbsp;&nbsp;    row_number() over() as row_no,<br>&nbsp;&nbsp;&nbsp;&nbsp;    a_x,<br>&nbsp;&nbsp;&nbsp;&nbsp;    10 * sin(a_x) :: float as ods_ele<br>&nbsp;&nbsp;  from generate_series(0.0::decimal, pi()::decimal, (pi() / 100)::decimal) tb_a(a_x)<br>)<br>,<br>-- construct error sequence<br>cte_test as<br>(<br>&nbsp;&nbsp;  select<br>&nbsp;&nbsp;&nbsp;&nbsp;    row_no,<br>&nbsp;&nbsp;&nbsp;&nbsp;    ods_ele + normal_rand(1, 0.0, 0.5) :: float as test_ele   -- assumed standard deviation of observation error 0.5<br>&nbsp;&nbsp;  from cte_ods<br>)<br>,<br>-- perform Kalman filtering<br>cte_predict as<br>(<br>&nbsp;&nbsp;  select <br>&nbsp;&nbsp;&nbsp;&nbsp;    sm_sc.fv_sgn_kalman<br>&nbsp;&nbsp;&nbsp;&nbsp;    (<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      array_agg(test_ele order by row_no),              <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      0.4 * 0.4                          ,      -- assumed standard deviation of process error  0.4<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      0.5 * 0.5                          ,      -- assumed standard deviation of observation error 0.5<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      0.5 * 0.5<br>&nbsp;&nbsp;&nbsp;&nbsp;    ) as predict_arr<br>&nbsp;&nbsp;  from cte_test<br>)<br>-- comparison results<br>select <br>&nbsp;&nbsp;  a_x, a_ods_ele, a_test_ele, a_predict_ele<br>&nbsp;&nbsp;  -- sum(abs(a_test_ele - a_ods_ele) - abs(a_predict_ele - a_ods_ele))<br>from unnest<br>(<br>&nbsp;&nbsp;  (select array_agg(a_x) from cte_ods),<br>&nbsp;&nbsp;  (select array_agg(ods_ele) from cte_ods),<br>&nbsp;&nbsp;  (select array_agg(test_ele) from cte_test),<br>&nbsp;&nbsp;  (select predict_arr from cte_predict)<br>)&nbsp; tb_a(a_x, a_ods_ele, a_test_ele, a_predict_ele) <br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;a_x&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a_ods_ele&nbsp;&nbsp;&nbsp;a_test_ele&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a_predict_ele<br>&nbsp;&nbsp;&nbsp;&nbsp;0.00000000	0.00000000	-0.07238570	-0.07238570<br>&nbsp;&nbsp;&nbsp;&nbsp;0.03141593	0.31410760	0.06269807	0.01152998<br>&nbsp;&nbsp;&nbsp;&nbsp;0.06283186	0.62790530	0.90212888	0.50826981<br>&nbsp;&nbsp;&nbsp;&nbsp;0.09424779	0.94108320	1.73031760	1.17427501<br>&nbsp;&nbsp;&nbsp;&nbsp;0.12566372	1.25333250	0.59666534	0.86101867<br>&nbsp;&nbsp;&nbsp;&nbsp;0.15707965	1.56434480	1.01786649	0.94599484<br>&nbsp;&nbsp;&nbsp;&nbsp;0.18849558	1.87381340	2.09580312	1.56879714<br>&nbsp;&nbsp;&nbsp;&nbsp;0.21991151	2.18143270	2.11825104	1.86639946<br>&nbsp;&nbsp;&nbsp;&nbsp;0.25132744	2.48689910	2.55075390	2.23706481<br>&nbsp;&nbsp;&nbsp;&nbsp;0.28274337	2.78991140	2.60066296	2.43399926<br>&nbsp;&nbsp;&nbsp;&nbsp;0.31415930	3.09017030	2.45377347	2.44470949<br>&nbsp;&nbsp;&nbsp;&nbsp;0.34557523	3.38737960	3.72143761	3.13621913<br>&nbsp;&nbsp;&nbsp;&nbsp;0.37699116	3.68124590	2.99812719	3.06142489<br>&nbsp;&nbsp;&nbsp;&nbsp;0.40840709	3.97147930	3.42792734	3.25993228<br>&nbsp;&nbsp;&nbsp;&nbsp;0.43982302	4.25779340	4.31830536	3.83317505<br>&nbsp;&nbsp;&nbsp;&nbsp;0.47123895	4.53990550	4.56145212	4.22762911<br>&nbsp;&nbsp;&nbsp;&nbsp;0.50265488	4.81753720	5.06905005	4.68336488<br>&nbsp;&nbsp;&nbsp;&nbsp;0.53407081	5.09041470	5.45222292	5.09979867<br>&nbsp;&nbsp;&nbsp;&nbsp;0.56548674	5.35826850	5.81926993	5.48948327<br>&nbsp;&nbsp;&nbsp;&nbsp;0.59690267	5.62083430	5.97156983	5.75059406<br>&nbsp;&nbsp;&nbsp;&nbsp;0.62831860	5.87785310	6.19611739	5.99190124<br>&nbsp;&nbsp;&nbsp;&nbsp;0.65973453	6.12907110	6.98108901	6.52767142<br>&nbsp;&nbsp;&nbsp;&nbsp;0.69115046	6.37424050	5.99603643	6.23972389<br>&nbsp;&nbsp;&nbsp;&nbsp;0.72256639	6.61311930	6.28735803	6.26552380<br>&nbsp;&nbsp;&nbsp;&nbsp;0.75398232	6.84547170	5.74105687	5.98145868<br>&nbsp;&nbsp;&nbsp;&nbsp;0.78539825	7.07106840	6.77602618	6.41181739<br>&nbsp;&nbsp;&nbsp;&nbsp;0.81681418	7.28968690	6.85151619	6.64996985<br>&nbsp;&nbsp;&nbsp;&nbsp;0.84823011	7.50111130	6.38093462	6.50425328<br>&nbsp;&nbsp;&nbsp;&nbsp;0.87964604	7.70513300	7.66786487	7.13449600<br>&nbsp;&nbsp;&nbsp;&nbsp;0.91106197	7.90155070	8.07567306	7.64426231<br>&nbsp;&nbsp;&nbsp;&nbsp;0.94247790	8.09017060	7.87262345	7.76794872<br>&nbsp;&nbsp;&nbsp;&nbsp;0.97389383	8.27080630	8.31935211	8.06660333<br>&nbsp;&nbsp;&nbsp;&nbsp;1.00530976	8.44327980	8.86376730	8.49836836<br>&nbsp;&nbsp;&nbsp;&nbsp;1.03672569	8.60742090	8.06964373	8.26615980<br>&nbsp;&nbsp;&nbsp;&nbsp;1.06814162	8.76306740	9.41039761	8.88590917<br>&nbsp;&nbsp;&nbsp;&nbsp;1.09955755	8.91006580	8.92835762	8.90890037<br>&nbsp;&nbsp;&nbsp;&nbsp;1.13097348	9.04827110	8.68164490	8.78581282<br>&nbsp;&nbsp;&nbsp;&nbsp;1.16238941	9.17754680	9.27285046	9.04960525<br>&nbsp;&nbsp;&nbsp;&nbsp;1.19380534	9.29776530	9.65991384	9.38016447<br>&nbsp;&nbsp;&nbsp;&nbsp;1.22522127	9.40880810	9.37680006	9.37834222<br>&nbsp;&nbsp;&nbsp;&nbsp;1.25663720	9.51056560	9.17752884	9.26957640<br>&nbsp;&nbsp;&nbsp;&nbsp;1.28805313	9.60293730	8.78020289	9.00451880<br>&nbsp;&nbsp;&nbsp;&nbsp;1.31946906	9.68583200	9.57998840	9.31620831<br>&nbsp;&nbsp;&nbsp;&nbsp;1.35088499	9.75916790	9.27521528	9.29400540<br>&nbsp;&nbsp;&nbsp;&nbsp;1.38230092	9.82287280	9.78488155	9.55987687<br>&nbsp;&nbsp;&nbsp;&nbsp;1.41371685	9.87688360	10.32271604	9.97305068<br>&nbsp;&nbsp;&nbsp;&nbsp;1.44513278	9.92114720	10.41625520	10.21310193<br>&nbsp;&nbsp;&nbsp;&nbsp;1.47654871	9.95561980	9.17540882	9.65105998<br>&nbsp;&nbsp;&nbsp;&nbsp;1.50796464	9.98026740	9.11660215	9.36158353<br>&nbsp;&nbsp;&nbsp;&nbsp;1.53938057	9.99506570	10.34605545	9.89479948<br>&nbsp;&nbsp;&nbsp;&nbsp;1.57079650	10.00000000	10.35161797	10.14222442<br>&nbsp;&nbsp;&nbsp;&nbsp;1.60221243	9.99506550	9.88948807	10.00533575<br>&nbsp;&nbsp;&nbsp;&nbsp;1.63362836	9.98026720	10.25024667	10.13798596<br>&nbsp;&nbsp;&nbsp;&nbsp;1.66504429	9.95561950	9.73600478	9.92026235<br>&nbsp;&nbsp;&nbsp;&nbsp;1.69646022	9.92114680	10.15850842	10.04930270<br>&nbsp;&nbsp;&nbsp;&nbsp;1.72787615	9.87688310	11.01563448	10.57269347<br>&nbsp;&nbsp;&nbsp;&nbsp;1.75929208	9.82287210	10.40551561	10.48214553<br>&nbsp;&nbsp;&nbsp;&nbsp;1.79070801	9.75916720	10.51001749	10.49724172<br>&nbsp;&nbsp;&nbsp;&nbsp;1.82212394	9.68583110	9.82572837	10.13353238<br>&nbsp;&nbsp;&nbsp;&nbsp;1.85353987	9.60293630	10.46598167	10.31359568<br>&nbsp;&nbsp;&nbsp;&nbsp;1.88495580	9.51056450	9.31690803	9.77376337<br>&nbsp;&nbsp;&nbsp;&nbsp;1.91637173	9.40880700	9.12661892	9.42325287<br>&nbsp;&nbsp;&nbsp;&nbsp;1.94778766	9.29776410	9.47494195	9.45124904<br>&nbsp;&nbsp;&nbsp;&nbsp;1.97920359	9.17754540	8.83534106	9.11765704<br>&nbsp;&nbsp;&nbsp;&nbsp;2.01061952	9.04826960	8.42020767	8.73990007<br>&nbsp;&nbsp;&nbsp;&nbsp;2.04203545	8.91006420	9.31084849	9.04914079<br>&nbsp;&nbsp;&nbsp;&nbsp;2.07345138	8.76306570	8.58320157	8.79677582<br>&nbsp;&nbsp;&nbsp;&nbsp;2.10486731	8.60741910	8.81196387	8.80500207<br>&nbsp;&nbsp;&nbsp;&nbsp;2.13628324	8.44327800	8.83431931	8.82088106<br>&nbsp;&nbsp;&nbsp;&nbsp;2.16769917	8.27080440	8.69911763	8.75493078<br>&nbsp;&nbsp;&nbsp;&nbsp;2.19911510	8.09016850	8.49712138	8.61529441<br>&nbsp;&nbsp;&nbsp;&nbsp;2.23053103	7.90154860	7.52881785	8.02683006<br>&nbsp;&nbsp;&nbsp;&nbsp;2.26194696	7.70513080	7.73606742	7.86934535<br>&nbsp;&nbsp;&nbsp;&nbsp;2.29336289	7.50110900	6.90444588	7.34673036<br>&nbsp;&nbsp;&nbsp;&nbsp;2.32477882	7.28968450	7.20778566	7.27147425<br>&nbsp;&nbsp;&nbsp;&nbsp;2.35619475	7.07106600	7.04657670	7.14966381<br>&nbsp;&nbsp;&nbsp;&nbsp;2.38761068	6.84546910	6.19311495	6.63157173<br>&nbsp;&nbsp;&nbsp;&nbsp;2.41902661	6.61311670	7.75630107	7.24075480<br>&nbsp;&nbsp;&nbsp;&nbsp;2.45044254	6.37423780	6.12010764	6.63378275<br>&nbsp;&nbsp;&nbsp;&nbsp;2.48185847	6.12906840	6.65184129	6.64356373<br>&nbsp;&nbsp;&nbsp;&nbsp;2.51327440	5.87785030	5.50467567	6.02671193<br>&nbsp;&nbsp;&nbsp;&nbsp;2.54469033	5.62083150	4.62287829	5.26635862<br>&nbsp;&nbsp;&nbsp;&nbsp;2.57610626	5.35826560	4.54544759	4.87589420<br>&nbsp;&nbsp;&nbsp;&nbsp;2.60752219	5.09041170	4.88442549	4.88051497<br>&nbsp;&nbsp;&nbsp;&nbsp;2.63893812	4.81753420	4.73826616	4.80346926<br>&nbsp;&nbsp;&nbsp;&nbsp;2.67035405	4.53990240	5.22731399	5.03303474<br>&nbsp;&nbsp;&nbsp;&nbsp;2.70176998	4.25779020	4.36847656	4.67309251<br>&nbsp;&nbsp;&nbsp;&nbsp;2.73318591	3.97147610	4.07646885	4.34994540<br>&nbsp;&nbsp;&nbsp;&nbsp;2.76460184	3.68124270	4.36259811	4.35679844<br>&nbsp;&nbsp;&nbsp;&nbsp;2.79601777	3.38737630	3.42840169	3.85395428<br>&nbsp;&nbsp;&nbsp;&nbsp;2.82743370	3.09016700	3.66048294	3.74916510<br>&nbsp;&nbsp;&nbsp;&nbsp;2.85884963	2.78990800	2.66247256	3.16058377<br>&nbsp;&nbsp;&nbsp;&nbsp;2.89026556	2.48689580	2.18075736	2.62988395<br>&nbsp;&nbsp;&nbsp;&nbsp;2.92168149	2.18142930	1.64406354	2.09593762<br>&nbsp;&nbsp;&nbsp;&nbsp;2.95309742	1.87380990	2.32140082	2.21805443<br>&nbsp;&nbsp;&nbsp;&nbsp;2.98451335	1.56434140	1.76298351	1.97157602<br>&nbsp;&nbsp;&nbsp;&nbsp;3.01592928	1.25332900	1.88917658	1.92694631<br>&nbsp;&nbsp;&nbsp;&nbsp;3.04734521	0.94107980	1.13478766	1.49789230<br>&nbsp;&nbsp;&nbsp;&nbsp;3.07876114	0.62790180	-0.13328145	0.61440559<br>&nbsp;&nbsp;&nbsp;&nbsp;3.11017707	0.31410420	-0.54286213	-0.01240112<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.31651924
>>
>>####   f. Slice granularity aggregation of matrices, matrix granularity aggregation, sorting by a certain slice value, slice granularity positioning extreme value position:
>>
>>> <table>
>>> <th>Function Name</th><th>Function Description</th><th>Call Example</th>
>>> <tr><td>sm_sc.fv_aggr_slice_sum            </td><td>SUM of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_sum             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;2.75      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_prod           </td><td>Direct product of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_prod            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;-19.305      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_avg            </td><td>Average of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;    </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_avg             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;0.6875      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_max            </td><td>Maximum of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_max             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;2.25      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_min            </td><td>Minimum of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_min             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;-3.0      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_chunk_sum            </td><td>SUM of chunk size aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Chunk specifications, length of each dimension;  </td><td>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_aggr_chunk_sum<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;6.4,&nbsp;&nbsp;6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;-8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;-3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;-6.4,&nbsp;&nbsp;-6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[5.12,&nbsp;&nbsp;10.20,&nbsp;&nbsp;-3.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[20.30,&nbsp;&nbsp;-5.30,&nbsp;&nbsp;-9.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;]</td></tr>
>>> <tr><td>sm_sc.fv_aggr_chunk_prod           </td><td>Direct product of chunk size aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Chunk specifications, length of each dimension;   </td><td>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_aggr_chunk_prod<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;6.4,&nbsp;&nbsp;6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;-8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;-3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;-6.4,&nbsp;&nbsp;-6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-34.668544,&nbsp;&nbsp;-288.422289,&nbsp;&nbsp;-242.7364]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[502.880625,&nbsp;&nbsp;-287.641600,&nbsp;&nbsp;-921.7296]<br>&nbsp;&nbsp;&nbsp;&nbsp;]</td></tr>
>>> <tr><td>sm_sc.fv_aggr_chunk_avg            </td><td>Average of chunk size aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Chunk specifications, length of each dimension;    </td><td>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_aggr_chunk_avg<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;6.4,&nbsp;&nbsp;6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;-8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;-3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;-6.4,&nbsp;&nbsp;-6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.280,&nbsp;&nbsp;2.550,&nbsp;&nbsp;-0.950]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[5.075,&nbsp;&nbsp;-1.325,&nbsp;&nbsp;-2.300]<br>&nbsp;&nbsp;&nbsp;&nbsp;]</td></tr>
>>> <tr><td>sm_sc.fv_aggr_chunk_max            </td><td>Maximum of chunk size aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Chunk specifications, length of each dimension;  </td><td>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_aggr_chunk_max<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;6.4,&nbsp;&nbsp;6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;-8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;-3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;-6.4,&nbsp;&nbsp;-6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.56,&nbsp;&nbsp;5.1,&nbsp;&nbsp;8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[6.9,&nbsp;&nbsp;6.4,&nbsp;&nbsp;6.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;]</td></tr>
>>> <tr><td>sm_sc.fv_aggr_chunk_min            </td><td>Minimum of chunk size aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Chunk specifications, length of each dimension;  </td><td>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_aggr_chunk_min<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;6.4,&nbsp;&nbsp;6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;-8.2,&nbsp;&nbsp;2.56,&nbsp;&nbsp;-3.33,&nbsp;&nbsp;-1.9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;-6.4,&nbsp;&nbsp;-6.6,&nbsp;&nbsp;6.9,&nbsp;&nbsp;-2.65,&nbsp;&nbsp;-4.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;-3.33,&nbsp;&nbsp;-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.25,&nbsp;&nbsp;-6.4,&nbsp;&nbsp;-6.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;]</td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_ptp            </td><td>Element aggregation span (maximum to minimum).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_ptp             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;5.25      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_var_pop        </td><td>Population variance of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_var_pop         <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;4.67546875      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_var_samp       </td><td>Sample variance of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_var_samp        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;6.2339583      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_stddev_pop     </td><td>Population standard deviation of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_stddev_pop      <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;2.162283      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_stddev_samp    </td><td>Sample standard deviation of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_stddev_samp     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;2.4967896      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_mode           </td><td>Mode of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_mode            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.25, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;2.25      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_median         </td><td>Median of element aggregation.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_median          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;1.75      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_concat         </td><td>Element aggregation bit concat.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_concat          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[['abc', 'de'],['fghi', 'jklmn']]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;'abcdefghijklmn'      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_or             </td><td>Element aggregation bit OR, logical OR.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_or             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;true<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_or             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;B'111' </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_and            </td><td>Element aggregation bit AND, logical AND.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_and             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;false<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_and             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;B'000' </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_coalesce       </td><td>Element aggregation null replace.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_coalesce        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[null, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;2.25      </td></tr>
>>> <tr><td>sm_sc.fv_aggr_slice_is_exists_null       </td><td>Empty element determination.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_aggr_slice_is_exists_null        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[null, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;true      </td></tr>
>>> <tr><td>sm_sc.fv_idx_1d_max                </td><td>Retrieve the position of the maximum value in a one-dimensional array.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_idx_1d_max                 <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.2, 2.3, 5.6, 52.1]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;4      </td></tr>
>>> <tr><td>sm_sc.fv_idx_y_max                 </td><td>Retrieve the maximum value position in each row of a 2D array.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_idx_y_max                  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, 2.3, 5.6, 52.1],[-1.2, 2.3, -5.6, 2.1]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[[4],[2]]      </td></tr>
>>> <tr><td>sm_sc.fv_idx_x_max                 </td><td>Retrieve the maximum value position in each column of a 2D array.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_idx_x_max                  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[[2, 1]]      </td></tr>
>>> <tr><td>sm_sc.fv_idx_mx_max                </td><td>Retrieve the maximum value position in a 2D array.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_idx_mx_max                 <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[1, 2]      </td></tr>
>>> <tr><td>sm_sc.fv_idx_1d_min                </td><td>Retrieve the minimum value position in a 1D array.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_idx_1d_min                 <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.2, 2.3, 5.6, 52.1]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;1      </td></tr>
>>> <tr><td>sm_sc.fv_idx_y_min                 </td><td>Retrieve the minimum value position in each row of a 2D array.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_idx_y_min                  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[[1],[2]]      </td></tr>
>>> <tr><td>sm_sc.fv_idx_x_min                 </td><td>Retrieve the minimum value position in each column of a 2D array.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_idx_x_min                  <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[[1, 2]]      </td></tr>
>>> <tr><td>sm_sc.fv_idx_mx_min                </td><td>Retrieve the minimum value position in a 2D array.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix;  </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_idx_mx_min                 <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[2, 2]      </td></tr>
>>> <tr><td>sm_sc.fv_ord_by_col                </td><td>Sort elements vertically, arranging rows according to a certain column value.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Column number, sorted according to the size of the element values in the column; <br>&nbsp;&nbsp;3. Descending control, default: ascending;   </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_ord_by_col                 <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[[2.2, -3.0],[1.3, 2.25]]      </td></tr>
>>> <tr><td>sm_sc.fv_ord_by_row                </td><td>Sort elements horizontal,arranging columns according to a certain row value.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Row number, sorted according to the size of the element values in the row; <br>&nbsp;&nbsp;3. Descending control, default: ascending;    </td><td>select<br>&nbsp;&nbsp;sm_sc.fv_ord_by_row                 <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, 2.25],[2.2, -3.0]]<br>&nbsp;&nbsp;, 1<br>&nbsp;&nbsp;, true<br>&nbsp;&nbsp;);   <br><br>return:<br>&nbsp;&nbsp;array[[2.25, 1.3],[-3.0, 2.2]]      </td></tr>
>>> <tr><td>sm_sc.fa_mx_sum                    </td><td>SUM of matrix size aggregation                             </td><td rowspan="6">select<br>&nbsp;&nbsp;sm_sc.fa_mx_sum(a_arr) as a_mx_sum,<br>&nbsp;&nbsp;sm_sc.fa_mx_prod(a_arr) as a_mx_prod,<br>&nbsp;&nbsp;sm_sc.fa_mx_avg(a_arr :: float[]) as a_mx_avg,<br>&nbsp;&nbsp;sm_sc.fa_mx_max(a_arr) as a_mx_max,<br>&nbsp;&nbsp;sm_sc.fa_mx_min(a_arr) as a_mx_min,<br>&nbsp;&nbsp;sm_sc.fa_mx_inner_prod(a_arr) as a_mx_inner_prod     <br>from <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;select array[[1.3, 2.25],[2.2, -3.0]] as a_arr<br>&nbsp;&nbsp;&nbsp;&nbsp;union all<br>&nbsp;&nbsp;&nbsp;&nbsp;select array[[2.3, 5.6],[0.47, 2.6]] as a_arr<br>&nbsp;&nbsp;&nbsp;&nbsp;union all<br>&nbsp;&nbsp;&nbsp;&nbsp;select array[[2.6, -0.8],[0.63, 3.71]] as a_arr<br>&nbsp;&nbsp;) tb_a_arr(a_arr);   <br><br>return:<br>&nbsp;&nbsp;a_mx_sum: <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[6.2, 7.05],[3.30, 3.31]] <br>&nbsp;&nbsp;a_mx_prod: <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[7.774, -10.08],[0.65142, -28.938]]<br>&nbsp;&nbsp;a_mx_avg: <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.0667, 2.35],[1.1, 1.1033]]<br>&nbsp;&nbsp;a_mx_max: <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.6, 5.6],[2.2, 3.71]]<br>&nbsp;&nbsp;a_mx_min: <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3, -0.8],[0.47, -3.0]]<br>&nbsp;&nbsp;a_mx_inner_prod: <br>&nbsp;&nbsp;&nbsp;&nbsp;-30.59258     </td></tr>
>>> <tr><td>sm_sc.fa_mx_prod                   </td><td>Direct product of matrix size aggregation                             </td></tr>
>>> <tr><td>sm_sc.fa_mx_avg                    </td><td>Average of matrix size aggregation                             </td></tr>
>>> <tr><td>sm_sc.fa_mx_max                    </td><td>Maximum of matrix size aggregation                             </td></tr>
>>> <tr><td>sm_sc.fa_mx_min                    </td><td>Minimum of matrix size aggregation                             </td></tr>
>>> <tr><td>sm_sc.fa_mx_inner_prod             </td><td>Inner product of matrix size aggregation                              </td></tr>
>>> <tr><td>sm_sc.fa_array_concat              </td><td>Matrix size 1D array aggregation concat                     </td><td>select<br>&nbsp;&nbsp;sm_sc.fa_array_concat(a_arr)<br>from <br>(<br>&nbsp;&nbsp;select array[1.2, 1.5, 1.8] as a_val<br>&nbsp;&nbsp;union all<br>&nbsp;&nbsp;select array[11.2, 21.5, 31.8] as a_val<br>&nbsp;&nbsp;union all<br>&nbsp;&nbsp;select array[31.2, 1.5, 21.8] as a_val<br>) tb_a_arr(a_arr);   <br><br>return:<br>&nbsp;&nbsp;array[1.2, 1.5, 1.8, 11.2, 21.5, 31.8, 31.2, 1.5, 21.8]      </td></tr>
>>> <tr><td>sm_sc.fa_mx_concat_y               </td><td>Matrix size aggregation in first dimension concat                     </td><td rowspan="5">&nbsp;&nbsp;select<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fa_mx_concat_y(a_arr)&nbsp;&nbsp;as&nbsp;&nbsp;a_mx_concat_y,<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fa_mx_concat_x(a_arr)&nbsp;&nbsp;as&nbsp;&nbsp;a_mx_concat_x,<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fa_mx_concat_x3(array[a_arr])&nbsp;&nbsp;as&nbsp;&nbsp;a_mx_concat_x3,<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fa_mx_concat_x4(array[[a_arr]])&nbsp;&nbsp;as&nbsp;&nbsp;a_mx_concat_x4,<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fa_mx_concat_per_ele(a_arr)&nbsp;&nbsp;as&nbsp;&nbsp;a_mx_concat_per_ele<br>from&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;array[[B'110',&nbsp;&nbsp;B'100',&nbsp;&nbsp;B'101'],[B'101',&nbsp;&nbsp;B'1010',&nbsp;&nbsp;B'100']]&nbsp;&nbsp;as&nbsp;&nbsp;a_val<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;array[[B'1010',&nbsp;&nbsp;B'110',&nbsp;&nbsp;B'1010'],[B'010',&nbsp;&nbsp;B'1010',&nbsp;&nbsp;B'110']]&nbsp;&nbsp;as&nbsp;&nbsp;a_val<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;array[[B'1010',&nbsp;&nbsp;B'10',&nbsp;&nbsp;B'1010'],[B'10',&nbsp;&nbsp;B'101001010',&nbsp;&nbsp;B'00010']]&nbsp;&nbsp;as&nbsp;&nbsp;a_val<br>)&nbsp;&nbsp;tb_a_arr(a_arr)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;a_mx_concat_y:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[B'110',B'100',B'101']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'101',B'1010',B'100']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'1010',B'110',B'1010']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'010',B'1010',B'110']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'1010',B'10',B'1010']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'10',B'101001010',B'00010']]<br>&nbsp;&nbsp;&nbsp;&nbsp;a_mx_concat_x:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[B'110',B'100',B'101',B'1010',B'110',B'1010',B'1010',B'10',B'1010']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'101',B'1010',B'100',B'010',B'1010',B'110',B'10',B'101001010',B'00010']]<br>&nbsp;&nbsp;&nbsp;&nbsp;a_mx_concat_x3:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[[B'110',B'100',B'101',B'1010',B'110',B'1010',B'1010',B'10',B'1010']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'101',B'1010',B'100',B'010',B'1010',B'110',B'10',B'101001010',B'00010']][<br>&nbsp;&nbsp;&nbsp;&nbsp;a_mx_concat_x4:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[[[B'110',B'100',B'101',B'1010',B'110',B'1010',B'1010',B'10',B'1010']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'101',B'1010',B'100',B'010',B'1010',B'110',B'10',B'101001010',B'00010']]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;a_mx_concat_per_ele:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[B'11010101010',B'10011010',B'10110101010']<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[B'10101010',B'10101010101001010',B'10011000010']]&nbsp;&nbsp;</td></tr>
>>> <tr><td>sm_sc.fa_mx_concat_x               </td><td>Matrix size aggregation in the second dimension concat                     </td>
>>> <tr><td>sm_sc.fa_mx_concat_x3              </td><td>Matrix size aggregation in the third dimension concat                     </td>
>>> <tr><td>sm_sc.fa_mx_concat_x4              </td><td>Matrix size aggregation in the fourth dimension concat                     </td>
>>> <tr><td>sm_sc.fa_mx_concat_per_ele         </td><td> Corresponding element aggregation bit at the same position of each matrix concat            </td>
>>> <tr><td>sm_sc.fa_mx_or                     </td><td>Element aggregation bit OR,  logical OR of corresponding  at the same position for each matrix       </td><td rowspan="2">select<br>&nbsp;&nbsp;sm_sc.fa_mx_or(a_arr) as a_mx_or,<br>&nbsp;&nbsp;sm_sc.fa_mx_and(a_arr) as a_mx_and<br>from <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;select array[[B'1101', B'100', B'101'],[B'10', B'100101010', B'10010']] as a_arr<br>&nbsp;&nbsp;&nbsp;&nbsp;union all<br>&nbsp;&nbsp;&nbsp;&nbsp;select array[[B'1010', B'110', B'100'],[B'00', B'100101010', B'10010']] as a_arr<br>&nbsp;&nbsp;&nbsp;&nbsp;union all<br>&nbsp;&nbsp;&nbsp;&nbsp;select array[[B'1010', B'101', B'010'],[B'10', B'101001010', B'00010']] as a_arr<br>&nbsp;&nbsp;) tb_a_arr(a_arr);   <br><br>return:<br>&nbsp;&nbsp;a_mx_or: <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'1111', B'111', B'111'],[B'10', B'101101010', B'10010']] <br>&nbsp;&nbsp;a_mx_and: <br>&nbsp;&nbsp;&nbsp;&nbsp;array[[B'1000', B'100', B'000'],[B'00', B'100001010', B'00010']]      </td></tr>
>>> <tr><td>sm_sc.fa_mx_and                    </td><td>Element aggregation bit AND, logical AND of corresponding  at the same position for each matrix       </td>
>>> <tr><td>sm_sc.fa_mx_coalesce               </td><td>Matrix size aggregation null replace                             </td><td>select<br>&nbsp;&nbsp;sm_sc.fa_mx_coalesce(a_arr)<br>from <br>(<br>&nbsp;&nbsp;select array[[null, 12, null],[14, null, 16]] as a_val<br>&nbsp;&nbsp;union all<br>&nbsp;&nbsp;select array[[null, 22, 23],[null, null, 26]] as a_val<br>&nbsp;&nbsp;union all<br>&nbsp;&nbsp;select array[[31, 1.5, 1.8],[11.2, 11.5, 11.8]] as a_val<br>) tb_a_arr(a_arr);   <br><br>return:<br>&nbsp;&nbsp;array[[31, 12, 23],[14, 11.5, 16]]      </td></tr>
>>> </table>

>>
>>####   g. Redistribution function for machine learning:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_redistr_0_1          |Standardize the entire matrix from 0 to 1 intervals (linear).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. group spec.  default the whole array; |select<br>&nbsp;&nbsp;sm_sc.fv_redistr_0_1       <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.3, 25.1],[2.56, 3.25]]<br>&nbsp;&nbsp;);  <br><br>return:<br>&nbsp;&nbsp;array[[0.432121, 1.0],[0.0, 0.030612]]
>>> sm_sc.fv_redistr_ne1_1        |Standardize the entire matrix from -1 to 1 intervals (linear).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. group spec.  default the whole array; |select<br>&nbsp;&nbsp;sm_sc.fv_redistr_ne1_1     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.3, 25.1],[2.56, 3.25]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[0.0664, 0.6343],[-0.3657, -0.3351]]
>>> sm_sc.fv_redistr_l1l2ln       |Normalization of the entire matrix l1, l2... ln norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Orders of Norms; <br>&nbsp;&nbsp;3. group spec.  default the whole array; |select<br>&nbsp;&nbsp;sm_sc.fv_redistr_l1l2ln    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.3, 25.1],[2.56, 3.25]]<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[0.4353, 0.8883],[0.0906, 0.1150]]
>>> sm_sc.fv_redistr_log          |Entire matrix logarithmic (nonlinear) standardization.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. group spec.  default the whole array; |select<br>&nbsp;&nbsp;sm_sc.fv_redistr_log       <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.3, 25.1],[2.56, 3.25]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[0.7787, 1.0000],[0.2917, 0.3657]]
>>> sm_sc.fv_redistr_softmax      |Standardization of entire matrix softmax.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. group spec.  default the whole array; |select<br>&nbsp;&nbsp;sm_sc.fv_redistr_softmax   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[4.3, 2.1],[2.56, 3.25]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[0.6111, 0.0677],[0.1073, 0.2139]]
>>> sm_sc.fv_redistr_softmax_ex   |Standardization of entire matrix softmax_ex.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. group spec.  default the whole array; |select<br>&nbsp;&nbsp;sm_sc.fv_redistr_softmax_ex   <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[4.3, 2.1],[2.56, 3.25]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[0.379, 0.042],[0.067, 0.133]]
>>> sm_sc.fv_redistr_zscore       |Standardization of entire matrix z_score.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. group spec.  default the whole array; |select<br>&nbsp;&nbsp;sm_sc.fv_redistr_zscore    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.3, 25.1],[2.56, 3.25]]<br>&nbsp;&nbsp;);        <br><br>return:<br>&nbsp;&nbsp;array[[0.1424, 1.3598],[-0.7839, -0.7183]]
>>> sm_sc.fv_redistr_centralize   |Entire matrix centralization.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. group spec.  default the whole array; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_redistr_centralize<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10,&nbsp;&nbsp;20,&nbsp;&nbsp;30,&nbsp;&nbsp;40,&nbsp;&nbsp;50,&nbsp;&nbsp;60]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100,&nbsp;&nbsp;200,&nbsp;&nbsp;300,&nbsp;&nbsp;400,&nbsp;&nbsp;500,&nbsp;&nbsp;600]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-128.5,&nbsp;&nbsp;-127.5,&nbsp;&nbsp;-126.5,&nbsp;&nbsp;-125.5,&nbsp;&nbsp;-124.5,&nbsp;&nbsp;-123.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-119.5,&nbsp;&nbsp;-109.5,&nbsp;&nbsp;-99.5,&nbsp;&nbsp;-89.5,&nbsp;&nbsp;-79.5,&nbsp;&nbsp;-69.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-29.5,&nbsp;&nbsp;70.5,&nbsp;&nbsp;170.5,&nbsp;&nbsp;270.5,&nbsp;&nbsp;370.5,&nbsp;&nbsp;470.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>
>>####   h. Activation function of machine learning:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_activate_elu         |Activation function elu.<br>Input param:<br>&nbsp;&nbsp;1. Original value; <br>&nbsp;&nbsp;2. α value;   |select<br>&nbsp;&nbsp;sm_sc.fv_activate_elu               <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, -1.0],[0.0, -8.3]]<br>&nbsp;&nbsp;, 0.1<br>&nbsp;&nbsp;);    <br><br>return:<br>&nbsp;&nbsp;array[[1.2000, -0.0632],[0.0000, -0.1000]]
>>> sm_sc.fv_activate_gelu        |Activation function gelu.<br>Input param:<br>&nbsp;&nbsp;1. Original value;   |select<br>&nbsp;&nbsp;sm_sc.fv_activate_gelu              <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, -1.0],[0.0, -8.3]]<br>&nbsp;&nbsp;);             <br><br>return:<br>&nbsp;&nbsp;array[[1.0617, -0.1588],[0.0000, 0.0000]]
>>> sm_sc.fv_activate_leaky_relu  |Activation function leaky_relu.<br>Input param:<br>&nbsp;&nbsp;1. Original value; <br>&nbsp;&nbsp;2. α value;   |select<br>&nbsp;&nbsp;sm_sc.fv_activate_leaky_relu        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, -1.0],[0.0, -8.3]]<br>&nbsp;&nbsp;, 0.1<br>&nbsp;&nbsp;);       <br><br>return:<br>&nbsp;&nbsp;array[[1.2000, -0.1000],[0.0000, -0.8300]]
>>> sm_sc.fv_activate_relu        |Activation function relu.<br>Input param:<br>&nbsp;&nbsp;1. Original value;   |select<br>&nbsp;&nbsp;sm_sc.fv_activate_relu              <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, -1.0],[0.0, -8.3]]<br>&nbsp;&nbsp;);             <br><br>return:<br>&nbsp;&nbsp;array[[1.2, 0.0],[0.0, 0.0]]
>>> sm_sc.fv_activate_selu        |Activation function selu.<br>Input param:<br>&nbsp;&nbsp;1. Original value;   |select<br>&nbsp;&nbsp;sm_sc.fv_activate_selu              <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, -1.0],[0.0, -8.3]]<br>&nbsp;&nbsp;);             <br><br>return:<br>&nbsp;&nbsp;array[[1.2608, -1.1113],[0.0, -1.7577]]
>>> sm_sc.fv_activate_sigmoid     |Activation function sigmoid.<br>Input param:<br>&nbsp;&nbsp;1. Original value;   |select<br>&nbsp;&nbsp;sm_sc.fv_activate_sigmoid           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, -1.0],[0.0, -8.3]]<br>&nbsp;&nbsp;);          <br><br>return:<br>&nbsp;&nbsp;array[[0.7685, 0.2689],[0.5000, 0.0002]]
>>> sm_sc.fv_activate_softplus    |Activation function softplus.<br>Input param:<br>&nbsp;&nbsp;1. Original value;   |select<br>&nbsp;&nbsp;sm_sc.fv_activate_softplus          <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, -1.0],[0.0, -8.3]]<br>&nbsp;&nbsp;);         <br><br>return:<br>&nbsp;&nbsp;array[[1.4633, 0.3133],[0.6931, 0.0002]]
>>> sm_sc.fv_activate_swish       |Activation function swish,also known as silu.<br>Input param:<br>&nbsp;&nbsp;1. Original value;   |select<br>&nbsp;&nbsp;sm_sc.fv_activate_swish             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2, -1.0],[0.0, -8.3]]<br>&nbsp;&nbsp;);            <br><br>return:<br>&nbsp;&nbsp;array[[0.9222, -0.2689],[0.0000, -0.0021]]
>>> sm_sc.fv_activate_boxcox      |Activation function boxcox.<br>Input param:<br>&nbsp;&nbsp;1. Original value; <br>&nbsp;&nbsp;2. lambda value;   |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_activate_boxcox<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10,&nbsp;&nbsp;20,&nbsp;&nbsp;30,&nbsp;&nbsp;40,&nbsp;&nbsp;50,&nbsp;&nbsp;60]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100,&nbsp;&nbsp;200,&nbsp;&nbsp;300,&nbsp;&nbsp;400,&nbsp;&nbsp;500,&nbsp;&nbsp;600]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,1.5,4,7.5,12,17.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[49.5,199.5,449.5,799.5,1249.5,1799.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4999.5,19999.5,44999.5,79999.5,124999.5,179999.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>
>>####   i. Serializing and deserializing the json of the calculated graph:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.ft_computational_graph_deserialize       |JSON computational graph configuration or algebraic/equation text deserialization to table records.<br>This function is implemented based on temporary table sm_sc._vt_fn_compu_graph_deseri__graph.<br>Input param:<br>&nbsp;&nbsp;1. Original value; JSON calculation diagram or algebraic/equation text;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The format of the JSON calculation chart is explained as follows:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a. Curly braces and their contents:expression.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Expressions refer to variables or operational relationships composed of variables, expressions, and operational types; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;b. out_param: Dependent variable name.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The expression has only one dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;c. opr: Basic operation type name or registered custom operation type name; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;d. in_params: Independent variable name.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Its value is in square brackets and its contents include: independent variable or set of independent variables; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;e. out_value: constant value.<br><br>Output param column: <br>&nbsp;&nbsp;1. o_out_param: Intermediate output dependent variable or final dependent variable;<br>&nbsp;&nbsp;2. o_in_param: Intermediate output independent variable or final independent variable; <br>&nbsp;&nbsp;3. o_in_value: constant value; <br>&nbsp;&nbsp;4. o_param_loc: The input param of the independent variable in this step of operation:Sequence Number; <br>&nbsp;&nbsp;5. o_out_opr: Operation type name;   |-- Taking the following function expression as an example: "v_z_in = (v_x_in * v_y_in) + exp(2 * v_y_in)"<br>select * from sm_sc.ft_computational_graph_deserialize('<br>{<br>&nbsp;&nbsp;  "out_param": "v_z_in",<br>&nbsp;&nbsp;  "opr": "sm_sc.fv_opr_add",<br>&nbsp;&nbsp;  "in_params": <br>&nbsp;&nbsp;  [<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;    {<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      "opr": "sm_sc.fv_opr_mul",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      "in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      [<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;        {<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;          "out_param": "v_x_in"<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;        },<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;        {<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;          "out_param": "v_y_in"<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;        }<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      ]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;    },<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;    {<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      "opr": "sm_sc.fv_opr_exp",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      "in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      [<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;        {<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;          "opr": "sm_sc.fv_opr_mul",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;          "in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;          [<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;            {<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;              "out_value": 2 <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;            },<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;            {<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;              "out_param": "v_y_in"<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;            }<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;          ]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;        }<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;      ]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;    }<br>&nbsp;&nbsp;  ]<br>}<br>'::jsonb);<br><br>return:<br>&nbsp;&nbsp;o_out_param&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; o_in_param&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; o_in_value   o_param_loc   o_out_opr<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;v_z_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;5fba12605bdc0e173138d1053c5e2d71&nbsp;&nbsp;&nbsp;&nbsp;v_x_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;5fba12605bdc0e173138d1053c5e2d71&nbsp;&nbsp;&nbsp;&nbsp;v_y_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;85a10bcd87dd33ec3668cf8719912cb6&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2.0&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;85a10bcd87dd33ec3668cf8719912cb6&nbsp;&nbsp;&nbsp;&nbsp;v_y_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;c37766ac76959dc0ab92af50ad916b13&nbsp;&nbsp;&nbsp;&nbsp;85a10bcd87dd33ec3668cf8719912cb6&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_exp<br>&nbsp;&nbsp;v_z_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;5fba12605bdc0e173138d1053c5e2d71&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_add<br>&nbsp;&nbsp;v_z_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;c37766ac76959dc0ab92af50ad916b13&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_add<br>--------------------------<br>select * from sm_sc.ft_computational_graph_deserialize('v_z_in = (v_x_in * v_y_in) + exp(2 * v_y_in)');<br><br>return:<br>&nbsp;&nbsp;o_out_param&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; o_in_param&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; o_in_value   o_param_loc   o_out_opr<br>&nbsp;&nbsp;3ad526261da71248975a90310b0da8f7&nbsp;&nbsp;&nbsp;&nbsp;e4dd85e5f4b1451eaef35f03a5e44c5f&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_exp<br>&nbsp;&nbsp;8efff79710388489d3fba3597afa8aac&nbsp;&nbsp;&nbsp;&nbsp;v_x_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;8efff79710388489d3fba3597afa8aac&nbsp;&nbsp;&nbsp;&nbsp;v_y_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;ce2101d23207b306073a470e7832654e&nbsp;&nbsp;&nbsp;&nbsp;8efff79710388489d3fba3597afa8aac&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_add<br>&nbsp;&nbsp;ce2101d23207b306073a470e7832654e&nbsp;&nbsp;&nbsp;&nbsp;3ad526261da71248975a90310b0da8f7&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_add<br>&nbsp;&nbsp;e4dd85e5f4b1451eaef35f03a5e44c5f&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2.0&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;e4dd85e5f4b1451eaef35f03a5e44c5f&nbsp;&nbsp;&nbsp;&nbsp;v_y_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;v_z_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;ce2101d23207b306073a470e7832654e&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;
>>
>>####   j. Convolutional, pooling, matrix multicast for machine learning:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_conv_2d_im2col_py    |Matrix convolution (im2col method).<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Convolutional kernel matrix; <br>&nbsp;&nbsp;3. Convolutional kernel offset; <br>&nbsp;&nbsp;4. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;6. Fill in the element values; |select <br>&nbsp;&nbsp;  sm_sc.fv_conv_2d_im2col_py<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100,200,300,400,500,600]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10,-20,-30,-40,-50,-60]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;, array[[1.1, 1.1, 1.1],[1.1, 2.1, 1.1],[1.1, 1.1, 1.1]]<br>&nbsp;&nbsp;, array[0.0]<br>&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;, array[1, 1, 1, 0]<br>&nbsp;&nbsp;, 0.0<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[37.3000, 111.9000, 186.5000]<br>&nbsp;&nbsp;,[459.7000, 1379.1000, 2298.5000]<br>&nbsp;&nbsp;,[-46.3000, -138.9000, -231.5000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_none<br>sm_sc.fv_pool_none_py           |Element remains unchanged pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Pooled window height and width size; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select&nbsp;&nbsp;sm_sc.fv_pool_none<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,3.0,4.0,5.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,30.0,40.0,50.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,300.0,400.0,500.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,300.0,400.0,500.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-3.0,-4.0,-5.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-30.0,-40.0,-50.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_pool_avg<br>sm_sc.fv_pool_avg_py            |Average pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Pooled window height and width size; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_avg<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[3.6667, 11.0000, 18.3333, 23.8333]<br>&nbsp;&nbsp;,[36.3333, 109.0000, 181.6667, 236.1667]<br>&nbsp;&nbsp;,[-5.5000, -16.5000, -27.5000, -35.7500]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_max<br>sm_sc.fv_pool_max_py            |Maximum pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Size of pooling window height and width; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_max<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[300.0000, 500.0000, 700.0000]<br>&nbsp;&nbsp;,[300.0000, 500.0000, 700.0000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_min<br>sm_sc.fv_pool_min_py            |Minimum pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Size of pooling window height and width; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_min<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.0000, 3.0000, 5.0000]<br>&nbsp;&nbsp;,[-30.0000, -50.0000, -70.0000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_sum<br>sm_sc.fv_pool_sum_py            |Total sum pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Size of pooling window height and width; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_sum<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[666.0000, 1332.0000, 1998.0000]<br>&nbsp;&nbsp;,[534.0000, 1068.0000, 1602.0000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_ptp            |Span pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Size of pooling window height and width; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_ptp<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[299.0000, 497.0000, 695.0000]<br>&nbsp;&nbsp;,[330.0000, 550.0000, 770.0000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_prod<br>sm_sc.fv_pool_prod_py           |Direct product pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Size of pooling window height and width; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_prod<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[216000000000, 216000000000000, 9261000000000000]<br>&nbsp;&nbsp;,[216000000000, 216000000000000, 9261000000000000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_mode           |Mode pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Size of pooling window height and width; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_mode<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.0000, 3.0000, 5.0000]<br>&nbsp;&nbsp;,[-30.0000, -50.0000, -70.0000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_median         |Median pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Size of pooling window height and width; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_median<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[20.0000, 40.0000, 60.0000]<br>&nbsp;&nbsp;,[-2.0000, -4.0000, -6.0000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_coalesce       |Null replace pooling.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix; <br>&nbsp;&nbsp;2. Size of pooling window height and width; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select sm_sc.fv_pool_coalesce<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[null,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10.0,20.0,null,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100.0,200.0,300.0,400.0,500.0,null,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1.0,null,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,null]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[2.0, 3.0, 5.0]<br>&nbsp;&nbsp;,[100.0, 300.0, 500.0]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_conv_2d_grp_x       |Matrix x direction convolution.<br>Support offset.<br>Input param:<br>&nbsp;&nbsp;1. The 2D original matrix composed of each row of vectors converted from 2D to 1D again; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, 100 images with 28\*14 pixels,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;First, convert 2D pixels into a 1D vector with a length of 392,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Reaggregate to synthesize a 2D matrix with a height and width of 100\*392; <br>&nbsp;&nbsp;2. Width of 2D samples.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Usually the width of the image, which is 14 in the param 1 example; <br>&nbsp;&nbsp;3. Single row matrix after dimensionality reduction with convolutional kernel.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If there is an offset, then the length of the one-dimensional array is 1 more.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, after reducing the dimensionality of a 3\*4 convolutional kernel matrix,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If there is no offset, it is a 1\*12 matrix,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If there is an offset, it is 1\*12 + 1 = 13 matrix; <br>&nbsp;&nbsp;4. The width of the convolutional kernel window.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, 4 in the parameter three examples; <br>&nbsp;&nbsp;5. Vertical and horizontal step sizes of convolutional kernel sliding.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;6. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;7. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_conv_2d_grp_x<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, 6<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[[1.1, 1.1, 1.1, 1.1, 2.1, 1.1, 1.1, 1.1, 1.1, 0.5]]<br>&nbsp;&nbsp;&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[1, 1, 1, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;, 0<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[37.30,111.90,186.50,459.70,1379.10,2298.50,-46.30,-138.90,-231.50]<br>&nbsp;&nbsp;,[1.00,3.00,5.00,369.70,1109.10,1848.50,287.00,861.00,1435.00]<br>&nbsp;&nbsp;,[1.00,3.00,5.00,6.70,20.10,33.50,397.00,1191.00,1985.00]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_avg_2d_grp_x   |Average pooling in the x-direction of the matrix.<br>Input param:<br>&nbsp;&nbsp;1. The 2D original matrix composed of each row of vectors converted from 2D to 1D again; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, 100 images with 28\*14 pixels,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;First, convert 2D pixels into a 1D vector with a length of 392,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Reaggregate to synthesize a 2D matrix with a height and width of 100\*392; <br>&nbsp;&nbsp;2. Width of 2D samples.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Usually the width of the image, which is 14 in the param 1 example; <br>&nbsp;&nbsp;3. Size of pooling window height and width.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Longitudinal and transverse step sizes of pooling sliding.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;6. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_pool_avg_2d_grp_x<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0,70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0,700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0,70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0,700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0,70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0,700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, 7<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[3.67,11.00,18.33,23.83,36.33,109.00,181.67,236.17,-5.50,-16.50,-27.50,-35.75]<br>&nbsp;&nbsp;,[0.00,0.00,0.00,0.00,36.33,109.00,181.67,236.17,45.00,135.00,225.00,292.50]<br>&nbsp;&nbsp;,[0.00,0.00,0.00,0.00,-0.33,-1.00,-1.67,-2.17,45.00,135.00,225.00,292.50]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_pool_max_2d_grp_x   |Pooling of maximum values in the x direction of a matrix.<br>Input param:<br>&nbsp;&nbsp;1. The 2D original matrix composed of each row of vectors converted from 2D to 1D again; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, 100 images with 28\*14 pixels,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;First, convert 2D pixels into a 1D vector with a length of 392,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Reaggregate to synthesize a 2D matrix with a height and width of 100\*392; <br>&nbsp;&nbsp;2. Width of 2D samples.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Usually the width of the image, which is 14 in the param 1 example; <br>&nbsp;&nbsp;3. Size of pooling window height and width.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Longitudinal and transverse step sizes of pooling sliding.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;6. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_pool_max_2d_grp_x<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0,70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0,700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0,70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0,700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 10.0,20.0,30.0,40.0,50.0,60.0,70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, -10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;, 100.0,200.0,300.0,400.0,500.0,600.0,700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;, 7<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[3, 3]<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[300.0,500.0,700.0,300.0,500.0,700.0]<br>&nbsp;&nbsp;,[30.0,50.0,70.0,300.0,500.0,700.0]<br>&nbsp;&nbsp;,[30.0,50.0,70.0,300.0,500.0,700.0]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_conv_add            |Matrix multicast - convolute add.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_conv_add<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100,200,300,400,500,600]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10,-20,-30,-40,-50,-60]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;, array[[1.1, 1.1, 1.1],[1.1, 2.1, 1.1],[1.1, 1.1, 1.1]]<br>&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;, array[1, 1, 1, 0]<br>&nbsp;&nbsp;, 0.0<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1.1,1.1,1.1,1.1,1.1,1.1,1.1,1.1,1.1]<br>&nbsp;&nbsp;,[1.1,3.1,3.1,3.1,5.1,5.1,5.1,7.1,7.1]<br>&nbsp;&nbsp;,[1.1,11.1,21.1,21.1,31.1,41.1,41.1,51.1,61.1]<br>&nbsp;&nbsp;,[1.1,11.1,21.1,21.1,31.1,41.1,41.1,51.1,61.1]<br>&nbsp;&nbsp;,[1.1,102.1,201.1,201.1,302.1,401.1,401.1,502.1,601.1]<br>&nbsp;&nbsp;,[1.1,0.1,-0.9,-0.9,-1.9,-2.9,-2.9,-3.9,-4.9]<br>&nbsp;&nbsp;,[1.1,0.1,-0.9,-0.9,-1.9,-2.9,-2.9,-3.9,-4.9]<br>&nbsp;&nbsp;,[1.1,-7.9,-18.9,-18.9,-27.9,-38.9,-38.9,-47.9,-58.9]<br>&nbsp;&nbsp;,[1.1,1.1,1.1,1.1,1.1,1.1,1.1,1.1,1.1]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_conv_sub            |Matrix multicast - convolute sub.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_conv_sub<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100,200,300,400,500,600]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10,-20,-30,-40,-50,-60]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;, array[[1.1, 1.1, 1.1],[1.1, 2.1, 1.1],[1.1, 1.1, 1.1]]<br>&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;, array[1, 1, 1, 0]<br>&nbsp;&nbsp;, 0.0<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[-1.1,-1.1,-1.1,-1.1,-1.1,-1.1,-1.1,-1.1,-1.1]<br>&nbsp;&nbsp;,[-1.1,-1.1,0.9,0.9,0.9,2.9,2.9,2.9,4.9]<br>&nbsp;&nbsp;,[-1.1,8.9,18.9,18.9,28.9,38.9,38.9,48.9,58.9]<br>&nbsp;&nbsp;,[-1.1,8.9,18.9,18.9,28.9,38.9,38.9,48.9,58.9]<br>&nbsp;&nbsp;,[-1.1,97.9,198.9,198.9,297.9,398.9,398.9,497.9,598.9]<br>&nbsp;&nbsp;,[-1.1,-2.1,-3.1,-3.1,-4.1,-5.1,-5.1,-6.1,-7.1]<br>&nbsp;&nbsp;,[-1.1,-2.1,-3.1,-3.1,-4.1,-5.1,-5.1,-6.1,-7.1]<br>&nbsp;&nbsp;,[-1.1,-12.1,-21.1,-21.1,-32.1,-41.1,-41.1,-52.1,-61.1]<br>&nbsp;&nbsp;,[-1.1,-1.1,-1.1,-1.1,-1.1,-1.1,-1.1,-1.1,-1.1]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_conv_mul            |Matrix multicast - convolute mul.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_conv_mul<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100,200,300,400,500,600]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10,-20,-30,-40,-50,-60]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;, array[[1.1, 1.1, 1.1],[1.1, 2.1, 1.1],[1.1, 1.1, 1.1]]<br>&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;, array[1, 1, 1, 0]<br>&nbsp;&nbsp;, 0.0<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00]<br>&nbsp;&nbsp;,[0.00,2.10,2.2,2.2,6.3,4.4,4.4,10.5,6.6]<br>&nbsp;&nbsp;,[0.00,11.0,22.0,22.0,33.0,44.0,44.0,55.0,66.0]<br>&nbsp;&nbsp;,[0.00,11.0,22.0,22.0,33.0,44.0,44.0,55.0,66.0]<br>&nbsp;&nbsp;,[0.00,210.0,220.0,220.0,630.0,440.0,440.0,1050.0,660.0]<br>&nbsp;&nbsp;,[0.00,-1.1,-2.2,-2.2,-3.3,-4.4,-4.4,-5.5,-6.6]<br>&nbsp;&nbsp;,[0.00,-1.1,-2.2,-2.2,-3.3,-4.4,-4.4,-5.5,-6.6]<br>&nbsp;&nbsp;,[0.00,-21.0,-22.0,-22.0,-63.0,-44.0,-44.0,-105.0,-66.0]<br>&nbsp;&nbsp;,[0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_conv_div            |Matrix multicast - convolute div.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_conv_div<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[100,200,300,400,500,600]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10,-20,-30,-40,-50,-60]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;, array[[1.1, 1.1, 1.1],[1.1, 2.1, 1.1],[1.1, 1.1, 1.1]]<br>&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;, array[1, 1, 1, 0]<br>&nbsp;&nbsp;, 0.0<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000]<br>&nbsp;&nbsp;,[0.000,0.476,1.818,1.818,1.429,3.636,3.636,2.381,5.455]<br>&nbsp;&nbsp;,[0.000,9.091,18.182,18.182,27.273,36.364,36.364,45.455,54.545]<br>&nbsp;&nbsp;,[0.000,9.091,18.182,18.182,27.273,36.364,36.364,45.455,54.545]<br>&nbsp;&nbsp;,[0.000,47.619,181.818,181.818,142.857,363.636,363.636,238.095,545.455]<br>&nbsp;&nbsp;,[0.000,-0.909,-1.818,-1.818,-2.727,-3.636,-3.636,-4.545,-5.455]<br>&nbsp;&nbsp;,[0.000,-0.909,-1.818,-1.818,-2.727,-3.636,-3.636,-4.545,-5.455]<br>&nbsp;&nbsp;,[0.000,-4.762,-18.182,-18.182,-14.286,-36.364,-36.364,-23.810,-54.545]<br>&nbsp;&nbsp;,[0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_conv_pow            |Matrix multicast - convolute pow.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_conv_pow<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10,-20,-30,-40,-50,-60]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-1,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[-10,-20,-30,-40,-50,-60]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;, array[[1.0, 2.0, 3.0],[2.0, 3.0, 1.0],[1.0, 3.0, 2.0]]<br>&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;, array[1, 1, 1, 0]<br>&nbsp;&nbsp;, 0.0<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00]<br>&nbsp;&nbsp;,[0.00,1.00,2.00,4.00,27.00,4.00,16.00,125.00,6.00]<br>&nbsp;&nbsp;,[0.00,1000.00,400.00,20.00,27000.00,1600.00,40.00,125000.00,3600.00]<br>&nbsp;&nbsp;,[0.00,100.00,8000.00,20.00,900.00,64000.00,40.00,2500.00,216000.00]<br>&nbsp;&nbsp;,[0.00,-1000.00,-20.00,400.00,-27000.00,-40.00,1600.00,-125000.00,-60.00]<br>&nbsp;&nbsp;,[0.00,-1.00,4.00,-2.00,-27.00,16.00,-4.00,-125.00,36.00]<br>&nbsp;&nbsp;,[0.00,1.00,-8.00,-2.00,9.00,-64.00,-4.00,25.00,-216.00]<br>&nbsp;&nbsp;,[0.00,-1000.00,-20.00,400.00,-27000.00,-40.00,1600.00,-125000.00,-60.00]<br>&nbsp;&nbsp;,[0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_conv_log            |Matrix multicast - convolute log.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select <br>&nbsp;&nbsp;  sm_sc.fv_conv_log<br>&nbsp;&nbsp;  (<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.0,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;, array[[1.0, 2.0, 3.0],[2.0, 3.0, 1.0],[1.0, 3.0, 2.0]]<br>&nbsp;&nbsp;, array[2, 2]<br>&nbsp;&nbsp;, array[1, 1, 1, 0]<br>&nbsp;&nbsp;, 2.0<br>&nbsp;&nbsp;  );<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[0.00,1.00,1.58,0.00,1.00,1.58,0.00,1.00,1.58]<br>&nbsp;&nbsp;,[1.00,1.58,0.00,1.00,1.00,0.00,0.50,0.68,0.00]<br>&nbsp;&nbsp;,[0.00,0.48,0.23,0.00,0.32,0.19,0.00,0.28,0.17]<br>&nbsp;&nbsp;,[0.00,0.30,0.37,0.00,0.20,0.30,0.00,0.18,0.27]<br>&nbsp;&nbsp;,[1.00,0.48,0.00,0.23,0.32,0.00,0.19,0.28,0.00]<br>&nbsp;&nbsp;,[0.00,1.58,1.00,0.00,1.58,1.00,0.00,1.58,1.00]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_conv_de_sub         |Matrix multicast - deconvolute sub.<br>The window is the first input parameter, and the original matrix is the second input parameter.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select&nbsp;&nbsp;sm_sc.fv_conv_de_sub<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.5,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[1.4,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.7]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6,6.2,7.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3,6.3,7.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4,6.4,7.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5,6.5,7.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.4,0.0,0.0,-1.5,-2.0,-2.0,-3.5,-4.0,-4.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.2,-0.2,-0.3,-1.9,-2.1,-2.6,-4.2,-4.2,-4.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.7,-0.3,-1.6,-0.3,-2.3,-3.6,-2.3,-4.3,-5.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.2,-0.3,-0.3,-1.8,-2.3,-2.3,-3.8,-4.3,-4.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.0,-0.4,-0.4,-2.0,-2.4,-2.4,-4.0,-4.4,-4.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,-0.5,-1.8,-0.5,-2.5,-3.8,-2.5,-4.5,-5.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_conv_de_div         |Matrix multicast - deconvolute div.<br>The window is the first input parameter, and the original matrix is the second input parameter.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select&nbsp;&nbsp;sm_sc.fv_conv_de_div<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.5,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[1.4,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.7]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6,6.2,7.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3,6.3,7.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4,6.4,7.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5,6.5,7.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.364,1.000,1.000,0.500,0.500,0.600,0.300,0.333,0.429]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.167,0.909,0.909,0.424,0.488,0.536,0.250,0.323,0.405]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.308,0.870,0.515,0.909,0.465,0.321,0.566,0.317,0.233]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.154,0.870,0.909,0.455,0.465,0.566,0.283,0.317,0.411]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.000,0.833,0.882,0.412,0.455,0.556,0.259,0.313,0.405]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.000,0.800,0.486,0.857,0.444,0.309,0.545,0.308,0.227]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_conv_de_pow         |Matrix multicast - deconvolute pow.<br>The window is the first input parameter, and the original matrix is the second input parameter.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_conv_de_pow<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],[2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;1.0],[1.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;2.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1,-2,-3,-4,-5,-6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.1,-0.2,-0.3,-0.4,-0.5,-0.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[1,&nbsp;&nbsp;1,&nbsp;&nbsp;1,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;0.0<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.000,1.000,1.000,1.000,1.000,1.000,1.000,1.000,1.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.000,3.000,1.000,4.000,27.000,1.000,16.000,243.000,1.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.000,0.333,0.250,1.000,0.037,0.063,1.000,0.004,0.016]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_conv_de_log         |Matrix multicast - deconvolute log.<br>The window is the first input parameter, and the original matrix is the second input parameter.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;5. Fill in the element values;  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_conv_de_log<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.5,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],[2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;1.5],[1.5,&nbsp;&nbsp;3.0,&nbsp;&nbsp;2.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.0,2,3,4,5,6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10,20,30,40,50,60]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[1,&nbsp;&nbsp;1,&nbsp;&nbsp;1,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2.0<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.710,1.000,0.631,1.710,1.000,0.631,1.710,1.000,0.631]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.000,0.631,1.710,1.000,1.000,3.419,2.000,1.465,4.419]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.710,2.096,4.322,7.388,3.096,5.322,9.098,3.561,5.907]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_conv_prod_mx        |Matrix multicast - convolute matrix multiplication.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Background matrix; <br>&nbsp;&nbsp;2.&nbsp;Window matrix; <br>&nbsp;&nbsp;3.&nbsp;Background matrix sliding window height and width specifications.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This window specification is different from the second independent variable height width specification, and the two are matrix multiplication relationships; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;6.&nbsp;Fill in the element values; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_conv_prod_mx<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-3.0,&nbsp;&nbsp;2.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[1,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,-4,1,-1,-5,0,-1,-6,-1,-1,-7,-2,-1,-8,-3,-1,-9,-4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10,-40,10,-10,-50,0,-10,-60,-10,-10,-70,-20,-10,-80,-30,-10,-90,-40]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10,-40,10,-10,-50,0,-10,-60,-10,-10,-70,-20,-10,-80,-30,-10,-90,-40]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-100,-400,100,-100,-500,0,-100,-600,-100,-100,-700,-200,-100,-800,-300,-100,-900,-400]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-100,-400,100,-100,-500,0,-100,-600,-100,-100,-700,-200,-100,-800,-300,-100,-900,-400]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,4,-1,1,5,0,1,6,1,1,7,2,1,8,3,1,9,4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,4,-1,1,5,0,1,6,1,1,7,2,1,8,3,1,9,4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10,40,-10,10,50,0,10,60,10,10,70,20,10,80,30,10,90,40]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_conv_de_prod_mx     |Matrix multicast - deconvolute matrix multiplication.<br>The window is the first input parameter, and the original matrix is the second input parameter.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Window matrix; <br>&nbsp;&nbsp;2.&nbsp;Background matrix; <br>&nbsp;&nbsp;3.&nbsp;Background matrix sliding window height and width specifications.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This window specification is different from the first independent variable height width specification, and the two are matrix multiplication relationships; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;6.&nbsp;Fill in the element values; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_conv_de_prod_mx<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0],&nbsp;&nbsp;[-3.0,&nbsp;&nbsp;-1.0],&nbsp;&nbsp;[-3.0,&nbsp;&nbsp;2.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,42,42,63,63,84,84,105,105,126,126,147]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-13,-26,-26,-39,-39,-52,-52,-65,-65,-78,-78,-91]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[17,34,34,51,51,68,68,85,85,102,102,119]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[210,420,420,630,630,840,840,1050,1050,1260,1260,1470]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-130,-260,-260,-390,-390,-520,-520,-650,-650,-780,-780,-910]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[170,340,340,510,510,680,680,850,850,1020,1020,1190]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[98,196,196,294,294,392,392,490,490,588,588,686]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-299,-598,-598,-897,-897,-1196,-1196,-1495,-1495,-1794,-1794,-2093]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-302,-604,-604,-906,-906,-1208,-1208,-1510,-1510,-1812,-1812,-2114]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-21,-42,-42,-63,-63,-84,-84,-105,-105,-126,-126,-147]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[13,26,26,39,39,52,52,65,65,78,78,91]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-17,-34,-34,-51,-51,-68,-68,-85,-85,-102,-102,-119]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_add_stride_1        |Matrix multicast - convolute add with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;                  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_add_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.0,4.0,6.0,3.0,5.0,7.0,4.0,6.0,8.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[9.0,18.0,27.0,19.0,28.0,37.0,29.0,38.0,47.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[103.0,198.0,301.0,203.0,298.0,401.0,303.0,398.0,501.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11.0,22.0,33.0,21.0,32.0,43.0,31.0,42.0,53.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[99.0,198.0,297.0,199.0,298.0,397.0,299.0,398.0,497.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.0,-4.0,-2.0,1.0,-5.0,-3.0,0.0,-6.0,-4.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_sub_stride_1        |Matrix multicast - convolute sub with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;                  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_sub_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.0,0.0,0.0,1.0,1.0,1.0,2.0,2.0,2.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11.0,22.0,33.0,21.0,32.0,43.0,31.0,42.0,53.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[97.0,202.0,299.0,197.0,302.0,399.0,297.0,402.0,499.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[9.0,18.0,27.0,19.0,28.0,37.0,29.0,38.0,47.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[101.0,202.0,303.0,201.0,302.0,403.0,301.0,402.0,503.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-4.0,0.0,-4.0,-5.0,-1.0,-5.0,-6.0,-2.0,-6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_de_sub_stride_1     |Matrix multicast - deconvolute sub with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;                |select&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_de_sub_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;&nbsp;&nbsp;2.0,&nbsp;&nbsp;&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;&nbsp;&nbsp;-2.0,&nbsp;&nbsp;&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;&nbsp;&nbsp;-2.0,&nbsp;&nbsp;&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.0,0.0,0.0,-1.0,-1.0,-1.0,-2.0,-2.0,-2.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-11.0,-22.0,-33.0,-21.0,-32.0,-43.0,-31.0,-42.0,-53.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-97.0,-202.0,-299.0,-197.0,-302.0,-399.0,-297.0,-402.0,-499.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-9.0,-18.0,-27.0,-19.0,-28.0,-37.0,-29.0,-38.0,-47.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-101.0,-202.0,-303.0,-201.0,-302.0,-403.0,-301.0,-402.0,-503.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4.0,0.0,4.0,5.0,1.0,5.0,6.0,2.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_mul_stride_1        |Matrix multicast - convolute mul with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;                  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_mul_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.00,4.00,9.00,2.00,6.00,12.00,3.00,8.00,15.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.00,-40.00,-90.00,-20.00,-60.00,-120.00,-30.00,-80.00,-150.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[300.00,-400.00,300.00,600.00,-600.00,400.00,900.00,-800.00,500.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.00,40.00,90.00,20.00,60.00,120.00,30.00,80.00,150.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-100.00,-400.00,-900.00,-200.00,-600.00,-1200.00,-300.00,-800.00,-1500.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3.00,4.00,-3.00,-6.00,6.00,-4.00,-9.00,8.00,-5.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_div_stride_1        |Matrix multicast - convolute div with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;                  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_div_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.000,1.000,1.000,2.000,1.500,1.333,3.000,2.000,1.667]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.000,-10.000,-10.000,-20.000,-15.000,-13.333,-30.000,-20.000,-16.667]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[33.333,-100.000,300.000,66.667,-150.000,400.000,100.000,-200.000,500.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.000,10.000,10.000,20.000,15.000,13.333,30.000,20.000,16.667]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-100.000,-100.000,-100.000,-200.000,-150.000,-133.333,-300.000,-200.000,-166.667]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.333,1.000,-3.000,-0.667,1.500,-4.000,-1.000,2.000,-5.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_de_div_stride_1     |Matrix multicast - deconvolute div with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;                |select&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_de_div_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.000,1.000,1.000,0.500,0.667,0.750,0.333,0.500,0.600]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.100,-0.100,-0.100,-0.050,-0.067,-0.075,-0.033,-0.050,-0.060]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.030,-0.010,0.003,0.015,-0.007,0.003,0.010,-0.005,0.002]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.100,0.100,0.100,0.050,0.067,0.075,0.033,0.050,0.060]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.010,-0.010,-0.010,-0.005,-0.007,-0.008,-0.003,-0.005,-0.006]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3.000,1.000,-0.333,-1.500,0.667,-0.250,-1.000,0.500,-0.200]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_pow_stride_1        |Matrix multicast - convolute pow with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;                  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_pow_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.100,4.000,27.000,2.000,9.000,64.000,3.000,16.000,125.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.833,0.207,0.028,0.455,0.092,0.015,0.303,0.059,0.006]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.197,0.189,3.300,12.167,0.092,4.300,35.937,0.054,5.300]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.200,4.840,35.937,2.200,10.890,68.921,3.300,16.810,175.616]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.769,0.189,0.028,0.435,0.092,0.013,0.303,0.054,0.007]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.744,0.174,3.400,13.824,0.087,4.400,39.304,0.052,5.400]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.300,5.290,35.937,2.300,10.890,79.507,3.300,18.490,148.877]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.714,0.174,0.025,0.417,0.087,0.012,0.294,0.052,0.006]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.375,0.160,3.500,15.625,0.082,4.500,42.875,0.049,5.500]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_de_pow_stride_1     |Matrix multicast - deconvolute pow with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;                |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_de_pow_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,-3.0,5.0,-6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,-3.3,5.6,-6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,-3.3,5.3,-6.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,-3.4,5.4,-6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,-3.5,5.5,-6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{1.000,4.000,0.037,1.000,0.125,243.000,1.000,32.000,0.001}<br>&nbsp;&nbsp;&nbsp;&nbsp;,{1.000,4.595,0.027,1.000,0.102,469.763,1.000,48.503,0.001}<br>&nbsp;&nbsp;&nbsp;&nbsp;,{4.171,4.925,1.000,12.514,0.102,1.000,0.027,39.397,1.000}<br>&nbsp;&nbsp;&nbsp;&nbsp;,{1.000,4.595,0.027,1.000,0.102,469.763,1.000,48.503,0.001}<br>&nbsp;&nbsp;&nbsp;&nbsp;,{1.000,4.925,0.027,1.000,0.102,337.865,1.000,39.397,0.001}<br>&nbsp;&nbsp;&nbsp;&nbsp;,{4.656,5.278,1.000,13.967,0.095,1.000,0.024,42.224,1.000}<br>&nbsp;&nbsp;&nbsp;&nbsp;,{1.000,4.925,0.027,1.000,0.102,337.865,1.000,39.397,0.001}<br>&nbsp;&nbsp;&nbsp;&nbsp;,{1.000,5.278,0.024,1.000,0.095,377.098,1.000,42.224,0.001}<br>&nbsp;&nbsp;&nbsp;&nbsp;,{5.196,5.657,1.000,15.588,0.088,1.000,0.021,45.255,1.000}<br>&nbsp;&nbsp;&nbsp;&nbsp;}
>>> sm_sc.fv_opr_conv_log_stride_1        |Matrix multicast - convolute log with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;                |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_log_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.000,1.000,1.000,0.000,0.631,0.792,0.000,0.500,0.683]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.000,0.879,0.920,0.000,0.581,0.779,0.000,0.491,0.638]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4.187,0.832,0.000,1.319,0.581,0.000,0.920,0.475,0.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.000,0.879,0.920,0.000,0.581,0.779,0.000,0.491,0.638]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.000,0.832,0.920,0.000,0.581,0.753,0.000,0.475,0.659]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.265,0.792,0.000,1.255,0.566,0.000,0.898,0.468,0.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.000,0.832,0.920,0.000,0.581,0.753,0.000,0.475,0.659]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.000,0.792,0.898,0.000,0.566,0.742,0.000,0.468,0.651]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.710,0.756,0.000,1.199,0.553,0.000,0.877,0.461,0.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_de_log_stride_1     |Matrix multicast - deconvolute log with stride 1.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_de_log_stride_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.5,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[1.8,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.2]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.235,1.000,1.000,1.710,1.585,1.262,2.710,2.000,1.465]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.310,1.138,1.087,1.341,1.722,1.284,2.031,2.036,1.568]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.239,1.202,6.548,0.758,1.722,8.000,1.087,2.104,9.147]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.450,1.138,1.087,1.945,1.722,1.284,2.945,2.036,1.568]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.446,1.202,1.087,1.417,1.722,1.328,2.031,2.104,1.518]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.306,1.263,6.712,0.797,1.766,8.126,1.114,2.138,9.250]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.647,1.202,1.087,2.054,1.722,1.328,2.945,2.104,1.518]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.572,1.263,1.114,1.489,1.766,1.349,2.082,2.138,1.535]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.369,1.322,6.871,0.834,1.807,8.250,1.140,2.170,9.350]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_add_stride_window        |Matrix multicast - convolute add with stride = window\'s length.<br>Co parameter default, step size is 1.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;                |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_add_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.0,4.0,6.0,5.0,7.0,9.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[9.0,18.0,27.0,39.0,48.0,57.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[103.0,198.0,301.0,403.0,498.0,601.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.0,0.0,0.0,-3.0,-3.0,-3.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-11.0,-22.0,-33.0,-41.0,-52.0,-63.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-7.0,-22.0,-29.0,-37.0,-52.0,-59.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_sub_stride_window        |Matrix multicast - convolute sub with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;      |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_sub_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.0,0.0,0.0,3.0,3.0,3.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11.0,22.0,33.0,41.0,52.0,63.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[97.0,202.0,299.0,397.0,502.0,599.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.0,-4.0,-6.0,-5.0,-7.0,-9.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-9.0,-18.0,-27.0,-39.0,-48.0,-57.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-13.0,-18.0,-31.0,-43.0,-48.0,-61.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_de_sub_stride_window     |Matrix multicast - deconvolute sub with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;    |select&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_de_sub_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6,6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3,6.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4,6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5,6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5,6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-0.1,0.0,0.0,-3.0,-3.0,-3.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.2,-4.2,-6.3,-5.1,-7.6,-9.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.7,-4.3,-2.3,-1.3,-7.3,-5.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.4,-0.4,-0.4,-3.4,-3.4,-3.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.5,-4.5,-6.5,-5.5,-7.5,-9.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,-4.5,-2.5,-1.5,-7.5,-5.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_mul_stride_window        |Matrix multicast - convolute mul with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;      |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_mul_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.00,4.00,9.00,4.00,10.00,18.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.00,-40.00,-90.00,-40.00,-100.00,-180.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[300.00,-400.00,300.00,1200.00,-1000.00,600.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.00,-4.00,-9.00,-4.00,-10.00,-18.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.00,40.00,90.00,40.00,100.00,180.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-30.00,40.00,-30.00,-120.00,100.00,-60.00]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_div_stride_window        |Matrix multicast - convolute div with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;      |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_div_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.000,1.000,1.000,4.000,2.500,2.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.000,-10.000,-10.000,-40.000,-25.000,-20.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[33.333,-100.000,300.000,133.333,-250.000,600.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.000,-1.000,-1.000,-4.000,-2.500,-2.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.000,10.000,10.000,40.000,25.000,20.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-3.333,10.000,-30.000,-13.333,25.000,-60.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_de_div_stride_window     |Matrix multicast - deconvolute div with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;    |select&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_de_div_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6,6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3,6.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4,6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5,6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5,6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.909,1.000,1.000,0.250,0.400,0.500]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.833,-0.909,-0.909,-0.244,-0.357,-0.484]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.308,-0.870,0.303,0.698,-0.377,0.159]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.714,0.833,0.882,0.227,0.370,0.469]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.667,-0.800,-0.857,-0.222,-0.364,-0.462]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.000,-0.800,0.286,0.667,-0.364,0.154]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_pow_stride_window        |Matrix multicast - convolute pow with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;      |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_pow_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6,6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3,6.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4,6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4,6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5,6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[-1.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;-3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;-2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.100,4.000,27.000,4.000,25.000,216.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.833,0.207,0.028,0.244,0.032,0.004]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.197,0.189,3.300,79.507,0.036,6.300]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.400,5.760,39.304,4.400,29.160,262.144]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.714,0.174,0.025,0.227,0.034,0.004]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3.375,0.160,3.500,91.125,0.033,6.500]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_de_pow_stride_window     |Matrix multicast - deconvolute pow with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;    |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_de_pow_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,-5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,-5.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,-5.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,-5.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,-5.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,-5.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.000,4.000,27.000,1.000,0.031,81.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.000,4.595,37.541,1.000,0.021,90.406]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4.171,4.925,1.000,112.622,0.025,1.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.000,5.278,41.900,1.000,0.024,125.699]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.000,5.657,46.765,1.000,0.022,140.296]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[5.196,5.657,1.000,140.296,0.022,1.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_log_stride_window        |Matrix multicast - convolute log with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;    |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_log_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,4.0,5.0,6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,4.1,5.6,6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,4.3,5.3,6.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4,6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,4.4,5.4,6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,4.5,5.5,6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.000,1.000,1.000,0.000,0.431,0.613]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.000,0.879,0.920,0.000,0.402,0.602]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4.187,0.832,0.000,0.753,0.416,0.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.000,0.792,0.898,0.000,0.411,0.592]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.000,0.792,0.898,0.000,0.411,0.592]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.710,0.756,0.000,0.730,0.407,0.000]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_opr_conv_de_log_stride_window     |Matrix multicast - deconvolute log with stride = window\'s length.<br>Co parameter default, step size is the same as the window specification.<br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;  |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_conv_de_log_stride_window<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[1.2,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0],&nbsp;&nbsp;[3.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;1.5]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,2.0,3.0,0.9,5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,2.2,3.3,0.1,5.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,2.3,3.3,0.3,5.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,2.4,3.4,0.4,5.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,0.5,5.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.5,2.5,3.5,0.5,5.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.363,1.000,1.000,-0.402,2.322,-0.096]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.000,1.138,1.087,-12.629,2.485,-2.096]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.239,1.202,2.945,-1.096,2.406,-2.969]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.282,1.263,1.114,-3.492,2.433,-0.834]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.224,1.322,1.140,-3.802,2.459,-0.631]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0.369,1.322,3.090,-0.631,2.459,-1.710]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>
>>####   k. Lambda for machine learning:
>>
>>> Function Name | Function Description | Call Example
>>> :-|:-|:-
>>> sm_sc.fv_algebra_execute             |For basic or custom operations, calculate numerical values using algebraic or functional formulas.<br>Overload 1: <br>&nbsp;&nbsp;Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Algebraic expression; <br>Overload 2: <br>&nbsp;&nbsp;Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Function name; <br>&nbsp;&nbsp;&nbsp;&nbsp;1. Input param list; |select<br>&nbsp;&nbsp;sm_sc.fv_algebra_execute               <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;'2.5 + 3.6'<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;6.1<br>--------------------------<br>select<br>&nbsp;&nbsp;sm_sc.fv_algebra_execute               <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;'sm_sc.fv_opr_mul'<br>&nbsp;&nbsp;, array[1.5, 2.0, 3.4]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;10.2
>>> sm_sc.fv_lambda_arr<br>sm_sc.fv_lambda_arr_p                  |Forward propagation node output dependent variable lambda.<br>Input param:<br>&nbsp;&nbsp;1. The sequence number of neural network nodes.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter does not participate in actual calculations and is only used for control, debugging, tracking, and logging; <br>&nbsp;&nbsp;2. Lambda naming, i.e. operation type.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The scope of support can be found in chapter 3, section 3, neural network node support operation types; <br>&nbsp;&nbsp;3. The first input param independent variable, or its address point.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This param is a 2D array; <br>&nbsp;&nbsp;4. The second input param independent variable, or its address point.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This param is a 2D array; For monocular operations, this param is null; <br>&nbsp;&nbsp;5. Hyperparameter configuration of node granularity in neural networks.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, activating a function's α value, sliding step size of convolutional/pooled windows; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Please refer to the dictionary table for details: <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;select \* from sm_sc.tb_dic_enum <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;where enum_name = 'node_fn_asso_value'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This param is a 2D array; For monocular operations, this param is null;  |<br>select <br>&nbsp;&nbsp;sm_sc.fv_lambda_arr<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;100001<br>&nbsp;&nbsp;&nbsp;&nbsp;, '03_elu'<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[[1.23, 3.34],[-7.2, -0.25]]<br>&nbsp;&nbsp;&nbsp;&nbsp;, null<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[0.1]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.2300, 3.3400],[-0.0999, -0.0221]]
>>> sm_sc.fv_lambda_arr_ddepdt_dindepdt<br>sm_sc.fv_lambda_arr_ddepdt_dindepdt_p  |Backpropagation node dependent variable derivation to independent variable lambda.<br>This lambda is suitable for the type of operation where the derivative can be obtained by forward propagation.<br>This lambda takes the derivative of the dependent variable over the independent variable, without broadcasting the inverse of the independent variable.<br>Input param:<br>&nbsp;&nbsp;1. The sequence number of neural network nodes.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This param does not participate in actual calculations and is only used for control, debugging, tracking, and logging; <br>&nbsp;&nbsp;2. Lambda naming, i.e. operation type.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The scope of support can be found in chapter 3, section 3, neural network node support operation types; <br>&nbsp;&nbsp;3. Independent variable input matrix which is taking derivative for, or its address point; <br>&nbsp;&nbsp;4. Independent variable input position which is taking derivative for; <br>&nbsp;&nbsp;5. Another independent variable input matrix, or its address point.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For monocular operations, this param is null; <br>&nbsp;&nbsp;6. Hyperparameter configuration of node granularity in neural networks.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, activating a function's α value, sliding step size of convolutional/pooled windows; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Please refer to the dictionary table for details: <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;select \* from sm_sc.tb_dic_enum <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;where enum_name = 'node_fn_asso_value'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This param is a 2D array; For monocular operations, this param is null; <br>&nbsp;&nbsp;7. Dependent variable matrix, or its address point; <br>&nbsp;&nbsp;8. Height and width specifications of independent variables; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Some types of operations can omit passing arguments and only pass specifications.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;包括: agg_concat_x, agg_concat_y, agg_sum, agg_avg, <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;slice_x, slice_y, add, sub, etc |select <br>&nbsp;&nbsp;sm_sc.fv_lambda_arr_ddepdt_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;100001<br>&nbsp;&nbsp;&nbsp;&nbsp;, '01_pow'<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[[1.23, 3.34],[2.2, 0.25]]<br>&nbsp;&nbsp;&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[[1.5, 1.3],[0.7, 0.25]]<br>&nbsp;&nbsp;&nbsp;&nbsp;, null<br>&nbsp;&nbsp;&nbsp;&nbsp;, array[[1.5, 1.3],[0.7, 0.25]] ^\` array[[1.23, 3.34],[2.2, 0.25]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.6676, 0.6302],[-0.1627, -0.9803]]
>>> sm_sc.fv_lambda_arr_dloss_dindepdt<br>sm_sc.fv_lambda_arr_dloss_dindepdt_p   |Backpropagation loss function derivative lambda for node independent variables.<br>This lambda is suitable for the type of operation in which the derivative can only be obtained during the backpropagation stage.<br>This lambda calculates the derivative of the loss function with respect to the independent variable, without broadcasting the inverse of the independent variable.<br>Input param:<br>&nbsp;&nbsp;1. The sequence number of neural network nodes.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This param is not involved in actual calculations and is only used for control, debugging, tracking, and logging purposes; <br>&nbsp;&nbsp;2. Lambda naming, i.e. operation type.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The scope of support can be found in chapter 3, section 3, neural network node support operation types; <br>&nbsp;&nbsp;3. Independent variable input matrix which is taking derivative for, or its address point; <br>&nbsp;&nbsp;4. Independent variable input position which is taking derivative for; <br>&nbsp;&nbsp;5. Another independent variable input matrix, or its address point.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For monocular operations, this param is null; <br>&nbsp;&nbsp;6. Hyperparameter configuration of node granularity in neural networks.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, activating a function's α value, sliding step size of convolutional/pooled windows; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Please refer to the dictionary table for details: <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;select \* from sm_sc.tb_dic_enum <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;where enum_name = 'node_fn_asso_value'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This param is a 2D array; For monocular operations, this param is null; <br>&nbsp;&nbsp;7. Dependent variable matrix, or its address point; <br>&nbsp;&nbsp;8. The derivative of the loss function with respect to the dependent variable of the neuron;   |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_lambda_arr_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;100001<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'03_softmax'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.23,&nbsp;&nbsp;3.34,&nbsp;&nbsp;0.96],[2.2,&nbsp;&nbsp;0.25,&nbsp;&nbsp;7.7]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;null<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_redistr_softmax(array[[1.23,&nbsp;&nbsp;3.34,&nbsp;&nbsp;0.96],[2.2,&nbsp;&nbsp;0.25,&nbsp;&nbsp;7.7]],&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]&nbsp;&nbsp;::&nbsp;&nbsp;int[])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;(-\`&nbsp;&nbsp;array[[0.0,&nbsp;&nbsp;0.0,&nbsp;&nbsp;1.0],[0.0,&nbsp;&nbsp;0.0,&nbsp;&nbsp;1.0]])&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;/\`&nbsp;&nbsp;sm_sc.fv_redistr_softmax(array[[1.23,&nbsp;&nbsp;3.34,&nbsp;&nbsp;0.96],[2.2,&nbsp;&nbsp;0.25,&nbsp;&nbsp;7.7]],&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]&nbsp;&nbsp;::&nbsp;&nbsp;int[])<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[[0.003,0.025,-0.998[0.008,0.001,0.960]]
>>> sm_sc.fv_lambda_arr_len              |Review of height and width of neural network nodes (independent and dependent variables) matrix.<br>This review occurs during the training preparation process.<br>Based on the input parameter specifications and hyperparameter configurations, obtain the output parameter height and width specifications. <br>Input param:<br>&nbsp;&nbsp;1. The sequence number of neural network nodes.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This param is not involved in actual calculations and is only used for control, debugging, tracking, and logging purposes; <br>&nbsp;&nbsp;2. Lambda naming, i.e. operation type; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The scope of support can be found in chapter 3, section 3, neural network node support operation types; <br>&nbsp;&nbsp;3. Height and width specifications of the first input objective matrix; <br>&nbsp;&nbsp;4. Height and width specifications of the second input objective matrix; <br>&nbsp;&nbsp;5. Hyperparameter configuration;  |select <br>&nbsp;&nbsp;sm_sc.fv_lambda_arr_len<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;1000001<br>&nbsp;&nbsp;, 'conv_2d'<br>&nbsp;&nbsp;, array[70, 784]<br>&nbsp;&nbsp;, array[1, 25]<br>&nbsp;&nbsp;, array[28, 5, 5, 1, 1, 2, 2, 2, 2, 0]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[73, 763]
>>> sm_sc.fv_lambda_loss       |Get value the loss function.<br>Input param:<br>&nbsp;&nbsp;1. Loss function type; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Please refer to the dictionary table for details: <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;select \* from sm_sc.tb_dic_enum <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;where enum_name = 'loss_fn_type'<br>&nbsp;&nbsp;2. Training output predictive value matrix; <br>&nbsp;&nbsp;3. Real value encoding of training sets;     |select <br>&nbsp;&nbsp;sm_sc.fv_lambda_loss<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;'101'<br>&nbsp;&nbsp;, array[[1, 2, 3],[2, 3, 4]]<br>&nbsp;&nbsp;, array[[1.1, 2.1, 3.2],[2.2, 3.3, 4.2]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;0.3391
>>> sm_sc.fv_lambda_dloss_dz   |Derivation of predicted values by loss function.<br>Input param:<br>&nbsp;&nbsp;1. Loss function type; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Please refer to the dictionary table for details: <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;select \* from sm_sc.tb_dic_enum <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;where enum_name = 'loss_fn_type'<br>&nbsp;&nbsp;2. Training output predictive value matrix; <br>&nbsp;&nbsp;3. Real value encoding of training sets;<br>&nbsp;&nbsp;4. multi modal output position; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Please refer to the dictionary table for details: <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;select \* from sm_sc.tb_dic_enum <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;where enum_name = 'node_type' and enum_key like 'output%'     |select <br>&nbsp;&nbsp;sm_sc.fv_lambda_dloss_dz<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;'101'<br>&nbsp;&nbsp;, array[[1, 2, 3],[2, 3, 4]]<br>&nbsp;&nbsp;, array[[1.1, 2.1, 3.2],[2.2, 3.3, 4.2]]<br>&nbsp;&nbsp;, 'output_01'<br>&nbsp;&nbsp;)<br><br>return: <br>&nbsp;&nbsp;array[[-0.0707, -0.0707, -0.1414],[-0.1414, -0.2121, -0.1414]]
>>
>>####   l. Specialized neural networks for machine learning: 
>>
>>> Function name | Function description | Call example
>>> :-|:-|:-
>>> sm_sc.fv_idx_samp_by_samp         |Sampling and resampling sequence number tracking.<br>Input param:<br>&nbsp;&nbsp;1. The multi interval range of the original serial number of the sample before this sampling; <br>&nbsp;&nbsp;2. The multi interval range of the position and sequence numbers of the samples collected this time in the old set;  |select <br>&nbsp;&nbsp;sm_sc.fv_idx_samp_by_samp<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;int4multirange<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(1, 3, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(8, 12, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(13, 15, '[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;),<br>&nbsp;&nbsp;&nbsp;&nbsp;int4multirange<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(4, 6, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(8, 12, '[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;int4multirange<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;int4range(8, 10, '[]'), <br>&nbsp;&nbsp;&nbsp;&nbsp;int4range(12, 12, '[]'), <br>&nbsp;&nbsp;&nbsp;&nbsp;int4range(13, 15, '[]')<br>&nbsp;&nbsp;)
>>> sm_sc.ft_nn_buff_slice_rand_pick  |Randomly select a small batch of single training from the specified buff table.<br>No duplicate records in the current batch.<br>Input param:<br>&nbsp;&nbsp;1. Training task identification; <br>&nbsp;&nbsp;2. Define the sample number range after sorting all samples in the training set according to classification groups; <br>&nbsp;&nbsp;3. Quantity of each classification group used in the training set.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Different groups can be configured with different sampling quantities; <br>Output param column: <br>&nbsp;&nbsp;1. Sampling serial number; <br>&nbsp;&nbsp;2. Extract the input param vector of the sample;   |-- Taking the MNIST dataset and configuring training tasks as an example,<br>-- Assuming the configuration training task number is:2022030501; <br>select <br>&nbsp;&nbsp;o_ord_no,<br>&nbsp;&nbsp;o_slice_rand_pick<br>from<br>&nbsp;&nbsp;sm_sc.ft_nn_buff_slice_rand_pick<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;2022030501,<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(1, 5923, '[]'), <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(5924, 12665, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(12666, 18623, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(18624, 24754, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(24755, 30596, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(30597, 36017, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(36018, 41935, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(41936, 48200, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(48201, 54052, '[]'),<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4range(54053, 60001, '[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;array_fill(7, array[10])<br>&nbsp;&nbsp;) tb_a;<br><br>return:<br>&nbsp;&nbsp;-- 10 classification groups, with 7 samples randomly selected from each group,<br>&nbsp;&nbsp;-- 70 samples in total.
>>> sm_sc.ft_nn_in_out<br>sm_sc.ft_nn_in_out_p                |Training set and testing set, single input, return judgment or prediction result vector.<br>This function is implemented based on temporary table sm_sc.__vt_tmp_nn_node.<br>Input param:<br>&nbsp;&nbsp;1. Identification of training tasks for output models; <br>&nbsp;&nbsp;2. session No.. as a model instance in __vt_tmp_nn_node. <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;when inputing null，create a new model instance; <br>&nbsp;&nbsp;3 - 6. Test or validation set single sample input parameter vector.  4 - 6 of them is default null。 <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The incoming independent variable is one dimension larger than the records in the training set, used to request multiple prediction data to be passed in at once; <br>&nbsp;&nbsp;7. Specify the sequence number of the dependent variable to be returned.  default: array[1]; |-- Assuming there are already trained model tasks: 2022030501.<br>select <br>&nbsp;&nbsp;ord_no,<br>&nbsp;&nbsp;i_depdt_01<br>from sm_sc.tb_nn_train_input_buff,<br>&nbsp;&nbsp;sm_sc.ft_nn_in_out<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;2022030501, <br>&nbsp;&nbsp;&nbsp;&nbsp;null, <br>&nbsp;&nbsp;&nbsp;&nbsp;array[i_indepdt]<br>&nbsp;&nbsp;) as a_pred_out_arr<br>where work_no = 2022030501<br>&nbsp;&nbsp;and ord_no = round((random() * 60001) + 0.5)<br><br>return(among them, a_pred_out_arr is the return value, i_depdt_01 as a true value comparison): <br>&nbsp;&nbsp;i_depdt_01:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]<br>&nbsp;&nbsp;a_pred_out_arr:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0.0001, 0.9193, 0.0028, 0.0337, 0.0010, 0.0048, 0.0013, 0.0125, 0.0190, 0.0053]]
>>> sm_sc.ft_nn_node2node_path        |Forward path between two nodes.<br>Input param:<br>&nbsp;&nbsp;1. Identification of training tasks for output models; <br>&nbsp;&nbsp;2. Starting node; <br>&nbsp;&nbsp;3. Arrival node;  |-- Assuming there are already trained model tasks: 2022030501.<br>select <br>&nbsp;&nbsp;o_back_node_no,<br>&nbsp;&nbsp;o_fore_node_no,<br>&nbsp;&nbsp;o_path_ord_no,<br>&nbsp;&nbsp;o_node_type<br>from sm_sc.ft_nn_node2node_path(2022030501, 103010025, 104030008)<br><br>Explanatory note: <br>&nbsp;&nbsp;Enumeration values for return field o_node_type: <br>&nbsp;&nbsp;&nbsp;&nbsp;1. main: nodes along the path; <br>&nbsp;&nbsp;&nbsp;&nbsp;2. leaf: Not a node along the way, but a supporting input parameter;<br><br>return:<br>&nbsp;&nbsp;o_back_node_no&nbsp;&nbsp;&nbsp;&nbsp;o_fore_node_no&nbsp;&nbsp;&nbsp;&nbsp;o_path_ord_no&nbsp;&nbsp;&nbsp;&nbsp;o_node_type<br>&nbsp;&nbsp;103010015&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;104010008&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;leaf<br>&nbsp;&nbsp;103010025&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;104010008&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;main<br>&nbsp;&nbsp;103010035&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;104010008&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;3&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;leaf<br>&nbsp;&nbsp;103010044&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;104010008&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;4&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;leaf<br>&nbsp;&nbsp;104010008&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;104020008&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;main<br>&nbsp;&nbsp;104020008&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;104030008&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;main
>>> sm_sc.fv_nn_node2node_val         |Forward propagation values between two nodes.<br>Input param:<br>&nbsp;&nbsp;1. Identification of training tasks for output models; <br>&nbsp;&nbsp;2. Starting node; <br>&nbsp;&nbsp;3. Arrival node;  |-- Assuming there are already trained model tasks: 2022030501.<br>select sm_sc.fv_nn_node2node_val(2022030501, 103010025, 104030008)<br><br>return(tensor output of an operator):  
>>> sm_sc.fv_nn_weight_len            |For the weight parameter operator, the weight independent variable specification is inferred through the dependent variable specification, hyperparameter configuration, and independent variable order position.<br>Input param:<br>&nbsp;&nbsp;1. Operator Type; <br>&nbsp;&nbsp;2. The position number of the independent variable for the weight specification to be determined; <br>&nbsp;&nbsp;3. Operator hyperparameter; <br>&nbsp;&nbsp;4. Operator dependent variable specifications; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_nn_weight_len<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'01_prod_mx'<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;5.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[8,&nbsp;&nbsp;10,&nbsp;&nbsp;2,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[8,&nbsp;&nbsp;10,&nbsp;&nbsp;3,&nbsp;&nbsp;5]
>>> sm_sc.fv_nn_none                  |Empty operator, direct propagation operator, nn node occupying operator.<br>The single independent variable remains unchanged as the dependent variable output.<br>When it can be used to design neural networks,<br>network node occupancy,<br>operator granularity hyperparameter configuration position extension or independent variable position tracking, etc.|select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_nn_none(array[1.0])<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.0]
>>> sm_sc.fv_nn_node_pick_depdt_idx   |After the operator is executed, the dataset sampling sequence number of the dependent variable is tracked.|select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_nn_node_pick_depdt_idx<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'04_slice_y'<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4multirange(int4range(2,&nbsp;&nbsp;5,&nbsp;&nbsp;'[]')),&nbsp;&nbsp;int4multirange(int4range(7,&nbsp;&nbsp;11,&nbsp;&nbsp;'[]'))]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[3.0],&nbsp;&nbsp;[6.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int4multirange(int4range(4,&nbsp;&nbsp;5,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(7,&nbsp;&nbsp;8,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>
>>####   m. Encoding and decoding of machine learning:
>>
>>> Function name | Function description | Call example
>>> :-|:-|:-
>>> sm_sc.fv_huffman      |Huffman encoding.<br>This function relies on the temporary table sm_sc.__vt_tmp_huffman implementation.<br>Input param:<br>&nbsp;&nbsp;1. A 1D array composed of the original set of numerical values;   |select<br>&nbsp;&nbsp;sm_sc.fv_huffman    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3, 8, 9, 13, 19]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[B'00101', B'00100',B'0011',B'001',B'11',B'10',B'01']
>>> sm_sc.fv_onehot       |One-hot encoding.<br>Input param:<br>&nbsp;&nbsp;1. A 1D array composed of the original set of numerical values or a 2D array with several rows and one colume; <br>&nbsp;&nbsp;2. hot encode value.  Default 1; <br>&nbsp;&nbsp;3. cold encode value.  Default 0;  |select<br>&nbsp;&nbsp;sm_sc.fv_onehot     <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3, 1, 4, 3, 5, 6]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1.0, 0.0, 0.0, 0.0, 0.0, 0.0]<br>&nbsp;&nbsp;,[0.0, 1.0, 0.0, 0.0, 0.0, 0.0]<br>&nbsp;&nbsp;,[0.0, 0.0, 1.0, 0.0, 0.0, 0.0]<br>&nbsp;&nbsp;,[1.0, 0.0, 0.0, 0.0, 0.0, 0.0]<br>&nbsp;&nbsp;,[0.0, 0.0, 0.0, 1.0, 0.0, 0.0]<br>&nbsp;&nbsp;,[0.0, 0.0, 1.0, 0.0, 0.0, 0.0]<br>&nbsp;&nbsp;,[0.0, 0.0, 0.0, 0.0, 1.0, 0.0]<br>&nbsp;&nbsp;,[0.0, 0.0, 0.0, 0.0, 0.0, 1.0]]
>>> sm_sc.ft_onehot_dic   |One-hot encoding dictionary.<br>Input param:<br>&nbsp;&nbsp;1. A 1D array composed of the original set of numerical values or a 2D array with several rows and one colume; <br>&nbsp;&nbsp;2. hot encode value.  Default 1; <br>&nbsp;&nbsp;3. cold encode value.  Default 0;  |select<br>&nbsp;&nbsp;&nbsp;&nbsp;o_ele<br>,&nbsp;&nbsp;o_onehot<br>from&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.ft_onehot_dic<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;1,&nbsp;&nbsp;4,&nbsp;&nbsp;3,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return: <br>&nbsp;&nbsp;&nbsp;&nbsp;o_ele&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;o_onehot<br>&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{1,0,0,0,0,0}<br>&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{0,1,0,0,0,0}<br>&nbsp;&nbsp;&nbsp;&nbsp;3&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{0,0,1,0,0,0}<br>&nbsp;&nbsp;&nbsp;&nbsp;4&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{0,0,0,1,0,0}<br>&nbsp;&nbsp;&nbsp;&nbsp;5&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{0,0,0,0,1,0}<br>&nbsp;&nbsp;&nbsp;&nbsp;6&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{0,0,0,0,0,1}
>>> sm_sc.ft_bpe          |BPE encoding.<br>Input param:<br>&nbsp;&nbsp;1. Original input sequence array;<br>&nbsp;&nbsp;2. Constraint: The number of iterations merging token is used as the threshold for the end of the loop;<br>&nbsp;&nbsp;3. Constraint: The dictionary size of the token list is used as the threshold for the end of the loop,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Default 0; <br>&nbsp;&nbsp;4. Expectation: After encoding, the maximum number of tokens contained in each seq serves as the threshold for the end of the loop,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The default value of 0 represents the unconstrained loop end threshold;<br>&nbsp;&nbsp;5. Preliminary evaluation: For a certain seq, when the number of token included is already less than i_threshold_seq_token_len,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;So, if the average length of the token it contains exceeds this threshold, it will no longer participate in BPE.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Default 0 represents not making this judgment;<br>&nbsp;&nbsp;6. Whether to distinguish the start and end of matching seq.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If distinguished, special treatment should be applied to the two escape characters \'\^\$\'in seq;<br>Output param column:<br>&nbsp;&nbsp;1. Output sequence array.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The order of elements corresponds one-to-one with the original input;<br>&nbsp;&nbsp;2. List, dictionary table;|select<br>&nbsp;&nbsp;&nbsp;&nbsp;o_ret_seq_code<br>,&nbsp;&nbsp;o_ret_token_list<br>from&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.ft_bpe<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array['estern',&nbsp;&nbsp;'widest',&nbsp;&nbsp;'longest']<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;o_ret_seq_code:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'est&nbsp;&nbsp;e&nbsp;&nbsp;r&nbsp;&nbsp;n'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'w&nbsp;&nbsp;i&nbsp;&nbsp;d&nbsp;&nbsp;est'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'l&nbsp;&nbsp;o&nbsp;&nbsp;n&nbsp;&nbsp;g&nbsp;&nbsp;est'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;o_ret_token_list:<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'r'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'n'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'d'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'g'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'i'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'w'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'l'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'o'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'est'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;'e'<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]
>>
>>####   n. The loss function of machine learning:
>>
>>> Function name | Function description | Call example
>>> :-|:-|:-
>>> sm_sc.fv_loss_least_square       |Least squares method.<br>&nbsp;&nbsp;1. Train the output prediction value matrix;<br>&nbsp;&nbsp;2. Encoding the true values of the training set;    |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_loss_least_square<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.1,&nbsp;&nbsp;2.1,&nbsp;&nbsp;3.2],&nbsp;&nbsp;[2.2,&nbsp;&nbsp;3.3,&nbsp;&nbsp;4.2]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;0.3391164991562635
>>> sm_sc.fv_loss_cross_entropy      |Cross entropy.<br>&nbsp;&nbsp;1. Train the output prediction value matrix;<br>&nbsp;&nbsp;2. Encoding the true values of the training set;    |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_loss_cross_entropy<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[0.3,&nbsp;&nbsp;0.5,&nbsp;&nbsp;0.2],&nbsp;&nbsp;[0.1,&nbsp;&nbsp;0.8,&nbsp;&nbsp;0.1]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[0.0,&nbsp;&nbsp;1.0,&nbsp;&nbsp;0.0],&nbsp;&nbsp;[0.0,&nbsp;&nbsp;1.0,&nbsp;&nbsp;0.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;0.4581453659370775
>>
>>####   o. Complex operations:
>>
>>> Function name | Function description | Call example
>>> :-|:-|:-
>>> sm_sc.fv_opr_add             |Complex add.          |select<br>&nbsp;&nbsp;sm_sc.fv_opr_add             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3, -12.3),<br>&nbsp;&nbsp;&nbsp;&nbsp;(-45.6, -45.6)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(-33.3, -57.9)
>>> sm_sc.fv_opr_sub             |Complex sub.         |select<br>&nbsp;&nbsp;sm_sc.fv_opr_sub             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3, -12.3),<br>&nbsp;&nbsp;&nbsp;&nbsp;(-45.6, -45.6)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(57.9, 33.3)
>>> sm_sc.fv_opr_sub             |Complex opposite number.        |select<br>&nbsp;&nbsp;sm_sc.fv_opr_sub             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3, -12.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(-12.3, 12.3)
>>> sm_sc.fv_opr_mul             |Complex mul.          |select<br>&nbsp;&nbsp;sm_sc.fv_opr_mul             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3, -12.3),<br>&nbsp;&nbsp;&nbsp;&nbsp;(-45.6, -45.6)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(1121.76, 0.0)
>>> sm_sc.fv_opr_div             |Complex div.          |select<br>&nbsp;&nbsp;sm_sc.fv_opr_div             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3, -12.3),<br>&nbsp;&nbsp;&nbsp;&nbsp;(-45.6, -45.6)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(0.0 ,0.2697)
>>> sm_sc.fv_opr_div             |Complex reciprocal.         |select<br>&nbsp;&nbsp;sm_sc.fv_opr_div             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3, -12.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(0.0407, 0.0407)
>>> sm_sc.fv_opr_pow             |Complex power.            |select<br>&nbsp;&nbsp;sm_sc.fv_opr_pow             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(2.3, -0.8),<br>&nbsp;&nbsp;&nbsp;&nbsp;(1.6, -3.6)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(-1.0288, 0.7009)
>>> sm_sc.fv_opr_exp             |Complex natural constant power.    |select<br>&nbsp;&nbsp;sm_sc.fv_opr_exp             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(2.3, -0.8)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(6.9491, -7.1550)
>>> sm_sc.fv_opr_log             |Complex log.          |select<br>&nbsp;&nbsp;sm_sc.fv_opr_log             <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3, -12.3),<br>&nbsp;&nbsp;&nbsp;&nbsp;(-45.6, -45.6)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(1.5671, -0.3940)
>>> sm_sc.fv_opr_ln              |Complex natural logarithm.      |select<br>&nbsp;&nbsp;sm_sc.fv_opr_ln              <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3, -12.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(2.8562, -0.7854)
>>> sm_sc.fv_sin                 |Complex sin.          |select<br>&nbsp;&nbsp;sm_sc.fv_sin                 <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(2.3, -2.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(3.7563, 3.2894)
>>> sm_sc.fv_cos                 |Complex cos.          |select<br>&nbsp;&nbsp;sm_sc.fv_cos                 <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(3.3, -1.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(-0.3109, 1.6771)
>>> sm_sc.fv_opr_real            |Complex real part.          |select<br>&nbsp;&nbsp;sm_sc.fv_opr_real            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(2.3, -2.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;2.3
>>> sm_sc.fv_opr_imaginary       |Complex imaginary part.          |select<br>&nbsp;&nbsp;sm_sc.fv_opr_imaginary       <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(2.3, -2.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;-2.3
>>> sm_sc.fv_opr_norm            |Complex norm.          |select<br>&nbsp;&nbsp;sm_sc.fv_opr_norm            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(2.3, -2.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;3.2527
>>> sm_sc.fv_opr_conjugate       |Complex conjugate.          |select<br>&nbsp;&nbsp;sm_sc.fv_opr_conjugate       <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(2.4, -6.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(2.4, 6.3)
>>> sm_sc.fv_opr_conjugate_45    |Complex real imaginary swap.      |select<br>&nbsp;&nbsp;sm_sc.fv_opr_conjugate_45    <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(2.4, -6.3)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(-6.3, 2.4)
>>> sm_sc.fv_opr_round           |Complex round.      |select<br>&nbsp;&nbsp;sm_sc.fv_opr_round           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.34854, -12.3483)<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(12, -12)
>>> sm_sc.fv_opr_floor           |Complex digit floor value.    |select<br>&nbsp;&nbsp;sm_sc.fv_opr_floor           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(1482.37878, -178672.3543)<br>&nbsp;&nbsp;, -1<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(1480, -178680)
>>> sm_sc.fv_opr_ceil            |Complex digit ceiling value.  |select<br>&nbsp;&nbsp;sm_sc.fv_opr_ceil            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.3233333, -12.3788)<br>&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(12.33, -12.37)
>>> sm_sc.fv_opr_trunc           |Complex digit truncation value.    |select<br>&nbsp;&nbsp;sm_sc.fv_opr_trunc           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;(12.31231, -12.32333)<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(12.312, -12.323)
>>
>>####   p. UDF public functions added to postgresql:
>>
>>> Function name | Function description | Call example
>>> :-|:-|:-
>>> sm_sc.fa_coalesce           |Null replace aggregation.                |select <br>&nbsp;&nbsp;sm_sc.fa_coalesce(a_val) as a_coal_agg,<br>&nbsp;&nbsp;sm_sc.fa_coalesce(a_val order by a_no) as a_coal_agg_asc,<br>&nbsp;&nbsp;sm_sc.fa_coalesce(a_val order by a_no desc) as a_coal_agg_desc<br>from <br>(<br>&nbsp;&nbsp;select null as a_val, 1 as a_no <br>&nbsp;&nbsp;union all <br>&nbsp;&nbsp;select 'abc', 3<br>&nbsp;&nbsp;union all <br>&nbsp;&nbsp;select 'cde', 2<br>&nbsp;&nbsp;union all <br>&nbsp;&nbsp;select 'fgh', 4<br>&nbsp;&nbsp;union all <br>&nbsp;&nbsp;select null, 5<br>) tb_a(a_val)<br><br>return:<br>&nbsp;&nbsp;a_coal_agg: 'abc'<br>&nbsp;&nbsp;a_coal_agg_asc: 'cde'<br>&nbsp;&nbsp;a_coal_agg_desc: 'fgh'
>>> sm_sc.fa_prod               |Multiplicative aggregation.                |select <br>&nbsp;&nbsp;sm_sc.fa_prod(a_val) as a_prod_agg<br>from <br>(<br>&nbsp;&nbsp;select 3.0 as a_val<br>&nbsp;&nbsp;union all <br>&nbsp;&nbsp;select 2.5<br>&nbsp;&nbsp;union all <br>&nbsp;&nbsp;select 4.8<br>) tb_a(a_val)<br><br>return:36.0
>>> sm_sc.fa_concat             |Connection aggregation.                |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fa_concat(a_val)&nbsp;&nbsp;as&nbsp;&nbsp;a_concat_agg<br>from&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;'abc'&nbsp;&nbsp;as&nbsp;&nbsp;a_val<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;'faera'<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;'4564'<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;'kuk7'<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;'66yy'<br>)&nbsp;&nbsp;tb_a(a_val)<br><br>return:abcfaera4564kuk766yy
>>> sm_sc.fv_enum_rand_pick     |Enumerate random picking without repetition.<br>Input param:<br>&nbsp;&nbsp;1. Enumeration type name; <br>&nbsp;&nbsp;2. Random sampling quantity;     |-- create type sm_sc.typ_enum_day as <br>-- enum('SUN','MON','TUE','WED','THU','FRI','SAT');<br>select <br>&nbsp;&nbsp;sm_sc.fv_enum_rand_pick<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;'sm_sc.typ_enum_day'<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);<br><br>return:array['TUE', 'MON', 'SUN']
>>> sm_sc.fv_enum_rand          |Enumerate random values, repeatable.<br>Input param:<br>&nbsp;&nbsp;1. Enumeration type name; <br>&nbsp;&nbsp;2. Random sampling quantity;     |-- create type sm_sc.typ_enum_day as <br>-- enum('SUN','MON','TUE','WED','THU','FRI','SAT');<br>select <br>&nbsp;&nbsp;sm_sc.fv_enum_rand<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;'sm_sc.typ_enum_day'<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);<br><br>return:array['TUE', 'MON', 'MON']
>>> sm_sc.fv_o_distance         |Two vector Euclidean distance.<br>Input param:<br>&nbsp;&nbsp;1. Vector 1; <br>&nbsp;&nbsp;2. Vector 2;           |select<br>&nbsp;&nbsp;sm_sc.fv_o_distance       <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.0, 2.0]<br>&nbsp;&nbsp;, array[4.0, 6.0]<br>&nbsp;&nbsp;);<br><br>return:5.0
>>> sm_sc.fa_range_or           |OR aggregate multiple intervals.<br>&nbsp;&nbsp;Aggregated field type: <br>&nbsp;&nbsp;&nbsp;&nbsp;range type or range array type; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fa_range_or(a_val)<br>from&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;int8range(1,&nbsp;&nbsp;2,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange&nbsp;&nbsp;as&nbsp;&nbsp;a_val<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;int8range(5,&nbsp;&nbsp;6,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;int8range(3,&nbsp;&nbsp;7,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;int8range(9,&nbsp;&nbsp;13,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;int8range(11,&nbsp;&nbsp;17,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;null<br>)&nbsp;&nbsp;t<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;int8multirange(int8range(1,&nbsp;&nbsp;7,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int8range(9,&nbsp;&nbsp;17,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>--&nbsp;&nbsp;-----------------------------------------------------<br>select&nbsp;&nbsp;sm_sc.fa_range_or(a_val)<br>from&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;array[int8range(1,&nbsp;&nbsp;2,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange,&nbsp;&nbsp;int8range(-2,&nbsp;&nbsp;-1,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange]&nbsp;&nbsp;as&nbsp;&nbsp;a_val<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;array[int8range(5,&nbsp;&nbsp;6,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange,&nbsp;&nbsp;int8range(-6,&nbsp;&nbsp;-5,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange]<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;array[int8range(3,&nbsp;&nbsp;7,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange,&nbsp;&nbsp;int8range(-7,&nbsp;&nbsp;-3,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange]<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;array[int8range(9,&nbsp;&nbsp;13,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange,&nbsp;&nbsp;int8range(-13,&nbsp;&nbsp;-9,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange]<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;array[int8range(11,&nbsp;&nbsp;17,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange,&nbsp;&nbsp;int8range(-17,&nbsp;&nbsp;-11,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange]<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;null<br>)&nbsp;&nbsp;t<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int8multirange(int8range(1,&nbsp;&nbsp;7,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int8range(9,&nbsp;&nbsp;17,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;int8multirange(int8range(-17,&nbsp;&nbsp;-7,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int8range(-7,&nbsp;&nbsp;-1,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fa_range_and          |AND aggregate multiple intervals.<br>&nbsp;&nbsp;Aggregated field type: <br>&nbsp;&nbsp;&nbsp;&nbsp;range type or range array type; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fa_range_and(a_val)<br>from&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;int8range(9,&nbsp;&nbsp;15,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange&nbsp;&nbsp;as&nbsp;&nbsp;a_val<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;int8range(11,&nbsp;&nbsp;17,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;int8range(7,&nbsp;&nbsp;12,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange<br>)&nbsp;&nbsp;t<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;int8multirange(int8range(11,&nbsp;&nbsp;12,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>--&nbsp;&nbsp;-----------------------------------------------------<br>select&nbsp;&nbsp;sm_sc.fa_range_and(a_val)<br>from&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;array[int8range(1,&nbsp;&nbsp;5,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange,&nbsp;&nbsp;int8range(-4,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange]&nbsp;&nbsp;as&nbsp;&nbsp;a_val<br>&nbsp;&nbsp;&nbsp;&nbsp;union&nbsp;&nbsp;all&nbsp;&nbsp;select&nbsp;&nbsp;array[int8range(3,&nbsp;&nbsp;6,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange,&nbsp;&nbsp;int8range(-6,&nbsp;&nbsp;-1,&nbsp;&nbsp;'[]')&nbsp;&nbsp;::&nbsp;&nbsp;int8multirange]<br>)&nbsp;&nbsp;t<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int8multirange(int8range(3,&nbsp;&nbsp;5,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;int8multirange(int8range(-4,&nbsp;&nbsp;-1,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.fv_multirange_len     |The length of multiple intervals is different from the span, excluding intervals.<br>&nbsp;&nbsp;Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Multi segment interval; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_multirange_len<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int8multirange(int8range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int8range(9,&nbsp;&nbsp;10,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;5
>>> sm_sc.fv_range_move         |Single interval move.<br>&nbsp;&nbsp;Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Single interval; <br>&nbsp;&nbsp;&nbsp;&nbsp;2. Moving distance; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_range_move<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int8range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]')<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;6&nbsp;&nbsp;::&nbsp;&nbsp;bigint<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;int8range(7,&nbsp;&nbsp;9,&nbsp;&nbsp;'[]')
>>> sm_sc.fv_multirange_move    |Multi segment interval move.<br>&nbsp;&nbsp;Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Multi segment interval; <br>&nbsp;&nbsp;&nbsp;&nbsp;2. Moving distance; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_multirange_move<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;int8multirange(int8range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int8range(9,&nbsp;&nbsp;10,&nbsp;&nbsp;'[]'))<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;10&nbsp;&nbsp;::&nbsp;&nbsp;bigint<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;int8multirange(int8range(11,&nbsp;&nbsp;13,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int8range(19,&nbsp;&nbsp;20,&nbsp;&nbsp;'[]'))
>>> sm_sc.fv_get_global_seq     |Obtain a global bigint unique identifier.<br>&nbsp;&nbsp;Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Obtain the number of unique identifiers. The default is 1; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_get_global_seq(3)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;int8range(8,&nbsp;&nbsp;10,&nbsp;&nbsp;'[]')
>>
>>####   q. Derivative formula derivation:
>>
>>> Function name | Function description | Call example
>>> :-|:-|:-
>>> sm_sc.fv_gradient                |Derivative derivation of algebraic expression formulas.<br>This function relies on the following temporary table implementation.<br>&nbsp;&nbsp;1. sm_sc._vt_fn_grad__graph; <br>&nbsp;&nbsp;2. sm_sc._vt_fn_grad__algebra; <br>&nbsp;&nbsp;3. sm_sc._vt_fn_grad__forward; <br>&nbsp;&nbsp;4. sm_sc._vt_fn_grad__chain; <br>Reload 1. <br>&nbsp;&nbsp;Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Derivation target independent variable name; <br>&nbsp;&nbsp;&nbsp;&nbsp;2. Serialization of algebraic expressions json;  <br>Reload 2.<br>&nbsp;&nbsp;Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1. Derivation target independent variable name; <br>&nbsp;&nbsp;&nbsp;&nbsp;2. Algebraic expression text expression;      |select <br>&nbsp;&nbsp;sm_sc.fv_gradient<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;'v_x' :: text,<br>&nbsp;&nbsp;&nbsp;&nbsp;'<br>&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_z",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"opr": "sm_sc.fv_opr_pow",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_x"<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;},<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_x"<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;}<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;}<br>&nbsp;&nbsp;&nbsp;&nbsp;' :: jsonb<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;(((v_x ^ v_x) * ln(v_x)) + (v_x * (v_x ^ (v_x - 1.00000000))))<br>--------------------------<br>select <br>&nbsp;&nbsp;sm_sc.fv_gradient<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;'v_x'<br>&nbsp;&nbsp;, '(v_x * v_y) + exp(v_x * v_y) + power(v_x, 2)'::text<br>&nbsp;&nbsp;);<br><br>return:(v_y + (2.00000000 * v_x) + (exp((v_x * v_y)) * v_y))
>>> sm_sc.ft_gradient                |Numerical calculation of algebraic expression derivative.<br>Input param:<br>&nbsp;&nbsp;1. Derivation target independent variable name list; <br>&nbsp;&nbsp;2.Serialized JSON of algebraic expressions containing independent variables substituted into numerical values; <br>Output param column:<br>&nbsp;&nbsp;1. Derivation target independent variable name; <br>&nbsp;&nbsp;2.Derivative value;       |select<br>&nbsp;&nbsp;o_indepdt_vars_name,<br>&nbsp;&nbsp;o_grad<br>from <br>sm_sc.ft_gradient<br>(<br>&nbsp;&nbsp;array['v_x_in', 'v_y_in'],<br>&nbsp;&nbsp;'<br>&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_z_in",<br>&nbsp;&nbsp;&nbsp;&nbsp;"opr": "sm_sc.fv_opr_add",<br>&nbsp;&nbsp;&nbsp;&nbsp;"in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"opr": "sm_sc.fv_opr_mul",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_x_in",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_value": 1.5<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;},<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_y_in",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_value": 1.5<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;}<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;},<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"opr": "sm_sc.fv_opr_exp",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"opr": "sm_sc.fv_opr_mul",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_x_in",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_value": 1.5<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;},<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_y_in",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_value": 1.5<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;}<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;}<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;},<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"opr": "sm_sc.fv_opr_pow",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"in_params": <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_param": "v_x_in",<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_value": 1.5<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;},<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;{<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"out_value": 2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;}<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;}<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;}<br>&nbsp;&nbsp;'::jsonb<br>);<br><br>return:<br>&nbsp;&nbsp;o_indepdt_vars_name&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;o_grad<br>&nbsp;&nbsp;v_x_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;18.73160376<br>&nbsp;&nbsp;v_y_in&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;15.73160376
>>> sm_sc.ft_gradient_graph          |Algebraic expression text derivative derivation (output as a tree structure).<br>Input param:<br>&nbsp;&nbsp;1. Derivation target independent variable name; <br>&nbsp;&nbsp;2.Algebraic expression text expression; <br>Output param column: <br>&nbsp;&nbsp;1. Intermediate dependent variable; <br>&nbsp;&nbsp;2. Intermediate independent variable; <br>&nbsp;&nbsp;3. Constant value; <br>&nbsp;&nbsp;4. Input param position sequence number; <br>&nbsp;&nbsp;5. Operation type;      |select<br>&nbsp;&nbsp;o_out_param_x,<br>&nbsp;&nbsp;o_in_param_x,<br>&nbsp;&nbsp;o_in_value_x,<br>&nbsp;&nbsp;o_param_loc_x,<br>&nbsp;&nbsp;o_out_opr_x  <br>from sm_sc.ft_gradient_graph('v_x', '3 * v_x * v_y'::text)<br><br>return:<br>&nbsp;&nbsp;o_out_param_x&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;o_in_param_x&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;o_in_value_x&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;o_param_loc_x&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;o_out_opr_x<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1ecb1aeee9bba546541394669cfbb68e&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;04460d66750dc2c536e3aef4a66df70a&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;3a98ec1e00cca8e4d330b49c896699c3&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;04460d66750dc2c536e3aef4a66df70a&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;e38869fbedbaa92e1aef2ee286e1a751&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;1ecb1aeee9bba546541394669cfbb68e&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;04460d66750dc2c536e3aef4a66df70a&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_add<br>&nbsp;&nbsp;3a98ec1e00cca8e4d330b49c896699c3&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.0&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;e38869fbedbaa92e1aef2ee286e1a751&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;v_y&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;e38869fbedbaa92e1aef2ee286e1a751&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;3.0&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul
>>> sm_sc.ft_gradient_opr_graph      |Derivative derivation of fundamental operators.<br>Input param:<br>&nbsp;&nbsp;1. Derivation target independent variable name; <br>&nbsp;&nbsp;2.Operation type; <br>&nbsp;&nbsp;3.Operation type; The sequence number of the input position of the derivative target independent variable; <br>&nbsp;&nbsp;4.List of other input param names.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If it is a constant, fill in null; <br>&nbsp;&nbsp;5.List of other input param constants.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If it is a variable, fill in null; <br>&nbsp;&nbsp;6.Specify the name of the derivative result variable;      |select <br>&nbsp;&nbsp;o_out_param,<br>&nbsp;&nbsp;o_in_param, <br>&nbsp;&nbsp;o_in_value, <br>&nbsp;&nbsp;o_param_loc,<br>&nbsp;&nbsp;o_out_opr  <br>from sm_sc.ft_gradient_opr_graph<br>(<br>&nbsp;&nbsp;'x',<br>&nbsp;&nbsp;'sm_sc.fv_opr_div',<br>&nbsp;&nbsp;2,<br>&nbsp;&nbsp;null,<br>&nbsp;&nbsp;array[1.2]<br>);<br><br>return:<br>&nbsp;&nbsp;o_out_param&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;o_in_param&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;o_in_value&nbsp;&nbsp;&nbsp;&nbsp;o_param_loc&nbsp;&nbsp;&nbsp;&nbsp;o_out_opr<br>&nbsp;&nbsp;b0ae53f5ac4c3091753f8e4e4c352039&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;x&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_pow<br>&nbsp;&nbsp;b0ae53f5ac4c3091753f8e4e4c352039&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;-2.0&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_pow<br>&nbsp;&nbsp;6da6d6058c859b21588e829d71a7e6b3&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.2&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;6da6d6058c859b21588e829d71a7e6b3&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;b0ae53f5ac4c3091753f8e4e4c352039&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_mul<br>&nbsp;&nbsp;8881836ea7405adb788660db93f5f116&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.0&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_sub<br>&nbsp;&nbsp;8881836ea7405adb788660db93f5f116&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;6da6d6058c859b21588e829d71a7e6b3&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_sub<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;8881836ea7405adb788660db93f5f116&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
>>> sm_sc.fv_gradient_jacobi         |Derivative derivation of jacobian matrix.<br>Input param:<br>&nbsp;&nbsp;1. List of independent variable names; <br>&nbsp;&nbsp;2. Algebraic expression list;      |select <br>&nbsp;&nbsp;sm_sc.fv_gradient_jacobi<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array['r', 'a', 'b']<br>&nbsp;&nbsp;, array['r*cos(a)\*sin(b)', 'r\*sin(a)\*sin(b)', 'r\*cos(b)']<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'(sin(b) \* cos(a))',<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'((sin(b) \* r) \* (0.0 - sin(a)))',<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'((r \* cos(a)) \* cos(b))'<br>&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'(sin(b) \* sin(a))',<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'((r \* sin(b)) \* cos(a))',<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'((r \* sin(a)) \* cos(b))'<br>&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'cos(b)',<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'0.0',<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'(r \* (0.0 - sin(b)))'<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_gradient_opr            |Numerical derivation of basic operators.<br>Input param:<br>&nbsp;&nbsp;1. Derivation target independent variable value; <br>&nbsp;&nbsp;2.Operation type; <br>&nbsp;&nbsp;3. Derivation target independent variable input param order position; <br>&nbsp;&nbsp;4. Dependent variable value; <br>&nbsp;&nbsp;5. Another independent variable enters the input param value;      |select sm_sc.fv_gradient_opr<br>(<br>&nbsp;&nbsp;8.0,<br>&nbsp;&nbsp;'sm_sc.fv_opr_log',<br>&nbsp;&nbsp;2,<br>&nbsp;&nbsp;null,<br>&nbsp;&nbsp;2.0<br>);<br><br>return:0.18033688
>>
>>####   r. Derivation of some matrix operations: 
>>
>> Explanation: 1: In the PG4ML framework, the support of derivative functions for backpropagation of independent variable broadcasting typically relies on tracking the specifications of independent variable arrays during the training preparation stage;
>>
>>> Function name | Function description | Call example
>>> :-|:-|:-
>>> sm_sc.<br>fv_d_nn_none_dloss_dindepdt  |Derivation of empty functions.<br>Input param:<br>&nbsp;&nbsp;1. Derivative of loss function with respect to dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_nn_none_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]]
>>> sm_sc.<br>fv_d_new_dloss_dindepdt      |Numerical derivative new.<br>Input param:<br>&nbsp;&nbsp;1. Derivative of loss function with respect to dependent variable; <br>&nbsp;&nbsp;2. The number of repetitions of grouping in each dimension direction of the new function;        |select <br>&nbsp;&nbsp;sm_sc.fv_d_new_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0.5, -0.2, 0.25, 1.3, -0.9, -0.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-0.5, -0.2, -0.25, 1.3, 0.9, 0.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.5, 0.2, 0.25, -1.3, 0.9, 0.05]],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 3]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-0.15, 1.0],[0.15, 1.2],[1.65, -1.05]]
>>> sm_sc.<br>fv_d_repeat_axis_dloss_dindepdt      |Numerical derivative repeat_axis。<br>Input: <br>&nbsp;&nbsp;1. Derivative of loss function with respect to dependent variable; <br>&nbsp;&nbsp;2. multiple of axes for repeat elements; <br>&nbsp;&nbsp;3. Element repetitions on each axis; |select&nbsp;&nbsp;array_dims(<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_repeat_axis_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[10,8])<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[1]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2]<br>&nbsp;&nbsp;&nbsp;&nbsp;));<br><br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>return: <br>&nbsp;&nbsp;&nbsp;&nbsp;[1:5][1:8]
>>> sm_sc.<br>fv_d_activate_elu            |Numerical derivative elu.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix; <br>&nbsp;&nbsp;2. α value;            |select <br>&nbsp;&nbsp;sm_sc.fv_d_activate_elu<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, -2.0],[3.0, 4.0]]<br>&nbsp;&nbsp;, 0.3<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.0, 0.0406],[1.0, 1.0]]
>>> sm_sc.<br>fv_d_activate_gelu           |Numerical derivative gelu.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix;            |select <br>&nbsp;&nbsp;sm_sc.fv_d_activate_gelu<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, -2.0],[3.0, 4.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.0830, -0.0861],[1.0116, 1.0003]]
>>> sm_sc.<br>fv_d_activate_leaky_relu     |Numerical derivative leaky_relu.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix; <br>&nbsp;&nbsp;2. α value;            |select <br>&nbsp;&nbsp;sm_sc.fv_d_activate_leaky_relu<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, -2.0],[3.0, 4.0]]<br>&nbsp;&nbsp;, 0.3<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.0000, 0.3000],[1.0000, 1.0000]]
>>> sm_sc.<br>fv_d_activate_relu           |Numerical derivative relu.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix;            |select <br>&nbsp;&nbsp;sm_sc.fv_d_activate_relu<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, -2.0],[3.0, 4.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.0000, 0.0000],[1.0000, 1.0000]]
>>> sm_sc.<br>fv_d_activate_selu           |Numerical derivative selu.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix;            |select <br>&nbsp;&nbsp;sm_sc.fv_d_activate_selu<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, -2.0],[3.0, 4.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.0507, 0.2379],[1.0507, 1.0507]]
>>> sm_sc.<br>fv_d_activate_sigmoid        |Numerical derivative sigmoid.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix;            |select <br>&nbsp;&nbsp;sm_sc.fv_d_activate_sigmoid<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, -2.0],[3.0, 4.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.1966, 0.1050],[0.0452, 0.0177]]
>>> sm_sc.<br>fv_d_activate_softplus       |Numerical derivative softplus.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix;            |select <br>&nbsp;&nbsp;sm_sc.fv_d_activate_softplus<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, -2.0],[3.0, 4.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.7311, 0.1192],[0.9526, 0.9820]]
>>> sm_sc.<br>fv_d_activate_swish          |Numerical derivative swish.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix;            |select <br>&nbsp;&nbsp;sm_sc.fv_d_activate_swish<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0, -2.0],[3.0, 4.0]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.9277, -0.0908],[1.0881, 1.0527]]
>>> sm_sc.<br>fv_d_activate_boxcox         |Numerical derivative boxcox.<br>Input param:<br>&nbsp;&nbsp;1. Independent variable matrix; <br>&nbsp;&nbsp;2. Lambda value;            |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_activate_boxcox<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;0.2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0.927,&nbsp;&nbsp;0.574,&nbsp;&nbsp;0.415],[0.330,&nbsp;&nbsp;0.276,&nbsp;&nbsp;0.238]]
>>> sm_sc.<br>fv_d_redistr_softmax_dloss_dindepdt            |Derivation of softmax's loss function over independent variables.<br>Input param:<br>&nbsp;&nbsp;1. Dependent variable matrix: <br>&nbsp;&nbsp;2. Derivation matrix of loss function over dependent variable; <br>&nbsp;&nbsp;3. Independent variable matrix:       |select <br>&nbsp;&nbsp;sm_sc.fv_d_redistr_softmax_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_redistr_softmax(array[[1, 2, 3, 4, 5]]),<br>&nbsp;&nbsp;&nbsp;&nbsp;-\` array[[0.0, 0.0, 0.0, 0.0, 1.0]] :: float[] /\` sm_sc.fv_redistr_softmax(array[[1, 2, 3, 4, 5]]),<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4, 5]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.0117, 0.0317, 0.0861, 0.2341, -0.3636]]
>>> sm_sc.<br>fv_d_redistr_zscore                            |Numerical derivative zscore .<br>Input param:<br>&nbsp;&nbsp;1. Dependent variable matrix: <br>&nbsp;&nbsp;2. Independent variable matrix:       |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_redistr_zscore<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_redistr_zscore(array[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]])<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.138,&nbsp;&nbsp;1.771,&nbsp;&nbsp;2.403,&nbsp;&nbsp;3.036,&nbsp;&nbsp;3.668]
>>> sm_sc.<br>fv_d_conv_2d_dloss_dindepdt_1_ex               |Derivation of the loss function of convolution over the first independent variable (i.e., background matrix).<br>Input param:<br>&nbsp;&nbsp;1. Derivation matrix of loss function over dependent variable; <br>&nbsp;&nbsp;2. Another independent variable,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;the second independent variable, convolutional kernel window matrix; <br>&nbsp;&nbsp;3. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a 1D array with a length of 2; <br>&nbsp;&nbsp;4. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a 1D array with a length of 4; |select <br>&nbsp;&nbsp;sm_sc.fv_d_conv_2d_dloss_dindepdt_1_ex<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.1, 1.1, 1.1],[1.1, 1.1, 1.1],[1.1, 1.1, 1.1]],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0.5, 1.1, 0.5],[1.1, 2.1, 1.1],[0.5, 1.1, 0.5]],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2,2],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,1,1,0]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[0.257, 0.269, 0.257, 0.269, 0.257, 0.134]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.269, 0.244, 0.269, 0.244, 0.269, 0.122]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.257, 0.269, 0.257, 0.269, 0.257, 0.134]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.269, 0.244, 0.269, 0.244, 0.269, 0.122]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.257, 0.269, 0.257, 0.269, 0.257, 0.134]]
>>> sm_sc.<br>fv_d_conv_2d_dloss_dindepdt_2_py               |Derivation of the loss function of convolution over the second independent variable (i.e., window matrix).<br>Input param:<br>&nbsp;&nbsp;1. Another independent variable matrix,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;the first param, the background matrix; <br>&nbsp;&nbsp;2. Derivation matrix of loss function over dependent variable; <br>&nbsp;&nbsp;3. The height and width of this independent variable,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;the second specification, convolutional kernel window matrix height and width; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;this parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;4. Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;6. Fill in the element values; |select <br>&nbsp;&nbsp;sm_sc.fv_d_conv_2d_dloss_dindepdt_2_py<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4, 5, 6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[10, 20, 30, 40, 50, 60]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[100, 200, 300, 400, 500, 600]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-1, -2, -3, -4, -5, -6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-10, -20, -30, -40, -50, -60]],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0.5, 1.1, 0.5],[1.1, 2.1, 1.1],[0.5, 1.1, 0.5]],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[3,3],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2,2],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,1,1,0],<br>&nbsp;&nbsp;&nbsp;&nbsp;0<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[66.800, 50.100, 33.400]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1644.400, 1233.300, 822.200]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[163.600, 122.700, 81.800]]
>>> sm_sc.<br>fv_d_conv_2d_dloss_dindepdt_3                  |Derivation of the loss function of convolution over the third independent variable (i.e. offset).<br>Input param:<br>&nbsp;&nbsp;1.  Derivation matrix of loss function over dependent variable; <br>&nbsp;&nbsp;2. Specification of this independent variable,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;the third independent variable specification; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_conv_2d_dloss_dindepdt_3<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,2.0,3.0,4.0,5.0,6.0,7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[10.0,20.0,30.0,40.0,50.0,60.0,70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[100.0,200.0,300.0,400.0,500.0,600.0,700.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-1.0,-2.0,-3.0,-4.0,-5.0,-6.0,-7.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-10.0,-20.0,-30.0,-40.0,-50.0,-60.0,-70.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[[2800.0]],&nbsp;&nbsp;[[2800.0]]]
>>> sm_sc.<br>fv_d_conv_2d_grp_x_dloss_dindepdt_1            |Derivation of the loss function of matrix x direction convolution over the first independent variable,<br>That is,derivation of the loss function over the background matrix.<br>Input param:<br>&nbsp;&nbsp;1. Width of 2D samples.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Usually the width of the image, which is 14 in the following example; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;for example, 100 images with 28\*14 pixels,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;first, convert 2D pixels into a 1D vector with a length of 392,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;reaggregate to synthesize a 2D matrix with a height and width of 100\*392; <br>&nbsp;&nbsp;2. The matrix after the loss function takes the derivative of the dependent variable and then reduces the dimensionality of the stacked matrix; <br>&nbsp;&nbsp;3. Single row matrix after dimensionality reduction with convolutional kernel.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If there is an offset, then the length of the one-dimensional array is 1 more.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;For example, after reducing the dimensionality of a 3\*4 convolutional kernel matrix,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;if there is no offset, it is a 1\*12 matrix,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;if there is an offset, it is 1\*12 + 1 = 13 matrix; <br>&nbsp;&nbsp;4. The width of the convolutional kernel window.<br>&nbsp;&nbsp;5. Vertical and horizontal step sizes of convolutional kernel sliding.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;6. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; |<br>select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_conv_2d_grp_x_dloss_dindepdt_1<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;6,<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2.1,&nbsp;3.1,&nbsp;0.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;-1.1,&nbsp;-0.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.1,&nbsp;1.1,&nbsp;-1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.1,&nbsp;-2.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.1,&nbsp;3.1,&nbsp;2.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;-1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.5,&nbsp;1.1,&nbsp;0.5<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;2.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.5,&nbsp;1.1,&nbsp;0.5<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[2,&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[1,&nbsp;1,&nbsp;1,&nbsp;0]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.257,&nbsp;0.269,&nbsp;0.257,&nbsp;0.269,&nbsp;0.257,&nbsp;0.134<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.269,&nbsp;0.244,&nbsp;0.269,&nbsp;0.244,&nbsp;0.269,&nbsp;0.122<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.257,&nbsp;0.269,&nbsp;0.257,&nbsp;0.269,&nbsp;0.257,&nbsp;0.134<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.269,&nbsp;0.244,&nbsp;0.269,&nbsp;0.244,&nbsp;0.269,&nbsp;0.122<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.257,&nbsp;0.269,&nbsp;0.257,&nbsp;0.269,&nbsp;0.257,&nbsp;0.134<br>&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.490,&nbsp;0.636,&nbsp;0.723,&nbsp;0.391,&nbsp;0.023,&nbsp;0.012<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.391,&nbsp;0.289,&nbsp;0.244,&nbsp;0.111,&nbsp;0.000,&nbsp;0.000<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.257,&nbsp;0.000,&nbsp;-0.257,&nbsp;-0.147,&nbsp;-0.023,&nbsp;-0.012<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;-0.067,&nbsp;-0.147,&nbsp;-0.067<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-0.257,&nbsp;0.000,&nbsp;0.257,&nbsp;0.000,&nbsp;-0.257,&nbsp;-0.134<br>&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.023,&nbsp;-0.244,&nbsp;-0.490,&nbsp;-0.122,&nbsp;0.257,&nbsp;0.134<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-0.122,&nbsp;0.000,&nbsp;0.122,&nbsp;0.233,&nbsp;0.391,&nbsp;0.178<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-0.257,&nbsp;0.244,&nbsp;0.723,&nbsp;0.636,&nbsp;0.490,&nbsp;0.257<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.000,&nbsp;0.111,&nbsp;0.244,&nbsp;0.289,&nbsp;0.391,&nbsp;0.178<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.257,&nbsp;0.000,&nbsp;-0.257,&nbsp;0.000,&nbsp;0.257,&nbsp;0.134<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_2d_grp_x_dloss_dindepdt_2            |Derivation of the loss function of matrix x direction convolution over the second independent variable,<br>that is,derivation of the loss function over the convolutional kernel window matrix.<br>Input param:<br>&nbsp;&nbsp;1. The 2D original matrix composed of each row of vectors converted from 2D to 1D again; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;for example, 100 images with 28\*14 pixels,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;first, convert 2D pixels into a 1D vector with a length of 392,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;reaggregate to synthesize a 2D matrix with a height and width of 100\*392; <br>&nbsp;&nbsp;2. Width of 2D samples.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Usually the width of the image, which is 14 in the param 1 example; <br>&nbsp;&nbsp;3. The matrix after the loss function takes the derivative of the dependent variable and then reduces the dimensionality of the stacked matrix; <br>&nbsp;&nbsp;4. The width of the convolutional kernel window.<br>&nbsp;&nbsp;5. Does the convolutional kernel attach an offset.<br>&nbsp;&nbsp;6. Vertical and horizontal step sizes of convolutional kernel sliding.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;7. Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;8. Fill in the element values;  |<br>select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_conv_2d_grp_x_dloss_dindepdt_2<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.0,&nbsp;2.0,&nbsp;3.0,&nbsp;4.0,&nbsp;5.0,&nbsp;6.0,&nbsp;7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;10.0,&nbsp;20.0,&nbsp;30.0,&nbsp;40.0,&nbsp;50.0,&nbsp;60.0,&nbsp;70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;100.0,&nbsp;200.0,&nbsp;300.0,&nbsp;400.0,&nbsp;500.0,&nbsp;600.0,&nbsp;700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.0,&nbsp;-2.0,&nbsp;-3.0,&nbsp;-4.0,&nbsp;-5.0,&nbsp;-6.0,&nbsp;-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-10.0,&nbsp;-20.0,&nbsp;-30.0,&nbsp;-40.0,&nbsp;-50.0,&nbsp;-60.0,&nbsp;-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.0,&nbsp;2.0,&nbsp;3.0,&nbsp;4.0,&nbsp;5.0,&nbsp;6.0,&nbsp;7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.0,&nbsp;-2.0,&nbsp;-3.0,&nbsp;-4.0,&nbsp;-5.0,&nbsp;-6.0,&nbsp;-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;10.0,&nbsp;20.0,&nbsp;30.0,&nbsp;40.0,&nbsp;50.0,&nbsp;60.0,&nbsp;70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;100.0,&nbsp;200.0,&nbsp;300.0,&nbsp;400.0,&nbsp;500.0,&nbsp;600.0,&nbsp;700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-10.0,&nbsp;-20.0,&nbsp;-30.0,&nbsp;-40.0,&nbsp;-50.0,&nbsp;-60.0,&nbsp;-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.0,&nbsp;2.0,&nbsp;3.0,&nbsp;4.0,&nbsp;5.0,&nbsp;6.0,&nbsp;7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.0,&nbsp;-2.0,&nbsp;-3.0,&nbsp;-4.0,&nbsp;-5.0,&nbsp;-6.0,&nbsp;-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;10.0,&nbsp;20.0,&nbsp;30.0,&nbsp;40.0,&nbsp;50.0,&nbsp;60.0,&nbsp;70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-10.0,&nbsp;-20.0,&nbsp;-30.0,&nbsp;-40.0,&nbsp;-50.0,&nbsp;-60.0,&nbsp;-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;100.0,&nbsp;200.0,&nbsp;300.0,&nbsp;400.0,&nbsp;500.0,&nbsp;600.0,&nbsp;700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;7,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2.1,&nbsp;3.1,&nbsp;0.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;-1.1,&nbsp;-0.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.1,&nbsp;-2.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.1,&nbsp;3.1,&nbsp;2.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[3,&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;true&nbsp;&nbsp;&nbsp;&nbsp;--&nbsp;false<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[2,&nbsp;2]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;4404.000,&nbsp;3652.000,&nbsp;2900.000<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-430.500,&nbsp;-404.800,&nbsp;-379.100<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1926.500,&nbsp;1548.800,&nbsp;1171.100<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;15.000<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_pool_none_dloss_dindepdt |Derivation of the loss function of element preservation invariant pooling over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The width and height of the independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Size of pooling window height and width; <br>&nbsp;&nbsp;3.&nbsp;Derivation matrix of loss function over dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; |select&nbsp;&nbsp;sm_sc.fv_d_pool_none_dloss_dindepdt<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[4,&nbsp;&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>,&nbsp;&nbsp;array[3,&nbsp;&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;<br>,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8,&nbsp;&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17,&nbsp;&nbsp;18,&nbsp;&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25,&nbsp;&nbsp;26,&nbsp;&nbsp;27,&nbsp;&nbsp;28,&nbsp;&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35,&nbsp;&nbsp;36,&nbsp;&nbsp;37,&nbsp;&nbsp;38,&nbsp;&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45,&nbsp;&nbsp;46,&nbsp;&nbsp;47,&nbsp;&nbsp;48,&nbsp;&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[51,&nbsp;&nbsp;52,&nbsp;&nbsp;53,&nbsp;&nbsp;54,&nbsp;&nbsp;55,&nbsp;&nbsp;56,&nbsp;&nbsp;57,&nbsp;&nbsp;58,&nbsp;&nbsp;59]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>,&nbsp;&nbsp;array[1,&nbsp;&nbsp;1]<br>,&nbsp;&nbsp;array[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;6,&nbsp;&nbsp;15,&nbsp;&nbsp;14,&nbsp;&nbsp;9]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[42,&nbsp;&nbsp;92,&nbsp;&nbsp;150,&nbsp;&nbsp;108,&nbsp;&nbsp;58]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[62,&nbsp;&nbsp;132,&nbsp;&nbsp;210,&nbsp;&nbsp;148,&nbsp;&nbsp;78]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[51,&nbsp;&nbsp;106,&nbsp;&nbsp;165,&nbsp;&nbsp;114,&nbsp;&nbsp;59]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_pool_avg_dloss_dindepdt<br>fv_d_pool_avg_dloss_dindepdt_ex         |Derivation of average pooling loss function over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The width and height of the independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Size of pooling window height and width; <br>&nbsp;&nbsp;3.&nbsp;Derivation matrix of loss function over dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_pool_avg_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[5,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;array[3,&nbsp;3]<br>&nbsp;&nbsp;,&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,&nbsp;1.1,&nbsp;1.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[1.1,&nbsp;1.1,&nbsp;1.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;,&nbsp;array[2,&nbsp;2]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[0.020,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.020]<br>&nbsp;&nbsp;,&nbsp;[0.020,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.020]<br>&nbsp;&nbsp;,&nbsp;[0.041,&nbsp;0.041,&nbsp;0.081,&nbsp;0.041,&nbsp;0.081,&nbsp;0.041,&nbsp;0.041]<br>&nbsp;&nbsp;,&nbsp;[0.020,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.020]<br>&nbsp;&nbsp;,&nbsp;[0.020,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.020]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_pool_avg_grp_x_dloss_dindepdt             |Derivation of the loss function of average pooling in the x-direction of matrix over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Width of 2D samples.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Usually, it is the width of the image, which is &nbsp;14 in the parameter example; <br>&nbsp;&nbsp;2.&nbsp;The matrix after the loss function takes the derivative of the dependent variable and then reduces the dimensionality of the stacked matrix; <br>&nbsp;&nbsp;3.&nbsp;The width of the pooling window.<br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_pool_avg_grp_x_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;7,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2.1,&nbsp;3.1,&nbsp;0.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;-1.1,&nbsp;-0.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.1,&nbsp;-2.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.1,&nbsp;3.1,&nbsp;2.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[3,&nbsp;3],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;2]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.020,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.020<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.020,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.020<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.041,&nbsp;0.041,&nbsp;0.081,&nbsp;0.041,&nbsp;0.081,&nbsp;0.041,&nbsp;0.041<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.020,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.020<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.020,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.041,&nbsp;0.020,&nbsp;0.020<br>&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.039,&nbsp;0.039,&nbsp;0.096,&nbsp;0.057,&nbsp;0.059,&nbsp;0.002,&nbsp;0.002<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.039,&nbsp;0.039,&nbsp;0.096,&nbsp;0.057,&nbsp;0.059,&nbsp;0.002,&nbsp;0.002<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.059,&nbsp;0.059,&nbsp;0.096,&nbsp;0.037,&nbsp;0.037,&nbsp;0.000,&nbsp;0.000<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.020,&nbsp;0.020,&nbsp;0.000,&nbsp;-0.020,&nbsp;-0.022,&nbsp;-0.002,&nbsp;-0.002<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.020,&nbsp;0.020,&nbsp;0.000,&nbsp;-0.020,&nbsp;-0.022,&nbsp;-0.002,&nbsp;-0.002<br>&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.002,&nbsp;0.002,&nbsp;-0.037,&nbsp;-0.039,&nbsp;-0.019,&nbsp;0.020,&nbsp;0.020<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0.002,&nbsp;0.002,&nbsp;-0.037,&nbsp;-0.039,&nbsp;-0.019,&nbsp;0.020,&nbsp;0.020<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-0.019,&nbsp;-0.019,&nbsp;0.000,&nbsp;0.019,&nbsp;0.078,&nbsp;0.059,&nbsp;0.059<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-0.020,&nbsp;-0.020,&nbsp;0.037,&nbsp;0.057,&nbsp;0.096,&nbsp;0.039,&nbsp;0.039<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-0.020,&nbsp;-0.020,&nbsp;0.037,&nbsp;0.057,&nbsp;0.096,&nbsp;0.039,&nbsp;0.039<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_pool_max_dloss_dindepdt<br>fv_d_pool_max_dloss_dindepdt_ex         |Derivation of the Maximum pooling loss function over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;Size of pooling window height and width; <br>&nbsp;&nbsp;4.&nbsp;Derivation matrix of loss function over dependent variable; <br>&nbsp;&nbsp;5.&nbsp;Dependent variable matrix; <br>&nbsp;&nbsp;6.&nbsp;Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;7.&nbsp;Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;8.&nbsp;Fill in the element values; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_pool_max_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[10,&nbsp;20,&nbsp;30,&nbsp;40,&nbsp;50,&nbsp;60]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[100,&nbsp;200,&nbsp;300,&nbsp;400,&nbsp;500,&nbsp;600]&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1,&nbsp;-2,&nbsp;-3,&nbsp;-4,&nbsp;-5,&nbsp;-6]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-10,&nbsp;-20,&nbsp;-30,&nbsp;-40,&nbsp;-50,&nbsp;-60]<br>&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,&nbsp;1.1,&nbsp;1.1],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,&nbsp;1.1,&nbsp;1.1],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,&nbsp;1.1,&nbsp;1.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;null&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;2]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;1,&nbsp;1,&nbsp;0]&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;0<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[0.122,&nbsp;0.000,&nbsp;0.122,&nbsp;0.000,&nbsp;0.122,&nbsp;0.000]<br>&nbsp;&nbsp;,&nbsp;[0.122,&nbsp;0.000,&nbsp;0.122,&nbsp;0.000,&nbsp;0.122,&nbsp;0.000]<br>&nbsp;&nbsp;,&nbsp;[0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000]<br>&nbsp;&nbsp;,&nbsp;[0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000]<br>&nbsp;&nbsp;,&nbsp;[0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000,&nbsp;0.000]]
>>> sm_sc.<br>fv_d_pool_max_grp_x_dloss_dindepdt             |Derivation of the loss function of maximum pooling in the x-direction of a matrix over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The 2D original matrix composed of each row of vectors converted from 2D to 1D again; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;for example: 100 &nbsp;2814&nbsp; pixel images,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;first, convert 2D pixels into a 1D vector with a length of 392,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;reaggregate to synthesize a 2D matrix with height and width &nbsp;100392&nbsp;; <br>&nbsp;&nbsp;2.&nbsp;Width of 2D samples.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Usually, it is the width of the image, which is &nbsp;14 in the parameter example; <br>&nbsp;&nbsp;3.&nbsp;The matrix after the loss function takes the derivative of the dependent variable and then reduces the dimensionality of the stacked matrix; <br>&nbsp;&nbsp;4.&nbsp;The width of the pooling window; <br>&nbsp;&nbsp;5.&nbsp;Matrix after dimensionality reduction and stacking of dependent variables; <br>&nbsp;&nbsp;6.&nbsp;Sliding window step sizes at vertical and horizontal.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 2; <br>&nbsp;&nbsp;7.&nbsp;Complete the number of rows/columns at top, bottom, left and right.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This parameter is a one-dimensional array with a length of 4; <br>&nbsp;&nbsp;8.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_pool_max_grp_x_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.0,&nbsp;2.0,&nbsp;3.0,&nbsp;4.0,&nbsp;5.0,&nbsp;6.0,&nbsp;7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;10.0,&nbsp;20.0,&nbsp;30.0,&nbsp;40.0,&nbsp;50.0,&nbsp;60.0,&nbsp;70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;100.0,&nbsp;200.0,&nbsp;300.0,&nbsp;400.0,&nbsp;500.0,&nbsp;600.0,&nbsp;700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.0,&nbsp;-2.0,&nbsp;-3.0,&nbsp;-4.0,&nbsp;-5.0,&nbsp;-6.0,&nbsp;-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-10.0,&nbsp;-20.0,&nbsp;-30.0,&nbsp;-40.0,&nbsp;-50.0,&nbsp;-60.0,&nbsp;-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.0,&nbsp;2.0,&nbsp;3.0,&nbsp;4.0,&nbsp;5.0,&nbsp;6.0,&nbsp;7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.0,&nbsp;-2.0,&nbsp;-3.0,&nbsp;-4.0,&nbsp;-5.0,&nbsp;-6.0,&nbsp;-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;10.0,&nbsp;20.0,&nbsp;30.0,&nbsp;40.0,&nbsp;50.0,&nbsp;60.0,&nbsp;70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;100.0,&nbsp;200.0,&nbsp;300.0,&nbsp;400.0,&nbsp;500.0,&nbsp;600.0,&nbsp;700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-10.0,&nbsp;-20.0,&nbsp;-30.0,&nbsp;-40.0,&nbsp;-50.0,&nbsp;-60.0,&nbsp;-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.0,&nbsp;2.0,&nbsp;3.0,&nbsp;4.0,&nbsp;5.0,&nbsp;6.0,&nbsp;7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.0,&nbsp;-2.0,&nbsp;-3.0,&nbsp;-4.0,&nbsp;-5.0,&nbsp;-6.0,&nbsp;-7.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;10.0,&nbsp;20.0,&nbsp;30.0,&nbsp;40.0,&nbsp;50.0,&nbsp;60.0,&nbsp;70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-10.0,&nbsp;-20.0,&nbsp;-30.0,&nbsp;-40.0,&nbsp;-50.0,&nbsp;-60.0,&nbsp;-70.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;100.0,&nbsp;200.0,&nbsp;300.0,&nbsp;400.0,&nbsp;500.0,&nbsp;600.0,&nbsp;700.0<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;7,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;1.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2.1,&nbsp;3.1,&nbsp;0.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1.1,&nbsp;-1.1,&nbsp;-0.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.1,&nbsp;-2.1,&nbsp;1.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;-1.1,&nbsp;3.1,&nbsp;2.1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[3,&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;null<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[2,&nbsp;2]<br>&nbsp;&nbsp;);<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0.367,&nbsp;0,&nbsp;0.367,&nbsp;0,&nbsp;0.367<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0.350,&nbsp;0,&nbsp;0.517,&nbsp;0,&nbsp;0.017<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0.183,&nbsp;0,&nbsp;-0.183,&nbsp;0,&nbsp;-0.017<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0]<br>&nbsp;&nbsp;,&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0.017,&nbsp;0,&nbsp;-0.350,&nbsp;0,&nbsp;0.183<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0,&nbsp;0<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;0,&nbsp;0,&nbsp;-0.183,&nbsp;0,&nbsp;0.517,&nbsp;0,&nbsp;0.350<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_add_dloss_dindepdt_1                 |Convolutional add derivative, loss function differentiation of the first independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_add_dloss_dindepdt_1<br>(&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1,&nbsp;6,&nbsp;15,&nbsp;14,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[42,&nbsp;92,&nbsp;150,&nbsp;108,&nbsp;58]<br>&nbsp;&nbsp;,&nbsp;[62,&nbsp;132,&nbsp;210,&nbsp;148,&nbsp;78]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;106,&nbsp;165,&nbsp;114,&nbsp;59]]
>>> sm_sc.<br>fv_d_conv_add_dloss_dindepdt_2                 |Convolutional add derivative, loss function derivative of the second independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right;                        |select&nbsp;sm_sc.fv_d_conv_add_dloss_dindepdt_2<br>(&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[]<br>);<br><br>return:<br>&nbsp;&nbsp;array[[114,120,126],[174,180,186],[234,240,246]]         
>>> sm_sc.<br>fv_d_conv_sub_dloss_dindepdt_1                 |Convolutional sub derivative, loss function differentiation of the first independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_sub_dloss_dindepdt_1<br>(&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1,&nbsp;6,&nbsp;15,&nbsp;14,&nbsp;9]<br>&nbsp;&nbsp;,[42,&nbsp;92,&nbsp;150,&nbsp;108,&nbsp;58]<br>&nbsp;&nbsp;,[62,&nbsp;132,&nbsp;210,&nbsp;148,&nbsp;78]<br>&nbsp;&nbsp;,[51,&nbsp;106,&nbsp;165,&nbsp;114,&nbsp;59]]
>>> sm_sc.<br>fv_d_conv_sub_dloss_dindepdt_2                 |Convolutional sub derivative, loss function derivative of the second independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_sub_dloss_dindepdt_2<br>(&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[-114,&nbsp;-120,&nbsp;-126]<br>&nbsp;&nbsp;,[-174,&nbsp;-180,&nbsp;-186]<br>&nbsp;&nbsp;,[-234,&nbsp;-240,&nbsp;-246]]
>>> sm_sc.<br>fv_d_conv_mul_dloss_dindepdt_1                 |Convolutional mul derivative, loss function differentiation of the first independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_mul_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,<br>&nbsp;&nbsp;array[[1,4,7],[2,5,8],[3,6,9]]&nbsp;::&nbsp;float[],<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[]<br>);<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1,&nbsp;12,&nbsp;48,&nbsp;74,&nbsp;63]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[53,&nbsp;250,&nbsp;621,&nbsp;622,&nbsp;425]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[145,&nbsp;502,&nbsp;1101,&nbsp;1010,&nbsp;653]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[153,&nbsp;474,&nbsp;978,&nbsp;852,&nbsp;531]]
>>> sm_sc.<br>fv_d_conv_mul_dloss_dindepdt_2                 |Convolutional mul derivative, loss function derivative of the second independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;6.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_mul_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array[[1,2,3,4,5],[3,4,5,6,7],[2,4,6,8,5],[7,5,3,4,8]]&nbsp;::&nbsp;float[],<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]<br>&nbsp;&nbsp;,[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]<br>&nbsp;&nbsp;,[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]<br>&nbsp;&nbsp;,[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]<br>&nbsp;&nbsp;,[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;,<br>&nbsp;&nbsp;0.0&nbsp;&nbsp;::&nbsp;float<br>);<br><br>return:<br>&nbsp;&nbsp;array[[444,582,732],[714,1053,1165],[1098,1119,1346]]
>>> sm_sc.<br>fv_d_conv_div_dloss_dindepdt_1                 |Convolutional div derivative, loss function differentiation of the first independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_div_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[[1,4,7],[2,5,8],[3,6,9]]&nbsp;::&nbsp;float[],&nbsp;&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1.000,&nbsp;4.500,&nbsp;8.679,&nbsp;2.857,&nbsp;1.286]<br>&nbsp;&nbsp;,[36.500,&nbsp;51.400,&nbsp;63.589,&nbsp;20.243,&nbsp;7.946]<br>&nbsp;&nbsp;,[27.500,&nbsp;42.067,&nbsp;53.597,&nbsp;22.906,&nbsp;9.347]<br>&nbsp;&nbsp;,[17.000,&nbsp;26.667,&nbsp;34.056,&nbsp;15.889,&nbsp;6.556]]
>>> sm_sc.<br>fv_d_conv_div_dloss_dindepdt_2                 |Convolutional div derivative, loss function derivative of the second independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;6.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_div_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array[[1,2,3,4,5],[3,4,5,6,7],[2,4,6,8,5],[7,5,3,4,8]]&nbsp;::&nbsp;float[],<br>&nbsp;&nbsp;array[[1,4,7],[2,5,8],[3,6,9]]&nbsp;::&nbsp;float[],<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],<br>&nbsp;&nbsp;array[1,&nbsp;1]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[-9.827,&nbsp;-24.044,&nbsp;-29.290]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[-34.166,&nbsp;-32.375,&nbsp;-41.803]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,[-51.102,&nbsp;-85.938,&nbsp;-112.643]]
>>> sm_sc.<br>fv_d_conv_pow_dloss_dindepdt_1                 |Convolutional pow derivative, loss function differentiation of the first independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;3.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; <br>&nbsp;&nbsp;4.&nbsp;Height and width specifications of the dependent variable matrix; <br>&nbsp;&nbsp;5.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;6.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;7.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_pow_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[[1,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_conv_pow<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[]<br>&nbsp;&nbsp;),&nbsp;<br>&nbsp;&nbsp;array[6,&nbsp;9],&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[1.1,&nbsp;1.2,&nbsp;1.3,&nbsp;1.4,&nbsp;1.5,&nbsp;1.6,&nbsp;1.7,&nbsp;1.8,&nbsp;1.9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[2.1,&nbsp;2.2,&nbsp;2.3,&nbsp;2.4,&nbsp;2.5,&nbsp;2.6,&nbsp;2.7,&nbsp;2.8,&nbsp;2.9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[3.1,&nbsp;3.2,&nbsp;3.3,&nbsp;3.4,&nbsp;3.5,&nbsp;3.6,&nbsp;3.7,&nbsp;3.8,&nbsp;3.9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[4.1,&nbsp;4.2,&nbsp;4.3,&nbsp;4.4,&nbsp;4.5,&nbsp;4.6,&nbsp;4.7,&nbsp;4.8,&nbsp;4.9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[5.1,&nbsp;5.2,&nbsp;5.3,&nbsp;5.4,&nbsp;5.5,&nbsp;5.6,&nbsp;5.7,&nbsp;5.8,&nbsp;5.9]]&nbsp;::&nbsp;float[]<br>)&nbsp;<br><br>return:<br>&nbsp;&nbsp;array&nbsp;<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.000,&nbsp;68.000,&nbsp;15856.000,&nbsp;174080.000,&nbsp;984375.000]<br>&nbsp;&nbsp;,&nbsp;[9.700,&nbsp;2369.800,&nbsp;1179895.700,&nbsp;4773859.200,&nbsp;15729671.300]<br>&nbsp;&nbsp;,&nbsp;[41.600,&nbsp;19043.200,&nbsp;44543997.600,&nbsp;470410854.400,&nbsp;13257812.500]<br>&nbsp;&nbsp;,&nbsp;[749.700,&nbsp;97905.000,&nbsp;321132.600,&nbsp;3338649.600,&nbsp;890870169.600]<br>&nbsp;&nbsp;]<br>
>>> sm_sc.<br>fv_d_conv_pow_dloss_dindepdt_2                 |Convolutional pow derivative, loss function derivative of the second independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;3.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;4.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If passing the input param, the window matrix of the second param can be omitted; <br>&nbsp;&nbsp;5.&nbsp;Height and width specifications of the dependent variable matrix; <br>&nbsp;&nbsp;6.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;7.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;8.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;9.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_pow_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;null,&nbsp;--&nbsp;array[[1,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_conv_pow<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;0.0&nbsp;&nbsp;::&nbsp;float<br>&nbsp;&nbsp;),&nbsp;<br>&nbsp;&nbsp;array[6,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;0.0&nbsp;&nbsp;::&nbsp;float<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[617.069,&nbsp;138109.798,&nbsp;85980348.256]<br>&nbsp;&nbsp;,&nbsp;[5224.588,&nbsp;4300539.139,&nbsp;2034488125.210]<br>&nbsp;&nbsp;,&nbsp;[59289.615,&nbsp;19159053.803,&nbsp;24251286231.684]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_log_dloss_dindepdt_1                 |Convolutional log derivative, loss function differentiation of the first independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;The second independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;4.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;5.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; <br>&nbsp;&nbsp;6.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;7.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;8.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_log_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[[1.3,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_conv_log<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;0.5&nbsp;&nbsp;::&nbsp;float<br>&nbsp;&nbsp;),&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;--&nbsp;,&nbsp;<br>&nbsp;&nbsp;--&nbsp;0.5&nbsp;&nbsp;::&nbsp;float<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[7.893,&nbsp;3.978,&nbsp;4.034,&nbsp;2.962,&nbsp;1.352]<br>&nbsp;&nbsp;,&nbsp;[4.352,&nbsp;10.706,&nbsp;14.315,&nbsp;9.603,&nbsp;4.354]<br>&nbsp;&nbsp;,&nbsp;[53.585,&nbsp;21.318,&nbsp;16.582,&nbsp;8.100,&nbsp;12.787]<br>&nbsp;&nbsp;,&nbsp;[2.114,&nbsp;11.774,&nbsp;76.673,&nbsp;29.525,&nbsp;3.748]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_log_dloss_dindepdt_2                 |Convolutional log derivative, loss function derivative of the second independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;6.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_log_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[[1.3,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;0.5&nbsp;&nbsp;::&nbsp;float<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[72.659,&nbsp;19.811,&nbsp;10.470]<br>&nbsp;&nbsp;,&nbsp;[73.898,&nbsp;21.303,&nbsp;12.917]<br>&nbsp;&nbsp;,&nbsp;[58.107,&nbsp;27.916,&nbsp;17.819]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_sub_dloss_dindepdt_1              |Deconvolutional sub derivative, loss function derivative of the first independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_de_sub_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[114,&nbsp;120,&nbsp;126]<br>&nbsp;&nbsp;,&nbsp;[174,&nbsp;180,&nbsp;186]<br>&nbsp;&nbsp;,&nbsp;[234,&nbsp;240,&nbsp;246]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_sub_dloss_dindepdt_2              |Deconvolutional sub derivative, loss function derivative of the second independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_de_sub_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array[3,&nbsp;3]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[-1,&nbsp;-6,&nbsp;-15,&nbsp;-14,&nbsp;-9]<br>&nbsp;&nbsp;,&nbsp;[-42,&nbsp;-92,&nbsp;-150,&nbsp;-108,&nbsp;-58]<br>&nbsp;&nbsp;,&nbsp;[-62,&nbsp;-132,&nbsp;-210,&nbsp;-148,&nbsp;-78]<br>&nbsp;&nbsp;,&nbsp;[-51,&nbsp;-106,&nbsp;-165,&nbsp;-114,&nbsp;-59]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_div_dloss_dindepdt_1              |Deconvolutional div derivative, loss function derivative of the first independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_de_div_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array[3,&nbsp;3],&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;-1,&nbsp;-1,&nbsp;1],&nbsp;<br>&nbsp;&nbsp;2.0&nbsp;&nbsp;::&nbsp;float<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[42.083,&nbsp;34.517,&nbsp;43.950]<br>&nbsp;&nbsp;,&nbsp;[41.300,&nbsp;34.107,&nbsp;54.188]<br>&nbsp;&nbsp;,&nbsp;[43.425,&nbsp;39.708,&nbsp;69.373]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_div_dloss_dindepdt_2              |Deconvolutional div derivative, loss function derivative of the second independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;6.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_de_div_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array[[1,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[-1.000,&nbsp;-8.250,&nbsp;-22.121,&nbsp;-2.490,&nbsp;-0.918]<br>&nbsp;&nbsp;,&nbsp;[-101.250,&nbsp;-159.920,&nbsp;-224.570,&nbsp;-24.478,&nbsp;-7.650]<br>&nbsp;&nbsp;,&nbsp;[-25.167,&nbsp;-63.831,&nbsp;-109.202,&nbsp;-29.900,&nbsp;-5.618]<br>&nbsp;&nbsp;,&nbsp;[-39.667,&nbsp;-37.222,&nbsp;-25.546,&nbsp;-9.210,&nbsp;-5.827]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_pow_dloss_dindepdt_1              |Deconvolutional pow derivative, loss function derivative of the first independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;3.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; <br>&nbsp;&nbsp;4.&nbsp;Height and width specifications of the dependent variable matrix; <br>&nbsp;&nbsp;5.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;6.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;7.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_de_pow_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array[[1.2,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_conv_de_pow<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,&nbsp;4,&nbsp;7],&nbsp;[2,&nbsp;5,&nbsp;8],&nbsp;[3,&nbsp;6,&nbsp;9]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;1]<br>&nbsp;&nbsp;),&nbsp;<br>&nbsp;&nbsp;array[6,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[793.584,&nbsp;288768.000,&nbsp;36261372.000]<br>&nbsp;&nbsp;,&nbsp;[12536.000,&nbsp;31255125.000,&nbsp;819486720.000]<br>&nbsp;&nbsp;,&nbsp;[325746.000,&nbsp;64284084.000,&nbsp;3261695202.000]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_pow_dloss_dindepdt_2              |Deconvolutional pow derivative, loss function derivative of the second independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;3.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;4.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If the input parameter is passed, the second parameter &nbsp;-&nbsp;Window matrix may not be passed; <br>&nbsp;&nbsp;5.&nbsp;Height and width specifications of the dependent variable matrix; <br>&nbsp;&nbsp;6.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;7.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;8.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;9.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_de_pow_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.3,&nbsp;4,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;5,&nbsp;8]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;6,&nbsp;9]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5],&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_conv_de_pow<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.3,&nbsp;4,&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,&nbsp;5,&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[3,&nbsp;6,&nbsp;9]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;0.5&nbsp;::&nbsp;float<br>&nbsp;&nbsp;),&nbsp;<br>&nbsp;&nbsp;array[6,&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;0.5&nbsp;::&nbsp;float<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[0.431,&nbsp;46.135,&nbsp;2449.991,&nbsp;30871.912,&nbsp;294344.207]<br>&nbsp;&nbsp;,&nbsp;[78.866,&nbsp;23608.050,&nbsp;2090611.225,&nbsp;17631876.126,&nbsp;145356281.366]<br>&nbsp;&nbsp;,&nbsp;[321.314,&nbsp;95958.069,&nbsp;53542107.772,&nbsp;4178421244.111,&nbsp;7101391.389]<br>&nbsp;&nbsp;,&nbsp;[122535.919,&nbsp;738917.515,&nbsp;107871.033,&nbsp;941978.441,&nbsp;5580415487.941]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_log_dloss_dindepdt_1              |Deconvolutional log derivative, loss function derivative of the first independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;3.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; <br>&nbsp;&nbsp;4.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;5.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;6.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;7.&nbsp;Fill in the element values; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_conv_de_log_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.3,&nbsp;&nbsp;4,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;5,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;6,&nbsp;&nbsp;9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;4,&nbsp;&nbsp;6,&nbsp;&nbsp;8,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[7,&nbsp;&nbsp;5,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_conv_de_log<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.3,&nbsp;&nbsp;4,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;5,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;6,&nbsp;&nbsp;9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;::&nbsp;&nbsp;float[],&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;4,&nbsp;&nbsp;6,&nbsp;&nbsp;8,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[7,&nbsp;&nbsp;5,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;::&nbsp;&nbsp;float[],&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[1,&nbsp;&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;0.5&nbsp;&nbsp;::&nbsp;&nbsp;float<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8,&nbsp;&nbsp;9]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16,&nbsp;&nbsp;17,&nbsp;&nbsp;18,&nbsp;&nbsp;19]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25,&nbsp;&nbsp;26,&nbsp;&nbsp;27,&nbsp;&nbsp;28,&nbsp;&nbsp;29]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35,&nbsp;&nbsp;36,&nbsp;&nbsp;37,&nbsp;&nbsp;38,&nbsp;&nbsp;39]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45,&nbsp;&nbsp;46,&nbsp;&nbsp;47,&nbsp;&nbsp;48,&nbsp;&nbsp;49]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[51,&nbsp;&nbsp;52,&nbsp;&nbsp;53,&nbsp;&nbsp;54,&nbsp;&nbsp;55,&nbsp;&nbsp;56,&nbsp;&nbsp;57,&nbsp;&nbsp;58,&nbsp;&nbsp;59]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;::&nbsp;&nbsp;float[]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[1,&nbsp;&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;0.5&nbsp;&nbsp;::&nbsp;&nbsp;float<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[-1692.138,&nbsp;&nbsp;-24.293,&nbsp;&nbsp;-8.285]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-241.940,&nbsp;&nbsp;-24.066,&nbsp;&nbsp;-9.775]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-95.275,&nbsp;&nbsp;-18.587,&nbsp;&nbsp;-9.217]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_log_dloss_dindepdt_2              |Deconvolutional log derivative, loss function derivative of the second independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;3.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;4.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;5.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;6.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_de_log_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.3,&nbsp;4,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;5,&nbsp;8]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;6,&nbsp;9]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1.2,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]&nbsp;&nbsp;,&nbsp;<br>&nbsp;&nbsp;0.5&nbsp;&nbsp;::&nbsp;float<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[5.063,&nbsp;5.160,&nbsp;6.429,&nbsp;2.061,&nbsp;0.799[<br>&nbsp;&nbsp;,&nbsp;[26.712,&nbsp;31.417,&nbsp;34.205,&nbsp;11.298,&nbsp;4.084]<br>&nbsp;&nbsp;,&nbsp;[39.674,&nbsp;30.345,&nbsp;29.913,&nbsp;11.015,&nbsp;5.808]<br>&nbsp;&nbsp;,&nbsp;[8.736,&nbsp;16.569,&nbsp;30.999,&nbsp;11.461,&nbsp;3.153]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_prod_mx_dloss_dindepdt_1             |Convolutional matrix multiplication derivative, loss function differentiation of the first independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;The height of the background matrix sliding window height width specification.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This window specification is different from the second independent variable height width specification, and the two are matrix multiplication specifications; <br>&nbsp;&nbsp;3.&nbsp;The second independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;4.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;5.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;6.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_prod_mx_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array[4,&nbsp;5],&nbsp;<br>&nbsp;&nbsp;3,&nbsp;<br>&nbsp;&nbsp;array[[1,&nbsp;4,&nbsp;7,&nbsp;6],&nbsp;[2,&nbsp;5,&nbsp;8,&nbsp;7],&nbsp;[3,&nbsp;6,&nbsp;9,&nbsp;8]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9,&nbsp;-1,&nbsp;-2,&nbsp;-3]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19,&nbsp;-11,&nbsp;-12,&nbsp;-13]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29,&nbsp;-21,&nbsp;-22,&nbsp;-23]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39,&nbsp;-31,&nbsp;-32,&nbsp;-33]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49,&nbsp;-41,&nbsp;-42,&nbsp;-43]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59,&nbsp;-51,&nbsp;-52,&nbsp;-53]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[1,&nbsp;1],&nbsp;<br>&nbsp;&nbsp;array[0,&nbsp;0,&nbsp;0,&nbsp;0]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[54,&nbsp;190,&nbsp;199,&nbsp;154,&nbsp;-21]<br>&nbsp;&nbsp;,&nbsp;[828,&nbsp;1980,&nbsp;1678,&nbsp;628,&nbsp;-842]<br>&nbsp;&nbsp;,&nbsp;[1188,&nbsp;2780,&nbsp;2318,&nbsp;788,&nbsp;-1242]<br>&nbsp;&nbsp;,&nbsp;[954,&nbsp;2190,&nbsp;1799,&nbsp;554,&nbsp;-1021]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_prod_mx_dloss_dindepdt_2             |Convolutional matrix multiplication derivative, loss function derivative of the second independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;2.&nbsp;The height of the background matrix sliding window height width specification.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This window specification is different from the second independent variable height width specification, and the two are matrix multiplication relationships; <br>&nbsp;&nbsp;3.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;4.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;5.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;6.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;7.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_prod_mx_dloss_dindepdt_2<br>(<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;3,&nbsp;<br>&nbsp;&nbsp;array[3,&nbsp;4],&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9,&nbsp;-1,&nbsp;-2,&nbsp;-3]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19,&nbsp;-11,&nbsp;-12,&nbsp;-13]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29,&nbsp;-21,&nbsp;-22,&nbsp;-23]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39,&nbsp;-31,&nbsp;-32,&nbsp;-33]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49,&nbsp;-41,&nbsp;-42,&nbsp;-43]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59,&nbsp;-51,&nbsp;-52,&nbsp;-53]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1758,&nbsp;1992,&nbsp;1743]<br>&nbsp;&nbsp;,&nbsp;[2478,&nbsp;2932,&nbsp;2653]<br>&nbsp;&nbsp;,&nbsp;[3198,&nbsp;3872,&nbsp;3563]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_prod_mx_dloss_dindepdt_1          |Deconvolutional matrix multiplication derivative, loss function derivative of the first independent variable, i.e. derivative of the window matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Height and width specifications of the first independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;The second independent variable matrix, i.e. the background matrix; <br>&nbsp;&nbsp;3.&nbsp;The width of the background matrix sliding window height and width specifications.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This window specification is different from the first independent variable height width specification, and the two are matrix multiplication relationships; <br>&nbsp;&nbsp;4.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;5.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;6.&nbsp;Complete the number of rows/columns at top, bottom, left and right; <br>&nbsp;&nbsp;7.&nbsp;Fill in the element values; |select&nbsp;sm_sc.fv_d_conv_de_prod_mx_dloss_dindepdt_1<br>(<br>&nbsp;&nbsp;array[4,&nbsp;3],&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[2,&nbsp;4,&nbsp;6,&nbsp;8,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[7,&nbsp;5,&nbsp;3,&nbsp;4,&nbsp;8]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;3,&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]<br>&nbsp;&nbsp;,&nbsp;[61,&nbsp;62,&nbsp;63,&nbsp;64,&nbsp;65,&nbsp;66,&nbsp;67,&nbsp;68,&nbsp;69]<br>&nbsp;&nbsp;,&nbsp;[71,&nbsp;72,&nbsp;73,&nbsp;74,&nbsp;75,&nbsp;76,&nbsp;77,&nbsp;78,&nbsp;79]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[2208,&nbsp;2482,&nbsp;2163]<br>&nbsp;&nbsp;,&nbsp;[3378,&nbsp;3912,&nbsp;3493]<br>&nbsp;&nbsp;,&nbsp;[4098,&nbsp;4852,&nbsp;4403]<br>&nbsp;&nbsp;,&nbsp;[3468,&nbsp;4322,&nbsp;4053]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_conv_de_prod_mx_dloss_dindepdt_2          |Deconvolutional matrix multiplication derivative, loss function derivative of the second independent variable, i.e. derivative of the background matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;The first independent variable matrix, i.e. the window matrix; <br>&nbsp;&nbsp;2.&nbsp;Height and width specifications of the second independent variable matrix; <br>&nbsp;&nbsp;3.&nbsp;The width of the background matrix sliding window height and width specifications.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;This window specification is different from the first independent variable height width specification, and the two are matrix multiplication relationships; <br>&nbsp;&nbsp;4.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;5.&nbsp;Sliding window step sizes at vertical and horizontal; <br>&nbsp;&nbsp;6.&nbsp;Complete the number of rows/columns at top, bottom, left and right; |select&nbsp;sm_sc.fv_d_conv_de_prod_mx_dloss_dindepdt_2<br>(&nbsp;&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;4,&nbsp;7]<br>&nbsp;&nbsp;,&nbsp;[6,&nbsp;2,&nbsp;5]<br>&nbsp;&nbsp;,&nbsp;[8,&nbsp;7,&nbsp;3]<br>&nbsp;&nbsp;,&nbsp;[6,&nbsp;9,&nbsp;8]]&nbsp;::&nbsp;float[],&nbsp;<br>&nbsp;&nbsp;array[4,&nbsp;5],&nbsp;<br>&nbsp;&nbsp;3,&nbsp;<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4,&nbsp;5,&nbsp;6,&nbsp;7,&nbsp;8,&nbsp;9]<br>&nbsp;&nbsp;,&nbsp;[11,&nbsp;12,&nbsp;13,&nbsp;14,&nbsp;15,&nbsp;16,&nbsp;17,&nbsp;18,&nbsp;19]<br>&nbsp;&nbsp;,&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24,&nbsp;25,&nbsp;26,&nbsp;27,&nbsp;28,&nbsp;29]<br>&nbsp;&nbsp;,&nbsp;[31,&nbsp;32,&nbsp;33,&nbsp;34,&nbsp;35,&nbsp;36,&nbsp;37,&nbsp;38,&nbsp;39]<br>&nbsp;&nbsp;,&nbsp;[41,&nbsp;42,&nbsp;43,&nbsp;44,&nbsp;45,&nbsp;46,&nbsp;47,&nbsp;48,&nbsp;49]<br>&nbsp;&nbsp;,&nbsp;[61,&nbsp;62,&nbsp;63,&nbsp;64,&nbsp;65,&nbsp;66,&nbsp;67,&nbsp;68,&nbsp;69]<br>&nbsp;&nbsp;,&nbsp;[71,&nbsp;72,&nbsp;73,&nbsp;74,&nbsp;75,&nbsp;76,&nbsp;77,&nbsp;78,&nbsp;79]<br>&nbsp;&nbsp;,&nbsp;[51,&nbsp;52,&nbsp;53,&nbsp;54,&nbsp;55,&nbsp;56,&nbsp;57,&nbsp;58,&nbsp;59]<br>&nbsp;&nbsp;]&nbsp;::&nbsp;float[]<br>)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[30,&nbsp;91,&nbsp;197,&nbsp;236,&nbsp;256]<br>&nbsp;&nbsp;,&nbsp;[540,&nbsp;1182,&nbsp;2114,&nbsp;2632,&nbsp;2152]<br>&nbsp;&nbsp;,&nbsp;[780,&nbsp;1682,&nbsp;2974,&nbsp;3712,&nbsp;2972]<br>&nbsp;&nbsp;,&nbsp;[750,&nbsp;1591,&nbsp;2777,&nbsp;3476,&nbsp;2716]<br>&nbsp;&nbsp;]
>>> sm_sc.fv_d_mx_sum                         |Derivation of sum of aggregate functions.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable height and width specifications; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_mx_sum(array[2,&nbsp;3])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[1,&nbsp;1,&nbsp;1],&nbsp;[1,&nbsp;1,&nbsp;1]]
>>> sm_sc.fv_d_mx_avg                         |Derivation of average aggregation function.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Number of independent variable matrices participating in aggregation; <br>&nbsp;&nbsp;2.&nbsp;Independent variable height and width specifications; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_mx_avg(3,&nbsp;array[2,&nbsp;3])<br><br>return:<br>&nbsp;&nbsp;array[[0.333,&nbsp;0.333,&nbsp;0.333],&nbsp;[0.333,&nbsp;0.333,&nbsp;0.333]]
>>> sm_sc.fv_d_mx_max                         |Derivation of the maximum value of an aggregate function.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Dependent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_mx_max<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,&nbsp;-2.3,&nbsp;3.3],&nbsp;[1.4,&nbsp;2.3,&nbsp;3.8]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[-1.2,&nbsp;-2.3,&nbsp;-3.3],&nbsp;[1.3,&nbsp;2.2,&nbsp;3.8]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[0.000,&nbsp;1.000,&nbsp;0.000],&nbsp;[0.000,&nbsp;0.000,&nbsp;1.000]]
>>> sm_sc.fv_d_mx_min                         |Derivation of the minimum value of an aggregate function.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Dependent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_mx_min<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,&nbsp;-2.3,&nbsp;3.3],&nbsp;[1.4,&nbsp;-2.3,&nbsp;3.8]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3,&nbsp;-2.3,&nbsp;3.8],&nbsp;[1.5,&nbsp;2.3,&nbsp;3.9]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[0.000,&nbsp;1.000,&nbsp;0.000],&nbsp;[0.000,&nbsp;0.000,&nbsp;0.000]]
>>> sm_sc.fv_d_mx_prod                        |Derivation of direct product of aggregate functions.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Dependent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_mx_prod<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.2,&nbsp;-2.3,&nbsp;3.3],&nbsp;[1.4,&nbsp;-2.3,&nbsp;3.8]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.3,&nbsp;-2.3,&nbsp;3.8],&nbsp;[1.5,&nbsp;2.3,&nbsp;3.9]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[1.083,&nbsp;1.000,&nbsp;1.152],&nbsp;[1.071,&nbsp;-1.000,&nbsp;1.026]]
>>> sm_sc.fv_d_add                            |Derivation of matrix element addition.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable height and width specifications; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_add(array[2,&nbsp;3])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[1,1,1],[1,1,1]]
>>> sm_sc.fv_d_sub_1                          |Derivation of first input parameter over matrix element subtraction.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable height and width specifications; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_sub_1(array[2,&nbsp;3])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[1,1,1],[1,1,1]]
>>> sm_sc.fv_d_sub_2                          |Derivation of second input parameter over matrix element subtraction.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable height and width specifications; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_sub_2(array[2,&nbsp;3])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[-1,-1,-1],[-1,-1,-1]]
>>> sm_sc.fv_d_mul                            |Derivation of matrix element multiplication.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Another independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_mul(array[[2.8,&nbsp;3.6],&nbsp;[2.8,&nbsp;3.6]])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[2.8,&nbsp;3.6],&nbsp;[2.8,&nbsp;3.6]]
>>> sm_sc.<br>fv_d_div_1                      |Derivation of first input parameter over matrix element division.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Another independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_div_1(array[[2.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;-1.6]])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[0.357,&nbsp;0.278],&nbsp;[0.417,&nbsp;-0.625]]
>>> sm_sc.fv_d_div_2<br>sm_sc.<br>fv_d_div_2_un_de_broadcast  |Derivation of the second input parameter over matrix element division, the latter does not perform inverse operations on broadcasting.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix, i.e. the second input parameter; <br>&nbsp;&nbsp;2.&nbsp;Another independent variable matrix, i.e. the first input parameter; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_div_2<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;-1.6]]<br>&nbsp;&nbsp;,&nbsp;array[[1.8,&nbsp;-4.6],&nbsp;[1.4,&nbsp;3.6]]<br>&nbsp;&nbsp;)<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[-0.230,&nbsp;0.355],&nbsp;[-0.243,&nbsp;-1.406]]
>>> sm_sc.fv_d_pow_1<br>sm_sc.<br>fv_d_pow_1_un_de_broadcast  |Derivation of the first input parameter over matrix element power, the latter does not perform inverse operations on broadcasting.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix, i.e. the first input parameter; <br>&nbsp;&nbsp;2.&nbsp;Another independent variable matrix, i.e. the second input parameter; <br>&nbsp;&nbsp;3.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_pow_1<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.8,&nbsp;-4.6],&nbsp;[1.4,&nbsp;3.6]],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]]&nbsp;\^\`&nbsp;array[[1.8,&nbsp;-4.6],&nbsp;[1.4,&nbsp;3.6]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[4.102,&nbsp;-0.004],&nbsp;[1.987,&nbsp;12.218]]
>>> sm_sc.fv_d_pow_2<br>sm_sc.<br>fv_d_pow_2_un_de_broadcast  |Derivation of the second input parameter over matrix element power, the latter does not perform inverse operations on broadcasting.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix, i.e. the second input parameter; <br>&nbsp;&nbsp;2.&nbsp;Another independent variable matrix, i.e. the first input parameter; <br>&nbsp;&nbsp;3.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_pow_2<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.8,&nbsp;-4.6],&nbsp;[1.4,&nbsp;3.6]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[6.570,&nbsp;0.004],&nbsp;[2.982,&nbsp;2.552]]
>>> sm_sc.fv_d_exp                            |Derivation of matrix element natural constant power.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Dependent variable matrix; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If the dependent variable is not empty, the independent variable matrix may not be passed; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_exp<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;null,<br>&nbsp;&nbsp;&nbsp;&nbsp;\^\`&nbsp;array[[1.8,&nbsp;-4.6],&nbsp;[1.4,&nbsp;3.6]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[6.050,&nbsp;0.010],&nbsp;[4.055,&nbsp;36.598]]
>>> sm_sc.fv_d_log_1<br>sm_sc.<br>fv_d_log_1_un_de_broadcast  |Derivation of the first input parameter over matrix element logarithm, the latter does not perform inverse operations on broadcasting.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix, i.e. the first input parameter; <br>&nbsp;&nbsp;2.&nbsp;Another independent variable matrix, i.e. the second input parameter; <br>&nbsp;&nbsp;3.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_log_1<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;3.6]],<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]]&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\^\!\`&nbsp;array[[1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;3.6]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[-0.198,&nbsp;-0.258],&nbsp;[-0.183,&nbsp;-3.624]]
>>> sm_sc.fv_d_log_2<br>sm_sc.<br>fv_d_log_2_un_de_broadcast  |Derivation of the second input parameter over matrix element logarithm, the latter does not perform inverse operations on broadcasting.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix, i.e. the second input parameter; <br>&nbsp;&nbsp;2.&nbsp;Another Independent variable matrix, i.e. the first input parameter; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_log_2<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;3.6]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[2.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[0.540,&nbsp;0.170],&nbsp;[0.816,&nbsp;0.591]]
>>> sm_sc.fv_d_ln                             |Derivation of matrix element natural constant logarithmic.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_ln(array[[1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;3.6]])<br><br>return:<br>&nbsp;&nbsp;array[[0.556,&nbsp;0.217],&nbsp;[0.714,&nbsp;0.278]]
>>> sm_sc.fv_d_abs                            |Derivation of absolute values of matrix points.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix;|select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_abs(array[[-1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;-3.6]])<br><br>return:<br>&nbsp;&nbsp;array[[-1.0,&nbsp;1.0],&nbsp;[1.0,&nbsp;-1.0]]
>>> sm_sc.fv_d_sin                            |Derivation of matrix sin.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_sin(array[[1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;3.6]])<br><br>return:<br>&nbsp;&nbsp;array[[-0.227,&nbsp;-0.112],&nbsp;[0.170,&nbsp;-0.897]]
>>> sm_sc.fv_d_cos                            |Derivation of matrix cos.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_cos(array[[1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;3.6]])<br><br>return:<br>&nbsp;&nbsp;array[[-0.974,&nbsp;0.994],&nbsp;[-0.985,&nbsp;0.443]]
>>> sm_sc.fv_d_tan                            |Derivation of matrix tan.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_tan(array[[1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;3.6]])<br><br>return:<br>&nbsp;&nbsp;array[[19.372,&nbsp;79.503],&nbsp;[34.615,&nbsp;1.244]]
>>> sm_sc.fv_d_cot                            |Derivation of matrix cot.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_cot(array[[1.8,&nbsp;4.6],&nbsp;[1.4,&nbsp;3.6]])<br><br>return:<br>&nbsp;&nbsp;array[[-1.054,&nbsp;-1.013],&nbsp;[-1.030,&nbsp;-5.107]]
>>> sm_sc.fv_d_sec                            |Derivation of matrix sec.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_sec(array[[0.8,&nbsp;-0.6],&nbsp;[0.4,&nbsp;0.3]])<br><br>return:<br>&nbsp;&nbsp;array[[1.478,&nbsp;-0.829],&nbsp;[0.459,&nbsp;0.324]]
>>> sm_sc.fv_d_csc                            |Derivation of matrix csc.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_csc<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0.8,&nbsp;-0.6],&nbsp;[0.4,&nbsp;0.3]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_csc(array[[0.8,&nbsp;-0.6],&nbsp;[0.4,&nbsp;0.3]])<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[-1.354,&nbsp;-2.589],&nbsp;[-6.074,&nbsp;-10.939]]
>>> sm_sc.fv_d_sinh                           |Derivation of the hyperbolic sine of a matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_sinh(array[[1.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]])<br><br>return:<br>&nbsp;&nbsp;array[[3.107,&nbsp;18.313],&nbsp;[5.557,&nbsp;2.577]]
>>> sm_sc.fv_d_cosh                           |Derivation of the hyperbolic cosine of a matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_cosh(array[[1.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]])<br><br>return:<br>&nbsp;&nbsp;array[[2.942,&nbsp;18.285],&nbsp;[5.466,&nbsp;2.376]]
>>> sm_sc.fv_d_tanh                           |Derivation of the hyperbolic tangent of a matrix.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; <br>&nbsp;&nbsp;2.&nbsp;Dependent variable matrix.<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;It is not necessary to pass in the parameter, try to pass it in as much as possible to save costs; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;If the dependent variable is not empty, the independent variable matrix may not be passed; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_tanh<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;null,&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_tanh(array[[1.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]])<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[0.104,&nbsp;0.003],&nbsp;[0.032,&nbsp;0.151]]
>>> sm_sc.fv_d_asin                           |Derivation of matrix arcsine.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_asin(array[[0.8,&nbsp;0.6],&nbsp;[0.4,&nbsp;-0.6]])<br><br>return:<br>&nbsp;&nbsp;array[[1.667,&nbsp;1.250],&nbsp;[1.091,&nbsp;1.250]]
>>> sm_sc.fv_d_acos                           |Derivation of matrix arccosine.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_acos(array[[0.8,&nbsp;0.6],&nbsp;[0.4,&nbsp;-0.6]])<br><br>return:<br>&nbsp;&nbsp;array[[-1.667,&nbsp;-1.250],&nbsp;[-1.091,&nbsp;-1.250]]
>>> sm_sc.fv_d_atan                           |Derivation of matrix arctangent.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_atan(array[[0.8,&nbsp;0.6],&nbsp;[0.4,&nbsp;-0.6]])<br><br>return:<br>&nbsp;&nbsp;array[[0.610,&nbsp;0.735],&nbsp;[0.862,&nbsp;0.735]]
>>> sm_sc.fv_d_acot                           |Derivation of matrix anti cotangent.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_acot(array[[0.8,&nbsp;0.6],&nbsp;[0.4,&nbsp;-0.6]])<br><br>return:<br>&nbsp;&nbsp;array[[-0.610,&nbsp;-0.735],&nbsp;[-0.862,&nbsp;-0.735]]
>>> sm_sc.fv_d_asec                           |Derivation of matrix arc secant.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_asec(array[[1.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]])<br><br>return:<br>&nbsp;&nbsp;array[[0.371,&nbsp;0.080],&nbsp;[0.191,&nbsp;0.500]]
>>> sm_sc.fv_d_acsc                           |Derivation of matrix arc cosecant.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_acsc(array[[1.8,&nbsp;3.6],&nbsp;[2.4,&nbsp;1.6]])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[-0.371,&nbsp;-0.080],&nbsp;[-0.191,&nbsp;-0.500]]
>>> sm_sc.fv_d_asinh                          |Derivation of matrix inverse hyperbolic sine.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_asinh(array[[0.8,&nbsp;0.6],&nbsp;[0.4,&nbsp;-0.6]])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[0.781,&nbsp;0.857],&nbsp;[0.928,&nbsp;0.857]]
>>> sm_sc.fv_d_acosh                          |Derivation of matrix inverse hyperbolic cosine.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_acosh(array[[1.8,&nbsp;1.6],&nbsp;[1.4,&nbsp;2.6]])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[0.668,&nbsp;0.801],&nbsp;[1.021,&nbsp;0.417]]
>>> sm_sc.fv_d_atanh                          |Derivation of matrix inverse hyperbolic tangent.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Independent variable matrix; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_atanh(array[[1.8,&nbsp;1.6],&nbsp;[1.4,&nbsp;2.6]])<br>&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;array[[-0.446,&nbsp;-0.641],&nbsp;[-1.042,&nbsp;-0.174]]
>>> sm_sc.<br>fv_d_transpose_dloss_dindepdt       |Derivation of the loss function of matrix transposition over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_transpose_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[12.3,&nbsp;-12.3],&nbsp;[45.6,&nbsp;-45.6],&nbsp;[1.2,&nbsp;2.3]]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array[[12.3,&nbsp;45.6,&nbsp;1.2],&nbsp;[-12.3,&nbsp;-45.6,&nbsp;2.3]]
>>> sm_sc.<br>fv_d_transpose_i_dloss_dindepdt       |Derivation of the loss function of matrix inversion transpose over the independent variable.<br>Input: <br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_transpose_i_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_opr_transpose<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,&nbsp;&nbsp;5,&nbsp;&nbsp;5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:2][1:5][1:5]
>>> sm_sc.<br>fv_d_chunk_transpose_dloss_dindepdt       |Derivation of the loss function of matrix chunk transpose over the independent variable.<br>Input: <br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Chunk height and width specifications;|select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_chunk_transpose_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_chunk_transpose<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,&nbsp;&nbsp;6,&nbsp;&nbsp;15])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:2][1:6][1:15]
>>> sm_sc.<br>fv_d_transpose_nd_dloss_dindepdt    |Derivation of the loss function of 3d matrix transpose over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Two dimensions involved in transposition; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_transpose_nd_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[5,&nbsp;6,&nbsp;7,&nbsp;8],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[9,&nbsp;10,&nbsp;11,&nbsp;12]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[25,&nbsp;26,&nbsp;27,&nbsp;28],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[29,&nbsp;30,&nbsp;31,&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[1,&nbsp;2]<br>&nbsp;&nbsp;)<br><br>return:<br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;2,&nbsp;3,&nbsp;4],&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[[5,&nbsp;6,&nbsp;7,&nbsp;8],&nbsp;[25,&nbsp;26,&nbsp;27,&nbsp;28]],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[[9,&nbsp;10,&nbsp;11,&nbsp;12],&nbsp;[29,&nbsp;30,&nbsp;31,&nbsp;32]]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_turn_90_dloss_dindepdt<br>fv_d_turn_90_dloss_dindepdt_py  |Two dimensional and above arrays turn 90 degrees in the given starting and ending directions along the positive half axis of the dimension.<br>Input: <br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Given the starting and ending directions of the positive half axis of the axis; |select&nbsp;&nbsp;sm_sc.fv_d_turn_90_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[[[1,2,3,4],[11,12,13,14]]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return: &nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[[[4,14],[3,13],[2,12],[1,11]]]]
>>> sm_sc.<br>fv_d_turn_180_dloss_dindepdt<br>fv_d_turn_180_dloss_dindepdt_py  |Two dimensional and above arrays turn 180 ° along the given axis. <br>Input: <br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Given two dimensional axes; |select&nbsp;&nbsp;sm_sc.fv_d_turn_180_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[[[1,2,3,4],[11,12,13,14]]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return: &nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[[[14,13,12,11],[4,3,2,1]]]]
>>> sm_sc.<br>fv_d_mx_ele_2d_2_3d_dloss_dindepdt  |Derivation of the loss function of transforming a 2D array into a 3D array over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Split dimension; <br>&nbsp;&nbsp;3.&nbsp;New dimension; <br>&nbsp;&nbsp;4.&nbsp;Should the element order be retained in the split dimension, otherwise in the new dimension; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_ele_2d_2_3d_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_mx_ele_2d_2_3d<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2],&nbsp;&nbsp;[3,&nbsp;&nbsp;4],&nbsp;&nbsp;[5,&nbsp;&nbsp;6],&nbsp;&nbsp;[7,&nbsp;&nbsp;8],&nbsp;&nbsp;[9,&nbsp;&nbsp;10],&nbsp;&nbsp;[11,&nbsp;&nbsp;12],&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;&nbsp;22],&nbsp;&nbsp;[23,&nbsp;&nbsp;24],&nbsp;&nbsp;[25,&nbsp;&nbsp;26],&nbsp;&nbsp;[27,&nbsp;&nbsp;28],&nbsp;&nbsp;[29,&nbsp;&nbsp;30],&nbsp;&nbsp;[31,&nbsp;&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;true<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;true<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[1,2,7,8,21,22,27,28]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[3,4,9,10,23,24,29,30]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[5,6,11,12,25,26,31,32]]
>>> sm_sc.<br>fv_d_mx_ele_3d_2_2d_dloss_dindepdt  |Derivation of the loss function of transforming a 3D array into a 2D array over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Two dimensions where independent variables are merged; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The merged new dimension is in the order and position of to; <br>&nbsp;&nbsp;3.&nbsp;The dimension in which the independent variable is fixed; <br>&nbsp;&nbsp;4.&nbsp;Independent variable dimension specification; |select&nbsp;<br>&nbsp;&nbsp;sm_sc.fv_d_mx_ele_3d_2_2d_dloss_dindepdt<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;21,&nbsp;2,&nbsp;22,&nbsp;3,&nbsp;23,&nbsp;4,&nbsp;24]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[5,&nbsp;25,&nbsp;6,&nbsp;26,&nbsp;7,&nbsp;27,&nbsp;8,&nbsp;28]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[9,&nbsp;29,&nbsp;10,&nbsp;30,&nbsp;11,&nbsp;31,&nbsp;12,&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[1,&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;1<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;array[2,3,4]<br>&nbsp;&nbsp;)<br><br>return: <br>&nbsp;&nbsp;array<br>&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;2,&nbsp;3,&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[5,&nbsp;6,&nbsp;7,&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[9,&nbsp;10,&nbsp;11,&nbsp;12]<br>&nbsp;&nbsp;&nbsp;&nbsp;],&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;22,&nbsp;23,&nbsp;24]&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[25,&nbsp;26,&nbsp;27,&nbsp;28]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;[29,&nbsp;30,&nbsp;31,&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_mx_ele_3d_2_4d_dloss_dindepdt  |Derivation of the loss function of transforming a 3D array into a 4D array over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Split dimension; <br>&nbsp;&nbsp;3.&nbsp;New dimension; <br>&nbsp;&nbsp;4.&nbsp;Should the element order be retained in the split dimension, otherwise in the new dimension; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_ele_3d_2_4d_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,3,7,5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;4<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;false<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;"[1:2][1:15][1:7]"
>>> sm_sc.<br>fv_d_mx_ele_4d_2_3d_dloss_dindepdt  |Derivation of the loss function of transforming a 4D array into a 3D array over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Two dimensions where independent variables are merged; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The merged new dimension is in the order and position of to; <br>&nbsp;&nbsp;3.&nbsp;The dimension in which the independent variable is fixed; <br>&nbsp;&nbsp;4.&nbsp;Independent variable dimension specification; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_ele_4d_2_3d_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,21,5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[1,3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,2,7,5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;"[1:3][1:2][1:7][1:5]"<br>
>>> sm_sc.<br>fv_d_mx_slice_3d_2_2d_dloss_dindepdt  |Derivation of the loss function for 3D to 2D shelling over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Independent variable specification; <br>&nbsp;&nbsp;3.&nbsp;Sliced dimension; <br>&nbsp;&nbsp;4.&nbsp;Slice position number; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_slice_3d_2_2d_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25],&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,2,3,4,5],[11,12,13,14,15],[21,22,23,24,25],[31,32,33,34,35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_mx_ele_flatten_2dims_dloss_dindepdt<br>fv_d_mx_ele_flatten_2dims_dloss_dindepdt_py  |Derivation of the loss function of flattening a N_D array into a array with the same ndims over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Two dimensions where independent variables are merged; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;The merged new dimension is in the order and position of to; <br>&nbsp;&nbsp;3.&nbsp;The dimension in which the independent variable is fixed; <br>&nbsp;&nbsp;4.&nbsp;Independent variable dimension specification; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_ele_flatten_2dims_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,15,1])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,3,5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>返回:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:2][1:3][1:5]
>>> sm_sc.<br>fv_d_mx_slice_4d_2_2d_dloss_dindepdt  |Derivation of the loss function for 4D to 2D Shelling over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Independent variable specification; <br>&nbsp;&nbsp;3.&nbsp;Two dimensions sliced; <br>&nbsp;&nbsp;4.&nbsp;Slice position number on two dimensions; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_slice_4d_2_2d_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25],&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[1,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,2,3,4,5],[11,12,13,14,15],[21,22,23,24,25],[31,32,33,34,35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_mx_slice_4d_2_3d_dloss_dindepdt  |Derivation of the loss function for 4D to 3D Shelling over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Independent variable specification; <br>&nbsp;&nbsp;3.&nbsp;Sliced dimension; <br>&nbsp;&nbsp;4.&nbsp;Slice position number; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_slice_4d_2_3d_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25],&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0],[0,0,0,0,0]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[[1,2,3,4,5],[11,12,13,14,15],[21,22,23,24,25],[31,32,33,34,35]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_mx_ascend_dim_dloss_dindepdt     |Derivation of the loss function of adding shell to increase dimension over the independent variable.<br>Input param:<br>&nbsp;&nbsp;1.&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;2.&nbsp;Number of dimensionality upgrades; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_ascend_dim_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[[[1,&nbsp;&nbsp;2]]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2]]
>>> sm_sc.<br>fv_d_apad_dloss_dindepdt_1      |Derivation of the loss function filled in with bottom alignment over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The height of the independent variable in the visual plane; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_apad_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_apad_dloss_dindepdt_2      |Derivation of the loss function filled in with bottom alignment over the second independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The height of the independent variable in the visual plane; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Fill Repetitions; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_apad_dloss_dindepdt_2<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[22,&nbsp;&nbsp;24,&nbsp;&nbsp;26,&nbsp;&nbsp;28,&nbsp;&nbsp;30]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[42,&nbsp;&nbsp;44,&nbsp;&nbsp;46,&nbsp;&nbsp;48,&nbsp;&nbsp;50]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_bpad_dloss_dindepdt_1      |Derivation of the loss function for top aligned and bottom filled over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The height of the independent variable in the visual plane; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_bpad_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_bpad_dloss_dindepdt_2      |Derivation of the loss function for top aligned and bottom filled over the second independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The height of the independent variable in the visual plane; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Fill repetitions; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_bpad_dloss_dindepdt_2<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[42,&nbsp;&nbsp;44,&nbsp;&nbsp;46,&nbsp;&nbsp;48,&nbsp;&nbsp;50]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[62,&nbsp;&nbsp;64,&nbsp;&nbsp;66,&nbsp;&nbsp;68,&nbsp;&nbsp;70]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_lpad_dloss_dindepdt_1      |Derivation of the loss function for right aligned left filled over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The width of the independent variable in the visual plane; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_lpad_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_lpad_dloss_dindepdt_2      |Derivation of the loss function for right aligned left filled over the second independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The width of the independent variable in the visual plane; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Fill repetitions; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_lpad_dloss_dindepdt_2<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[4,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[24,&nbsp;&nbsp;26]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[44,&nbsp;&nbsp;46]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[64,&nbsp;&nbsp;66]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[84,&nbsp;&nbsp;86]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_rpad_dloss_dindepdt_1      |Derivation of the loss function for left aligned right filled over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The width of the independent variable in the visual plane; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_rpad_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_rpad_dloss_dindepdt_2      |Derivation of the loss function for left aligned right filled over the second independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The width of the independent variable in the visual plane; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Fill repetitions; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_rpad_dloss_dindepdt_2<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44,&nbsp;&nbsp;45]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;],<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2,<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[6,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[26,&nbsp;&nbsp;28]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[46,&nbsp;&nbsp;48]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[66,&nbsp;&nbsp;68]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[86,&nbsp;&nbsp;88]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_chunk_dloss_dindepdt     |Derivation of the loss function of slice(chunk) over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Independent variable specification; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Slice (chunk) position with subscripts in all dimensions; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_chunk_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[2,1],[3,2]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[4,3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,2]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0,0,0],[0,2,1],[0,3,2],[0,0,0]]
>>> sm_sc.<br>fv_d_slice_y_dloss_dindepdt     |Derivation of the loss function of the first dimensional slice over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The first dimension length of the independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Slice position number range; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_slice_y_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1,&nbsp;&nbsp;2],&nbsp;&nbsp;[3,&nbsp;&nbsp;4],&nbsp;&nbsp;[5,&nbsp;&nbsp;11],&nbsp;&nbsp;[12,&nbsp;&nbsp;13],&nbsp;&nbsp;[14,&nbsp;&nbsp;15]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;7<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(2,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]')]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[1,2],[15,17],[19,26],[0,0],[0,0],[0,0],[0,0]]
>>> sm_sc.<br>fv_d_slice_x_dloss_dindepdt     |Derivation of the loss function of the second dimension slice over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The second dimension length of the independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Slice position number range; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_slice_x_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5],&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;7<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(2,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]')]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[1,6,8,0,0,0,0],[11,26,28,0,0,0,0]]
>>> sm_sc.<br>fv_d_slice_x3_dloss_dindepdt    |Derivation of the loss function of the third dimension slice over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The third dimension length of the independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Slice position number range; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_slice_x3_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;7<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(2,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]')]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;6,&nbsp;&nbsp;8,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;26,&nbsp;&nbsp;28,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;&nbsp;46,&nbsp;&nbsp;48,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;66,&nbsp;&nbsp;68,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_slice_x4_dloss_dindepdt    |Derivation of the loss function of the fourth dimension slice over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The fourth dimension length of the independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Slice position number range; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_slice_x4_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5],[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24,&nbsp;&nbsp;25],[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34,&nbsp;&nbsp;35]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;7<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(2,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]')]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;6,&nbsp;&nbsp;8,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;26,&nbsp;&nbsp;28,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[21,&nbsp;&nbsp;46,&nbsp;&nbsp;48,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;66,&nbsp;&nbsp;68,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_sample_y_dloss_dindepdt_1     |Derivation of the loss function of the first dimensional sampling over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The first dimension length of the independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Constrain window width for each sampling period; <br>&nbsp;&nbsp;&nbsp;&nbsp;4.&nbsp;&nbsp;Sampling sequence number (multiple) intervals for each sampling period; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_sample_y_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[21,&nbsp;&nbsp;22,&nbsp;&nbsp;23,&nbsp;&nbsp;24]<br><br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[31,&nbsp;&nbsp;32,&nbsp;&nbsp;33,&nbsp;&nbsp;34]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[41,&nbsp;&nbsp;42,&nbsp;&nbsp;43,&nbsp;&nbsp;44]<br><br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[51,&nbsp;&nbsp;52,&nbsp;&nbsp;53,&nbsp;&nbsp;54]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[61,&nbsp;&nbsp;62,&nbsp;&nbsp;63,&nbsp;&nbsp;64]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[71,&nbsp;&nbsp;72,&nbsp;&nbsp;73,&nbsp;&nbsp;74]<br><br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[81,&nbsp;&nbsp;82,&nbsp;&nbsp;83,&nbsp;&nbsp;84]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[91,&nbsp;&nbsp;92,&nbsp;&nbsp;93,&nbsp;&nbsp;94]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;15<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(2,&nbsp;&nbsp;5,&nbsp;&nbsp;'[]')]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,2,3,4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[42,44,46,48]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[62,64,66,68]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,2,3,4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[42,44,46,48]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[62,64,66,68]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1,2,3,4]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[42,44,46,48]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[62,64,66,68]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,0,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_sample_x_dloss_dindepdt_1     |Derivation of the loss function of the second dimensional sampling over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The second dimension length of the independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Constrain window width for each sampling period; <br>&nbsp;&nbsp;&nbsp;&nbsp;4.&nbsp;&nbsp;Sampling sequence number (multiple) intervals for each sampling period; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_sample_x_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8,&nbsp;&nbsp;9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16&nbsp;&nbsp;,17&nbsp;&nbsp;,18&nbsp;&nbsp;,19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;15<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(2,&nbsp;&nbsp;5,&nbsp;&nbsp;'[]')]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,6,8,0,0,1,6,8,0,0,1,6,8,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,26,28,0,0,11,26,28,0,0,11,26,28,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_sample_x3_dloss_dindepdt_1    |Derivation of the loss function of the third dimensional sampling over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The third dimension length of the independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Constrain window width for each sampling period; <br>&nbsp;&nbsp;&nbsp;&nbsp;4.&nbsp;&nbsp;Sampling sequence number (multiple) intervals for each sampling period; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_sample_x3_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8,&nbsp;&nbsp;9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16&nbsp;&nbsp;,17&nbsp;&nbsp;,18&nbsp;&nbsp;,19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;15<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(2,&nbsp;&nbsp;5,&nbsp;&nbsp;'[]')]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,6,8,0,0,1,6,8,0,0,1,6,8,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,26,28,0,0,11,26,28,0,0,11,26,28,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]]
>>> sm_sc.<br>fv_d_sample_x4_dloss_dindepdt_1    |Derivation of the loss function of the fourth dimensional sampling over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The fourth dimension length of the independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Constrain window width for each sampling period; <br>&nbsp;&nbsp;&nbsp;&nbsp;4.&nbsp;&nbsp;Sampling sequence number (multiple) intervals for each sampling period; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_sample_x4_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8,&nbsp;&nbsp;9]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,&nbsp;&nbsp;12,&nbsp;&nbsp;13,&nbsp;&nbsp;14,&nbsp;&nbsp;15,&nbsp;&nbsp;16&nbsp;&nbsp;,17&nbsp;&nbsp;,18&nbsp;&nbsp;,19]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;15<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[int4range(1,&nbsp;&nbsp;3,&nbsp;&nbsp;'[]'),&nbsp;&nbsp;int4range(2,&nbsp;&nbsp;5,&nbsp;&nbsp;'[]')]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,6,8,0,0,1,6,8,0,0,1,6,8,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[11,26,28,0,0,11,26,28,0,0,11,26,28,0,0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]]]
>>> sm_sc.<br>fv_d_lower_tri_mx_dloss_dindepdt    |Derivation of the loss function of the lower triangular matrix over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_lower_tri_mx_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[&nbsp;&nbsp;[1,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_upper_tri_mx_dloss_dindepdt    |Derivation of the loss function of the upper triangular matrix over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_upper_tri_mx_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,2,3],[0,-2,-3],[0,0,-3],[0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,2,3],[0,-2,-3],[0,0,-3],[0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,2,3],[0,-2,-3],[0,0,-3],[0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,2,3],[0,-2,-3],[0,0,-3],[0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_lower_tri_mx_ex_dloss_dindepdt    |Derivation of the loss function of the degenerate lower triangular matrix over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_lower_tri_mx_ex_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;0],[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;0],[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[&nbsp;&nbsp;[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;0],[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;0,&nbsp;&nbsp;0],[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;0],[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_upper_tri_mx_ex_dloss_dindepdt    |Derivation of the loss function of the degenerate upper triangular matrix over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_upper_tri_mx_ex_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;-2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[-1,&nbsp;&nbsp;2,&nbsp;&nbsp;-3],&nbsp;&nbsp;[1,&nbsp;&nbsp;-2,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;);<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[0,2,3],[0,0,-3],[0,0,0],[0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[0,2,3],[0,0,-3],[0,0,0],[0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[0,2,3],[0,0,-3],[0,0,0],[0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[0,2,3],[0,0,-3],[0,0,0],[0,0,0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_lmask_dloss_dindepdt_1    |Derivation of the loss function of the left mask on high wide 2D faces over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Left mask length for each line; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_lmask_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[3],&nbsp;&nbsp;[0],&nbsp;&nbsp;[2],&nbsp;&nbsp;[1]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,&nbsp;&nbsp;0,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_rmask_dloss_dindepdt_1    |Derivation of the loss function of the right mask on high wide 2D faces over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Right mask length for each line; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_rmask_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[3],&nbsp;&nbsp;[0],&nbsp;&nbsp;[2],&nbsp;&nbsp;[1]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_amask_dloss_dindepdt_1    |Derivation of the loss function of the upper mask on high wide 2D faces over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Upper mask length for each column; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_amask_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[3,&nbsp;&nbsp;0,&nbsp;&nbsp;2,&nbsp;&nbsp;1,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0,&nbsp;&nbsp;2,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,&nbsp;&nbsp;3,&nbsp;&nbsp;0,&nbsp;&nbsp;5,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_bmask_dloss_dindepdt_1    |Derivation of the loss function of the lower mask on high wide 2D faces over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Lower mask length for each column; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_b_amask_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[4,&nbsp;&nbsp;5,&nbsp;&nbsp;6,&nbsp;&nbsp;7,&nbsp;&nbsp;8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[3,&nbsp;&nbsp;0,&nbsp;&nbsp;2,&nbsp;&nbsp;1,&nbsp;&nbsp;3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1,&nbsp;&nbsp;2,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,&nbsp;&nbsp;3,&nbsp;&nbsp;4,&nbsp;&nbsp;5,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,&nbsp;&nbsp;4,&nbsp;&nbsp;0,&nbsp;&nbsp;6,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[0,&nbsp;&nbsp;5,&nbsp;&nbsp;0,&nbsp;&nbsp;0,&nbsp;&nbsp;0]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_aggr_slice_sum_dloss_dindepdt |Derivation of the loss function of slice total aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Slice grouping specifications; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_slice_sum_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;-5.1,&nbsp;&nbsp;-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;-5.1,&nbsp;&nbsp;8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,2.3,2.3,5.1,5.1,5.1,8.2,8.2,8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,-2.3,-2.3,5.1,5.1,5.1,-8.2,-8.2,-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,2.3,2.3,5.1,5.1,5.1,8.2,8.2,8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,-2.3,-2.3,5.1,5.1,5.1,-8.2,-8.2,-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,2.3,2.3,-5.1,-5.1,-5.1,-8.2,-8.2,-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,-2.3,-2.3,-5.1,-5.1,-5.1,8.2,8.2,8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,2.3,2.3,-5.1,-5.1,-5.1,-8.2,-8.2,-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,-2.3,-2.3,-5.1,-5.1,-5.1,8.2,8.2,8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_aggr_slice_avg_dloss_dindepdt |Derivation of the loss function of slice average aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Slice grouping specifications; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_slice_avg_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;5.1,&nbsp;&nbsp;-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.3,&nbsp;&nbsp;-5.1,&nbsp;&nbsp;-8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-2.3,&nbsp;&nbsp;-5.1,&nbsp;&nbsp;8.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.383,0.383,0.383,0.850,0.850,0.850,1.367,1.367,1.367]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.383,-0.383,-0.383,0.850,0.850,0.850,-1.367,-1.367,-1.367]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.383,0.383,0.383,0.850,0.850,0.850,1.367,1.367,1.367]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.383,-0.383,-0.383,0.850,0.850,0.850,-1.367,-1.367,-1.367]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.383,0.383,0.383,-0.850,-0.850,-0.850,-1.367,-1.367,-1.367]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.383,-0.383,-0.383,-0.850,-0.850,-0.850,1.367,1.367,1.367]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[0.383,0.383,0.383,-0.850,-0.850,-0.850,-1.367,-1.367,-1.367]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[-0.383,-0.383,-0.383,-0.850,-0.850,-0.850,1.367,1.367,1.367]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_aggr_slice_prod_dloss_dindepdt |Derivation of the loss function of slice direct product aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;4.&nbsp;&nbsp;Slice grouping specifications; |with&nbsp;&nbsp;<br>cte_rand&nbsp;&nbsp;as&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[4,&nbsp;&nbsp;3])&nbsp;&nbsp;as&nbsp;&nbsp;a_indepdt<br>)<br>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_slice_prod_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a_indepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_aggr_slice_prod(a_indepdt,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,&nbsp;&nbsp;3])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>from&nbsp;&nbsp;cte_rand<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:4][1:3]
>>> sm_sc.<br>fv_d_aggr_slice_max_dloss_dindepdt |Derivation of the loss function of slice maximum aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;4.&nbsp;&nbsp;Slice grouping specifications; |with&nbsp;&nbsp;<br>cte_rand&nbsp;&nbsp;as&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[4,&nbsp;&nbsp;3])&nbsp;&nbsp;as&nbsp;&nbsp;a_indepdt<br>)<br>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_slice_max_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a_indepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_aggr_slice_max(a_indepdt,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,&nbsp;&nbsp;3])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>from&nbsp;&nbsp;cte_rand<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:4][1:3]
>>> sm_sc.<br>fv_d_aggr_slice_min_dloss_dindepdt |Derivation of the loss function of slice minimum aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;4.&nbsp;&nbsp;Slice grouping specifications; |with&nbsp;&nbsp;<br>cte_rand&nbsp;&nbsp;as&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[4,&nbsp;&nbsp;3])&nbsp;&nbsp;as&nbsp;&nbsp;a_indepdt<br>)<br>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_slice_min_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a_indepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_aggr_slice_min(a_indepdt,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,&nbsp;&nbsp;3])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>from&nbsp;&nbsp;cte_rand<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:4][1:3]
>>> sm_sc.<br>fv_d_aggr_chunk_sum_dloss_dindepdt |Derivation of the loss function of chunk total aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Independent variable specification; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_chunk_sum_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[6,&nbsp;&nbsp;8,&nbsp;&nbsp;8,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2,&nbsp;&nbsp;3])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:6][1:8][1:8][1:6]
>>> sm_sc.<br>fv_d_aggr_chunk_prod_dloss_dindepdt |Derivation of the loss function of chunk direct product aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |with&nbsp;&nbsp;<br>cte_rand&nbsp;&nbsp;as&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[6,&nbsp;&nbsp;8,&nbsp;&nbsp;8])&nbsp;&nbsp;as&nbsp;&nbsp;a_indepdt<br>)<br>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_chunk_prod_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a_indepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_aggr_chunk_prod(a_indepdt,&nbsp;&nbsp;array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>from&nbsp;&nbsp;cte_rand<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:6][1:8][1:8]
>>> sm_sc.<br>fv_d_aggr_chunk_avg_dloss_dindepdt |Derivation of the loss function of chunk average aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Independent variable specification; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_chunk_avg_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[6,&nbsp;&nbsp;8,&nbsp;&nbsp;8,&nbsp;&nbsp;6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2,&nbsp;&nbsp;3])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:6][1:8][1:8][1:6]
>>> sm_sc.<br>fv_d_aggr_chunk_max_dloss_dindepdt |Derivation of the loss function of chunk maximum aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |with&nbsp;&nbsp;<br>cte_rand&nbsp;&nbsp;as&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[6,&nbsp;&nbsp;8,&nbsp;&nbsp;8])&nbsp;&nbsp;as&nbsp;&nbsp;a_indepdt<br>)<br>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_chunk_max_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a_indepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_aggr_chunk_max(a_indepdt,&nbsp;&nbsp;array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>from&nbsp;&nbsp;cte_rand<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:6][1:8][1:8]
>>> sm_sc.<br>fv_d_aggr_chunk_min_dloss_dindepdt |Derivation of the loss function of chunk minimum aggregation over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; |with&nbsp;&nbsp;<br>cte_rand&nbsp;&nbsp;as&nbsp;&nbsp;<br>(<br>&nbsp;&nbsp;&nbsp;&nbsp;select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[6,&nbsp;&nbsp;8,&nbsp;&nbsp;8])&nbsp;&nbsp;as&nbsp;&nbsp;a_indepdt<br>)<br>select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_aggr_chunk_min_dloss_dindepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;a_indepdt<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_aggr_chunk_min(a_indepdt,&nbsp;&nbsp;array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[3,&nbsp;&nbsp;2,&nbsp;&nbsp;2])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>from&nbsp;&nbsp;cte_rand<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:6][1:8][1:8]
>>> sm_sc.<br>fv_d_prod_mx_1 |Derivation of the dependent variable of matrix multiplication over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Another independent variable, the second independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Independent variable specification; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_prod_mx_1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[4,&nbsp;&nbsp;3,&nbsp;&nbsp;2,&nbsp;&nbsp;5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;6,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:3][1:2][1:5]
>>> sm_sc.<br>fv_d_prod_mx_2 |Derivation of the dependent variable of matrix multiplication over the second independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Another independent variable, the first independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Independent variable specification; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_prod_mx_2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[4,&nbsp;&nbsp;3,&nbsp;&nbsp;2,&nbsp;&nbsp;5])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[3,&nbsp;&nbsp;3,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:3][1:5][1:2]
>>> sm_sc.<br>fv_d_prod_mx_dloss_dindepdt_1 |Derivation of the loss function of matrix multiplication over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Another independent variable, the second independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Independent variable specification; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_prod_mx_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[5,&nbsp;&nbsp;6,&nbsp;&nbsp;2,&nbsp;&nbsp;4])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[6,&nbsp;&nbsp;3,&nbsp;&nbsp;4])&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[5,&nbsp;&nbsp;6,&nbsp;&nbsp;2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:5][1:6][1:2][1:3]
>>> sm_sc.<br>fv_d_prod_mx_dloss_dindepdt_2 |Derivation of the loss function of matrix multiplication over the second independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Another independent variable, the first independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Independent variable specification; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_prod_mx_dloss_dindepdt_2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[5,&nbsp;&nbsp;2,&nbsp;&nbsp;4])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[2,&nbsp;&nbsp;3])&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[5,&nbsp;&nbsp;3,&nbsp;&nbsp;4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:5][1:3][1:4]
>>> sm_sc.<br>fv_d_chunk_prod_mx_dloss_dindepdt_1 |Derivation of the loss function of chunk matrix multiplication over the first independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Another independent variable, the second independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Block size; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;the height of input parameter one, the width of input parameter one, and the width of input parameter two in matrix multiplication|select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_chunk_prod_mx_dloss_dindepdt_1<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[3,&nbsp;&nbsp;2&nbsp;&nbsp;*&nbsp;&nbsp;7,&nbsp;&nbsp;5&nbsp;&nbsp;*&nbsp;&nbsp;11])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[3,&nbsp;&nbsp;3&nbsp;&nbsp;*&nbsp;&nbsp;7,&nbsp;&nbsp;5&nbsp;&nbsp;*&nbsp;&nbsp;11])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:3][1:14][1:33]
>>> sm_sc.<br>fv_d_chunk_prod_mx_dloss_dindepdt_2 |Derivation of the loss function of chunk matrix multiplication over the second independent variable.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;Another independent variable, the first independent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;Block size; <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;the height of input parameter one, the width of input parameter one, and the width of input parameter two in matrix multiplication|select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array_dims<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_chunk_prod_mx_dloss_dindepdt_2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_new_rand(array[2&nbsp;&nbsp;*&nbsp;&nbsp;7,&nbsp;&nbsp;5&nbsp;&nbsp;*&nbsp;&nbsp;11])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;sm_sc.fv_new_rand(array[2&nbsp;&nbsp;*&nbsp;&nbsp;7,&nbsp;&nbsp;3&nbsp;&nbsp;*&nbsp;&nbsp;11])<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[2,&nbsp;&nbsp;3,&nbsp;&nbsp;5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;[1:21][1:55]
>>> sm_sc.<br>fv_d_mx_concat_y_dloss_dindepdt_n |Derivation of the loss function of the first dimensional concatenation aggregation over the independent variable of a certain position.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The lower limit of the position where the independent variable is concatenated into the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;The upper limit of the position where the independent variable is concatenated into the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_concat_y_dloss_dindepdt_n<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.1,&nbsp;&nbsp;2.1,&nbsp;&nbsp;3.1,&nbsp;&nbsp;4.1,&nbsp;&nbsp;5.1,&nbsp;&nbsp;6.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,&nbsp;&nbsp;2.2,&nbsp;&nbsp;3.2,&nbsp;&nbsp;4.2,&nbsp;&nbsp;5.2,&nbsp;&nbsp;6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,&nbsp;&nbsp;2.3,&nbsp;&nbsp;3.3,&nbsp;&nbsp;4.3,&nbsp;&nbsp;5.3,&nbsp;&nbsp;6.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,&nbsp;&nbsp;2.4,&nbsp;&nbsp;3.4,&nbsp;&nbsp;4.4,&nbsp;&nbsp;5.4,&nbsp;&nbsp;6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.1,&nbsp;&nbsp;2.1,&nbsp;&nbsp;3.1,&nbsp;&nbsp;4.1,&nbsp;&nbsp;5.1,&nbsp;&nbsp;6.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;,[1.2,&nbsp;&nbsp;2.2,&nbsp;&nbsp;3.2,&nbsp;&nbsp;4.2,&nbsp;&nbsp;5.2,&nbsp;&nbsp;6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_mx_concat_x_dloss_dindepdt_n |Derivation of the loss function of the second dimensional concatenation aggregation over the independent variable of a certain position.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The lower limit of the position where the independent variable is concatenated into the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;The upper limit of the position where the independent variable is concatenated into the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_concat_x_dloss_dindepdt_n<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.1,&nbsp;&nbsp;2.1,&nbsp;&nbsp;3.1,&nbsp;&nbsp;4.1,&nbsp;&nbsp;5.1,&nbsp;&nbsp;6.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,&nbsp;&nbsp;2.2,&nbsp;&nbsp;3.2,&nbsp;&nbsp;4.2,&nbsp;&nbsp;5.2,&nbsp;&nbsp;6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,&nbsp;&nbsp;2.3,&nbsp;&nbsp;3.3,&nbsp;&nbsp;4.3,&nbsp;&nbsp;5.3,&nbsp;&nbsp;6.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,&nbsp;&nbsp;2.4,&nbsp;&nbsp;3.4,&nbsp;&nbsp;4.4,&nbsp;&nbsp;5.4,&nbsp;&nbsp;6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.1,&nbsp;&nbsp;3.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.2,&nbsp;&nbsp;3.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.3,&nbsp;&nbsp;3.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.4,&nbsp;&nbsp;3.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_mx_concat_x3_dloss_dindepdt_n |Derivation of the loss function of the third dimensional concatenation aggregation over the independent variable of a certain position.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The lower limit of the position where the independent variable is concatenated into the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;The upper limit of the position where the independent variable is concatenated into the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_concat_x3_dloss_dindepdt_n<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[1.0],&nbsp;&nbsp;[2.0],&nbsp;&nbsp;[3.0],&nbsp;&nbsp;[4.0],&nbsp;&nbsp;[5.0],&nbsp;&nbsp;[6.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1.1],&nbsp;&nbsp;[2.1],&nbsp;&nbsp;[3.1],&nbsp;&nbsp;[4.1],&nbsp;&nbsp;[5.1],&nbsp;&nbsp;[6.1]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1.2],&nbsp;&nbsp;[2.2],&nbsp;&nbsp;[3.2],&nbsp;&nbsp;[4.2],&nbsp;&nbsp;[5.2],&nbsp;&nbsp;[6.2]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1.3],&nbsp;&nbsp;[2.3],&nbsp;&nbsp;[3.3],&nbsp;&nbsp;[4.3],&nbsp;&nbsp;[5.3],&nbsp;&nbsp;[6.3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[1.4],&nbsp;&nbsp;[2.4],&nbsp;&nbsp;[3.4],&nbsp;&nbsp;[4.4],&nbsp;&nbsp;[5.4],&nbsp;&nbsp;[6.4]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;4<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[2],&nbsp;&nbsp;[3],&nbsp;&nbsp;[4]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[2.1],&nbsp;&nbsp;[3.1],&nbsp;&nbsp;[4.1]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[2.2],&nbsp;&nbsp;[3.2],&nbsp;&nbsp;[4.2]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[2.3],&nbsp;&nbsp;[3.3],&nbsp;&nbsp;[4.3]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[[2.4],&nbsp;&nbsp;[3.4],&nbsp;&nbsp;[4.4]]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
>>> sm_sc.<br>fv_d_mx_concat_x4_dloss_dindepdt_n |Derivation of the loss function of the fourth dimensional concatenation aggregation over the independent variable of a certain position.<br>Input param:<br>&nbsp;&nbsp;&nbsp;&nbsp;1.&nbsp;&nbsp;Matrix of the derivative of the loss function with respect over the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;2.&nbsp;&nbsp;The lower limit of the position where the independent variable is concatenated into the dependent variable; <br>&nbsp;&nbsp;&nbsp;&nbsp;3.&nbsp;&nbsp;The upper limit of the position where the independent variable is concatenated into the dependent variable; |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_d_mx_concat_x4_dloss_dindepdt_n<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[1.0,&nbsp;&nbsp;2.0,&nbsp;&nbsp;3.0,&nbsp;&nbsp;4.0,&nbsp;&nbsp;5.0,&nbsp;&nbsp;6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.1,&nbsp;&nbsp;2.1,&nbsp;&nbsp;3.1,&nbsp;&nbsp;4.1,&nbsp;&nbsp;5.1,&nbsp;&nbsp;6.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.2,&nbsp;&nbsp;2.2,&nbsp;&nbsp;3.2,&nbsp;&nbsp;4.2,&nbsp;&nbsp;5.2,&nbsp;&nbsp;6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.3,&nbsp;&nbsp;2.3,&nbsp;&nbsp;3.3,&nbsp;&nbsp;4.3,&nbsp;&nbsp;5.3,&nbsp;&nbsp;6.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[1.4,&nbsp;&nbsp;2.4,&nbsp;&nbsp;3.4,&nbsp;&nbsp;4.4,&nbsp;&nbsp;5.4,&nbsp;&nbsp;6.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;]]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;2<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;3<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br><br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[[[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2,&nbsp;&nbsp;3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.1,&nbsp;&nbsp;3.1]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.2,&nbsp;&nbsp;3.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.3,&nbsp;&nbsp;3.3]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[2.4,&nbsp;&nbsp;3.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;]]]
>>> sm_sc.fv_dloss_dz_least_square       |Derivation of the loss function using the least squares method.<br>&nbsp;&nbsp;1. Train the output prediction value matrix;<br>&nbsp;&nbsp;2. Encoding the true values of the training set;    |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_dloss_dz_least_square<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[1.0,&nbsp;&nbsp;2,&nbsp;&nbsp;3],&nbsp;&nbsp;[2,&nbsp;&nbsp;3,&nbsp;&nbsp;4]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[1.1,&nbsp;&nbsp;2.1,&nbsp;&nbsp;3.2],&nbsp;&nbsp;[2.2,&nbsp;&nbsp;3.3,&nbsp;&nbsp;4.2]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[-0.071,-0.071,-0.141], [-0.141,-0.212,-0.141]]
>>> sm_sc.fv_dloss_dz_cross_entropy      |Derivation of cross entropy loss function.<br>&nbsp;&nbsp;1. Train the output prediction value matrix;<br>&nbsp;&nbsp;2. Encoding the true values of the training set;    |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;sm_sc.fv_dloss_dz_cross_entropy<br>&nbsp;&nbsp;&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;array[[0.3,&nbsp;&nbsp;0.5,&nbsp;&nbsp;0.2],&nbsp;&nbsp;[0.1,&nbsp;&nbsp;0.8,&nbsp;&nbsp;0.1]]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;array[[0.0,&nbsp;&nbsp;1.0,&nbsp;&nbsp;0.0],&nbsp;&nbsp;[0.0,&nbsp;&nbsp;1.0,&nbsp;&nbsp;0.0]]<br>&nbsp;&nbsp;&nbsp;&nbsp;)<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[0.000,-1.000,0.000],[0.000,-0.625,0.000]]

>>
>>####   s. Linear algebra:
>>
>>> Function name | Function description | Call example
>>> :-|:-|:-
>>> sm_sc.fv_mx_col_mix           |Elementary column transformation, resetting a position by referencing another position in the same row.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>&nbsp;&nbsp;2. The column number of the position to be zeroed; <br>&nbsp;&nbsp;3. The row number of the position to be zeroed: <br>&nbsp;&nbsp;4. Column number for multiples:   |select <br>&nbsp;&nbsp;sm_sc.fv_mx_col_mix<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.9, 34.5, 0.55, 45.7, 400.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[45.9, 4.6, 34.5, 0.55, 45.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[3.2, 7.7, 1.9, 34.5, 0.55]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4.7, 9.0, 4.6, 34.5, 0.55]]<br>&nbsp;&nbsp;&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;&nbsp;&nbsp;, 4<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.9000, 24.3003, 0.5500, 45.7000, 400.5000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[45.9000, 4.4772, 34.5000, 0.5500, 45.7000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[3.2000, 0.0000, 1.9000, 34.5000, 0.5500]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4.7000, 1.3000, 4.6000, 34.5000, 0.5500]]
>>> sm_sc.fv_mx_row_mix           |Elementary row transformation, resetting a position by referencing another position in the same column.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>&nbsp;&nbsp;2. The row number of the position to be zeroed; <br>&nbsp;&nbsp;3. The column number of the position to be zeroed: <br>&nbsp;&nbsp;4. Row number for multiples:   |select <br>&nbsp;&nbsp;sm_sc.fv_mx_row_mix<br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1.9, 34.5, 0.55, 45.7, 400.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[45.9, 4.6, 34.5, 0.55, 45.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[3.2, 7.7, 1.9, 34.5, 0.55]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4.7, 9.0, 4.6, 34.5, 0.55]]<br>&nbsp;&nbsp;&nbsp;&nbsp;, 2<br>&nbsp;&nbsp;&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;&nbsp;&nbsp;, 4<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.9000, 34.5000, 0.5500 ,45.7000, 400.5000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[10.6500, -62.9000, 0.0000, -258.2000, 41.5750]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[3.2000, 7.7000, 1.9000, 34.5000, 0.5500]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4.7000, 9.0000, 4.6000, 34.5000, 0.5500]]
>>> sm_sc.fv_mx_determinant       |Matrix determinant.<br>This function is implemented based on temporary table sm_sc.__vt_tmp_matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:  |select<br>&nbsp;&nbsp;sm_sc.fv_mx_determinant        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3],[0, 8, 9],[4, 5, 6]]<br>&nbsp;&nbsp;);<br><br>return:-21.0
>>> sm_sc.fv_mx_inversion         |Inverse matrix.<br>This function is implemented based on temporary table sm_sc.__vt_tmp_matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:  |select<br>&nbsp;&nbsp;sm_sc.fv_mx_inversion        <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3],[0, 8, 9],[4, 5, 6]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[-0.1429, -0.1429, 0.2857]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-1.7143, 0.2857, 0.4286]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[1.5238, -0.1429, -0.3810]]
>>> sm_sc.fv_arr_norm_1             |Vector 1 norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_arr_norm_1            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3]<br>&nbsp;&nbsp;);<br><br>return:6.0
>>> sm_sc.fv_arr_norm_2             |Vector 2 norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_arr_norm_2            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3]<br>&nbsp;&nbsp;);<br><br>return:3.7417
>>> sm_sc.fv_arr_norm_nega_inf      |Vector negative infinite norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_arr_norm_nega_inf            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3]<br>&nbsp;&nbsp;);<br><br>return:1.0
>>> sm_sc.fv_arr_norm_posi_inf      |Vector positive infinite norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_arr_norm_posi_inf            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3]<br>&nbsp;&nbsp;);<br><br>return:3.0
>>> sm_sc.fv_arr_norm_p             |Vector p-norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>&nbsp;&nbsp;2. Norm series p:   |select<br>&nbsp;&nbsp;sm_sc.fv_arr_norm_p            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3]<br>&nbsp;&nbsp;, 3<br>&nbsp;&nbsp;);<br><br>return:3.3079
>>> sm_sc.fv_mx_norm_col          |Column and norm 1-norm  column module.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_norm_col           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:17.0
>>> sm_sc.fv_mx_norm_inf          |Row and norm infinite - norm  row module.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_norm_inf           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:12.0
>>> sm_sc.fv_mx_norm_l0           |L0-norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_norm_l0           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:8.0
>>> sm_sc.fv_mx_norm_l1           |L1-norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_norm_l1           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:30.0
>>> sm_sc.fv_mx_norm_l2_f         |L2(F)-norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_norm_l2_f           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:12.0
>>> sm_sc.fv_mx_norm_l21          |L21-norm.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_norm_l21           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:20.9773
>>> sm_sc.fv_mx_is_symmetry         |Determine whether the matrix is symmetric or not.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_is_symmetry           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:false
>>> sm_sc.fv_mx_rank                |Rank of matrix.<br>This function is implemented based on temporary table sm_sc.__vt_tmp_matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_rank           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:3
>>> sm_sc.fv_mx_row_step            |Row echelon matrix, upper trapezoid.<br>This function is implemented based on temporary table sm_sc.__vt_tmp_matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_row_step            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1, 2, 3]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.0000, 2.0000, 3.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 8.0000, 9.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 0.0000, -2.6250]]
>>> sm_sc.fv_mx_rows_step_simple    |Row ladder minimalist matrix, that is, equivalent diagonal matrix, that is, solving the linear equation system of the matrix.<br>This function is implemented based on temporary table sm_sc.__vt_tmp_matrix.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix: |select<br>&nbsp;&nbsp;sm_sc.fv_mx_rows_step_simple           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1.0000, 0.0000, 0.0000, 13.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 1.0000, 0.0000, 3.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 0.0000, 1.0000, -20.3333]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 0.0000, 0.0000, 0.0000]]
>>> sm_sc.fv_mx_times_col           |Specify a column multiplication factor.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>&nbsp;&nbsp;2. Number of columns multiplied by the coefficient: <br>&nbsp;&nbsp;3. Multiplication coefficient value:  |select<br>&nbsp;&nbsp;sm_sc.fv_mx_times_col           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;, 4<br>&nbsp;&nbsp;, 1.5<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1, -2, 0, 10.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[5, 0, 3, 6.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0.0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, -2, 0, -9.0]]
>>> sm_sc.fv_mx_times_row           |Specify a row multiplication factor.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>&nbsp;&nbsp;2. Number of rows multiplied by the coefficient: <br>&nbsp;&nbsp;3. Multiplication coefficient value:  |select<br>&nbsp;&nbsp;sm_sc.fv_mx_times_row           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;, 4<br>&nbsp;&nbsp;, 1.5<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1, -2, 0, 7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0, -3.0, 0.0, -9.0]]
>>> sm_sc.fv_mx_reorder_col         |Reorder Columns.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>&nbsp;&nbsp;2. Rearrange the old column numbers and store them in the new column position:  |select<br>&nbsp;&nbsp;sm_sc.fv_mx_reorder_col           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;, array[null,3,4,2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1, 0, 7, -2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[5, 3, 4, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, -6, -2]]
>>> sm_sc.fv_mx_reorder_row         |Reorder Rows.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>&nbsp;&nbsp;2. Rearrange the old row numbers and store them in the new row position:  |select<br>&nbsp;&nbsp;sm_sc.fv_mx_reorder_row           <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, -2, 0, 7],[5, 0, 3, 4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0],[0, -2, 0, -6]]<br>&nbsp;&nbsp;, array[null,3,4,2]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;array[[1, -2, 0, 7]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, 0, 0, 0]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0, -2, 0, -6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[5, 0, 3, 4]]
>>> sm_sc.ft_mx_lu                  |LU decomposition.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>Output param column: <br>&nbsp;&nbsp;1. Q-matrix; <br>&nbsp;&nbsp;2. R-matrix;  |select<br>&nbsp;&nbsp;o_mx_l, o_mx_u<br> from <br>&nbsp;&nbsp;sm_sc.ft_mx_lu            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4],[0, 8, 9, 10]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4, 5, 6, 7],[5, 7, 2, 10]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;o_mx_l:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.0000, 0.0000, 0.0000, 0.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 1.0000, 0.0000, 0.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4.0000, -0.3750, 1.0000, 0.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[5.0000, -0.3750, 3.6667, 1.0000]]<br>&nbsp;&nbsp;o_mx_u:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.0000, 2.0000, 3.0000, 4.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 8.0000, 9.0000, 10.0000]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 0.0000, -2.6250, -5.2500]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 0.0000, 0.0000, 13.0000]]
>>> sm_sc.ft_mx_qr                  |QR decomposition.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:<br>Output param column: <br>&nbsp;&nbsp;1. Q-matrix; <br>&nbsp;&nbsp;2. R-matrix;  |select<br>&nbsp;&nbsp;o_mx_q, o_mx_r<br> from <br>&nbsp;&nbsp;sm_sc.ft_mx_qr            <br>&nbsp;&nbsp;(<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[1, 2, 3, 4],[0, 8, 9, 10]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4, 5, 6, 7],[5, 7, 2, 10]]<br>&nbsp;&nbsp;);<br><br>return:<br>&nbsp;&nbsp;o_mx_q:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.1543, 0.0800, 0.3341, 0.9264]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 0.9950, 0.0276, -0.0958]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.6172, -0.0533, 0.7020, -0.3514]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.7715, 0.0267, -0.6284, 0.0958]]<br>&nbsp;&nbsp;o_mx_r:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[6.4807, 8.7953, 5.7092, 12.6529]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 8.0401, 8.9285, 10.1634]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 0.0000, 4.2056, 0.2422]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.0000, 0.0000, 0.0000, 1.2458]]

>>
>###  2. Operator list
>
>> Reminder: in postgresql v14 and above, the definition of the operator on the right 1_Opr is not supported.
>>
>> Operator | Operator Position | Operator Description | Example
>> :-|:-|:-|:-
>> \+\      |2_Opr      |Element (broadcast) add.  The element type is not limited to float.                     <br>Equivalent function: sm_sc.fv_opr_add                       |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\+        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-1.3, 2.5],[-2.4, 2.1]]
>> \-\      |2_Opr      |Element (broadcast) sub.  The element type is not limited to float.                     <br>Equivalent function: sm_sc.fv_opr_sub                       |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\-        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[3.9, -20.1],[6.8, -8.1]]
>> \-\      |Left 1_Opr  |Element (broadcast) opposite number.  The element type is not limited to float.                 <br>Equivalent function: sm_sc.fv_opr_sub              |select                                                   <br>&nbsp;&nbsp;\-        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[2.6, -11.3],[4.6, -5.1]]
>> \*\      |2_Opr      |Element (broadcast) mul.  The element type is not limited to float.                     <br>Equivalent function: sm_sc.fv_opr_mul                       |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\*        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-3.38, -99.44],[-10.12, -15.30]]
>> \/\      |2_Opr      |Element (broadcast) div.  The element type is not limited to float.                     <br>Equivalent function: sm_sc.fv_opr_div                       |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\/        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-0.5000, -0.7788],[-0.4783, -0.5882]]
>> \^      |2_Opr      |Element (broadcast) power.  The element type is not limited to float.                     <br>Equivalent function: sm_sc.fv_opr_pow                      |select <br>&nbsp;&nbsp;array[[1.3, 8.8],[2.2, 3.0]]      <br>&nbsp;&nbsp;\^        <br>&nbsp;&nbsp;array[[-2.6, 1.3],[-4.6, 5.1]];     <br><br>return:<br>&nbsp;&nbsp;array[[0.5055, 16.8977],[0.0266, 271.2179]]
>> \^      |Left 1_Opr  |Element (broadcast) natural constant power.  The element type is not limited to float.             <br>Equivalent function: sm_sc.fv_opr_exp            |select                                                   <br>&nbsp;&nbsp;\^        <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]];     <br><br>return:<br>&nbsp;&nbsp;array[[3.6693, 0.0002],[9.0250, 0.0498]]
>> \^\!    |2_Opr      |Element (broadcast) logarithm.  The element type is not limited to float.                   <br>Equivalent function: sm_sc.fv_opr_log                    |select <br>&nbsp;&nbsp;array[[1.3, 8.8],[2.2, 3.0]]      <br>&nbsp;&nbsp;\^\!      <br>&nbsp;&nbsp;array[[2.6, 11.3],[4.6, 5.1]];      <br><br>return:<br>&nbsp;&nbsp;array[[3.6419, 1.1150],[1.9355, 1.4830]]
>> \^\!    |Left 1_Opr  |Element (broadcast) natural logarithm.  The element type is not limited to float.               <br>Equivalent function: sm_sc.fv_opr_ln                |select                                                   <br>&nbsp;&nbsp;\^\!      <br>&nbsp;&nbsp;array[[2.6, 11.3],[4.6, 5.1]];      <br><br>return:<br>&nbsp;&nbsp;array[[0.9555, 2.4248},{1.5261, 1.6292]]
>> \+\`      |2_Opr      |Element (broadcast) add.                       <br>Equivalent function: sm_sc.fv_opr_add_py                       |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\+\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-1.3, 2.5],[-2.4, 2.1]]
>> \-\`      |2_Opr      |Element (broadcast) sub.                       <br>Equivalent function: sm_sc.fv_opr_sub_py                       |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\-\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[3.9, -20.1],[6.8, -8.1]]
>> \-\`      |Left 1_Opr  |Element (broadcast) opposite number.                   <br>Equivalent function: sm_sc.fv_opr_sub_py                       |select                                                   <br>&nbsp;&nbsp;\-\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[2.6, -11.3],[4.6, -5.1]]
>> \*\`      |2_Opr      |Element (broadcast) mul.                       <br>Equivalent function: sm_sc.fv_opr_mul_py                       |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\*\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-3.38, -99.44],[-10.12, -15.30]]
>> \/\`      |2_Opr      |Element (broadcast) div.                       <br>Equivalent function: sm_sc.fv_opr_div_py                       |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\/\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-0.5000, -0.7788],[-0.4783, -0.5882]]
>> \/\`      |Left 1_Opr  |Element (broadcast) reciprocal.                     <br>Equivalent function: sm_sc.fv_opr_div_py                       |select                                                   <br>&nbsp;&nbsp;\/\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-0.3846, 0.0885],[-0.2174, 0.1961]]
>> \%\`      |2_Opr      |Element (broadcast) positive remainder.        <br>Equivalent function: sm_sc.fv_opr_mod_posi_py                       |select <br>&nbsp;&nbsp;array[[10, -29],[12, -33]]        <br>&nbsp;&nbsp;\%\`        <br>&nbsp;&nbsp;array[[-3, 11],[-5, 5]];            <br><br>return:<br>&nbsp;&nbsp;array[[1, 4],[2, 2]]
>> \%\`\-    |2_Opr      |Element (broadcast) negative remainder.        <br>Equivalent function: sm_sc.fv_opr_mod_nega_py               |select <br>&nbsp;&nbsp;array[[10, -29],[12, -33]]        <br>&nbsp;&nbsp;\%\`\-      <br>&nbsp;&nbsp;array[[-3, 11],[-5, 5]];            <br><br>return: <br>&nbsp;&nbsp;array[[-2, -7],[-3, -3]] 
>> \^\`      |2_Opr      |Element (broadcast) power.                       <br>Equivalent function: sm_sc.fv_opr_pow_py                       |select <br>&nbsp;&nbsp;array[[1.3, 8.8],[2.2, 3.0]]      <br>&nbsp;&nbsp;\^\`        <br>&nbsp;&nbsp;array[[-2.6, 1.3],[-4.6, 5.1]];     <br><br>return:<br>&nbsp;&nbsp;array[[0.5055, 16.8977],[0.0266, 271.2179]]
>> \^\`      |Left 1_Opr  |Element (broadcast) natural constant power.               <br>Equivalent function: sm_sc.fv_opr_exp_py                       |select                                                   <br>&nbsp;&nbsp;\^\`        <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]];     <br><br>return:<br>&nbsp;&nbsp;array[[3.6693, 0.0002],[9.0250, 0.0498]]
>> \^\!\`    |2_Opr      |Element (broadcast) logarithm.                     <br>Equivalent function: sm_sc.fv_opr_log_py                       |select <br>&nbsp;&nbsp;array[[1.3, 8.8],[2.2, 3.0]]      <br>&nbsp;&nbsp;\^\!\`      <br>&nbsp;&nbsp;array[[2.6, 11.3],[4.6, 5.1]];      <br><br>return:<br>&nbsp;&nbsp;array[[3.6419, 1.1150],[1.9355, 1.4830]]
>> \^\!\`    |Left 1_Opr  |Element (broadcast) natural logarithm.                 <br>Equivalent function: sm_sc.fv_opr_ln_py                        |select                                                   <br>&nbsp;&nbsp;\^\!\`      <br>&nbsp;&nbsp;array[[2.6, 11.3],[4.6, 5.1]];      <br><br>return:<br>&nbsp;&nbsp;array[[0.9555, 2.4248},{1.5261, 1.6292]]
>> \|\`      |2_Opr      |Element (broadcast) bit OR, logical OR operation.           <br>Equivalent function: sm_sc.fv_opr_or                        |select <br>&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;\|\`        <br>&nbsp;&nbsp;array[[false, true],[false, true]]; <br><br>return:<br>&nbsp;&nbsp;array[[true, true],[true, true]]                <br>--------------------------<br>select <br>&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;\|\`        <br>&nbsp;&nbsp;array[[B'011', B'101'],[B'011', B'101']]; <br><br>return:<br>&nbsp;&nbsp;array[[B'011', B'111'],[B'111', B'111']]    
>> \&\`      |2_Opr      |Element (broadcast) bit AND, logical AND operational.           <br>Equivalent function: sm_sc.fv_opr_and                       |select <br>&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;\&\`        <br>&nbsp;&nbsp;array[[false, true],[false, true]]; <br><br>return:<br>&nbsp;&nbsp;array[[false, false],[false, false]]            <br>--------------------------<br>select <br>&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;\&\`        <br>&nbsp;&nbsp;array[[B'011', B'101'],[B'011', B'101']]; <br><br>return:<br>&nbsp;&nbsp;array[[B'010', B'001'],[B'001', B'001']]
>> \~\`      |Left 1_Opr  |Element (broadcast) bit NOT, logical NOT operation.           <br>Equivalent function: sm_sc.fv_opr_not                       |select                                                   <br>&nbsp;&nbsp;\~\`        <br>&nbsp;&nbsp;array[[false, true],[false, true]]; <br><br>return:<br>&nbsp;&nbsp;array[[true, false],[true, false]]              <br>--------------------------<br>select                                                         <br>&nbsp;&nbsp;\~\`        <br>&nbsp;&nbsp;array[[B'011', B'101'],[B'011', B'101']]; <br><br>return:<br>&nbsp;&nbsp;array[[B'100', B'010'],[B'100', B'010']]
>> \#\`      |2_Opr      |Element (broadcast) bit XOR, logical XOR operation.         <br>Equivalent function: sm_sc.fv_opr_xor                       |select <br>&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;\#\`        <br>&nbsp;&nbsp;array[[false, true],[false, true]]; <br><br>return:<br>&nbsp;&nbsp;array[[true, true],[true, true]]                <br>--------------------------<br>select <br>&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;\#\`        <br>&nbsp;&nbsp;array[[B'011', B'101'],[B'011', B'101']]; <br><br>return:<br>&nbsp;&nbsp;array[[B'001', B'110'],[B'110', B'110']]
>> \!\#\`    |2_Opr      |Element (broadcast) bit XNOR, logical XNOR operation.         <br>Equivalent function: sm_sc.fv_opr_xnor                      |select <br>&nbsp;&nbsp;array[[true, false],[true, false]]<br>&nbsp;&nbsp;\!\#\`      <br>&nbsp;&nbsp;array[[false, true],[false, true]]; <br><br>return:<br>&nbsp;&nbsp;array[[false, false],[false, false]]            <br>--------------------------<br>select <br>&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]<br>&nbsp;&nbsp;\!\#\`      <br>&nbsp;&nbsp;array[[B'011', B'101'],[B'011', B'101']]; <br><br>return:<br>&nbsp;&nbsp;array[[B'110', B'001'],[B'001', B'001']]
>> \~\`      |2_Opr      |Element (broadcast) does the text regex match.         <br>Equivalent function: sm_sc.fv_opr_is_regexp_match           |select <br>&nbsp;&nbsp;array[['abbbbbc122223', 'abc123'],['abc123', 'ac13']]    <br>&nbsp;&nbsp;\~\`        <br>&nbsp;&nbsp;array[['a.c', '1.\*?3'],['1.3', 'a.\*?c']];    <br><br>return:<br>&nbsp;&nbsp;array[[false, true],[true, true]]
>> \~\`      |Left 1_Opr  |Element (broadcast) complex element conjugation.               <br>Equivalent function: sm_sc.fv_opr_conjugate                 |select                                                    <br>&nbsp;&nbsp;\~\`        <br>&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[];    <br><br>return:<br>&nbsp;&nbsp;array[[(12.3, 25.1), (-2.56, -3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, -3.25)]] :: sm_sc.typ_l_complex[]
>> \@\~\`    |Left 1_Opr  |Element (broadcast) complex element real part.               <br>Equivalent function: sm_sc.fv_opr_real                      |select                                                    <br>&nbsp;&nbsp;\@\~\`      <br>&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[];    <br><br>return:<br>&nbsp;&nbsp;array[[12.3000, -2.5600],[12.3000, 0.0000]]
>> \~\@\`    |Left 1_Opr  |Element (broadcast) complex element imaginary part.               <br>Equivalent function: sm_sc.fv_opr_imaginary                 |select                                                    <br>&nbsp;&nbsp;\~\@\`      <br>&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[];    <br><br>return:<br>&nbsp;&nbsp;array[[-25.1000, 3.2500],[0.0000, 3.2500]]
>> \~\^\`    |Left 1_Opr  |Element (broadcast) exchange of real and imaginary parts of complex elements.           <br>Equivalent function: sm_sc.fv_opr_conjugate_45              |select                                                    <br>&nbsp;&nbsp;\~\^\`      <br>&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[];    <br><br>return:<br>&nbsp;&nbsp;array[[(-25.1, 12.3), (3.25, -2.56)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(0.0, 12.3), (3.25, 0.0)]] :: sm_sc.typ_l_complex[]
>> \|\|\`    |2_Opr      |Element (broadcast) bit concat, character string concat operation.         <br>Equivalent function: sm_sc.fv_opr_concat                    |select <br>&nbsp;&nbsp;array[['abc', 'efgh'],['hijk', 'lm']]    <br>&nbsp;&nbsp;\|\|\`      <br>&nbsp;&nbsp;array[['no', 'pqr'],['stu', 'vwxyz']];    <br><br>return:<br>&nbsp;&nbsp;array[['abcno', 'efghpqr'],['hijkstu', 'lmvwxyz']]
>> \=\=\`    |2_Opr      |Element (broadcast) equality judgment.                 <br>Equivalent function: sm_sc.fv_opr_is_equal                  |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\=\=\`      <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, -3.0]];    <br><br>return:<br>&nbsp;&nbsp;array[[false, false],[false, true]]
>> \<\`      |2_Opr      |Element (broadcast) less than judgment.                 <br>Equivalent function: sm_sc.fv_opr_is_less                   |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\<\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, -3.0]];    <br><br>return:<br>&nbsp;&nbsp;array[[false, true],[false, false]]
>> \>\`      |2_Opr      |Element (broadcast) greater than judgment.                 <br>Equivalent function: sm_sc.fv_opr_is_greater                |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\>\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, -3.0]];    <br><br>return:<br>&nbsp;&nbsp;array[[true, false],[true, false]]
>> \<\=\`    |2_Opr      |Element (broadcast) less than or equal to judgment.             <br>Equivalent function: sm_sc.fv_opr_is_less_ex                |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\<\=\`      <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, -3.0]];    <br><br>return:<br>&nbsp;&nbsp;array[[false, true],[false, true]]
>> \>\=\`    |2_Opr      |Element (broadcast) greater than or equal to judgment.             <br>Equivalent function: sm_sc.fv_opr_is_greater_ex             |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\>\=\`      <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, -3.0]];    <br><br>return:<br>&nbsp;&nbsp;array[[true, false],[true, true]]
>> \<\>\`    |2_Opr      |Element (broadcast) compare.                 <br>Equivalent function: sm_sc.fv_opr_compare                   |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\<\>\`      <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, -3.0]];    <br><br>return:<br>&nbsp;&nbsp;array[[1, -1],[1, 0]]
>> \<\>\`    |Left 1_Opr  |Element (broadcast) positive and negative judgment.           <br>Equivalent function: sm_sc.fv_opr_sign                      |select                                                   <br>&nbsp;&nbsp;\<\>\`      <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 0.0]];    <br><br>return:<br>&nbsp;&nbsp;array[[-1, 1],[-1, 0]]
>> \@\<\`    |2_Opr      |Element (broadcast) take the minimum value.               <br>Equivalent function: sm_sc.fv_opr_least                     |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\@\<\`      <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-2.6, -8.8],[-4.6, -3.0]]
>> \@\>\`    |2_Opr      |Element (broadcast) take the maximum value.               <br>Equivalent function: sm_sc.fv_opr_greatest                  |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\@\>\`      <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[1.3, 11.3],[2.2, 5.1]]
>> \@\|\`      |Left 1_Opr  |Element (broadcast) absolute value of real element, absolute value of complex' real and imaginary part.   <br>Equivalent function: sm_sc.fv_opr_abs                       |select                                                   <br>&nbsp;&nbsp;\@\|\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[2.6, 11.3],[4.6, 5.1]]<br>--------------------------<br>select                                                   <br>&nbsp;&nbsp;\@\|\`        <br>&nbsp;&nbsp;array[[(12.3, 25.1), (2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[]
>> \@\`      |Left 1_Opr  |Element (broadcast) absolute value of real element, norm of complex numbers.   <br>Equivalent function: sm_sc.fv_opr_norm                       |select                                                   <br>&nbsp;&nbsp;\@\`        <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[2.6, 11.3],[4.6, 5.1]]<br>--------------------------<br>select                                                   <br>&nbsp;&nbsp;\@\`        <br>&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[];    <br><br>return:<br>&nbsp;&nbsp;array[[27.9517, 4.1372],[12.3, 3.25]]
>> \~\=\`    |2_Opr      |Element (broadcast) rounding.                 <br>Equivalent function: sm_sc.fv_opr_round                     |select <br>&nbsp;&nbsp;array[[1.327, -8.886],[2.2, -3.036]]    <br>&nbsp;&nbsp;\~\=\`      <br>&nbsp;&nbsp;2;                            <br><br>return:<br>&nbsp;&nbsp;array[[1.33, -8.89],[2.20, -3.04]]
>> \~\<\`    |2_Opr      |Element (broadcast) digital floor value.               <br>Equivalent function: sm_sc.fv_opr_floor                     |select <br>&nbsp;&nbsp;array[[1.327, -8.886],[2.2, -3.036]]    <br>&nbsp;&nbsp;\~\<\`      <br>&nbsp;&nbsp;2;                            <br><br>return:<br>&nbsp;&nbsp;array[[1.32, -8.89],[2.2, -3.04]]
>> \~\>\`    |2_Opr      |Element (broadcast) digital ceiling value.             <br>Equivalent function: sm_sc.fv_opr_ceil                      |select <br>&nbsp;&nbsp;array[[1.327, -8.886],[2.2, -3.036]]    <br>&nbsp;&nbsp;\~\>\`      <br>&nbsp;&nbsp;2;                            <br><br>return:<br>&nbsp;&nbsp;array[[1.33, -8.88],[2.2, -3.03]]
>> \>\<\`    |2_Opr      |Element (broadcast) digital truncation value.               <br>Equivalent function: sm_sc.fv_opr_trunc                     |select <br>&nbsp;&nbsp;array[[1.327, -8.886],[2.2, -3.036]]    <br>&nbsp;&nbsp;\>\<\`      <br>&nbsp;&nbsp;2;                            <br><br>return:<br>&nbsp;&nbsp;array[[1.32, -8.88],[2.20, -3.03]]
>> \|\|\~\|     |Left 1_Opr  |Up-down Symmetric Flipping of 2D Arrays.                     <br>Equivalent function: sm_sc.fv_opr_mirror_heigh                   |select                                                   <br>&nbsp;&nbsp;\|\|\~\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-4.6, 5.1],[-2.6, 11.3]]
>> \|\-\~\|     |Left 1_Opr  |Left-Right Symmetric Flipping of 2D Arrays.                     <br>Equivalent function: sm_sc.fv_opr_mirror_width                   |select                                                   <br>&nbsp;&nbsp;\|\-\~\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[11.3, -2.6],[5.1, -4.6]]
>> \|\+\~\|     |2_Opr      |Symmetrical flipping of high-dimensional arrays with specified dimensions.                 <br>Equivalent function: sm_sc.fv_opr_mirror                     |select                                                   <br>&nbsp;&nbsp;array[[[-2.6, 11.3],[-4.6, 5.1]],[[2.6, -11.3],[4.6, -5.1]]]<br>&nbsp;&nbsp;\|\+\~\|    <br>&nbsp;&nbsp;3;    <br><br>return:<br>&nbsp;&nbsp;array[[[11.3, -2.6],[5.1, -4.6]],[[-11.3, 2.6],[-5.1, 4.6]]]
>> \|\<\~\|     |Left 1_Opr  |Rotate 2D array 90 ° to the left (counterclockwise).              <br>Equivalent function: sm_sc.__fv_turn_x_y_90                |select                                                   <br>&nbsp;&nbsp;\|\<\~\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[11.3, 5.1],[-2.6, -4.6]]
>> \|\>\~\|     |Left 1_Opr  |Rotate 2D array 90 ° to the right (clockwise).              <br>Equivalent function: sm_sc.__fv_turn_y_x_90                |select                                                   <br>&nbsp;&nbsp;\|\>\~\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-4.6, -2.6],[5.1, 11.3]]
>> \|\~\~\|     |Left 1_Opr  |Rotate a 2D array 180 ° (equivalent to double diagonal transpose).       <br>Equivalent function: sm_sc.__fv_turn_y_x_180               |select                                                   <br>&nbsp;&nbsp;\|\~\~\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[5.1, -4.6],[11.3, -2.6]]
>> \|\*\~\|     |2_Opr      |Rotate a high-dimensional array 90 ° according to the specified start and end dimensions.                 <br>Equivalent function: sm_sc.fv_opr_turn_90                    |select                                                   <br>&nbsp;&nbsp;array[[[-2.6, 11.3],[-4.6, 5.1]],[[2.6, -11.3],[4.6, -5.1]]]<br>&nbsp;&nbsp;\|\*\~\|    <br>&nbsp;&nbsp;array[2, 3];    <br><br>return:<br>&nbsp;&nbsp;array[[[11.3, 5.1],[-2.6, -4.6]],[[-11.3, -5.1],[2.6, 4.6]]]
>> \|\~\~\|     |2_Opr      |Rotate a high-dimensional array 180 ° according to the specified start and end dimensions.                 <br>Equivalent function: sm_sc.fv_opr_turn_180                    |select                                                   <br>&nbsp;&nbsp;array[[[-2.6, 11.3],[-4.6, 5.1]],[[2.6, -11.3],[4.6, -5.1]]]<br>&nbsp;&nbsp;\|\~\~\|    <br>&nbsp;&nbsp;array[2, 3];    <br><br>return:<br>&nbsp;&nbsp;array[[[5.1, -4.6],[11.3, -2.6]],[[-5.1, 4.6],[-11.3, 2.6]]]
>> \|\^\~\|     |Left 1_Opr  |Forward transpose (swap left down right up).         <br>Equivalent function: sm_sc.fv_opr_transpose                  |select                                                   <br>&nbsp;&nbsp;\|\^\~\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[-2.6, -4.6],[11.3, 5.1]]
>> \|\^\~\|     |2_Opr      |3D array specifies two dimensions for forward transpose.                     <br>Equivalent function: sm_sc.fv_opr_transpose_nd               |select                                                   <br>&nbsp;&nbsp;array[[[1.3, -8.8, 2.25],[2.2, -3.0, -0.95]],[[-1.3, 8.8, -2.25],[-2.2, 3.0, 0.95]]]<br>&nbsp;&nbsp;\|\^\~\|    <br>&nbsp;&nbsp;array[2, 3];    <br><br>return:<br>&nbsp;&nbsp;array[[[1.3, 2.2],[-8.8, -3.0],[2.25, -0.95]],[[-1.3, -2.2],[8.8, 3.0],[-2.25, 0.95]]]
>> \|\^\~\`\|   |Left 1_Opr  |Conjugate matrix (different from forward transpose).       <br>Equivalent function: sm_sc.fv_opr_conjugate_i                |select                                                   <br>&nbsp;&nbsp;\|\^\~\`\|  <br>&nbsp;&nbsp;array[[(12.3, -25.1), (-2.56, 3.25)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(12.3, 0.0), (0.0, 3.25)]] :: sm_sc.typ_l_complex[];    <br><br>return:<br>&nbsp;&nbsp;array[[(12.30, 25.10), (12.30, 0.00)]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[(-2.56, -3.25), (0.00, -3.25)]] :: sm_sc.typ_l_complex[]
>> \|\~\^\|     |Left 1_Opr  |Right Down Left Up Swap Transpose.             <br>Equivalent function: sm_sc.fv_opr_transpose_i                |select                                                   <br>&nbsp;&nbsp;\|\~\^\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[5.1, 11.3],[-4.6, -2.6]]
>> \|\-\|\|     |2_Opr      |2D plane up and down (height direction) concat.   <br>Equivalent function: sm_sc.fv_opr_concat_heigh               |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\|\-\|\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0],[-2.6, 11.3],[-4.6, 5.1]]
>> \|\|\|\|     |2_Opr      |2D plane left and right (width direction) concat.   <br>Equivalent function: sm_sc.fv_opr_concat_width               |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\|\|\|\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[1.3, -8.8, -2.6, 11.3],[2.2, -3.0, -4.6, 5.1]]
>> \|\`\|       |2_Opr      |Inner product, corresponding element product total.       <br>Equivalent function: sm_sc.fv_opr_prod_inner                 |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\|\`\|      <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:-128.24
>> \|\`\^\|     |2_Opr      |Inner product power, corresponding to the sum of elements multiplied together.     <br>Equivalent function: sm_sc.fv_opr_prod_inner_pow             |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\|\`\^\|    <br>&nbsp;&nbsp;3.0;                                <br><br>return:-695.627
>> \|\*\*\|     |2_Opr      |Matrix multiplication: multiplication, dot product, quantity product.   <br>Equivalent function: sm_sc.fv_opr_prod_mx                    |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\|\*\*\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[37.1000, -30.1900],[8.0800, 9.5600]]
>> \|\*\*\|     |Left 1_Opr  |Perform matrix multiplication with the original matrix after matrix transposition.           <br>Equivalent function: sm_sc.fv_opr_prod_mx_right              |select                                                   <br>&nbsp;&nbsp;\|\*\*\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];    <br><br>return:<br>&nbsp;&nbsp;array[[27.9200, -52.8400],[-52.8400, 153.7000]]
>> \|\*\^\|     |2_Opr      |Square matrix multiplication power, square matrix multiplication with itself for N times.               <br>Equivalent function: sm_sc.fv_opr_prod_mx_pow                |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\|\*\^\|    <br>&nbsp;&nbsp;3;                                  <br><br>return:<br>&nbsp;&nbsp;array[[1.3000,-8.8000],[2.2000,-3.0000]]
>> \|\@\+\|       |Left 1_Opr<br>2_Opr  |SUM of element aggregation.                 <br>Equivalent function: sm_sc.fv_aggr_slice_sum               |select <br>&nbsp;&nbsp;\|\@\+\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:9.2
>> \|\@\*\|       |Left 1_Opr<br>2_Opr  |Direct product of element aggregation.                 <br>Equivalent function: sm_sc.fv_aggr_slice_prod              |select <br>&nbsp;&nbsp;\|\@\*\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:689.2548
>> \|\@\/\|       |Left 1_Opr<br>2_Opr  |Average of element aggregation.                 <br>Equivalent function: sm_sc.fv_aggr_slice_avg               |select <br>&nbsp;&nbsp;\|\@\/\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:2.3
>> \|\@\>\|       |Left 1_Opr<br>2_Opr  |Maximum of element aggregation.                 <br>Equivalent function: sm_sc.fv_aggr_slice_max               |select <br>&nbsp;&nbsp;\|\@\>\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:11.3
>> \|\@\<\|       |Left 1_Opr<br>2_Opr  |Minimum of element aggregation.                 <br>Equivalent function: sm_sc.fv_aggr_slice_min               |select <br>&nbsp;&nbsp;\|\@\<\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:-4.6
>> \|\@\/\=\|     |Left 1_Opr<br>2_Opr  |Element aggregation span (maximum to minimum).      <br>Equivalent function: sm_sc.fv_aggr_slice_ptp               |select <br>&nbsp;&nbsp;\|\@\/\=\|  <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:15.9
>> \|\@\#\|       |Left 1_Opr<br>2_Opr  |Population variance of element aggregation.             <br>Equivalent function: sm_sc.fv_aggr_slice_var_pop           |select <br>&nbsp;&nbsp;\|\@\#\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:40.115
>> \|\@\#\`\|     |Left 1_Opr<br>2_Opr  |Sample variance of element aggregation.             <br>Equivalent function: sm_sc.fv_aggr_slice_var_samp          |select <br>&nbsp;&nbsp;\|\@\#\`\|  <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:53.4867
>> \|\@\%\|       |Left 1_Opr<br>2_Opr  |Population standard deviation of element aggregation.           <br>Equivalent function: sm_sc.fv_aggr_slice_stddev_pop        |select <br>&nbsp;&nbsp;\|\@\%\|    <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:6.3336
>> \|\@\%\`\|     |Left 1_Opr<br>2_Opr  |Sample standard deviation of element aggregation.           <br>Equivalent function: sm_sc.fv_aggr_slice_stddev_samp       |select <br>&nbsp;&nbsp;\|\@\%\`\|  <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:7.3135
>> \|\@\/\>\|     |Left 1_Opr<br>2_Opr  |Mode of element aggregation.                 <br>Equivalent function: sm_sc.fv_aggr_slice_mode              |select <br>&nbsp;&nbsp;\|\@\/\>\|  <br>&nbsp;&nbsp;array[[-4.6, 11.3],[-4.6, 5.1]];         <br><br>return:-4.6
>> \|\@\/\<\|     |Left 1_Opr<br>2_Opr  |Median of element aggregation.               <br>Equivalent function: sm_sc.fv_aggr_slice_median            |select <br>&nbsp;&nbsp;\|\@\/\<\|  <br>&nbsp;&nbsp;array[[-2.6, 11.3],[-4.6, 5.1]];         <br><br>return:1.25
>> \|\@\|\|\|     |Left 1_Opr<br>2_Opr  |Element aggregation bit concat.             <br>Equivalent function: sm_sc.fv_aggr_slice_concat            |select <br>&nbsp;&nbsp;\|\@\|\|\|  <br>&nbsp;&nbsp;array[[B'10', B'001'],[B'01', B'1']];    <br><br>return:B'10001011'
>> \|\@\|\`\|     |Left 1_Opr<br>2_Opr  |Element aggregation bit OR, logical OR.       <br>Equivalent function: sm_sc.fv_aggr_slice_or                |select <br>&nbsp;&nbsp;\|\@\|\`\|  <br>&nbsp;&nbsp;array[[B'001', B'101'],[B'100', B'010']];<br><br>return:B'111'<br>--------------------------<br>select <br>&nbsp;&nbsp;\|\@\|\`\|  <br>&nbsp;&nbsp;array[[true, false],[true, true]];<br><br>return:true
>> \|\@\&\`\|     |Left 1_Opr<br>2_Opr  |Element aggregation bit AND, logical AND.       <br>Equivalent function: sm_sc.fv_aggr_slice_and               |select <br>&nbsp;&nbsp;\|\@\&\`\|  <br>&nbsp;&nbsp;array[[B'001', B'101'],[B'100', B'010']];<br><br>return:B'000'<br>--------------------------<br>select <br>&nbsp;&nbsp;\|\@\&\`\|  <br>&nbsp;&nbsp;array[[true, false],[true, true]];<br><br>return:false
>> \|\@\=\|       |Left 1_Opr<br>2_Opr  |Element aggregation null replace.                 <br>Equivalent function: sm_sc.fv_aggr_slice_coalesce          |select <br>&nbsp;&nbsp;\|\@\=\|    <br>&nbsp;&nbsp;array[[null, 11.3],[null, 5.1]];         <br><br>return:11.3
>> \+\`\|         |2_Opr      |Matrix multicast - convolute add with stride 1.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_add_stride_1    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]     <br>&nbsp;&nbsp;\+\`\|        <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]  <br><br>return:<br>&nbsp;&nbsp;array[[3.2, 4.16, 1.96, 6.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[9.28, 9.19, 5.37, 14.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[9.3, 7.59, 5.39, 12.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[9.28, 9.19, 5.37, 14.2]]<br>--------------------------<br>select <br>&nbsp;&nbsp;array[2.5, 1.26, 3.3, 8.6, 4.69, 9.7, 8.6, 4.69, 9.7]     <br>&nbsp;&nbsp;\+\`\|        <br>&nbsp;&nbsp;array[0.7, 2.9, 0.68, 4.5]  <br><br>return:<br>&nbsp;&nbsp;array[3.2, 4.16, 1.96, 6.2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,9.28, 9.19, 5.37, 14.2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,9.3, 7.59, 5.39, 12.6<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,9.28, 9.19, 5.37, 14.2]
>> \-\`\|         |2_Opr      |Matrix multicast - convolute sub with stride 1.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_sub_stride_1    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]     <br>&nbsp;&nbsp;\-\`\|        <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]  <br><br>return:<br>&nbsp;&nbsp;array[[1.8,-1.64,0.56,0.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.92,0.19,4.01,5.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.9,1.79,3.99,6.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[7.92,0.19,4.01,5.2]]
>> \|\-\`         |2_Opr      |Matrix multicast - deconvolute sub with stride 1.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_conv_de_sub_stride_1 <br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]     <br>&nbsp;&nbsp;\|\-\`        <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]  <br><br>return:<br>&nbsp;&nbsp;array[[-1.8, 1.64, -0.56, -0.4]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-7.92, -0.19, -4.01, -5.2]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-7.9, -1.79, -3.99, -6.8]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-7.92, -0.19, -4.01, -5.2]]<br>--------------------------<br>select <br>&nbsp;&nbsp;array[0.7, 2.9, 0.68, 4.5]     <br>&nbsp;&nbsp;\|\-\`        <br>&nbsp;&nbsp;array[2.5, 1.26, 3.3, 8.6, 4.69, 9.7, 8.6, 4.69, 9.7]  <br><br>return:<br>&nbsp;&nbsp;array[-1.8, 1.64, -0.56, -0.4<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,-7.92, -0.19, -4.01, -5.2<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,-7.9, -1.79, -3.99, -6.8<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,-7.92, -0.19, -4.01, -5.2]
>> \*\`\|         |2_Opr      |Matrix multicast - convolute mul with stride 1.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_mul_stride_1    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]     <br>&nbsp;&nbsp;\*\`\|        <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]  <br><br>return:<br>&nbsp;&nbsp;array[[1.75, 3.654, 0.882, 9.57]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[5.848, 21.105, 3.1892, 43.65]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[6.02, 13.601, 3.283, 28.13]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[5.848, 21.105, 3.1892, 43.65]]
>> \/\`\|         |2_Opr      |Matrix multicast - convolute div with stride 1.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_div_stride_1    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]     <br>&nbsp;&nbsp;\/\`\|        <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]  <br><br>return:<br>&nbsp;&nbsp;array[[3.571, 0.434, 1.800, 1.138]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[12.647, 1.042, 6.897, 2.156]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[12.286, 1.617, 6.700, 3.345]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[12.647, 1.042, 6.897, 2.156]]
>> \|\/\`         |2_Opr      |Matrix multicast - deconvolute div with stride 1.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_conv_de_div_stride_1 <br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]     <br>&nbsp;&nbsp;\|\/\`        <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]  <br><br>return:<br>&nbsp;&nbsp;array[[0.280, 2.302, 0.556, 0.879]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.079, 0.959, 0.145, 0.464]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.081, 0.618, 0.149, 0.299]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.079, 0.959, 0.145, 0.464]]
>> \^\`\|         |2_Opr      |Matrix multicast - convolute pow with stride 1.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_pow_stride_1    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]     <br>&nbsp;&nbsp;\^\`\|        <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]  <br><br>return:<br>&nbsp;&nbsp;array[[1.899, 1.955, 1.176, 31.893]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4.320, 1047.799, 2.860, 27572.288]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4.510, 88.390, 2.950, 727.173]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[4.320, 1047.799, 2.860, 27572.288]]
>> \|\^\`         |2_Opr      |Matrix multicast - deconvolute pow with stride 1.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_conv_de_pow_stride_1 <br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]     <br>&nbsp;&nbsp;\|\^\`        <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]  <br><br>return:<br>&nbsp;&nbsp;array[[0.410, 3.825, 0.638, 33.567]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.036, 1157.622, 0.164, 2168509.772]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.047, 147.451, 0.188, 30567.546]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[0.036, 1157.622, 0.164, 2168509.772]]
>> \^\!\`\|       |2_Opr      |Matrix multicast - convolute log with stride 1.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_conv_log_stride_1    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]     <br>&nbsp;&nbsp;\^\!\`\|      <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]  <br><br>return:<br>&nbsp;&nbsp;array[[-0.389, 4.607, -1.543, 0.892]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-0.179, 0.973, -0.250, 0.662]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-0.166, 0.689, -0.231, 0.469]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-0.179, 0.973, -0.250, 0.662]]
>> \|\^\!\`       |2_Opr      |Matrix multicast - deconvolute log with stride 1.<br>Co parameter default, supports one-dimensional. <br>Equivalent function: sm_sc.fv_opr_conv_de_log_stride_1 <br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]     <br>&nbsp;&nbsp;\|\^\!\`      <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]  <br><br>return:<br>&nbsp;&nbsp;array[[-2.569, 0.217, -0.648, 1.121]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-5.579, 1.027, -4.007, 1.511]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-6.033, 1.452, -4.333, 2.134]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[-5.579, 1.027, -4.007, 1.511]]
>> \*\*\|         |2_Opr      |Matrix multicast - convolute with stride 1.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_2d_stride_1     <br>Input param:<br>&nbsp;&nbsp;1. Background matrix;<br>&nbsp;&nbsp;2. Window matrix;             |select <br>&nbsp;&nbsp;array[[2.5, 1.26, 3.3], [8.6, 4.69, 9.7], [8.6, 4.69, 9.7]]     <br>&nbsp;&nbsp;\*\*\|        <br>&nbsp;&nbsp;array[[0.7, 2.9], [0.68, 4.5]]  <br><br>return:<br>&nbsp;&nbsp;array[[40.3298, 28.305]<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;,[70.1122, 48.092]]
>> \%\+\`\|         |2_Opr      |Matrix multicast - convolute add with stride = window\'s length.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_add_stride_window    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\%\+\`\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;0.68]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[3.2,&nbsp;&nbsp;4.16,&nbsp;&nbsp;3.98,&nbsp;&nbsp;9.3,&nbsp;&nbsp;7.59,&nbsp;&nbsp;10.38,&nbsp;&nbsp;9.3,&nbsp;&nbsp;7.59,&nbsp;&nbsp;10.38]
>> \%\-\`\|         |2_Opr      |Matrix multicast - convolute sub with stride = window\'s length.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_sub_stride_window    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\%\-\`\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;0.68]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.8,&nbsp;&nbsp;-1.64,&nbsp;&nbsp;2.62,&nbsp;&nbsp;7.9,&nbsp;&nbsp;1.79,&nbsp;&nbsp;9.02,&nbsp;&nbsp;7.9,&nbsp;&nbsp;1.79,&nbsp;&nbsp;9.02]
>> \|\-\`\%         |2_Opr      |Matrix multicast - deconvolute sub with stride = window\'s length.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_conv_de_sub_stride_window <br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;0.68,&nbsp;&nbsp;4.5]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\|\-\`\%&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[-1.8,&nbsp;&nbsp;1.64,&nbsp;&nbsp;-2.62,&nbsp;&nbsp;-4.1,&nbsp;&nbsp;-3.99,&nbsp;&nbsp;-6.8,&nbsp;&nbsp;-4.01,&nbsp;&nbsp;-5.2]
>> \%\*\`\|         |2_Opr      |Matrix multicast - convolute mul with stride = window\'s length.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_mul_stride_window    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\%\*\`\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;0.68,&nbsp;&nbsp;4.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.75,&nbsp;&nbsp;3.654,&nbsp;&nbsp;2.244,&nbsp;&nbsp;38.70,&nbsp;&nbsp;3.283,&nbsp;&nbsp;28.13,&nbsp;&nbsp;5.848,&nbsp;&nbsp;21.105]
>> \%\/\`\|         |2_Opr      |Matrix multicast - convolute div with stride = window\'s length.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_div_stride_window    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\%\/\`\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;0.68]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[3.571,&nbsp;&nbsp;0.434,&nbsp;&nbsp;4.853,&nbsp;&nbsp;12.286,&nbsp;&nbsp;1.617,&nbsp;&nbsp;14.265,&nbsp;&nbsp;12.286,&nbsp;&nbsp;1.617,&nbsp;&nbsp;14.265]
>> \|\/\`\%         |2_Opr      |Matrix multicast - deconvolute div with stride = window\'s length.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_conv_de_div_stride_window <br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;0.68,&nbsp;&nbsp;4.5]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\|\/\`\%&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.280,&nbsp;&nbsp;0.540,&nbsp;&nbsp;1.364,&nbsp;&nbsp;0.081,&nbsp;&nbsp;0.145,&nbsp;&nbsp;0.464,&nbsp;&nbsp;0.081,&nbsp;&nbsp;0.145,&nbsp;&nbsp;0.464]
>> \%\^\`\|         |2_Opr      |Matrix multicast - convolute pow with stride = window\'s length.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_pow_stride_window    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\%\^\`\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;4.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[1.899,&nbsp;&nbsp;1.955,&nbsp;&nbsp;215.433,&nbsp;&nbsp;4.510,&nbsp;&nbsp;88.390,&nbsp;&nbsp;27572.288,&nbsp;&nbsp;4.510,&nbsp;&nbsp;88.390,&nbsp;&nbsp;27572.288]
>> \|\^\`\%         |2_Opr      |Matrix multicast - deconvolute pow with stride = window\'s length.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_conv_de_pow_stride_window <br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;0.68,&nbsp;&nbsp;4.5]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\|\^\`\%&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.410,&nbsp;&nbsp;33.567,&nbsp;&nbsp;0.036,&nbsp;&nbsp;1157.622,&nbsp;&nbsp;0.031,&nbsp;&nbsp;9475.950,&nbsp;&nbsp;0.164,&nbsp;&nbsp;2168509.772]
>> \%\^\!\`\|       |2_Opr      |Matrix multicast - convolute log with stride = window\'s length.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_conv_log_stride_window    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;9.7]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\%\^\!\`\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;0.68,&nbsp;&nbsp;4.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[-0.389,&nbsp;&nbsp;4.607,&nbsp;&nbsp;-0.323,&nbsp;&nbsp;0.699,&nbsp;&nbsp;-0.231,&nbsp;&nbsp;0.469,&nbsp;&nbsp;-0.179,&nbsp;&nbsp;0.662]
>> \|\^\!\`\%       |2_Opr      |Matrix multicast - deconvolute log with stride = window\'s length.<br>Co parameter default, supports one-dimensional. <br>Equivalent function: sm_sc.fv_opr_conv_de_log_stride_window <br>Input param:<br>&nbsp;&nbsp;1. Window matrix; <br>&nbsp;&nbsp;2. Background matrix;              |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;4.5]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\|\^\!\`\%&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[-2.569,&nbsp;&nbsp;0.217,&nbsp;&nbsp;0.794,&nbsp;&nbsp;-6.033,&nbsp;&nbsp;1.452,&nbsp;&nbsp;1.511,&nbsp;&nbsp;-6.033,&nbsp;&nbsp;1.452,&nbsp;&nbsp;1.511]
>> \%\*\*\|         |2_Opr      |Matrix multicast - convolute with stride = window\'s length.<br>Co parameter default, supports one-dimensional.     <br>Equivalent function: sm_sc.fv_opr_conv_2d_stride_window     <br>Input param:<br>&nbsp;&nbsp;1. Background matrix;<br>&nbsp;&nbsp;2. Window matrix;             |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.6,&nbsp;&nbsp;4.69]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\%\*\*\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[0.7,&nbsp;&nbsp;2.9,&nbsp;&nbsp;0.68,&nbsp;&nbsp;4.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[46.348,&nbsp;&nbsp;58.366]
>> \@\>\|           |2_Opr      |Matrix max pooling - stride 1.<br>Co parameter default, supports one-dimensional.       <br>Equivalent function: sm_sc.fv_opr_pool_max_stride_1         <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window height and width specifications;          |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;2.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.55]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\@\>\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[8.6,9.7,9.7],[8.6,9.7,9.7]]
>> \@\/\|           |2_Opr      |Matrix average pooling - stride 1.<br>Co parameter default, supports one-dimensional.       <br>Equivalent function: sm_sc.fv_opr_pool_avg_stride_1         <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window height and width specifications;          |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;2.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.55]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\@\/\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array[[4.263,4.738,5.525],[6.645,7.195,8.613]]
>> \@\`\|           |2_Opr      |Matrix preservation invariant pooling - stride 1.<br>Co parameter default, supports one-dimensional.   <br>Equivalent function: sm_sc.fv_opr_pool_none_stride_1        <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window height and width specifications;          |select&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.5,&nbsp;&nbsp;1.26,&nbsp;&nbsp;3.3,&nbsp;&nbsp;2.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,&nbsp;&nbsp;4.69,&nbsp;&nbsp;9.7,&nbsp;&nbsp;8.55]<br>&nbsp;&nbsp;&nbsp;&nbsp;]&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;\@\`\|&nbsp;&nbsp;<br>&nbsp;&nbsp;&nbsp;&nbsp;array[2,&nbsp;&nbsp;2]<br>&nbsp;&nbsp;&nbsp;&nbsp;<br>return:<br>&nbsp;&nbsp;&nbsp;&nbsp;array<br>&nbsp;&nbsp;&nbsp;&nbsp;[<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[2.5,1.26,1.26,3.3,3.3,2.6]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,4.69,4.69,9.7,9.7,6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,4.69,4.69,9.7,9.7,6.5]<br>&nbsp;&nbsp;&nbsp;&nbsp;,&nbsp;&nbsp;[8.6,4.69,4.69,9.7,9.7,8.55]<br>&nbsp;&nbsp;&nbsp;&nbsp;]
<!--
>> \<\<\`    |Right 1_Opr  |Element (broadcast) bit shift left operation.               <br>Equivalent function: sm_sc.fv_opr_shift_left                |select <br>&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]    <br>&nbsp;&nbsp;\<\<\`      <br>&nbsp;&nbsp;array[[1, 2],[-1, -2]];    <br><br>return:<br>&nbsp;&nbsp;array[[B'100', B'100'],[B'010', B'000']]
>> \>\>\`    |Right 1_Opr  |Element (broadcast) bit shift right operation.               <br>Equivalent function: sm_sc.fv_opr_shift_right              |select <br>&nbsp;&nbsp;array[[B'010', B'011'],[B'101', B'011']]    <br>&nbsp;&nbsp;\>\>\`      <br>&nbsp;&nbsp;array[[1, 2],[-1, -2]];    <br><br>return:<br>&nbsp;&nbsp;array[[B'001', B'000'],[B'010', B'100']]
>> \~\=\`    |Right 1_Opr  |Element (broadcast) round to integer.             <br>Equivalent function: sm_sc.fv_opr_round                     |select <br>&nbsp;&nbsp;array[[1.327, -8.886],[2.2, -3.036]]    <br>&nbsp;&nbsp;\~\=\`;                                                   <br><br>return:<br>&nbsp;&nbsp;array[[1, -9],[2, -3]]
>> \~\<\`    |Right 1_Opr  |Element (broadcast) take an integer for the digital floor value.           <br>Equivalent function: sm_sc.fv_opr_floor                     |select <br>&nbsp;&nbsp;array[[1.327, -8.886],[2.2, -3.036]]    <br>&nbsp;&nbsp;\~\<\`;                                                   <br><br>return:<br>&nbsp;&nbsp;array[[1, -9],[2, -4]]
>> \~\>\`    |Right 1_Opr  |element (broadcast) take an integer for the digital ceiling values.         <br>Equivalent function: sm_sc.fv_opr_ceil                      |select <br>&nbsp;&nbsp;array[[1.327, -8.886],[2.2, -3.036]]    <br>&nbsp;&nbsp;\~\>\`;                                                   <br><br>return:<br>&nbsp;&nbsp;array[[2, -8],[3, -3]]
>> \>\<\`    |Right 1_Opr  |element (broadcast) digit truncation value takes an integer.           <br>Equivalent function: sm_sc.fv_opr_trunc                     |select <br>&nbsp;&nbsp;array[[1.327, -8.886],[2.2, -3.036]]    <br>&nbsp;&nbsp;\>\<\`;                                                   <br><br>return:<br>&nbsp;&nbsp;array[[1, -8],[2, -3]]
>> \|\*\*\|     |Right 1_Opr  |Matrix multiplication and its transposition.               <br>Equivalent function: sm_sc.fv_opr_prod_mx_left               |select <br>&nbsp;&nbsp;array[[1.3, -8.8],[2.2, -3.0]]    <br>&nbsp;&nbsp;\|\*\*\|;                                                       <br><br>return:<br>&nbsp;&nbsp;array[[79.1300, 29.2600],[29.2600, 13.8400]]
-->
<!--
>> \%\@\>\|         |2_Opr      |Matrix pooling - stride = window\'s length.<br>Co parameter default, supports one-dimensional.         <br>Equivalent function: sm_sc.fv_opr_pool_max_stride_window    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window height and width specifications;          |select <br>&nbsp;&nbsp;array
>> \%\@\/\|         |2_Opr      |Matrix pooling - stride = window\'s length.<br>Co parameter default, supports one-dimensional.         <br>Equivalent function: sm_sc.fv_opr_pool_avg_stride_window    <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window height and width specifications;          |select <br>&nbsp;&nbsp;array
>> \%\@\`\|         |2_Opr      |Matrix pooling - stride = window\'s length.<br>Co parameter default, supports one-dimensional.         <br>Equivalent function: sm_sc.fv_opr_pool_none_stride_window   <br>Input param:<br>&nbsp;&nbsp;1. Background matrix; <br>&nbsp;&nbsp;2. Window height and width specifications;          |select <br>&nbsp;&nbsp;array
-->
>

>###  3. Introduction to stored procedures
>>####   st_a. procedure for ML train:
>>
>>> Process name | Main operation table | Process description
>>> :-|:-|:-
>>> sm_sc.prc_linear_regression   |1. sm_sc.tb_classify_task: <br>&nbsp;&nbsp;&nbsp;&nbsp;Classification task training task table, used for planning tasks and storing an array of trained weight values; <br>2. sm_sc.tb_nn_train_input_buff:<br>&nbsp;&nbsp;&nbsp;&nbsp;Encoded training set data table, used for importing training sets;         |Linear regression.<br>Input param:<br>&nbsp;&nbsp;1. Training task identification; <br>&nbsp;&nbsp;2. Learning rate; <br>&nbsp;&nbsp;3. Plan training frequency; <br>&nbsp;&nbsp;4. Loss function gradient threshold; 
>>> sm_sc.prc_logistic_regression |1. sm_sc.tb_classify_task: <br>&nbsp;&nbsp;&nbsp;&nbsp;Classification task training task table, used for planning tasks and storing an array of trained weight values; <br>2. sm_sc.tb_nn_train_input_buff:<br>&nbsp;&nbsp;&nbsp;&nbsp;Encoded training set data table, used for importing training sets;         |Logistic regression.<br>Input param:<br>&nbsp;&nbsp;1. Training task identification; <br>&nbsp;&nbsp;2. Learning rate; <br>&nbsp;&nbsp;3. Plan training frequency; <br>&nbsp;&nbsp;4. Loss function gradient threshold; 
>>> sm_sc.prc_kmeans_pp           |1. sm_sc.tb_cluster_task: <br>&nbsp;&nbsp;&nbsp;&nbsp;Kmean clustering task data table, used to input kmeans hard clustering dataset and output clustering results; <br>2. sm_sc.__vt_kmean_ods_array:<br>&nbsp;&nbsp;&nbsp;&nbsp;Encoded training set data table, used to input data during execution; <br>3. sm_sc.tb_nn_train_input_buff: <br>&nbsp;&nbsp;&nbsp;&nbsp;Encoded training set data table, used for importing training sets;         |Kmeans hard clustering.<br>Input param:<br>&nbsp;&nbsp;1. Training task identification; 
>>> sm_sc.prc_dbscan_pp           |1. sm_sc.tb_cluster_task: <br>&nbsp;&nbsp;&nbsp;&nbsp;Clustering task data table, used to input the dbscan soft clustering dataset and output clustering results; <br>2. sm_sc.__vt_dbscan_ods_array:<br>&nbsp;&nbsp;&nbsp;&nbsp;Encoded training set data table, used to input data during execution; <br>3. sm_sc.tb_nn_train_input_buff: <br>&nbsp;&nbsp;&nbsp;&nbsp;Encoded training set data table, used for importing training sets;         |Dbscan soft clustering.<br>Input param:<br>&nbsp;&nbsp;1. Training task identification; 

>>
>>####   st_b. procedure for DL train:
>>
>>> Process name | Main operation table | Process description
>>> :-|:-|:-
>>> sm_sc.prc_nn_prepare<br>sm_sc.prc_nn_prepare_p          |1. sm_sc.tb_nn_node: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network nodes and attribute tables, used to configure neural network nodes.<br>&nbsp;&nbsp;&nbsp;&nbsp;During the training process, register attribute parameters and gradients;<br>2. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;          |Preparation for deep learning task training.<br>Input param:<br>&nbsp;&nbsp;1. Training task identification; <br>&nbsp;&nbsp;2. Plan training frequency; <br>&nbsp;&nbsp;3. Small batch random size for each selected range.  (default: null); <br>&nbsp;&nbsp;4. Small batch random selected ranges.  (default: null);
>>> sm_sc.prc_nn_train<br>sm_sc.prc_nn_train            |1. sm_sc.tb_nn_node: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network nodes and attribute tables, used to configure neural network nodes.<br>&nbsp;&nbsp;&nbsp;&nbsp;During the training process, register attribute parameters and gradients; <br>2. sm_sc.tb_nn_path: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network path table, used to configure neural network propagation paths; <br>3.sm_sc.__vt_nn_node: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network nodes and attribute tables, export models to store trained attribute parameters;<br>4.sm_sc.__vt_nn_path: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network path tables, export models to store net structure;<br>5. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;          |Deep learning task training execution.<br>Input param:<br>&nbsp;&nbsp;1. Training task identification; <br>&nbsp;&nbsp;2. Learning rate.  (default 0.0003); <br>&nbsp;&nbsp;3. Loss function gradient threshold.  (default 0.0001); <br>&nbsp;&nbsp;4. if use ADAM.  (default false); <br>&nbsp;&nbsp;5. L1 regularization retio.  (default null); <br>&nbsp;&nbsp;6. L2 regularization retio.  (default null); <br>&nbsp;&nbsp;6. If or not log middle tensor array during current propagation.  (default false); 

>>
>>####   st_c. procedure for manage DL session:
>>
>>> Process name | Main operation table | Process description
>>> :-|:-|:-
>>> sm_sc.prc_nn_add_sess<br>sm_sc.prc_nn_add_sess_p          |1. sm_sc.__vt_nn_sess: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network session resource pool, used to deploy session connection resources; <br>2. sm_sc.__vt_tmp_nn_node: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network nodes and attribute tables for predicting access;          |Deploy DL models to increase connection resources.<br>Input: <br>&nbsp;&nbsp;1. Neural network ID; <br>&nbsp;&nbsp;2. count of session to deploy; 
>>> sm_sc.prc_nn_del_sess<br>sm_sc.prc_nn_del_sess_p          |1. sm_sc.__vt_nn_sess: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network session resource pool, used to deploy session connection resources; <br>2. sm_sc.__vt_tmp_nn_node: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network nodes and attribute tables for predicting access;<br>3. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;          |Clean up DL models and log out connection resources. <br>Input: <br>&nbsp;&nbsp;1. Neural network ID; <br>&nbsp;&nbsp;2. numbric range of sessions to be deleted; 
>>> sm_sc.prc_nn_subscribe_sess    |1. sm_sc.__vt_nn_sess: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network session resource pool, used to deploy session connection resources;          |Access the model and apply for a session resource.  <br>Input: <br>&nbsp;&nbsp;1. Neural network ID; <br>Output: <br>&nbsp;&nbsp;1. session ID subscribed; 
>>> sm_sc.prc_nn_release_sess      |1. sm_sc.__vt_nn_sess: <br>&nbsp;&nbsp;&nbsp;&nbsp;Neural network session resource pool, used to deploy session connection resources;          |End of model access, release session resources.  <br>Input: <br>&nbsp;&nbsp;1. Neural network ID; <br>&nbsp;&nbsp;2. session ID to release; 

>>
>>####   st_d. procedure for manage DL model:
>>
>>> Process name | Main operation table | Process description
>>> :-|:-|:-
>>> sm_sc.prc_add_pretrain_from_infer<br>sm_sc.prc_add_pretrain_from_infer_p  |1. sm_sc.tb_classify_task：<br>&nbsp;&nbsp;&nbsp;&nbsp;task table for pretrain; <br>2. sm_sc.tb_nn_node：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network node table for pretrain; <br>3. sm_sc.tb_nn_path：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network path table for pretrain;<br>4. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;          |For building training model from infer model.  <br>Equivalent to call prc_import_pretrain_model without training dataset.  <br>Input param: <br>&nbsp;&nbsp;1. deep learn model ID; <br>&nbsp;&nbsp;2. Loss function type.  refer dictionary table: <br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;select * from sm_sc.tb_dic_enum where enum_name = 'loss_fn_type' 
>>> sm_sc.prc_import_pretrain_model<br>sm_sc.prc_import_pretrain_model_p  |1. sm_sc.tb_classify_task：<br>&nbsp;&nbsp;&nbsp;&nbsp;task table for pretrain; <br>2. sm_sc.tb_nn_node：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network node table for pretrain; <br>3. sm_sc.tb_nn_path：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network path table for pretrain; <br>4. sm_sc.tb_nn_train_input_buff：<br>&nbsp;&nbsp;&nbsp;&nbsp;train data table;<br>5. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;          |for import pretrain model. <br>Input param: <br>&nbsp;&nbsp;1. deep learn model ID; <br>&nbsp;&nbsp;2. task table file path; <br>&nbsp;&nbsp;3. node table file path; <br>&nbsp;&nbsp;4. path table file path; <br>&nbsp;&nbsp;5. train dataset table file path; 
>>> sm_sc.prc_import_infer_model<br>sm_sc.prc_import_infer_model     |1. sm_sc.__vt_nn_node：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network node table for infer; <br>2. sm_sc.__vt_nn_path：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network path table for infer; <br>3. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;          |for import infer model。<br>Input param: <br>&nbsp;&nbsp;1. deep learn model ID; <br>&nbsp;&nbsp;2. node table file path; <br>&nbsp;&nbsp;3. path table file path; 
>>> sm_sc.prc_export_pretrain_model<br>sm_sc.prc_export_pretrain_model_p  |1. sm_sc.tb_classify_task：<br>&nbsp;&nbsp;&nbsp;&nbsp;task table for pretrain; <br>2. sm_sc.tb_nn_node：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network node table for pretrain; <br>3. sm_sc.tb_nn_path：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network path table for pretrain; <br>4. sm_sc.tb_nn_train_input_buff：<br>&nbsp;&nbsp;&nbsp;&nbsp;train data table; <br>5. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;         |for export pretrain model. <br>Input param: <br>&nbsp;&nbsp;1. deep learn model ID; <br>&nbsp;&nbsp;2. task table file path; <br>&nbsp;&nbsp;3. node table file path; <br>&nbsp;&nbsp;4. path table file path; <br>&nbsp;&nbsp;5. train dataset table file path; 
>>> sm_sc.prc_export_infer_model<br>sm_sc.prc_export_infer_model_p     |1. sm_sc.__vt_nn_node：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network node table for infer; <br>2. sm_sc.__vt_nn_path：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network path table for infer; <br>3. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;         |for export pretrain model。<br>Input param: <br>&nbsp;&nbsp;1. deep learn model ID; <br>&nbsp;&nbsp;2. node table file path; <br>&nbsp;&nbsp;3. path table file path; 
>>> sm_sc.prc_del_pretrain_model<br>sm_sc.prc_del_pretrain_model_p     |1. sm_sc.tb_classify_task：<br>&nbsp;&nbsp;&nbsp;&nbsp;task table for pretrain; <br>2. sm_sc.tb_nn_node：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network node table for pretrain; <br>3. sm_sc.tb_nn_path：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network path table for pretrain; <br>4. sm_sc.tb_nn_train_input_buff：<br>&nbsp;&nbsp;&nbsp;&nbsp;train data table; <br>5. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;         |for delete pretrain model。<br>Input param: <br>&nbsp;&nbsp;1. deep learn model ID; 
>>> sm_sc.prc_del_infer_model<br>sm_sc.prc_del_infer_model_p        |1. sm_sc.__vt_nn_node：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network node table for infer; <br>2. sm_sc.__vt_nn_path：<br>&nbsp;&nbsp;&nbsp;&nbsp;neural network path table for infer; <br>3. sm_sc.__vt_array_kv: <br>&nbsp;&nbsp;&nbsp;&nbsp;tensor register，for getting/setting middle tensor array;         |for delete infer model。<br>Input param: <br>&nbsp;&nbsp;1. deep learn model ID; 

<!--
>>> sm_sc.prc_mx_evd_value         |1. sm_sc.__vt_tmp_eigen_arrays: <br>&nbsp;&nbsp;&nbsp;&nbsp;Table of eigenvalues and eigenvectors, used for feature decomposition and intermediate data storage;  |Eigenvalues of matrices.<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:
>>> sm_sc.prc_mx_evd               |1. sm_sc.__vt_tmp_eigen_arrays: <br>&nbsp;&nbsp;&nbsp;&nbsp;Table of eigenvalues and eigenvectors, used for feature decomposition and intermediate data storage;  |Eigenvalues and eigenvectors of matrices<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:
>>> sm_sc.prc_mx_max_abs_evd_value |1. sm_sc.__vt_tmp_eigen_arrays: <br>&nbsp;&nbsp;&nbsp;&nbsp;Table of eigenvalues and eigenvectors, used for feature decomposition and intermediate data storage;  |Main eigenvalue, modulus of maximum eigenvalue (absolute value).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:
>>> sm_sc.prc_mx_min_abs_evd_value |1. sm_sc.__vt_tmp_eigen_arrays: <br>&nbsp;&nbsp;&nbsp;&nbsp;Table of eigenvalues and eigenvectors, used for feature decomposition and intermediate data storage;  |Modulus of minimum eigenvalue (absolute value).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:
>>> sm_sc.prc_mx_norm_eigenvalue   |    |Spectral norm  2-norm spectral  modulus (maximum singular value).<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:
>>> sm_sc.prc_mx_svd               |    |Singular value decomposition<br>Input param:<br>&nbsp;&nbsp;1. Original matrix:
-->        

>###  4. Neural network nodes support operation types
> PG4ML neural network 2_Opr operators typically support broadcast types and typically require alignment of the dimensionality of all input parameters at or above the 2_Opr, that is, when the dimension length is not equal and can be broadcasted, the number of dimensions is equal.
>> Lambda naming, that is, operation type | Calling function | Count of operations | Whether to configure coparameters | Whether to support differentiation | Derivative function | Whether to calculate the derivative of dependent variable to independent variable | Whether to require independent variable for differentiation | Whether to require another independent variable for differentiation | Whether to require dependent variable for differentiation | Whether to use independent variable height and width specification for differentiation | Whether the input set of the training set has changed
>> :-|:-|:-|:-|:-|:-|:-|:-|:-|:-|:-|:-
>> 00_full_dataset        |                            |1 |Yes,<br>but manual configuration is not required|No|  |                |  |  |  |  |
>> 00_const               |                            |0 |Yes|No|  |                |  |  |  |  |
>> 00_buff_slice_rand_pick|sm_sc.<br>ft_nn_buff_slice_rand_pick|0|Yes|No|  |      |  |  |  |Yes
>> 00_none                |sm_sc.<br>fv_nn_none            |1 ||Yes|sm_sc.<br>fv_d_nn_none_dloss_dindepdt|    |  |  |  |Yes|
>> 01_add                 |sm_sc.<br>fv_opr_add            |2 ||Yes|sm_sc.<br>fv_d_add|Yes    |  |  |Yes|Yes|
>> 01_mul                 |sm_sc.<br>fv_opr_mul            |2 ||Yes|sm_sc.<br>fv_d_mul|Yes    |  |Yes|  |  |
>> 01_sub                 |sm_sc.<br>fv_opr_sub            |2 ||Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_sub_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_sub_2|Yes    |  |  |Yes|Yes|
>> 01_0sub                |sm_sc.<br>fv_opr_sub            |1 ||Yes|sm_sc.<br>fv_d_sub_2|Yes    |  |  |  |Yes|
>> 01_div                 |sm_sc.<br>fv_opr_div            |2 ||Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_div_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_div_2_un_de_broadcast|Yes    |2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|  |  |
>> 01_1div                |sm_sc.<br>fv_opr_div            |1 ||Yes|sm_sc.<br>fv_d_div_2_un_de_broadcast|Yes    |Yes|  |  |  |
>> 01_pow                 |sm_sc.<br>fv_opr_pow            |2 ||Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_pow_1_un_de_broadcast<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_pow_2_un_de_broadcast|Yes      |1st Opr: Yes<br>|1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|  |
>> 01_log                 |sm_sc.<br>fv_opr_log            |2 ||Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_log_1_un_de_broadcast<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_log_2_un_de_broadcast|Yes      |1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes|  |
>> 01_exp                 |sm_sc.<br>fv_opr_exp            |1 ||Yes|sm_sc.<br>fv_d_exp|Yes      |  |  |Yes|  |
>> 01_ln                  |sm_sc.<br>fv_opr_ln             |1 ||Yes|sm_sc.<br>fv_d_ln|Yes    |Yes|  |  |  |
>> 01_abs                 |sm_sc.<br>fv_opr_abs            |1 ||Yes|sm_sc.<br>fv_d_abs|Yes    |Yes|  |  |  |
>> 01_prod_mx             |sm_sc.<br>fv_opr_prod_mx_py     |2|Yes,<br>Only the height and width of the 2nd Opr need to be configured, that is, the 2nd and 3rd Opr of the co parameters|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_prod_mx_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_prod_mx_dloss_dindepdt_2|    |1st Opr: Yes<br>2nd Opr: Yes|  ||1st Opr: Yes<br>2nd Opr: Yes|
>> 01_chunk_prod_mx       |sm_sc.<br>fv_chunk_prod_mx      |2|Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_chunk_prod_mx_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_chunk_prod_mx_dloss_dindepdt_2|    |1st Opr: Yes<br>2nd Opr: Yes|  ||1st Opr: Yes<br>2nd Opr: Yes|
>> 02_sin                 |sm_sc.<br>fv_sin                |1 ||Yes|sm_sc.<br>fv_d_sin  |Yes    |Yes|  |  |  |
>> 02_cos                 |sm_sc.<br>fv_cos                |1 ||Yes|sm_sc.<br>fv_d_cos  |Yes    |Yes|  |  |  |
>> 02_tan                 |sm_sc.<br>fv_tan                |1 ||Yes|sm_sc.<br>fv_d_tan  |Yes    |Yes|  |  |  |
>> 02_cot                 |sm_sc.<br>fv_cot                |1 ||Yes|sm_sc.<br>fv_d_cot  |Yes    |Yes|  |  |  |
>> 02_sec                 |sm_sc.<br>fv_sec                |1 ||Yes|sm_sc.<br>fv_d_sec  |Yes    |Yes|  |  |  |
>> 02_csc                 |sm_sc.<br>fv_csc                |1 ||Yes|sm_sc.<br>fv_d_csc  |Yes    |Yes|  |  |  |
>> 02_asin                |sm_sc.<br>fv_asin               |1 ||Yes|sm_sc.<br>fv_d_asin |Yes    |Yes|  |  |  |
>> 02_acos                |sm_sc.<br>fv_acos               |1 ||Yes|sm_sc.<br>fv_d_acos |Yes    |Yes|  |  |  |
>> 02_atan                |sm_sc.<br>fv_atan               |1 ||Yes|sm_sc.<br>fv_d_atan |Yes    |Yes|  |  |  |
>> 02_acot                |sm_sc.<br>fv_acot               |1 ||Yes|sm_sc.<br>fv_d_acot |Yes    |Yes|  |  |  |
>> 02_sinh                |sm_sc.<br>fv_sinh               |1 ||Yes|sm_sc.<br>fv_d_sinh |Yes    |Yes|  |  |  |
>> 02_cosh                |sm_sc.<br>fv_cosh               |1 ||Yes|sm_sc.<br>fv_d_cosh |Yes    |Yes|  |  |  |
>> 02_tanh                |sm_sc.<br>fv_tanh               |1 ||Yes|sm_sc.<br>fv_d_tanh |Yes      |  |  |Yes|  |
>> 02_asinh               |sm_sc.<br>fv_asinh              |1 ||Yes|sm_sc.<br>fv_d_asinh|Yes    |Yes|  |  |  |
>> 02_acosh               |sm_sc.<br>fv_acosh              |1 ||Yes|sm_sc.<br>fv_d_acosh|Yes    |Yes|  |  |  |
>> 02_atanh               |sm_sc.<br>fv_atanh              |1 ||Yes|sm_sc.<br>fv_d_atanh|Yes    |Yes|  |  |  |
>> 03_sigmoid             |sm_sc.<br>fv_activate_sigmoid            |1 ||Yes|sm_sc.<br>fv_d_activate_sigmoid|Yes      |  |  |Yes|  |
>> 03_relu                |sm_sc.<br>fv_activate_relu               |1 ||Yes|sm_sc.<br>fv_d_activate_relu|Yes    |Yes|  |  |  |
>> 03_leaky_relu          |sm_sc.<br>fv_activate_leaky_relu         |1 |Yes|Yes|sm_sc.<br>fv_d_activate_leaky_relu|Yes    |Yes|  |  |  |
>> 03_elu                 |sm_sc.<br>fv_activate_elu                |1 |Yes|Yes|sm_sc.<br>fv_d_activate_elu|Yes    |Yes|  |  |  |
>> 03_selu                |sm_sc.<br>fv_activate_selu               |1 ||Yes|sm_sc.<br>fv_d_activate_selu|Yes    |Yes|  |  |  |
>> 03_gelu                |sm_sc.<br>fv_activate_gelu               |1 ||Yes|sm_sc.<br>fv_d_activate_gelu|Yes    |Yes|  |  |  |
>> 03_swish               |sm_sc.<br>fv_activate_swish              |1 ||Yes|sm_sc.<br>fv_d_activate_swish|Yes      |Yes|  |Yes|  |
>> 03_softplus            |sm_sc.<br>fv_activate_softplus           |1 ||Yes|sm_sc.<br>fv_d_activate_softplus|Yes    |Yes|  |  |  |
>> 03_boxcox              |sm_sc.<br>fv_activate_boxcox             |1 ||Yes|sm_sc.<br>fv_d_activate_boxcox|Yes    |Yes|  |  |  |
>> 03_softmax             |sm_sc.<br>fv_redistr_softmax             |1 ||Yes|sm_sc.<br>fv_d_redistr_softmax_dloss_dindepdt|        |Yes|  |Yes|  |
>> 03_softmax_ex          |sm_sc.<br>fv_redistr_softmax_ex           |1 ||Yes|sm_sc.<br>fv_d_redistr_softmax_dloss_dindepdt|        |Yes|  |Yes|  |
>> 03_zscore              |sm_sc.<br>fv_redistr_zscore              |1 ||Yes|sm_sc.<br>fv_d_redistr_zscore                |Yes      |Yes|  |Yes|  |
>> 04_new                 |sm_sc.<br>fv_new                |1 |Yes|Yes|sm_sc.<br>fv_d_new_dloss_dindepdt|        |  |  |  |  |Yes
>> 04_repeat_axis         |sm_sc.<br>fv_repeat_axis_py     |1 |Yes|Yes|sm_sc.<br>fv_d_repeat_axis_dloss_dindepdt||  |  |  |  |Yes
>> 04_apad                |sm_sc.<br>fv_apad|2|Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_apad_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_apad_dloss_dindepdt_2|||  ||Yes|
>> 04_bpad                |sm_sc.<br>fv_bpad|2|Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_bpad_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_bpad_dloss_dindepdt_2|||  ||Yes|
>> 04_lpad                |sm_sc.<br>fv_lpad|2|Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_lpad_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_lpad_dloss_dindepdt_2|||  ||Yes|
>> 04_rpad                |sm_sc.<br>fv_rpad|2|Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_rpad_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_rpad_dloss_dindepdt_2|||  ||Yes|
>> 04_transpose           |sm_sc.<br>fv_opr_transpose_py      |1 |No|Yes|sm_sc.<br>fv_d_transpose_dloss_dindepdt     |        |  |  |  |  |
>> 04_transpose_i         |sm_sc.<br>fv_opr_transpose_i_py    |1 |No|Yes|sm_sc.<br>fv_d_transpose_i_dloss_dindepdt   |        |  |  |  |  |
>> 04_chunk_transpose     |sm_sc.<br>fv_chunk_transpose    |1 |No|Yes|sm_sc.<br>fv_d_chunk_transpose_dloss_dindepdt |        |  |  |  |  |
>> 04_transpose_nd        |sm_sc.<br>fv_opr_transpose_nd_py   |1 |Yes|Yes|sm_sc.<br>fv_d_transpose_nd_dloss_dindepdt  |        |  |  |  |  |
>> 04_turn_90             |sm_sc.<br>fv_opr_turn_90            |1 |Yes|Yes|sm_sc.<br>fv_d_turn_90_dloss_dindepdt_py       |        |  |  |  |  |
>> 04_turn_180            |sm_sc.<br>fv_opr_turn_180           |1 |Yes|Yes|sm_sc.<br>fv_d_turn_180_dloss_dindepdt_py      |        |  |  |  |  |
>> 04_mirror              |sm_sc.<br>fv_opr_mirror_py         |1 |Yes|Yes|sm_sc.<br>fv_d_mirror_dloss_dindepdt        |        |  |  |  |  |
>> 04_mx_ele_3d_2_2d      |sm_sc.<br>fv_mx_ele_3d_2_2d_py     |1 |Yes|Yes|sm_sc.<br>fv_d_mx_ele_3d_2_2d_dloss_dindepdt|        |  |  |  |  |
>> 04_mx_ele_2d_2_3d      |sm_sc.<br>fv_mx_ele_2d_2_3d_py     |1 |Yes|Yes|sm_sc.<br>fv_d_mx_ele_2d_2_3d_dloss_dindepdt|        |  |  |  |  |
>> 04_mx_ele_4d_2_3d      |sm_sc.<br>fv_mx_ele_4d_2_3d_py     |1 |Yes|Yes|sm_sc.<br>fv_d_mx_ele_4d_2_3d_dloss_dindepdt|        |  |  |  |Yes|
>> 04_mx_ele_3d_2_4d      |sm_sc.<br>fv_mx_ele_3d_2_4d_py     |1 |Yes|Yes|sm_sc.<br>fv_d_mx_ele_3d_2_4d_dloss_dindepdt|        |  |  |  |  |
>> 04_mx_ele_flatten_2dims|sm_sc.<br>fv_mx_ele_flatten_2dims_py|1 |Yes|Yes|sm_sc.<br>fv_d_mx_ele_flatten_2dims_dloss_dindepdt_py|        |  |  |  |Yes|
>> 04_mx_slice_3d_2_2d    |sm_sc.<br>fv_mx_slice_3d_2_2d_py     |1 |Yes|Yes|sm_sc.<br>fv_d_mx_slice_3d_2_2d_dloss_dindepdt|        |  |  |  |Yes|
>> 04_mx_slice_4d_2_2d    |sm_sc.<br>fv_mx_slice_4d_2_2d_py     |1 |Yes|Yes|sm_sc.<br>fv_d_mx_slice_4d_2_2d_dloss_dindepdt|        |  |  |  |Yes|
>> 04_mx_slice_4d_2_3d    |sm_sc.<br>fv_mx_slice_4d_2_3d_py     |1 |Yes|Yes|sm_sc.<br>fv_d_mx_slice_4d_2_3d_dloss_dindepdt|        |  |  |  |Yes|
>> 04_mx_ascend_dim       |sm_sc.<br>fv_mx_ascend_dim        |1 |Yes|Yes|sm_sc.<br>fv_d_mx_ascend_dim_dloss_dindepdt|        |  |  |  |  |
>> 04_mx_descend_dim      |sm_sc.<br>fv_mx_descend_dim_py       |1 |Yes|Yes|sm_sc.<br>fv_d_mx_descend_dim_dloss_dindepdt|       |  |  |  |  |
>> 04_rand_pick_y         |sm_sc.<br>fv_rand_slice_y_pick           |1 |Yes|No||        |  |  |  |  |Yes
>> 04_rand_pick_x         |sm_sc.<br>fv_rand_slice_x_pick           |1 |Yes|No||        |  |  |  |  |
>> 04_rand_pick_x3        |sm_sc.<br>fv_rand_slice_x3_pick          |1 |Yes|No||        |  |  |  |  |
>> 04_rand_pick_x4        |sm_sc.<br>fv_rand_slice_x4_pick          |1 |Yes|No||        |  |  |  |  |
>> 04_chunk               |sm_sc.<br>fv_chunk                               |1 |Yes|Yes|sm_sc.<br>fv_d_chunk_dloss_dindepdt |    |  |  |  |Yes|Yes
>> 04_slice_y             |sm_sc.<br>fv_slice_y                             |1 |Yes|Yes|sm_sc.<br>fv_d_slice_y_dloss_dindepdt |    |  |  |  |Yes|Yes
>> 04_slice_x             |sm_sc.<br>fv_slice_x                             |1 |Yes|Yes|sm_sc.<br>fv_d_slice_x_dloss_dindepdt |    |  |  |  |Yes|
>> 04_slice_x3            |sm_sc.<br>fv_slice_x3                            |1 |Yes|Yes|sm_sc.<br>fv_d_slice_x3_dloss_dindepdt|    |  |  |  |Yes|
>> 04_slice_x4            |sm_sc.<br>fv_slice_x4                            |1 |Yes|Yes|sm_sc.<br>fv_d_slice_x4_dloss_dindepdt|    |  |  |  |Yes|
>> 04_sample_y            |sm_sc.<br>fv_sample_y              |1 |Yes|Yes, currently, the 2nd Opr is not supported|sm_sc.<br>fv_d_sample_y_dloss_dindepdt_1 |        |  |  |  |Yes|Yes
>> 04_sample_x            |sm_sc.<br>fv_sample_x              |1 |Yes|Yes, currently, the 2nd Opr is not supported|sm_sc.<br>fv_d_sample_x_dloss_dindepdt_1 |        |  |  |  |Yes|
>> 04_sample_x3           |sm_sc.<br>fv_sample_x3             |1 |Yes|Yes, currently, the 2nd Opr is not supported|sm_sc.<br>fv_d_sample_x3_dloss_dindepdt_1|        |  |  |  |Yes|
>> 04_sample_x4           |sm_sc.<br>fv_sample_x4             |1 |Yes|Yes, currently, the 2nd Opr is not supported|sm_sc.<br>fv_d_sample_x4_dloss_dindepdt_1|        |  |  |  |Yes|
>> 04_lower_tri_mx        |sm_sc.<br>fv_lower_tri_mx       |1 |Yes|Yes|sm_sc.<br>fv_d_lower_tri_mx_dloss_dindepdt|        |  |  |  |  |
>> 04_upper_tri_mx        |sm_sc.<br>fv_upper_tri_mx       |1 |Yes|Yes|sm_sc.<br>fv_d_upper_tri_mx_dloss_dindepdt|        |  |  |  |  |
>> 04_lower_tri_mx_ex     |sm_sc.<br>fv_lower_tri_mx_ex    |1 |Yes|Yes|sm_sc.<br>fv_d_lower_tri_mx_ex_dloss_dindepdt|     |  |  |  |  |
>> 04_upper_tri_mx_ex     |sm_sc.<br>fv_upper_tri_mx_ex    |1 |Yes|Yes|sm_sc.<br>fv_d_upper_tri_mx_ex_dloss_dindepdt|     |  |  |  |  |
>> 04_lmask               |sm_sc.<br>fv_lmask              |1 |Yes|Yes, currently, the 2nd Opr is not supported|sm_sc.<br>fv_d_lmask_dloss_dindepdt_1|        |  |1st Opr: Yes|  |  |
>> 04_rmask               |sm_sc.<br>fv_rmask              |1 |Yes|Yes, currently, the 2nd Opr is not supported|sm_sc.<br>fv_d_rmask_dloss_dindepdt_1|        |  |1st Opr: Yes|  |  |
>> 04_amask               |sm_sc.<br>fv_amask              |1 |Yes|Yes, currently, the 2nd Opr is not supported|sm_sc.<br>fv_d_amask_dloss_dindepdt_1|        |  |1st Opr: Yes|  |  |
>> 04_bmask               |sm_sc.<br>fv_bmask              |1 |Yes|Yes, currently, the 2nd Opr is not supported|sm_sc.<br>fv_d_bmask_dloss_dindepdt_1|        |  |1st Opr: Yes|  |  |
>> 05_pool_max_2d_grp_x   |sm_sc.<br>fv_pool_max_2d_grp_x  |1 |Yes|Yes|sm_sc.<br>fv_d_pool_max_grp_x_dloss_dindepdt|        |Yes|  |Yes|  |
>> 05_pool_avg_2d_grp_x   |sm_sc.<br>fv_pool_avg_2d_grp_x  |1 |Yes|Yes|sm_sc.<br>fv_d_pool_avg_grp_x_dloss_dindepdt|        |  |  |  |  |
>> 05_pool_max            |sm_sc.<br>fv_pool_max_py        |1 |Yes|Yes|sm_sc.<br>fv_d_pool_max_dloss_dindepdt_ex|        |Yes|  |Yes|  |
>> 05_pool_avg            |sm_sc.<br>fv_pool_avg_py        |1 |Yes|Yes|sm_sc.<br>fv_d_pool_avg_dloss_dindepdt_ex|     |  |  |  |  |
>> 05_pool_none           |sm_sc.<br>fv_pool_none_py       |1 |Yes|Yes|sm_sc.<br>fv_d_pool_none_dloss_dindepdt|        |  |  |  |Yes|
>> 05_conv_2d_grp_x       |sm_sc.<br>fv_conv_2d_grp_x      |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_2d_grp_x_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_2d_grp_x_dloss_dindepdt_2|        |  |1st Opr: Yes<br>2nd Opr: Yes|  |  |
>> 05_conv_2d             |sm_sc.<br>fv_conv_2d_im2col_py   |2 或 3|Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_2d_dloss_dindepdt_1_ex<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_2d_dloss_dindepdt_2<br>3rd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_2d_dloss_dindepdt_3|        |  |1st Opr: Yes<br>2nd Opr: Yes|  |3rd Opr: Yes|
>> 05_conv_add            |sm_sc.<br>fv_conv_add           |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_add_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_add_dloss_dindepdt_2|        ||  ||1st Opr: Yes<br>2nd Opr: Yes|
>> 05_conv_sub            |sm_sc.<br>fv_conv_sub           |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_sub_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_sub_dloss_dindepdt_2|        ||  ||1st Opr: Yes<br>2nd Opr: Yes|
>> 05_conv_mul            |sm_sc.<br>fv_conv_mul           |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_mul_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_mul_dloss_dindepdt_2|        |  |1st Opr: Yes<br>2nd Opr: Yes|  |1st Opr: Yes<br>2nd Opr: Yes|
>> 05_conv_div            |sm_sc.<br>fv_conv_div           |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_div_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_div_dloss_dindepdt_2|        |2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|  |1st Opr: Yes|
>> 05_conv_pow            |sm_sc.<br>fv_conv_pow           |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_pow_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_pow_dloss_dindepdt_2|        |1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|  |
>> 05_conv_log            |sm_sc.<br>fv_conv_log           |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_log_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_log_dloss_dindepdt_2|        |1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes|  |
>> 05_conv_prod_mx        |sm_sc.<br>fv_conv_prod_mx       |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_prod_mx_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_prod_mx_dloss_dindepdt_2|        |  |1st Opr: Yes<br>2nd Opr: Yes|  |1st Opr: Yes<br>2nd Opr: Yes|
>> 05_conv_de_sub         |sm_sc.<br>fv_conv_de_sub        |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_sub_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_sub_dloss_dindepdt_2|        ||  ||1st Opr: Yes<br>2nd Opr: Yes|
>> 05_conv_de_div         |sm_sc.<br>fv_conv_de_div        |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_div_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_div_dloss_dindepdt_2|        |2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|  |1st Opr: Yes|
>> 05_conv_de_pow         |sm_sc.<br>fv_conv_de_pow        |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_pow_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_pow_dloss_dindepdt_2|        |1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|  |
>> 05_conv_de_log         |sm_sc.<br>fv_conv_de_log        |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_log_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_log_dloss_dindepdt_2|        |1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes<br>2nd Opr: Yes|1st Opr: Yes|  |
>> 05_conv_de_prod_mx     |sm_sc.<br>fv_conv_de_prod_mx    |2 |Yes|Yes|1st Opr:<br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_prod_mx_dloss_dindepdt_1<br>2nd Opr: <br>&nbsp;&nbsp;sm_sc.<br>&nbsp;&nbsp;fv_d_conv_de_prod_mx_dloss_dindepdt_2|        |  |1st Opr: Yes<br>2nd Opr: Yes|  |1st Opr: Yes<br>2nd Opr: Yes|
>> 06_aggr_mx_sum         |sm_sc.<br>fa_mx_sum             |n ||Yes|sm_sc.<br>fv_d_mx_sum |Yes    |  |  |Yes|Yes|
>> 06_aggr_mx_prod        |sm_sc.<br>fa_mx_prod            |n ||Yes|sm_sc.<br>fv_d_mx_prod|Yes      |Yes|  |Yes|  |
>> 06_aggr_mx_avg         |sm_sc.<br>fa_mx_avg             |n |Yes,<br>but manual configuration is not required|Yes|sm_sc.<br>fv_d_mx_avg |Yes    |  |  |Yes|Yes|
>> 06_aggr_mx_max         |sm_sc.<br>fa_mx_max             |n ||Yes|sm_sc.<br>fv_d_mx_max |Yes      |Yes|  |Yes|  |
>> 06_aggr_mx_min         |sm_sc.<br>fa_mx_min             |n ||Yes|sm_sc.<br>fv_d_mx_min |Yes      |Yes|  |Yes|  |
>> 06_aggr_mx_concat_y    |sm_sc.<br>fa_mx_concat_y        |n |Yes,<br>but manual configuration is not required|Yes|sm_sc.<br>fv_d_mx_concat_y_dloss_dindepdt_n |        |  |  |  |Yes|Yes
>> 06_aggr_mx_concat_x    |sm_sc.<br>fa_mx_concat_x        |n |Yes,<br>but manual configuration is not required|Yes|sm_sc.<br>fv_d_mx_concat_x_dloss_dindepdt_n |        |  |  |  |Yes|
>> 06_aggr_mx_concat_x3   |sm_sc.<br>fa_mx_concat_x3       |n |Yes,<br>but manual configuration is not required|Yes|sm_sc.<br>fv_d_mx_concat_x3_dloss_dindepdt_n|        |  |  |  |Yes|
>> 06_aggr_mx_concat_x4   |sm_sc.<br>fa_mx_concat_x4       |n |Yes,<br>but manual configuration is not required|Yes|sm_sc.<br>fv_d_mx_concat_x4_dloss_dindepdt_n|        |  |  |  |Yes|
>> 07_aggr_slice_sum      |sm_sc.<br>fv_aggr_slice_sum     |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_slice_sum_dloss_dindepdt|        |  |  |  |  |
>> 07_aggr_slice_prod     |sm_sc.<br>fv_aggr_slice_prod    |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_slice_prod_dloss_dindepdt|       |Yes|  |Yes|  |
>> 07_aggr_slice_avg      |sm_sc.<br>fv_aggr_slice_avg     |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_slice_avg_dloss_dindepdt|        |  |  |  |  |
>> 07_aggr_slice_max      |sm_sc.<br>fv_aggr_slice_max     |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_slice_max_dloss_dindepdt|        |Yes|  |Yes|  |
>> 07_aggr_slice_min      |sm_sc.<br>fv_aggr_slice_min     |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_slice_min_dloss_dindepdt|        |Yes|  |Yes|  |
>> 07_aggr_chunk_sum      |sm_sc.<br>fv_aggr_chunk_sum     |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_chunk_sum_dloss_dindepdt|        |  |  |  |Yes|
>> 07_aggr_chunk_prod     |sm_sc.<br>fv_aggr_chunk_prod    |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_chunk_prod_dloss_dindepdt|       |Yes|  |Yes|  |
>> 07_aggr_chunk_avg      |sm_sc.<br>fv_aggr_chunk_avg     |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_chunk_avg_dloss_dindepdt|        |  |  |  |Yes|
>> 07_aggr_chunk_max      |sm_sc.<br>fv_aggr_chunk_max     |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_chunk_max_dloss_dindepdt|        |Yes|  |Yes|  |
>> 07_aggr_chunk_min      |sm_sc.<br>fv_aggr_chunk_min     |1 |Yes|Yes|sm_sc.<br>fv_d_aggr_chunk_min_dloss_dindepdt|        |Yes|  |Yes|  |
